/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <gm.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "orfs_protocol.h"
#include "orfs_common_network.h"
#include "interactive_messages.h"

#define ANSBUF_SIZE 31


unsigned long gm_server_port = ORFS_GM_SERVER_PORT;
unsigned long gm_server_board = ORFS_GM_SERVER_BOARD;
unsigned long gm_client_board = ORFS_GM_CLIENT_BOARD;
static unsigned int server_id;
static struct gm_port *port = NULL;

static struct orfs_request *static_req = NULL;
static struct orfs_answer *static_ans = NULL;


int GM_init(char *name)
{
  gm_status_t status;
  int port_id = ORFS_GM_CLIENT_FIRST_PORT_ID;

  /* Initialize GM */
  status = gm_init();
  if ( status != GM_SUCCESS ) {
    ERROR("GM INIT failed with error %s\n",
	  gm_strerror(status));
    goto error;
  }

  /* Find a free port and open it */
  while ( port == NULL ) {
    status = gm_open(&port,
		     gm_client_board, port_id,
		     "intcli", GM_API_VERSION);
    if ( status == GM_SUCCESS ) {
      ERROR("GM port %d successfully open\n", port_id);
      break;
    }

    ERROR("GM cannot open port %d (error %s)\n",
	  port_id, gm_strerror(status));

    if ( status != GM_BUSY ) {
      ERROR("Failed to open a GM port\n");
      goto error_with_init;
    }

    port = NULL;
    port_id++;
  }

  /* Prepare connection */
  if ( !strcmp(name, "localhost") ) {
    status = gm_get_node_id(port, &server_id);
    if ( status != GM_SUCCESS ) {
      ERROR("Cannot find GM id (error %s)\n", 
	    gm_strerror(status));
      goto error_with_port;
    }
    ERROR("Connected to localhost\n");
  } else {
    char name_with_board[GM_MAX_HOST_NAME_LEN+1];
    char *fullname = name;
    int count;
    if ( gm_server_board ) {
      count = snprintf(name_with_board, GM_MAX_HOST_NAME_LEN+1,
		       "%s:%ld", name, gm_server_board);
      if ( count < 0 || count > GM_MAX_HOST_NAME_LEN ) {
	ERROR("GM name too long\n");
	goto error_with_port;
      }
      fullname = name_with_board;
    }
    server_id = gm_host_name_to_node_id(port, fullname);
    if ( server_id == GM_NO_SUCH_NODE_ID ) {
      ERROR("Cannot find GM node id of node %s\n", fullname);
      goto error_with_port;
    }
    ERROR("Connected to %s\n", fullname);
  }

  /* Allocate Static Buffers */
  static_req = gm_dma_malloc(port, sizeof(*static_req));
  static_ans = gm_dma_malloc(port, sizeof(*static_ans));

  return 0;

 error_with_port:
  gm_close(port);
 error_with_init:
  gm_finalize();
 error:
  return -1;
}


static void
req_send_callback(struct gm_port *port, void *context,
		     gm_status_t status)
{
  if ( status != GM_SUCCESS )
    ERROR("Request send got status %s\n", gm_strerror(status));  
}


static void
data_send_callback(struct gm_port *port, void *context,
		      gm_status_t status)
{
  if ( status != GM_SUCCESS )
    ERROR("Data send got status %s\n", gm_strerror(status));  
}


#define MAX_LINE_SIZE 32
static char line[MAX_LINE_SIZE];


int GM_request(struct orfs_request *req,
	       const void *in_buf,
	       struct orfs_answer *ans,
	       void *out_buf,
	       size_t out_buf_len)
{
  size_t send_len;
  int to_recv = 1;

  /* initializes tag to 0 */
  req->tag = 0;

  /* Provide Receive Buffers */
  gm_provide_receive_buffer(port,
			    static_ans,
			    ORFS_GM_ANS_SIZE,
			    ORFS_GM_ANS_PRIO);
  if ( out_buf ) {
    gm_register_memory(port, out_buf, out_buf_len);

    printf("Use RDMA from the server to recv data (y/n) ? ");
    get_line(line, MAX_LINE_SIZE);
    if ( *line == 'y' ) {
      req->net.gm.rdma_ansbuf = 1;
      req->net.gm.ansbuf.rdma_ptr = (gm_remote_ptr_t) (unsigned long) out_buf;
    } else {
      req->net.gm.rdma_ansbuf = 0;
      req->net.gm.ansbuf.rdv_size = ANSBUF_SIZE;
      gm_provide_receive_buffer(port,
				out_buf,
				ANSBUF_SIZE,
				ORFS_GM_ANSBUF_PRIO);
      to_recv++;
    }
  }

  /* Send Buffer if Needed */
  if ( get_request_buffer_len(req, &send_len) ) {
    gm_register_memory(port, (void*)in_buf, send_len);

    printf("Use RDMA from the server to send data (y/n) ? ");
    get_line(line, MAX_LINE_SIZE);
    if ( *line == 'y' ) {
      req->net.gm.rdma_reqbuf = 1;
      req->net.gm.reqbuf.rdma_ptr = (gm_remote_ptr_t) (unsigned long) in_buf;
    } else {
      req->net.gm.rdma_reqbuf = 0;
      gm_send_with_callback(port,
			    (void*)in_buf,
			    ORFS_GM_REQBUF_SIZE(send_len), send_len,
			    ORFS_GM_REQBUF_PRIO,
			    server_id, gm_server_port,
			    data_send_callback, NULL);
    }
  }
  
  /* copy request to static_req */
  memcpy(static_req, req, sizeof(*static_req));

  /* Send Request */
  gm_send_with_callback(port,
			static_req,
			ORFS_GM_REQ_SIZE, sizeof(*static_req),
			ORFS_GM_REQ_PRIO,
			server_id, gm_server_port,
			req_send_callback, NULL);

  /* Wait for the answer and its buffer */
  while ( to_recv ) {
    gm_recv_event_t *ev = gm_blocking_receive(port);
    if ( ( gm_ntoh_u8(ev->recv.type) == ORFS_GM_ANS_RECV_EV
	   || gm_ntoh_u8(ev->recv.type) == ORFS_GM_ANSBUF_RECV_EV )
	 && !--to_recv)
      break;
    gm_unknown(port, ev);
  }

  /* copy answer from static_ans */
  memcpy(ans, static_ans, sizeof(*static_ans));
  
  /* Deregister Buffer if Needed */
  if ( in_buf )
    gm_deregister_memory(port, (void*)in_buf, send_len);
  if ( out_buf )
    gm_deregister_memory(port, out_buf, out_buf_len);

  return 0;
}


void GM_exit()
{
  gm_dma_free(port, static_ans);
  gm_dma_free(port, static_req);
  gm_close(port);
  gm_finalize();
}
