/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <myriexpress.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "orfs_protocol.h"
#include "orfs_common_network.h"
#include "interactive_network.h"
#include "interactive_messages.h"


#define ID 0xfedcba98


unsigned long mx_server_ep = ORFS_MX_SERVER_EP;
unsigned long mx_server_board = ORFS_MX_SERVER_BOARD;
unsigned long mx_client_board = 0; /* not used for now */
static mx_endpoint_t ep;
static mx_endpoint_addr_t server_addr;


int MX_init(char *name)
{
  uint64_t nic_id;
  mx_return_t ret;

  /* Initialize MX */
  ret = mx_init();
  if ( ret != MX_SUCCESS ) {
    ERROR("MX INIT failed with error %s\n",
	  mx_strerror(ret));
    goto error;
  }

  /* Open an endpoint */
  ret = mx_open_endpoint(MX_ANY_NIC, MX_ANY_ENDPOINT, 0, NULL, 0, &ep);
  if ( ret != MX_SUCCESS ) {
    ERROR("Failed to open a MX endpoint, error %s\n",
	  mx_strerror(ret));
    goto error_with_init;
  } else {
    ERROR("MX endpoint successfully open\n");  // which one ?
  }

  /* Get server nics */
  if ( !strcmp(name, "localhost") ) {
    ret = mx_board_number_to_nic_id(mx_server_board, &nic_id);
    if ( ret != MX_SUCCESS ) {
      ERROR("Cannot find MX local nic id (error %s)\n",
	    mx_strerror(ret));
      goto error_with_ep;
    }
  } else {
    uint32_t nics = 1;
    // use with board number
    uint64_t nic_ids[1];
    ret = mx_hostname_to_nic_ids(name, nic_ids, &nics);
    if ( ret != MX_SUCCESS ) {
      ERROR("Cannot find MX nic ids of node %s (error %s)\n",
	    name, mx_strerror(ret));
      goto error_with_ep;
    }
    nic_id = nic_ids[0];
  }

  ERROR("Using server nic id %llx port %ld filter %lx\n",
	nic_id, mx_server_ep, ORFS_MX_EP_FILTER);

  /* Connect to server's endpoint */
  ret = mx_connect(ep, nic_id, mx_server_ep, ORFS_MX_EP_FILTER,
		   &server_addr);
  if ( ret != MX_SUCCESS ) {
    ERROR("Cannot connect to server's endpoint (error %s)\n",
	  mx_strerror(ret));
    goto error_with_ep;
  }

  return 0;

 error_with_ep:
  mx_close_endpoint(ep);
 error_with_init:
  mx_finalize();
 error:
  return -1;
}


#define MAX_LINE_SIZE 32
static char line[MAX_LINE_SIZE];


#define ISEND_REQ 0x1
#define ISEND_REQBUF 0x2
#define RDMA_WIN_REQBUF 0x4
#define IRECV_ANS 0x10
#define IRECV_ANSBUF 0x20
#define RDMA_WIN_ANSBUF 0x40


int MX_request(struct orfs_request *req,
	       const void *in_buf,
	       struct orfs_answer *ans,
	       void *out_buf,
	       size_t out_buf_len)
{
  mx_segment_t req_seg;
  mx_request_t req_req;
  mx_segment_t ans_seg;
  mx_request_t ans_req;
  mx_segment_t reqbuf_seg;
  mx_request_t reqbuf_req;
  mx_segment_t ansbuf_seg;
  mx_request_t ansbuf_req;
  uint32_t reqbuf_window_handle;
  uint32_t ansbuf_window_handle;
  mx_return_t ret;
  mx_status_t status;
  uint32_t result;
  size_t in_buf_len;
  unsigned long to_cancel = 0;

  /* initializes tag to 0 */
  req->tag = 0;

  /* prepare request buffer */
  if ( in_buf ) {
    get_request_buffer_len(req, &in_buf_len);
    reqbuf_seg.segment_ptr = (void *) in_buf;
    reqbuf_seg.segment_length = in_buf_len;

    printf("Use RDMA from the server to send data (y/n) ? ");
    get_line(line, MAX_LINE_SIZE);

    req->net.mx.rdma_reqbuf = ( *line == 'y' );

    if ( *line == 'y' ) {
      /* provide a RDMA window to the server */
      ret = mx_create_rdma_window(ep,
				  &reqbuf_seg, 1,
				  MX_RDMA_READ,
				  &reqbuf_window_handle);
      if ( ret != MX_SUCCESS ) {
        ERROR("failed to create request buffer rdma window (error %s)\n",
              mx_strerror(ret));
        goto error;
      }
      req->net.mx.reqbuf.window_handle = reqbuf_window_handle;
      req->net.mx.reqbuf.offset = 0;
      to_cancel |= RDMA_WIN_REQBUF;
    } else {
      /* send the request buffer */
      if ( in_buf_len > ORFS_MX_PREPROVIDED_REQBUF_LEN ) {
	ret = mx_isend(ep,
		       &reqbuf_seg, 1,
		       server_addr,
		       ORFS_MX_LARGE_REQBUF_MATCH(ID),
		       NULL, &reqbuf_req);
      } else {
	ret = mx_isend(ep,
		       &reqbuf_seg, 1,
		       server_addr,
		       ORFS_MX_PREPROVIDED_REQBUF_MATCH(ID),
		       NULL, &reqbuf_req);
      }
      if ( ret != MX_SUCCESS ) {
	ERROR("failed to post request buffer send (error %s)\n",
	      mx_strerror(ret));
	goto error;
      }
      to_cancel |= ISEND_REQBUF;
    }
  }

  /* prepare answer buffer */
  if ( out_buf ) {
    ansbuf_seg.segment_ptr = out_buf;
    ansbuf_seg.segment_length = out_buf_len;

    printf("Use RDMA from the server to recv data (y/n) ? ");
    get_line(line, MAX_LINE_SIZE);

    req->net.mx.rdma_ansbuf = ( *line == 'y' );
    
    if ( *line == 'y' ) {
      /* provide a RDMA window to the server */
      ret = mx_create_rdma_window(ep,
				  &ansbuf_seg, 1,
				  MX_RDMA_WRITE,
				  &ansbuf_window_handle);
      if ( ret != MX_SUCCESS ) {
        ERROR("failed to create answer buffer rdma window (error %s)\n",
              mx_strerror(ret));
        goto error;
      }
      req->net.mx.ansbuf.window_handle = ansbuf_window_handle;
      req->net.mx.ansbuf.offset = 0;
      to_cancel |= RDMA_WIN_ANSBUF;
    } else {
      /* post a recv for the answer buffer */
      ret = mx_irecv(ep,
		     &ansbuf_seg, 1,
		     ORFS_MX_ANSBUF_RECV_MATCH(ID), ORFS_MX_ANSBUF_RECV_MASK,
		     NULL, &ansbuf_req);
      if ( ret != MX_SUCCESS ) {
	ERROR("failed to post answer buffer receive (error %s)\n",
	      mx_strerror(ret));
	goto error;
      }
      to_cancel |= IRECV_ANSBUF;
    }
  }

  /* prepare answer */
  ans_seg.segment_ptr = ans;
  ans_seg.segment_length = sizeof(*ans);
  ret = mx_irecv(ep,
		 &ans_seg, 1,
		 /* use the same ID to receive the answer */
		 ORFS_MX_ANS_RECV_MATCH(ID), ORFS_MX_ANS_RECV_MASK,
		 NULL, &ans_req);
  if ( ret != MX_SUCCESS ) {
    ERROR("failed to post answer receive (error %s)\n",
	  mx_strerror(ret));
    goto error;
  }
  to_cancel |= IRECV_ANS;

  /* send request */
  req_seg.segment_ptr = req;
  req_seg.segment_length = sizeof(*req);
  ret = mx_isend(ep,
		 &req_seg, 1,
		 server_addr,
		 ORFS_MX_REQ_MATCH(ID),
		 NULL, &req_req);
  if ( ret != MX_SUCCESS ) {
    ERROR("failed to post request send (error %s)\n",
	  mx_strerror(ret));
    goto error;
  }
  to_cancel |= ISEND_REQ;

  /* wait for completion of the request sending */
  to_cancel &= ~ISEND_REQ;
  ret = mx_wait(ep, &req_req, MX_INFINITE, &status, &result);
  if ( ret != MX_SUCCESS || result != 1 ) {
    ERROR("request send failed (return %s)\n", mx_strerror(ret));
    goto error;
  }

  /* wait for the answer */
  to_cancel &= ~IRECV_ANS;
  ret = mx_wait(ep, &ans_req, MX_INFINITE, &status, &result);
  if ( ret != MX_SUCCESS || result != 1 ) {
    ERROR("answer receive failed (return %s)\n", mx_strerror(ret));
    goto error;
  }

  /* wait for the answer buffer */
  if ( out_buf ) {
    if ( req->net.mx.rdma_ansbuf ) {
      /* RDMA has been done before the answer arrived */

      /* destroy rdma window */
      to_cancel &= ~RDMA_WIN_ANSBUF;
      ret = mx_destroy_rdma_window(ep, ansbuf_window_handle);
      if ( ret != MX_SUCCESS ) {
	ERROR("answer buffer RDMA window destroying failed (return %s)\n",
	      mx_strerror(ret));
	goto error;
      }
    } else {
      /* wait for the answer buffer */
      to_cancel &= ~IRECV_ANSBUF;
      ret = mx_wait(ep, &ansbuf_req, MX_INFINITE, &status, &result);
      if ( ret != MX_SUCCESS || result != 1 ) {
	ERROR("answer buffer receive failed (return %s)\n", mx_strerror(ret));
	goto error;
      }
      
    }
  }

  /* wait for the request buffer */
  if ( in_buf ) {
    if ( req->net.mx.rdma_reqbuf ) {
      /* RDMA has been done before the answer arrived */

      /* destroy rdma window */
      to_cancel &= ~RDMA_WIN_REQBUF;
      ret = mx_destroy_rdma_window(ep, reqbuf_window_handle);
      if ( ret != MX_SUCCESS ) {
	ERROR("request buffer RDMA window destroying failed (return %s)\n",
	      mx_strerror(ret));
	goto error;
      }
    } else {
      /* wait for the request buffer */
      to_cancel &= ~ISEND_REQBUF;
      ret = mx_wait(ep, &reqbuf_req, MX_INFINITE, &status, &result);
      if ( ret != MX_SUCCESS || result != 1 ) {
	ERROR("request send failed (return %s)\n", mx_strerror(ret));
	goto error;
      }
    }
  }

  return 0;

 error:
  if (to_cancel & RDMA_WIN_ANSBUF) {
    mx_destroy_rdma_window(ep, ansbuf_window_handle);
  } else if (to_cancel & IRECV_ANSBUF) {
    //    mx_cancel(ep, ansbuf_req);
    mx_wait(ep, &ansbuf_req, MX_INFINITE, &status, &result);
  }
  if (to_cancel & IRECV_ANS) {
    //  mx_cancel(ep, ans_req);
    mx_wait(ep, &ans_req, MX_INFINITE, &status, &result);
  }
  if (to_cancel & RDMA_WIN_REQBUF) {
    mx_destroy_rdma_window(ep, reqbuf_window_handle);
  } else if (to_cancel & ISEND_REQBUF) {
    //  mx_cancel(ep, reqbuf_req);
    mx_wait(ep, &reqbuf_req, MX_INFINITE, &status, &result);
  }
  if (to_cancel & ISEND_REQ) {
    //  mx_cancel(ep, req_req);
    mx_wait(ep, &req_req, MX_INFINITE, &status, &result);
  }
  return -EIO;
}


void MX_exit()
{
  mx_close_endpoint(ep);
  mx_finalize();
}
