/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "orfs_protocol.h"
#include "orfs_common_network.h"
#include "interactive_network.h"
#include "interactive_messages.h"


/*********************
 * Network functions *
 *********************/

static int (*network_init)(char *) = NULL;
static void (*network_exit)() = NULL;
static int (*network_request)(struct orfs_request *,
			      const void *,
			      struct orfs_answer *,
			      void *,
			      size_t) = NULL;


/*******************
 * Cmdline parsing *
 *******************/

static char *callname;
static char *server_name;


static
void exit_with_usage()
{
  printf("Usage: %s [options] [server]\n"
#if CONFIG_TCP
	 "  --tcp-port|-T N           set TCP server port\n"
#endif
#if CONFIG_GM
	 "  --gm-port|-G  N           set GM server port\n"
#endif
#if CONFIG_MX
	 "  --mx-port|-M  N           set MX server port\n"
#endif
#if CONFIG_GM || CONFIG_GM
	 "  --myri-board|-B N         set Myrinet server board (for GM/MX)\n"
	 "  --client-myri-board|-b N  set Myrinet client board (for GM/MX)\n"
#endif
 	 "  -v                        verbose\n"
	 , callname);
  exit(-1);
}


static inline
void parse_cmdline(int argc, char *const argv[])
{
  int verbose = 0;

  char c;
  static struct option long_opts[] = { 
#if CONFIG_TCP
    { "tcp", 0, NULL, 't' },
    { "tcp-port", 1, NULL, 'T' },
#endif
#if CONFIG_GM
    { "gm", 0, NULL, 'g' },
    { "gm-port", 1, NULL, 'G' },
#endif
#if CONFIG_MX
    { "mx", 0, NULL, 'm' },
    { "mx-port", 1, NULL, 'M' },
#endif
#if CONFIG_GM || CONFIG_MX
    { "myri-board", 1, NULL, 'B' },
    { "client-myri-board", 1, NULL, 'b' },
#endif
    { NULL, 0, NULL, 0 },
  };

  callname = strrchr(argv[0], '/')+1;

  /* Default network setup */
#if CONFIG_TCP
  TCP_SETUP();
#elif CONFIG_GM
  GM_SETUP();
#elif CONFIG_MX
  MX_SETUP();
#endif

  /* option parsing loop */
  while ( (c=getopt_long(argc, argv,
#if CONFIG_TCP
			 "tT:"
#endif
#if CONFIG_GM
			 "gG:"
#endif
#if CONFIG_MX
			 "mM:"
#endif
#if CONFIG_GM || CONFIG_MX
			 "B:b:"
#endif
			 "v", long_opts, NULL))
	  != EOF ) {
    switch ( c ) {

#if CONFIG_TCP
    case 't':
      TCP_SETUP();
      if ( verbose )
	fprintf(stderr, "Using TCP network layer\n");
      break;

    case 'T': {
      char *strtolres;
      unsigned long port = 0;

      port = strtol(optarg, &strtolres, 10);

      if ( !*optarg || *strtolres ) {
	fprintf(stderr, "Unrecognized TCP port %s\n", optarg);
	exit_with_usage();
      }

      if ( verbose )
	fprintf(stderr, "Using server TCP port %ld\n", port);

      tcp_server_port = port;
      break;
    }
#endif /* CONFIG_TCP */

#if CONFIG_GM
    case 'g':
      GM_SETUP();
      if ( verbose )
	fprintf(stderr, "Using GM network layer\n");
      break;

    case 'G': {
      char *strtolres;
      unsigned long port = 0;

      port = strtol(optarg, &strtolres, 10);

      if ( !*optarg || *strtolres ) {
	fprintf(stderr, "Unrecognized GM port %s\n", optarg);
	exit_with_usage();
      }

      if ( verbose )
	fprintf(stderr, "Using server GM port %ld\n", port);

      gm_server_port = port;
      break;
    }
#endif /* CONFIG_GM */

#if CONFIG_MX
    case 'm':
      MX_SETUP();
      if ( verbose )
	fprintf(stderr, "Using MX network layer\n");
      break;

    case 'M': {
      char *strtolres;
      unsigned long ep = 0;

      ep = strtol(optarg, &strtolres, 10);

      if ( !*optarg || *strtolres ) {
	fprintf(stderr, "Unrecognized MX endpoint %s\n", optarg);
	exit_with_usage();
      }

      if ( verbose )
	fprintf(stderr, "Using server MX endpoint %ld\n", ep);

      mx_server_ep = ep;
      break;
    }
#endif /* CONFIG_MX */

#if CONFIG_GM || CONFIG_MX
    case 'B': {
      char *strtolres;
      unsigned long board = 0;

      board = strtol(optarg, &strtolres, 10);

      if ( !*optarg || *strtolres ) {
	fprintf(stderr, "Unrecognized server Myrinet board %s\n", optarg);
	exit_with_usage();
      }

      if ( verbose )
	fprintf(stderr, "Using server Myrinet board %ld\n", board);

#if CONFIG_GM
      gm_server_board = board;
#endif
#if CONFIG_MX
      mx_server_board = board;
#endif
      break;
    }

    case 'b': {
      char *strtolres;
      unsigned long board = 0;

      board = strtol(optarg, &strtolres, 10);

      if ( !*optarg || *strtolres ) {
	fprintf(stderr, "Unrecognized client Myrinet board %s\n", optarg);
	exit_with_usage();
      }

      if ( verbose )
	fprintf(stderr, "Using client Myrinet board %ld\n", board);

#if CONFIG_GM
      gm_client_board = board;
#endif
#if CONFIG_MX
      mx_client_board = board;
#endif
      break;
    }
#endif /* CONFIG_GM || CONFIG_MX */

    case 'v':
      verbose = 1;
      break;

    case '?':
    default:
      exit_with_usage();
    }
  }

  /* Check network */
  if ( !network_init ) {
    fprintf(stderr, "No network layer available\n");
    exit(-1);
  }

  server_name = optind < argc ? argv[optind] : "localhost";
}


/******************
 * Various Macros *
 ******************/

#define OUT() printf(" ABORDED\n\n")


#define MAX_INT_SIZE 16

static
int get_d_arg(const char *arg_name, int *value)
{
  char string[MAX_INT_SIZE];
  printf(" %s = ", arg_name);
  memset(string, 0, MAX_INT_SIZE);
  get_line(string, MAX_INT_SIZE);
  if ( string[0] == '\0'  ||  string[0] == '\n' ||  sscanf(string, "%d", value) != 1 )
    return -1;
  return 0;
}


#define MAX_LONGLONG_SIZE 32

static
int get_lld_arg(const char *arg_name, long long *value)
{
  char string[MAX_LONGLONG_SIZE];
  printf(" %s = ", arg_name);
  memset(string, 0, MAX_LONGLONG_SIZE);
  get_line(string, MAX_LONGLONG_SIZE);
  if ( string[0] == '\0'  ||  string[0] == '\n'  ||  sscanf(string, "%lld", value) != 1 )
    return -1;
  return 0;
}


static
int get_s_arg(const char *arg_name, char *name, long long len)
{
  char *end;
  printf(" %s = ", arg_name);
  memset(name, 0, len+1);
  get_line(name, len+1);
  if ( name[0] == '\0'  ||  name[0] == '\n' )
    return -1;
  if ( (end=strchr(name, '\n')) != NULL )
    *end = '\0';
  return 0;
}


static
void get_buf_arg(const char *arg_name, char *buf, int size)
{
  printf(" %s = ", arg_name);
  get_line(buf, size);
}


static inline
char *str_orfs_error(orfs_result_t res)
{
  switch ( res ) {
  case ORFS_SUCCESS:
    return "Success";
  case ORFS_BAD_REQUEST:
    return "Bad Request";
  case ORFS_BAD_FILE:
    return "Bad File";
  case ORFS_INVALID_PARAMETER:
    return "Invalid Parameter";
  case ORFS_FUNCTION_NOT_IMPLEMENTED:
    return "Function Not Implemented";
  case ORFS_NO_FILE_LEFT:
    return "No File Left";
  case ORFS_NO_SPACE_LEFT:
    return "No Space Left";
  case ORFS_SERVER_INVALID_PARAMETER:
    return "Internal Server Error due to Invalid Parameter";
  case ORFS_SERVER_NOMEM:
    return "Internal Server Error due to Not Enough Memory";
  case ORFS_NOT_EMPTY_DIR:
    return "Not Empty Dir"; 
  case ORFS_EXISTING_DIR_ENTRY:
    return "Existing Dir Entry";
  case ORFS_NO_SUCH_DIR_ENTRY:
    return "No Such Dir Entry";
  case ORFS_BAD_INODE_TYPE:
    return "Bad Inode Type";
  case ORFS_NO_SUCH_INODE:
    return "No Such Inode";
  case ORFS_BAD_INODE_TYPE_2:
    return "Bad Inode Type (2nd)";
  case ORFS_NO_SUCH_INODE_2:
    return "No Such Inode (2nd)";
  case ORFS_BAD_INODE_TYPE_3:
    return "Bad Inode Type (3rd)";
  default:
    return "Error Code Not Recognized";
  }
}


static
void print_error(orfs_result_t res)
{
  printf("=> Result: %s\n\n", str_orfs_error(res));
}


/*************************
 * FS specific constants *
 *************************/

static orfs_path_len_t symlen;


/************
 * Requests *
 ************/

static
void lookup_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long ino;
  size_t len;
  
  printf("\nLOOKUP:\n");
  if ( get_lld_arg("directory ino", &ino) )
    goto out;
  if ( get_s_arg("dentry name", req.target.dir.name, ORFS_NAME_LEN) )
    goto out;
  len = strlen(req.target.dir.name);

  req.type = ORFS_DIR_REQ;
  req.subtype = ORFS_LOOKUP_DIR_REQ;
  req.target.dir.dir.ino = (orfs_ino_t)ino;
  req.target.dir.name_len = (orfs_name_len_t)len;
  network_request(&req, NULL, &ans, NULL, 0);

  if ( ans.result )
    print_error(ans.result);
  else
    printf("=> Received Ino %lld\n\n",
	   (long long)ans.target.dir.data.lookup.node_attr.id.ino);
  return;

 out:
  OUT();
  return;
}


static
void getattr_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long ino;
  
  printf("\nGETATTR:\n");
  if ( get_lld_arg("node ino", &ino) )
    goto out;

  req.type = ORFS_INODE_REQ;
  req.subtype = ORFS_GETATTR_INODE_REQ;
  req.target.inode.inode.ino = (orfs_ino_t)ino;
  network_request(&req, NULL, &ans, NULL, 0);

  if ( ans.result )
    print_error(ans.result);
  else
    printf("=> Inode attr:\n"
	   "    ino   %lld\n"
	   "    mode  %d\n"
	   "    uid   %d\n"
	   "    gid   %d\n"
	   "    nlink %d\n"
	   "    size  %lld\n"
	   "    blk   %lld\n"
	   "    atime %lld\n"
	   "    mtime %lld\n"
	   "    ctime %lld\n\n",
	   (long long)ans.target.inode.node_attr.id.ino,
	   (int)ans.target.inode.node_attr.mode,
	   (int)ans.target.inode.node_attr.uid,
	   (int)ans.target.inode.node_attr.gid,
	   (int)ans.target.inode.node_attr.nlink,
	   (long long)ans.target.inode.node_attr.size,
	   (long long)ans.target.inode.node_attr.blocks,
	   (long long)ans.target.inode.node_attr.atime,
	   (long long)ans.target.inode.node_attr.mtime,
	   (long long)ans.target.inode.node_attr.ctime);
  return;
  
 out:
  OUT();
  return;
}


static
void setattr_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long ino;
  long long llarg;
  int darg;

  printf("\nSETATTR:\n");
  if ( get_lld_arg("node ino", &ino) )
    goto out;

  req.target.inode.data.setattr.new_attr.valid = 0;
  if ( !get_d_arg("mode", &darg) ) {
    req.target.inode.data.setattr.new_attr.mode = (orfs_mode_t)darg;
    req.target.inode.data.setattr.new_attr.valid |= ORFS_ATTR_MODE;
  }
  if ( !get_d_arg("uid", &darg) ) {
    req.target.inode.data.setattr.new_attr.uid = (orfs_uid_t)darg;
    req.target.inode.data.setattr.new_attr.valid |= ORFS_ATTR_UID;
  }
  if ( !get_d_arg("gid", &darg) ) {
    req.target.inode.data.setattr.new_attr.gid = (orfs_gid_t)darg;
    req.target.inode.data.setattr.new_attr.valid |= ORFS_ATTR_GID;
  }
  if ( !get_lld_arg("size", &llarg) ) {
    req.target.inode.data.setattr.new_attr.size = (orfs_off_t)llarg;
    req.target.inode.data.setattr.new_attr.valid |= ORFS_ATTR_SIZE;
  }
  if ( !get_lld_arg("atime", &llarg) ) {
    req.target.inode.data.setattr.new_attr.atime = (orfs_time_t)llarg;
    req.target.inode.data.setattr.new_attr.valid |= ORFS_ATTR_ATIME|ORFS_ATTR_ATIME_SET;
  }
  if ( !get_lld_arg("mtime", &llarg) ) {
    req.target.inode.data.setattr.new_attr.mtime = (orfs_time_t)llarg;
    req.target.inode.data.setattr.new_attr.valid |= ORFS_ATTR_MTIME|ORFS_ATTR_MTIME_SET;
  }

  req.type = ORFS_INODE_REQ;
  req.subtype = ORFS_SETATTR_INODE_REQ;
  req.target.inode.inode.ino = (orfs_ino_t)ino;
  network_request(&req, NULL, &ans, NULL, 0);

  print_error(ans.result);
  return;

 out:
  OUT();
  return;
}


static
void mkreg_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long ino;
  size_t len;
  
  printf("\nMKREG:\n");
  if ( get_lld_arg("directory ino", &ino) )
    goto out;
  if ( get_s_arg("dentry name", req.target.dir.name, ORFS_NAME_LEN+1) )
    goto out;
  len = strlen(req.target.dir.name);

  req.type = ORFS_DIR_REQ;
  req.subtype = ORFS_MKREG_DIR_REQ;
  req.target.dir.dir.ino = (orfs_ino_t)ino;
  req.target.dir.name_len = (orfs_name_len_t)len;
  req.target.dir.data.create.mode = ORFS_IRWXUGO | ORFS_IFREG;
  req.target.dir.data.create.uid = 0;
  req.target.dir.data.create.gid = 0;
  network_request(&req, NULL, &ans, NULL, 0);

  if ( ans.result )
    print_error(ans.result);
  else
    printf("=> Received Ino %lld\n\n",
	   (long long)ans.target.dir.data.create.new_node_attr.id.ino);
  return;

 out:
  OUT();
  return;
}


static
void mkdir_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long ino;
  size_t len;
  
  printf("\nMKDIR:\n");
  if ( get_lld_arg("directory ino", &ino) )
    goto out;
  if ( get_s_arg("dentry name", req.target.dir.name, ORFS_NAME_LEN+1) )
    goto out;
  len = strlen(req.target.dir.name);

  req.type = ORFS_DIR_REQ;
  req.subtype = ORFS_MKDIR_DIR_REQ;
  req.target.dir.dir.ino = (orfs_ino_t)ino;
  req.target.dir.name_len = (orfs_name_len_t)len;
  req.target.dir.data.create.mode = ORFS_IRWXUGO | ORFS_IFDIR;
  req.target.dir.data.create.uid = 0;
  req.target.dir.data.create.gid = 0;
  network_request(&req, NULL, &ans, NULL, 0);

  if ( ans.result )
    print_error(ans.result);
  else
    printf("=> Received Ino %lld\n\n",
	   (long long)ans.target.dir.data.create.new_node_attr.id.ino);
  return;

 out:
  OUT();
  return;
}


static
void mknod_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long ino;
  size_t len;
  long long type;
  long long dev;

  printf("\nMKNOD:\n");
  if ( get_lld_arg("directory ino", &ino) )
    goto out;
  if ( get_s_arg("dentry name", req.target.dir.name, ORFS_NAME_LEN+1) )
    goto out;
  if ( get_lld_arg("chr/blk/fifo/sock (0/1/2/3)", &type) 
       ||  type < 0  ||  type > 3 )
    goto out;
  if ( type <= 1 )
    if ( get_lld_arg("dev", &dev) )
      goto out;
  len = strlen(req.target.dir.name);

  req.type = ORFS_DIR_REQ;
  req.subtype = ORFS_MKNOD_DIR_REQ;
  req.target.dir.dir.ino = (orfs_ino_t)ino;
  req.target.dir.name_len = (orfs_name_len_t)len;

  req.target.dir.data.create.mode = ORFS_IRWXUGO;
  switch ( type ) {
  case 0: req.target.dir.data.create.mode |= ORFS_IFCHR; break;
  case 1: req.target.dir.data.create.mode |= ORFS_IFBLK; break;
  case 2: req.target.dir.data.create.mode |= ORFS_IFIFO; break;
  case 3: req.target.dir.data.create.mode |= ORFS_IFSOCK; break;
  default: goto out;
  }

  req.target.dir.data.create.uid = 0;
  req.target.dir.data.create.gid = 0;
  req.target.dir.data.create.dev = dev;
  network_request(&req, NULL, &ans, NULL, 0);

  if ( ans.result )
    print_error(ans.result);
  else
    printf("=> Received Ino %lld\n\n",
	   (long long)ans.target.dir.data.create.new_node_attr.id.ino);
  return;

 out:
  OUT();
  return;
}


void link_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long old_dir_ino, new_dir_ino;
  size_t old_len, new_len;
  
  printf("\nLINK:\n");
  if ( get_lld_arg("old directory ino", &old_dir_ino) )
    goto out;
  if ( get_s_arg("old dentry name", req.target.twodir.old_name, ORFS_NAME_LEN+1) )
    goto out;
  old_len = strlen(req.target.twodir.old_name);
  if ( get_lld_arg("new directory ino", &new_dir_ino) )
    goto out;
  if ( get_s_arg("new dentry name", req.target.twodir.new_name, ORFS_NAME_LEN+1) )
    goto out;
  new_len = strlen(req.target.twodir.new_name);

  req.type = ORFS_2DIR_REQ;
  req.subtype = ORFS_LINK_2DIR_REQ;
  req.target.twodir.old_dir.ino = (orfs_ino_t)old_dir_ino;
  req.target.twodir.old_name_len = (orfs_name_len_t)old_len;
  req.target.twodir.new_dir.ino = (orfs_ino_t)new_dir_ino;
  req.target.twodir.new_name_len = (orfs_name_len_t)new_len;
  network_request(&req, NULL, &ans, NULL, 0);

  print_error(ans.result);
  return;

 out:
  OUT();
  return;
}


void symlink_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long dir_ino;
  size_t len, path_len;
  char *buffer = malloc(symlen+1);

  printf("\nSYMLINK:\n");
  if ( get_lld_arg("directory ino", &dir_ino) )
    goto out;
  if ( get_s_arg("dentry name", req.target.dir.name, ORFS_NAME_LEN+1) )
    goto out;
  len = strlen(req.target.dir.name);
  if ( get_s_arg("symlink path", buffer, symlen+1) )
    goto out;
  path_len = strlen(buffer);

  req.type = ORFS_DIR_REQ;
  req.subtype = ORFS_SYMLINK_DIR_REQ;
  req.target.dir.dir.ino = (orfs_ino_t)dir_ino;
  req.target.dir.name_len = (orfs_name_len_t)len;
  req.target.dir.data.symlink.path_len = (orfs_path_len_t)path_len;
  req.target.dir.data.symlink.uid = 0;
  req.target.dir.data.symlink.gid = 0;
  network_request(&req, buffer, &ans, NULL, 0);

  print_error(ans.result);
  free(buffer);
  return;

 out:
  OUT();
  return;
}


static
void move_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long old_dir_ino, new_dir_ino;
  size_t old_len, new_len;
  
  printf("\nMOVE:\n");
  if ( get_lld_arg("old directory ino", &old_dir_ino) )
    goto out;
  if ( get_s_arg("old dentry name", req.target.twodir.old_name, ORFS_NAME_LEN+1) )
    goto out;
  old_len = strlen(req.target.twodir.old_name);
  if ( get_lld_arg("new directory ino", &new_dir_ino) )
    goto out;
  if ( get_s_arg("new dentry name", req.target.twodir.new_name, ORFS_NAME_LEN+1) )
    goto out;
  new_len = strlen(req.target.twodir.new_name);

  req.type = ORFS_2DIR_REQ;
  req.subtype = ORFS_MOVE_2DIR_REQ;
  req.target.twodir.old_dir.ino = (orfs_ino_t)old_dir_ino;
  req.target.twodir.old_name_len = (orfs_name_len_t)old_len;
  req.target.twodir.new_dir.ino = (orfs_ino_t)new_dir_ino;
  req.target.twodir.new_name_len = (orfs_name_len_t)new_len;
  network_request(&req, NULL, &ans, NULL, 0);

  print_error(ans.result);
  return;

 out:
  OUT();
  return;
}


static
void unlink_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long ino;
  size_t len;
  
  printf("\nUNLINK:\n");
  if ( get_lld_arg("directory ino", &ino) )
    goto out;
  if ( get_s_arg("dentry name", req.target.dir.name, ORFS_NAME_LEN+1) )
    goto out;
  len = strlen(req.target.dir.name);

  req.type = ORFS_DIR_REQ;
  req.subtype = ORFS_UNLINK_DIR_REQ;
  req.target.dir.dir.ino = (orfs_ino_t)ino;
  req.target.dir.name_len = (orfs_name_len_t)len;
  network_request(&req, NULL, &ans, NULL, 0);

  print_error(ans.result);
  return;

 out:
  OUT();
  return;
}


static
void rmdir_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long ino;
  size_t len;
  
  printf("\nRMDIR:\n");
  if ( get_lld_arg("directory ino", &ino) )
    goto out;
  if ( get_s_arg("dentry name", req.target.dir.name, ORFS_NAME_LEN+1) )
    goto out;
  len = strlen(req.target.dir.name);

  req.type = ORFS_DIR_REQ;
  req.subtype = ORFS_RMDIR_DIR_REQ;
  req.target.dir.dir.ino = (orfs_ino_t)ino;
  req.target.dir.name_len = (orfs_name_len_t)len;
  network_request(&req, NULL, &ans, NULL, 0);

  print_error(ans.result);
  return;

 out:
  OUT();
  return;
}


static
void readlink_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long ino;
  char *buffer;

  printf("\nREADLINK:\n");
  if ( get_lld_arg("node ino", &ino) )
    goto out;

  buffer = malloc(symlen+1);

  req.type = ORFS_INODE_REQ;
  req.subtype = ORFS_READLINK_INODE_REQ;
  req.target.inode.inode.ino = (orfs_ino_t)ino;
  network_request(&req, NULL, &ans, buffer, symlen+1);

  if ( ans.result )
    print_error(ans.result);
  else {
    buffer[ans.target.inode.data.readlink.count] = '\0';
    printf("=> Readlink: %lld\n%s\n",
	   (long long)ans.target.inode.data.readlink.count,
	   buffer);
  }
  free(buffer);
  return;
  
 out:
  OUT();
  return;
}


static
void open_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long ino;
  
  printf("\nOPEN:\n");
  if ( get_lld_arg("node ino", &ino) )
    goto out;

  req.type = ORFS_INODE_REQ;
  req.subtype = ORFS_OPEN_INODE_REQ;
  req.target.inode.inode.ino = (orfs_ino_t)ino;
  network_request(&req, NULL, &ans, NULL, 0);

  if ( ans.result )
    print_error(ans.result);
  else
    printf("=> File %lld\n\n",
	   (long long)ans.target.inode.data.open.file.id);
  return;
  
 out:
  OUT();
  return;
}


static
void close_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long file;
  
  printf("\nCLOSE:\n");
  if ( get_lld_arg("file id", &file) )
    goto out;

  req.type = ORFS_FILE_REQ;
  req.subtype = ORFS_CLOSE_FILE_REQ;
  req.target.file.file.id = (orfs_file_id_t)file;
  network_request(&req, NULL, &ans, NULL, 0);

  print_error(ans.result);
  return;
  
 out:
  OUT();
  return;
}


static
void sync_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long file;
  long long datasync;

  printf("\nSYNC:\n");
  if ( get_lld_arg("file id", &file) )
    goto out;
  if ( get_lld_arg("datasync", &datasync) )
    goto out;

  req.type = ORFS_FILE_REQ;
  req.subtype = ORFS_SYNC_FILE_REQ;
  req.target.file.file.id = (orfs_file_id_t)file;
  req.target.file.data.fsync.datasync = datasync;
  network_request(&req, NULL, &ans, NULL, 0);

  print_error(ans.result);
  return;
  
 out:
  OUT();
  return;
}


static
void pread_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long file;
  long long offset;
  long long count;
  char *buffer;

  printf("\nPREAD:\n");
  if ( get_lld_arg("file id", &file) )
    goto out;
  if ( get_lld_arg("offset", &offset) )
    goto out;
  if ( get_lld_arg("count", &count) )
    goto out;

  buffer = malloc(count+1);

  req.type = ORFS_FILE_REQ;
  req.subtype = ORFS_PREAD_FILE_REQ;
  req.target.file.file.id = (orfs_file_id_t)file;
  req.target.file.data.data_access.offset = (orfs_off_t)offset;
  req.target.file.data.data_access.count = (orfs_off_t)count;
  network_request(&req, NULL, &ans, buffer, count);

  if ( ans.result )
    print_error(ans.result);
  else {
    buffer[ans.target.file.data.data_access.count] = '\0';
    printf("=> Read: %lld/%lld\n%s\n",
	   (long long)ans.target.file.data.data_access.count,
	   (long long)count,
	   buffer);
  }
  free(buffer);
  return;
  
 out:
  OUT();
  return;
}


static
void pwrite_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long file;
  long long offset;
  long long count;
  char *buffer;

  printf("\nPWRITE:\n");
  if ( get_lld_arg("file id", &file) )
    goto out;
  if ( get_lld_arg("offset", &offset) )
    goto out;
  if ( get_lld_arg("count", &count) )
    goto out;

  buffer = malloc(count+1);
  buffer[count] = '\0';
  get_buf_arg("data", buffer, count+1);

  req.type = ORFS_FILE_REQ;
  req.subtype = ORFS_PWRITE_FILE_REQ;
  req.target.file.file.id = (orfs_file_id_t)file;
  req.target.file.data.data_access.offset = (orfs_off_t)offset;
  req.target.file.data.data_access.count = (orfs_off_t)count;
  network_request(&req, buffer, &ans, NULL, 0);

  print_error(ans.result);
  free(buffer);
  return;
  
 out:
  OUT();
  return;
}


static
void preaddir_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long file;
  long long offset;
  long long data_max;
  char *buffer;

  printf("\nPREADDIR:\n");
  if ( get_lld_arg("file id", &file) )
    goto out;
  if ( get_lld_arg("offset", &offset) )
    goto out;
  if ( get_lld_arg("count", &data_max) )
    goto out;

  buffer = malloc(data_max);

  req.type = ORFS_FILE_REQ;
  req.subtype = ORFS_PREADDIR_FILE_REQ;
  req.target.file.file.id = (orfs_file_id_t)file;
  req.target.file.data.preaddir.offset = (orfs_doff_t)offset;
  req.target.file.data.preaddir.data_max = (orfs_data_len_t)data_max;
  network_request(&req, NULL, &ans, buffer, data_max);

  if ( ans.result )
    print_error(ans.result);
  else {
    char *d_ptr;
    long long count = (long long)ans.target.file.data.preaddir.count;
    long long i;

    printf("=> Received %lld dirents\n", count);
    d_ptr = buffer;
    for(i=0; i<count; i++) {
      struct orfs_dirent *d = (struct orfs_dirent *) d_ptr;
      printf("  => Dirent %lld is:\n"
	     "      ino      %lld\n"
	     "      type     %d\n"
	     "      offset   %lld\n"
	     "      name_len %d\n"
	     "      name     %s\n",
	     i,
	     (long long)d->attr.id.ino,
	     (int)d->type,
	     (long long)d->off,
	     (int)d->name_len,
	     d->name);
      d_ptr += sizeof(*d) + d->name_len+1;
    }
    if ( ans.target.file.data.preaddir.eof_reached )
      printf("  => EOF reached\n");
    printf("\n");
  }

  free(buffer);
  return;
  
 out:
  OUT();
  return;
}


static
void readpage_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long ino;
  long long offset;
  long long count;
  char *buffer;

  printf("\nREADPAGE:\n");
  if ( get_lld_arg("node ino", &ino) )
    goto out;
  if ( get_lld_arg("offset", &offset) )
    goto out;
  if ( get_lld_arg("count", &count) )
    goto out;

  buffer = malloc(count+1);

  req.type = ORFS_PAGE_REQ;
  req.subtype = ORFS_READ_FULL_PAGE_REQ;
  req.target.page.inode.ino = (orfs_ino_t)ino;
  req.target.page.data.data_access.offset = (orfs_off_t)offset;
  req.target.page.data.data_access.count = (orfs_off_t)count;
  network_request(&req, NULL, &ans, buffer, count);

  if ( ans.result )
    print_error(ans.result);
  else {
    buffer[ans.target.page.data.data_access.count] = '\0';
    printf("=> Read: %lld/%lld\n%s\n",
	   (long long)ans.target.page.data.data_access.count,
	   (long long)count,
	   buffer);
  }
  free(buffer);
  return;
  
 out:
  OUT();
  return;
}


static
void writepage_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long ino;
  long long offset;
  long long count;
  char *buffer;

  printf("\nWRITEPAGE:\n");
  if ( get_lld_arg("node ino", &ino) )
    goto out;
  if ( get_lld_arg("offset", &offset) )
    goto out;
  if ( get_lld_arg("count", &count) )
    goto out;

  buffer = malloc(count+1);
  buffer[count] = '\0';
  get_buf_arg("data", buffer, count+1);

  req.type = ORFS_PAGE_REQ;
  req.subtype = ORFS_WRITE_FULL_PAGE_REQ;
  req.target.page.inode.ino = (orfs_ino_t)ino;
  req.target.page.data.data_access.offset = (orfs_off_t)offset;
  req.target.page.data.data_access.count = (orfs_off_t)count;
  network_request(&req, buffer, &ans, NULL, 0);

  print_error(ans.result);
  free(buffer);
  return;
  
 out:
  OUT();
  return;
}


static
void prepare_write_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long ino;
  long long offset;
  long long count;

  printf("\nPREPARE_WRITE:\n");
  if ( get_lld_arg("node ino", &ino) )
    goto out;
  if ( get_lld_arg("offset", &offset) )
    goto out;
  if ( get_lld_arg("count", &count) )
    goto out;

  req.type = ORFS_PAGE_REQ;
  req.subtype = ORFS_PREPARE_WRITE_PAGE_REQ;
  req.target.page.inode.ino = (orfs_ino_t)ino;
  req.target.page.data.data_access.offset = (orfs_off_t)offset;
  req.target.page.data.data_access.count = (orfs_off_t)count;
  network_request(&req, NULL, &ans, NULL, 0);

  print_error(ans.result);
  return;
  
 out:
  OUT();
  return;
}


static
void commit_write_req()
{
  struct orfs_request req;
  struct orfs_answer ans;
  long long ino;
  long long offset;
  long long count;
  char *buffer;

  printf("\nCOMMIT_WRITE:\n");
  if ( get_lld_arg("node ino", &ino) )
    goto out;
  if ( get_lld_arg("offset", &offset) )
    goto out;
  if ( get_lld_arg("count", &count) )
    goto out;

  buffer = malloc(count+1);
  buffer[count] = '\0';
  get_buf_arg("data", buffer, count+1);

  req.type = ORFS_PAGE_REQ;
  req.subtype = ORFS_COMMIT_WRITE_PAGE_REQ;
  req.target.page.inode.ino = (orfs_ino_t)ino;
  req.target.page.data.data_access.offset = (orfs_off_t)offset;
  req.target.page.data.data_access.count = (orfs_off_t)count;
  network_request(&req, buffer, &ans, NULL, 0);

  print_error(ans.result);
  free(buffer);
  return;
  
 out:
  OUT();
  return;
}


static
void statfs_req()
{
  struct orfs_request req;
  struct orfs_answer ans;

  printf("\nSTATFS:\n");

  req.type = ORFS_FS_REQ;
  req.subtype = ORFS_STATFS_FS_REQ;
  network_request(&req, NULL, &ans, NULL, 0);

  if ( ans.result )
    print_error(ans.result);
  else {
    printf("=> %lld total blocks\n"
	   "   %lld free blocks\n"
	   "   %lld available to non-user blocks\n"
	   "   %lld total nodes\n"
	   "   %lld free nodes\n\n",
	   (long long)ans.target.fs.data.statfs.fsattr.blocks,
	   (long long)ans.target.fs.data.statfs.fsattr.bfree,
	   (long long)ans.target.fs.data.statfs.fsattr.bavail,
	   (long long)ans.target.fs.data.statfs.fsattr.files,
	   (long long)ans.target.fs.data.statfs.fsattr.ffree);
  }
  return;
}


void ping_req()
{
  struct orfs_request req;
  struct orfs_answer ans;

  printf("\nPING:\n");

  req.type = ORFS_FS_REQ;
  req.subtype = ORFS_PING_FS_REQ;
  network_request(&req, NULL, &ans, NULL, 0);

  print_error(ans.result);
  return;
}


void init_req()
{
  struct orfs_request req;
  struct orfs_answer ans;

  printf("\nINIT:\n");

  req.type = ORFS_FS_REQ;
  req.subtype = ORFS_INIT_FS_REQ;
  network_request(&req, NULL, &ans, NULL, 0);

  if ( ans.result )
    print_error(ans.result);
  else {
    printf("=> magic        %lld\n"
	   "   protocol     %lld\n"
	   "   blksize      %lld\n"
	   "   symlen       %lld\n"
	   "   capabilities %lld\n",
	   (long long)ans.target.fs.data.init.data.magic,
	   (long long)ans.target.fs.data.init.data.protocol,
	   (long long)ans.target.fs.data.init.data.blksize,
	   (long long)ans.target.fs.data.init.data.symlen,
	   (long long)ans.target.fs.data.init.data.capabilities);
    symlen = ans.target.fs.data.init.data.symlen;
  }
  return;
}


void fini_req()
{
  struct orfs_request req;
  struct orfs_answer ans;

  printf("\nFINI:\n");

  req.type = ORFS_FS_REQ;
  req.subtype = ORFS_FINI_FS_REQ;
  network_request(&req, NULL, &ans, NULL, 0);

  print_error(ans.result);
  return;
}


static
void out()
{
  fini_req();
  printf("\nExiting...\n");
  network_exit();
  exit(0);
}


/********
 * Main *
 ********/

struct request_type {
  void (*func)();
  int choice;
  char *name;
};


static struct request_type types[] = {
  { lookup_req, 1, "lookup" },
  { getattr_req, 2, "getattr" },
  { setattr_req, 3, "setattr" },
  { mkreg_req, 4, "mkreg" },
  { mkdir_req, 5, "mkdir" },
  { mknod_req, 6, "mknod" },
  { link_req, 7, "link" },
  { symlink_req, 8, "symlink" },
  { move_req, 9, "move" },
  { unlink_req, 10, "unlink" },
  { rmdir_req, 11, "rmdir" },
  { readlink_req, 12, "readlink" },
  { open_req, 13, "open" },
  { close_req, 14, "close" },
  { sync_req, 15, "sync" },
  { pread_req, 16, "pread" },
  { pwrite_req, 17, "pwrite" },
  { preaddir_req, 18, "preaddir" },
  { readpage_req, 19, "readpage" },
  { writepage_req, 20, "writepage" },
  { prepare_write_req, 21, "prepare_write" },
  { commit_write_req, 22, "commit_write" },
  { statfs_req, 23, "statfs" },
  { ping_req, 24, "ping" },
  { out, 0, "exit" },
};


int main(int argc, char *const argv[])
{
  int types_nr;

  parse_cmdline(argc, argv);

  for(types_nr=0; types[types_nr].func!=out; types_nr++);
  types_nr++;

  if ( network_init(server_name) < 0 )
    return -1;

  init_req();

  for(;;) {
    int choice = 0;
    int i;
    printf("NEW REQUEST:\n");
    for(i=0; i<types_nr; i++)
      printf("    %d -> %s\n", types[i].choice, types[i].name);
    while ( get_d_arg("Request type", &choice) );
    for(i=0; i<types_nr; i++)
      if ( choice == types[i].choice ) {
	types[i].func();
	continue;
      }
  }
}
