/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/net.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <asm/uaccess.h>

#ifdef CONFIG_ORFS_GMKRC
#include "gmkrc.h"
#define gm_open gmkrc_open
#define gm_close gmkrc_close
#define gm_send_with_callback gmkrc_send_with_callback
#define gm_provide_receive_buffer_with_tag gmkrc_provide_receive_buffer_with_tag
#define gm_register_memory gmkrc_register_memory
#define gm_deregister_memory gmkrc_deregister_memory
#define gm_get_remote_ptr gmkrc_get_remote_ptr
#else				/* CONFIG_ORFS_GMKRC */
#warning Not using GMKRC may lead to O_DIRECT transfer confusion
#define gm_get_remote_ptr(ptr) ((gm_remote_ptr_t)(unsigned long)(ptr))
#endif

#include "gm.h.h"
#include "gm_internal.h.h"

#include "orfs_fs.h"
#include "orfs_common_network.h"
#include "orfs_request.h"
#include "orfs_mount.h"
#include "orfs_messages.h"
#include "orfs_misc.h"

/*****************
 * GM connection *
 *****************/

struct orfs_gm_connection {
	/* connection port and target */
	struct gm_port *port;
	u16 server_node;
	u8 server_port;
	/* provided answer list */
	struct orfs_answer *ans1, *ans2;
	/* ansbuf params */
	struct orfs_gm_abp *abp_array;
	struct semaphore abp_sem;
	spinlock_t abp_lock;
	unsigned int abp_max;
	unsigned int last_abp;
	/* spinlock protecting concurrent gm_calls */
	spinlock_t lock;
};

#define CONNECTION(sbi) ((struct orfs_gm_connection*)(sbi)->connection)
#define SET_CONNECTION(sbi, connection) ((sbi)->connection = connection)

#define DECLARE_GMLOCK(sbi) spinlock_t *gm_lock = &CONNECTION(sbi)->lock
#define GM_LOCK() spin_lock(gm_lock)
#define GM_UNLOCK() spin_unlock(gm_lock)

/****************************
 * Answer Buffer Parameters *
 ****************************/

struct orfs_gm_abp {
	gm_u8_t size;
	unsigned char free;
	struct orfs_client_request *creq;
};

#define ABP(creq) ((struct orfs_gm_abp*)(creq)->net_pd)
#define SET_ABP(creq, abp) ((creq)->net_pd = abp)

#define ABP2TAG(connection, abp) ((abp) - (connection)->abp_array)
#define TAG2ABP(connection, tag) (&(connection)->abp_array[tag])

static inline struct orfs_gm_abp*
orfs_gm_get_abp(struct orfs_sb_info *sbi,
		struct orfs_client_request *creq)
{
	struct orfs_gm_connection *connection = CONNECTION(sbi);
	struct orfs_gm_abp *abp;
	int i;

	down(&connection->abp_sem);

	spin_lock(&connection->abp_lock);
	for(i=connection->last_abp; ; i++) {
		if ( i == connection->abp_max )
			i = 0;
		abp = TAG2ABP(connection, i);
		if ( abp->free ) {
			abp->free = 0;
			connection->last_abp = i+1;
			break;
		}
	}
	spin_unlock(&connection->abp_lock);

	SET_ABP(creq, abp);
	abp->creq = creq;

	ORFS_DEBUG(GM, "Returning abp %d (size %d)\n",
		   ABP2TAG(connection, abp), abp->size);

	return abp;
}

static inline void
orfs_gm_put_abp(struct orfs_sb_info *sbi,
		struct orfs_client_request *creq)
{
	struct orfs_gm_connection *connection = CONNECTION(sbi);
	struct orfs_gm_abp *abp = ABP(creq);
	int i = ABP2TAG(connection, abp);

	ORFS_DEBUG(GM, "Releasing abp %d\n", i);

	/* no need to lock here */
	abp->free = 1;
	connection->last_abp = i;

	up(&connection->abp_sem);
}

static inline int 
orfs_gm_abp_init(struct orfs_sb_info *sbi)
{
	struct orfs_gm_connection *connection = CONNECTION(sbi);
	struct orfs_gm_abp *abp;
	int i;

	connection->abp_max = 16;
	sema_init(&connection->abp_sem, connection->abp_max);
	spin_lock_init(&connection->abp_lock);
	connection->last_abp = 0;

	abp = kmalloc(connection->abp_max*sizeof(struct orfs_gm_abp),
		      GFP_KERNEL);
	if ( !abp )
		return -ENOMEM;
	connection->abp_array = abp;

	for(i=0; i<connection->abp_max; i++) {
		abp->free = 1;
		abp->size = 31-i;
		abp++;
	}

	return 0;
}

static inline void
orfs_gm_abp_exit(struct orfs_sb_info *sbi)
{
	struct orfs_gm_connection *connection = CONNECTION(sbi);
	kfree(connection->abp_array);
}

/*************
 * Callbacks *
 *************/

static void
orfs_gm_request_send_callback(struct gm_port *port,
			     void *context, gm_status_t status)
{
	struct orfs_sb_info *sbi = context;

	if (status != GM_SUCCESS) {
		gm_resume_sending(CONNECTION(sbi)->port,
				  ORFS_GM_REQ_PRIO,
				  CONNECTION(sbi)->server_node,
				  CONNECTION(sbi)->server_port,
				  NULL, NULL);
		if (status == GM_SEND_TIMED_OUT) {
			ORFS_ERROR("Request send timeout, try again\n");
		} else {
			ORFS_ERROR("Request send got erroneous status %s\n",
				   gm_strerror(status));
			orfs_set_connection_error(sbi);
		}
	}
}

static void
orfs_gm_buffer_send_callback(struct gm_port *port,
			     void *context, gm_status_t status)
{
	struct orfs_sb_info *sbi = context;

	if (status != GM_SUCCESS) {
		gm_resume_sending(CONNECTION(sbi)->port,
				  ORFS_GM_REQBUF_PRIO,
				  CONNECTION(sbi)->server_node,
				  CONNECTION(sbi)->server_port,
				  NULL, NULL);
		if (status == GM_SEND_TIMED_OUT) {
			ORFS_ERROR("Buffer send timeout, try again\n");
		} else {
			ORFS_ERROR("Buffer send got erroneous status %s\n",
				   gm_strerror(status));
			orfs_set_connection_error(sbi);
		}
	}
}

/**********************
 * Internal Functions *
 **********************/

static inline void
orfs_gm_provide(struct orfs_sb_info *sbi,
		struct orfs_client_request *creq)
{
	struct orfs_gm_connection *connection = CONNECTION(sbi);
	struct gm_port *port = connection->port;
	char *buffer = creq->out_buf;
	off_t off = creq->out_page_off;
	orfs_buffer_flags_t flags = creq->out_buf_flags;

	if (buffer) {
		gm_u8_t size = ABP(creq)->size;
		gm_u8_t tag = ABP2TAG(connection, ABP(creq));
		ORFS_CREQ2REQ(creq)->net.gm.rdma_ansbuf = 0;
		ORFS_CREQ2REQ(creq)->net.gm.ansbuf.rdv_size = size;
		switch (flags) {
		case ORFS_BUF_USER_ANY:
			gm_provide_receive_buffer_with_tag(port, buffer,
							   size,
							   ORFS_GM_ANSBUF_PRIO,
							   tag);
			break;
		case ORFS_BUF_KMALLOC:
			gm_provide_physical_receive_buffer_with_tag(port,
								    virt_to_phys(buffer),
								    size,
								    ORFS_GM_ANSBUF_PRIO,
								    tag);
			break;
		case ORFS_BUF_PAGE:
			gm_provide_physical_receive_buffer_with_tag(port,
								    page_to_phys((struct page *)buffer) + off,
								    size,
								    ORFS_GM_ANSBUF_PRIO,
								    tag);
			break;
		default:
			ORFS_ALERT("Out buffer flags incorrect %d\n", flags);
			BUG();
			break;
		}
		ORFS_CREQ_NEED_BUF(creq);
	}
	ORFS_CREQ_NEED_REQ(creq);
}

static inline void
orfs_gm_send(struct orfs_sb_info *sbi,
	     struct orfs_client_request *creq)
{
	struct orfs_gm_connection *connection = CONNECTION(sbi);
	struct gm_port *port = connection->port;
	u16 server_node = connection->server_node;
	u8 server_port = connection->server_port;
	char *buffer = creq->in_buf;
	size_t len = creq->in_buf_len;
	off_t off = creq->in_page_off;
	orfs_buffer_flags_t flags = creq->in_buf_flags;

	gm_send_physical_with_callback(port,
				       virt_to_phys(ORFS_CREQ2REQ(creq)),
				       ORFS_GM_REQ_SIZE, sizeof (*ORFS_CREQ2REQ(creq)),
				       ORFS_GM_REQ_PRIO,
				       server_node, server_port,
				       orfs_gm_request_send_callback, sbi);
	if (buffer) {
		ORFS_CREQ2REQ(creq)->net.gm.rdma_reqbuf = 0;
		switch (flags) {
		case ORFS_BUF_USER_ANY:
			gm_send_with_callback(port, buffer,
					      ORFS_GM_REQBUF_SIZE(len), len,
					      ORFS_GM_REQBUF_PRIO,
					      server_node, server_port,
					      orfs_gm_buffer_send_callback,
					      sbi);
			break;
		case ORFS_BUF_KMALLOC:
			gm_send_physical_with_callback(port,
						       virt_to_phys(buffer),
						       ORFS_GM_REQBUF_SIZE(len), len,
						       ORFS_GM_REQBUF_PRIO,
						       server_node, server_port,
						       orfs_gm_buffer_send_callback,
						       sbi);
			break;
		case ORFS_BUF_PAGE:
			gm_send_physical_with_callback(port,
						       page_to_phys((struct page
								     *) buffer)
						       + off,
						       ORFS_GM_REQBUF_SIZE(len), len,
						       ORFS_GM_REQBUF_PRIO,
						       server_node, server_port,
						       orfs_gm_buffer_send_callback,
						       sbi);
			break;
		default:
			ORFS_ALERT("Out buffer flags incorrect %d\n", flags);
			BUG();
			break;
		}
	}
}

static inline int
orfs_gm_register(struct orfs_sb_info *sbi, struct orfs_client_request *creq)
{
	struct orfs_gm_connection *connection = CONNECTION(sbi);
	struct gm_port *port = connection->port;
	gm_status_t status = GM_SUCCESS;

	void *in_buf = creq->in_buf;
	size_t in_buf_len = creq->in_buf_len;
	orfs_buffer_flags_t in_buf_flags = creq->in_buf_flags;

	void *out_buf = creq->out_buf;
	size_t out_buf_len = creq->out_buf_len;
	orfs_buffer_flags_t out_buf_flags = creq->out_buf_flags;

	/* register everything needed */
	if (in_buf && in_buf_flags == ORFS_BUF_USER_ANY) {
#ifndef CONFIG_ORFS_GMKRC
		ORFS_NOTICE_ONCE(should_use_GMKRC4, "Not using GMKRC with O_DIRECT transfer may lead to\n");
		ORFS_NOTICE_ONCE(should_use_GMKRC5, "confusion between address spaces of simultaneous clients,\n");
		ORFS_NOTICE_ONCE(should_use_GMKRC6, "resulting in data read at the wrong place.\n");
#endif
		status = gm_register_memory(port, in_buf, in_buf_len);
		if ( status != GM_SUCCESS )
			goto error;
	}

	if (out_buf && out_buf_flags == ORFS_BUF_USER_ANY) {
#ifndef CONFIG_ORFS_GMKRC
		ORFS_NOTICE_ONCE(should_use_GMKRC4, "Not using GMKRC with O_DIRECT transfer may lead to\n");
		ORFS_NOTICE_ONCE(should_use_GMKRC5, "confusion between address spaces of simultaneous clients,\n");
		ORFS_NOTICE_ONCE(should_use_GMKRC6, "resulting in data written at the wrong place.\n");
#endif
		status = gm_register_memory(port, out_buf, out_buf_len);
		if ( status != GM_SUCCESS )
			goto error_with_in_buf;
	}

	return GM_SUCCESS;

      error_with_in_buf:
	if (in_buf && in_buf_flags == ORFS_BUF_USER_ANY)
		gm_deregister_memory(port, in_buf, in_buf_len);
      error:
	return status;
}

static inline void
orfs_gm_deregister(struct orfs_sb_info *sbi, struct orfs_client_request *creq)
{
	struct orfs_gm_connection *connection = CONNECTION(sbi);
	struct gm_port *port = connection->port;

	void *in_buf = creq->in_buf;
	size_t in_buf_len = creq->in_buf_len;
	orfs_buffer_flags_t in_buf_flags = creq->in_buf_flags;

	void *out_buf = creq->out_buf;
	size_t out_buf_len = creq->out_buf_len;
	orfs_buffer_flags_t out_buf_flags = creq->out_buf_flags;

	/* unregister everything needed */
	if (in_buf && in_buf_flags == ORFS_BUF_USER_ANY)
		gm_deregister_memory(port, in_buf, in_buf_len);
	if (out_buf && out_buf_flags == ORFS_BUF_USER_ANY)
		gm_deregister_memory(port, out_buf, out_buf_len);
}

/* only used for mono remote call */
static inline int
orfs_gm_recv(struct orfs_sb_info *sbi, struct orfs_client_request *creq)
{
	struct orfs_gm_connection *connection = CONNECTION(sbi);
	struct gm_port *port = connection->port;
	int error = 0;
	ORFS_DECLARE_SIGMASK();
	ORFS_INIT_SIGMASK();
	ORFS_START_SIGMASK();

	while (IS_ORFS_CREQ_WAITING(creq)) {
		gm_recv_event_t *ev;
		ev = gm_blocking_receive_no_spin(port);
		switch (gm_ntoh_u8(ev->recv.type)) {
		case ORFS_GM_ANS_RECV_EV:
			ORFS_CREQ_RECV_REQ(creq);
			break;
		case ORFS_GM_ANSBUF_RECV_EV:
			ORFS_CREQ_RECV_BUF(creq);
			break;
		default:
			gm_unknown(port, ev);
			if (sbi->connection_status != ORFS_CONNECTION_OK) {
				error = -EIO;
				goto out;
			}
		}
	}

      out:
	ORFS_STOP_SIGMASK();
	return error;
}

/* This two functions are used to save and restore
 * the first byte of a buffer.
 * This allows to consume the posted receive with
 * a fake one byte message when nothing should be
 * because of a erroneous request.
 * This is only defined for user buffers.
 */

static inline char
orfs_gm_save_buffer(struct orfs_client_request *creq)
{
	char saved = 0;
	char *buf = creq->out_buf;
	if (buf && creq->out_buf_flags == ORFS_BUF_USER_ANY)
		get_user(saved, (char *) buf);
	return saved;
}

static inline void
orfs_gm_restore_buffer(struct orfs_client_request *creq, char saved)
{
	char *buf = creq->out_buf;
	if (buf && creq->out_buf_flags == ORFS_BUF_USER_ANY)
		put_user(saved, (char *) buf);
}

/***************************
 * Remote Calls Processing *
 ***************************/

static int
orfs_gm_mono_remote_call(struct orfs_sb_info *sbi, struct orfs_client_request *creq)
{
	struct orfs_answer *ans;
	char saved;
	int error = 0;

	if (sbi->connection_status != ORFS_CONNECTION_OK) {
		ORFS_DEBUG(GM, "Process refused\n");
		error = -EIO;
		goto disconnected;
	}
	/* The connection status may change after if this point if another
	 * thread encounters an error or if umount -f is called.
	 * This buggy thread will loop in orfs_wakeup_all_requests, waking up
	 * everybody and waiting for everybody to leave this routine.
	 */

	ans = orfs_get_client_answer(sbi);
	if ( !ans )
		return -ENOMEM;
	ORFS_CREQ2ANS(creq) = ans;

	saved = orfs_gm_save_buffer(creq);

	orfs_gm_get_abp(sbi, creq);

	if ( orfs_gm_register(sbi, creq) != GM_SUCCESS ) {
		error = -EIO;
		goto register_failed;
	}	

	ORFS_DEBUG(GM, "Providing Receive Buffers\n");
	gm_provide_physical_receive_buffer(CONNECTION(sbi)->port, virt_to_phys(ans),
					   ORFS_GM_ANS_SIZE, ORFS_GM_ANS_PRIO);
	orfs_gm_provide(sbi, creq);

	ORFS_DEBUG(GM, "Sending request\n");
	orfs_gm_send(sbi, creq);

	ORFS_DEBUG(GM, "Receiving answer\n");
	if ((error = orfs_gm_recv(sbi, creq)) < 0) {
		error = -EIO;
		goto disconnected_with_reg;
	}

      disconnected_with_reg:
	orfs_gm_deregister(sbi, creq);

	ans = ORFS_CREQ2ANS(creq);
	if (ans && ans->result != ORFS_SUCCESS)
		orfs_gm_restore_buffer(creq, saved);
      disconnected:
      register_failed:
	orfs_gm_put_abp(sbi, creq);
	return error;
}

static int
orfs_gm_multi_remote_call(struct orfs_sb_info *sbi, struct orfs_client_request *creq)
{
	struct orfs_answer *ans;
	char saved;
	int error;
	DECLARE_GMLOCK(sbi);

	if (sbi->connection_status != ORFS_CONNECTION_OK) {
		ORFS_DEBUG(GM, "Process refused\n");
		error = -EIO;
		goto disconnected;
	}
	/* The connection status may change after if this point if another
	 * thread encounters an error or if umount -f is called.
	 * This buggy thread will loop in orfs_wakeup_all_requests, waking up
	 * everybody and waiting for everybody to leave this routine.
	 */

	saved = orfs_gm_save_buffer(creq);

        orfs_gm_get_abp(sbi, creq);

	if ( orfs_gm_register(sbi, creq) != GM_SUCCESS ) {
		error = -EIO;
		goto register_failed;
	}	

	ORFS_DEBUG(GM, "Providing Receive Buffers\n");
	GM_LOCK();
	orfs_gm_provide(sbi, creq);
	GM_UNLOCK();

	ORFS_DEBUG(GM, "Sending request\n");
	GM_LOCK();
	orfs_gm_send(sbi, creq);
	GM_UNLOCK();

	ORFS_DEBUG(GM, "Sleeping until answer is received...\n");
	error = wait_event_interruptible(creq->queue,
					 /* wait until answer and buffer have been received */
					 !IS_ORFS_CREQ_WAITING(creq)
					 /* or client disconnected */
					 || (sbi->connection_status != ORFS_CONNECTION_OK));
	if (error < 0) {
		ORFS_DEBUG(GM, "Process woken up by a signal\n");
		ORFS_CREQ_IS_ORPHAN(creq);
		goto interrupted;
	}
	if (sbi->connection_status != ORFS_CONNECTION_OK) {
		ORFS_DEBUG(GM, "Process woken up by ORFSD network error\n");
		error = -EIO;
		goto disconnected_with_reg;
	}

	ORFS_DEBUG(GM, "Process woken up by ORFSD\n");
	/* we can now assume that ORFS_CREQ2ANS(creq) and its eventual
	 * buffer have been filled */

	orfs_gm_deregister(sbi, creq);

	orfs_gm_put_abp(sbi, creq);

	return 0;

      disconnected_with_reg:
      interrupted:
	orfs_gm_deregister(sbi, creq);

	ans = ORFS_CREQ2ANS(creq);
	if (ans && ans->result != ORFS_SUCCESS)
		orfs_gm_restore_buffer(creq, saved);
      disconnected:
      register_failed:
	orfs_gm_put_abp(sbi, creq);
	return error;
}

/*********************
 * ORFSD Receive Try *
 *********************/

static void
orfs_gm_update_request(struct orfs_sb_info *sbi,
		       struct orfs_client_request *creq)
{
	/* is this request done ? */
	if (!IS_ORFS_CREQ_WAITING(creq)) {
		if (!IS_ORFS_CREQ_ADOPTED(creq)) {
			/* wake up process */
			wake_up_interruptible(&creq->queue);
			ORFS_DEBUG(ORFSD, "ORFSD wakes up process (request received)\n");
		} else {
			/* ORFSD release this request */
			orfs_put_client_request(sbi, creq);
			ORFS_DEBUG(ORFSD, "ORFSD releases orphan request\n");
		}
	}
}

static void
orfs_gm_kt_loop(struct orfs_sb_info *sbi)
{
	struct orfs_gm_connection *connection = CONNECTION(sbi);
	struct gm_port *port = connection->port;
	struct orfs_answer *ans;
	DECLARE_GMLOCK(sbi);

	/* GMLOCK is initialized during orfs_gm_connect */

	/* Preprovide two answer.
	 * These are not preprovided in main init routine because
	 * the init request uses mono_remote_call and thus a specific
	 * orfs_answer.
	 */
	ans = orfs_get_client_answer(sbi);
	BUG_ON(!ans); // hum, bad idea ? add a kt_init ?
	GM_LOCK();
	gm_provide_physical_receive_buffer(port,
					   virt_to_phys(ans),
					   ORFS_GM_ANS_SIZE,
					   ORFS_GM_ANS_PRIO);
	GM_UNLOCK();
	connection->ans1 = ans;
	
	ans = orfs_get_client_answer(sbi);
	BUG_ON(!ans);
	GM_LOCK();
	gm_provide_physical_receive_buffer(port,
					   virt_to_phys(ans),
					   ORFS_GM_ANS_SIZE,
					   ORFS_GM_ANS_PRIO);
	GM_UNLOCK();
	connection->ans2 = ans;

	/* main loop */
	for (;;) {
		gm_recv_event_t *ev;
		struct orfs_client_request *creq = NULL;

		GM_LOCK();
		ev = gm_blocking_receive_no_spin(port);
		GM_UNLOCK();

		switch (gm_ntoh_u8(ev->recv.type)) {

		case ORFS_GM_ANS_RECV_EV:
			/* get ans */
#if 1
			ans = phys_to_virt((unsigned long) gm_ntohp(ev->recv.buffer));
			if (ans == connection->ans1)
				/* move the 2nd in first place */
				connection->ans1 = connection->ans2;
			else if (ans != connection->ans2)
				BUG();
#else
			if ( virt_to_phys(connection->ans1) == (unsigned long) gm_ntohp(ev->recv.buffer) ) {
				ans = connection->ans1;
				/* move the 2nd in first place */
				connection->ans1 = connection->ans2;
			}
			else if ( virt_to_phys(connection->ans2) == (unsigned long) gm_ntohp(ev->recv.buffer) ) {
				ans = connection->ans2;
				/* leave the first in first place */
			}
			else
				BUG();
#endif

			/* update creq */
			creq = (void*)(unsigned long)ans->tag;
			creq->ans = ans;
			ORFS_CREQ_RECV_REQ(creq);
			orfs_gm_update_request(sbi, creq);

			/* provide another one */
			ans = orfs_get_client_answer(sbi);
			BUG_ON(!ans);
			GM_LOCK();
			gm_provide_physical_receive_buffer(port,
							   virt_to_phys(ans),
							   ORFS_GM_ANS_SIZE,
							   ORFS_GM_ANS_PRIO);
			GM_UNLOCK();
			connection->ans2 = ans;
			break;

		case ORFS_GM_ANSBUF_RECV_EV: {
			gm_u8_t gm_tag = gm_ntoh_u8(ev->recv.tag);
			struct orfs_gm_abp *abp = TAG2ABP(connection, gm_tag);

			/* update creq */
			creq = abp->creq;
			ORFS_CREQ_RECV_BUF(creq);
			orfs_gm_update_request(sbi, creq);
		}

		case _GM_SLEEP_EVENT:
			/* no lock here !! */
			gm_unknown(port, ev);
			/* no unlock here !! */
			break;

		default:
			GM_LOCK();
			gm_unknown(port, ev);
			GM_UNLOCK();
			if (sbi->connection_status != ORFS_CONNECTION_OK)
				return;
		}

		if (sbi->param.multi.kt_finish)
			/* finish required */
			return;
	}
}

/**************
 * Connection *
 **************/

static int
orfs_gm_connect(struct orfs_sb_info *sbi,
		struct orfs_user_mount_options *options)
{
	struct orfs_gm_connection *connection;
	struct gm_port *port = NULL;
	int port_id = ORFS_GM_CLIENT_FIRST_PORT_ID;
	int len = strlen(options->common.name);
	unsigned int server_node;
	u8 server_port;
	unsigned client_board;
	unsigned server_board;
	gm_status_t status;
	int error;

	/* fix server port */
	server_port = options->specific.mask & ORFS_SERVER_PORT_MASK
	  ? options->gm.port : ORFS_GM_SERVER_PORT;

	/* fix server board */
	server_board = options->specific.mask & ORFS_SERVER_BOARD_MASK
	  ? options->gm.board : ORFS_GM_SERVER_BOARD;

	/* fix client board */
	client_board = options->specific.mask & ORFS_CLIENT_BOARD_MASK
	  ? options->gm.client_board : ORFS_GM_CLIENT_BOARD;

	ORFS_DEBUG(NET, "Trying to connect to server %s board %d through GM port %ld (from our board %d)\n",
		   options->common.name, server_board, (unsigned long) server_port, client_board);

	/* Init GM */
	error = -EIO;
	status = gm_init();
	if (status != GM_SUCCESS) {
		ORFS_ERROR("gm_init failed (%s)\n", gm_strerror(status));
		goto error;
	}

	/* Find a free port and open it */
	error = -EIO;
	while (port == NULL) {
		status = gm_open(&port,
				 client_board, port_id,
				 "kercli", GM_API_VERSION);
		if (status == GM_SUCCESS)
			break;
		ORFS_ERROR("GM cannot open port %d (error %s)\n",
			   port_id, gm_strerror(status));

		if (status != GM_BUSY) {
			ORFS_ERROR("Failed to open a GM port\n");
			goto error_with_init;
		}

		port = NULL;
		port_id++;
	}

	/* Set enable_nack_down to get GM_SEND_TIMED_OUT immedialty */
	gm_set_enable_nack_down(port, 1);

	/* Get Server Node and Port */
	error = -EINVAL;
	if ( !strcmp(options->common.name, "localhost") ) {
		status = gm_get_node_id(port, &server_node);
		if ( status != GM_SUCCESS ) {
		ORFS_ERROR("Cannot find GM node (error %s)\n", 
			   gm_strerror(status));
		goto error_with_port;
		}
	} else {
		char name_with_board[GM_MAX_HOST_NAME_LEN+1];
		char *fullname = options->common.name;
		int count;
		if ( server_board ) {
			count = snprintf(name_with_board, GM_MAX_HOST_NAME_LEN+1,
					 "%s:%d", options->common.name, server_board);
			if ( count < 0 || count > GM_MAX_HOST_NAME_LEN ) {
				ORFS_ERROR("GM name too long\n");
				goto error_with_port;
			}
			fullname = name_with_board;
		}
		server_node = gm_host_name_to_node_id(port, fullname);
		if ( server_node == GM_NO_SUCH_NODE_ID ) {
			ORFS_ERROR("Cannot find GM node %s\n", fullname);
			goto error_with_port;
		}
	}

	/* Setup connection */
	error = -ENOMEM;
	connection = kmalloc(sizeof (*connection), GFP_KERNEL);
	if (!connection)
		goto error_with_port;
	connection->port = port;
	connection->server_port = server_port;
	connection->server_node = server_node;

	connection->ans1 = NULL;
	connection->ans2 = NULL;

	/* end connection initializing */
	spin_lock_init(&connection->lock);	/* needed before init_buffers for gm_dma_malloc */
	SET_CONNECTION(sbi, connection);

	/* init ansbuf */
	error = orfs_gm_abp_init(sbi);
	if (error < 0)
		goto error_with_connection;

	/* Setup path */
	error = -ENOMEM;
	sbi->server_path = kmalloc(len + 4 + 10 + 1 + 10 + 1, GFP_KERNEL);
	if (!sbi->server_path)
		goto error_with_abp;
	sprintf(sbi->server_path, "%s:GM:%ld:%ld",
		options->common.name, (unsigned long) server_board, (unsigned long) server_port);

	return 0;

      error_with_abp:
	orfs_gm_abp_exit(sbi);
      error_with_connection:
	kfree(connection);
      error_with_port:
	gm_close(port);
      error_with_init:
	gm_finalize();
      error:
	return error;
}

static void
orfs_gm_disconnect(struct orfs_sb_info *sbi)
{
	struct orfs_gm_connection *connection = CONNECTION(sbi);
	struct gm_port *port = connection->port;

	/* remove preprovided answers */
	if ( connection->ans2 )
		orfs_put_client_answer(sbi, connection->ans2);
	if ( connection->ans1 )
		orfs_put_client_answer(sbi, connection->ans1);

	/* free other stuff */
	kfree(sbi->server_path);
	orfs_gm_abp_exit(sbi);
	kfree(connection);
	gm_close(port);
	gm_finalize();
}

/***********
 * Wake up *
 ***********/

static void
orfs_gm_wake_up(struct orfs_sb_info *sbi)
{
	gm_arch_wake(&CONNECTION(sbi)->port->kernel_port_state->sleep_sync);
}

/**********************
 * Network Operations *
 **********************/

static struct orfs_net_operations orfs_gm_nops = {
	.connect = orfs_gm_connect,
	.disconnect = orfs_gm_disconnect,

	.multi_remote_call = orfs_gm_multi_remote_call,
	.mono_remote_call = orfs_gm_mono_remote_call,

	.kt_loop = orfs_gm_kt_loop,

	.wake_up_waiters = orfs_gm_wake_up,
};

/*****************
 * Network Layer *
 *****************/

static DECLARE_ORFS_NET_LAYER(orfs_gm_layer, "gm", &orfs_gm_nops);

int
orfs_gm_layer_init(void)
{
	return orfs_register_net_layer(&orfs_gm_layer);
}

void
orfs_gm_layer_exit(void)
{
	orfs_unregister_net_layer(&orfs_gm_layer);
}

#ifdef CONFIG_ORFS_NETMODULES
MODULE_AUTHOR("Brice Goglin <Brice.Goglin@ens-lyon.fr>");
MODULE_LICENSE("GPL");
module_init(orfs_gm_layer_init);
module_exit(orfs_gm_layer_exit);
#endif				/* CONFIG_ORFS_NETMODULES */
