/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/net.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#if ORFS_LINUX_26
#include <linux/aio.h>
#endif
#include <asm/uaccess.h>
#include <asm/io.h>

#include "myriexpress.h"

#include "orfs_fs.h"
#include "orfs_common_network.h"
#include "orfs_request.h"
#include "orfs_mount.h"
#include "orfs_messages.h"
#include "orfs_misc.h"

/*****************
 * MX connection *
 *****************/

struct orfs_mx_connection {
	mx_endpoint_t ep;
	mx_endpoint_addr_t server_addr;
};

#define CONNECTION(sbi) ((struct orfs_mx_connection*)(sbi)->connection)
#define SET_CONNECTION(sbi, connection) ((sbi)->connection = connection)

/***************************
 * Remote Calls Processing *
 ***************************/

struct orfs_mx_request {
	mx_request_t req_req;
	mx_request_t ans_req;
	mx_request_t reqbuf_req;
	mx_request_t ansbuf_req;
	//uint32_t reqbuf_window_handle;
	//uint32_t ansbuf_window_handle;
	unsigned long to_cancel;
};

#define ISEND_REQ 0x1
#define ISEND_REQBUF 0x2
#define RDMA_WIN_REQBUF 0x4
#define IRECV_ANS 0x10
#define IRECV_ANSBUF 0x20
#define RDMA_WIN_ANSBUF 0x40

static inline int
orfs_mx_post_requests(struct orfs_client_request *creq,
		      struct orfs_mx_connection *connection,
		      struct orfs_mx_request *mxreq,
		      uint32_t id)
{
	mx_endpoint_t ep = connection->ep;
	mx_endpoint_addr_t addr = connection->server_addr;
	struct orfs_request *req = ORFS_CREQ2REQ(creq);
	struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
	void *in_buf = creq->in_buf;
        size_t in_buf_len = creq->in_buf_len;
	off_t in_page_off = creq->in_page_off;
	orfs_buffer_flags_t in_buf_flags = creq->in_buf_flags;
	char *out_buf = creq->out_buf;
        size_t out_buf_len = creq->out_buf_len;
	off_t out_page_off = creq->out_page_off;
	orfs_buffer_flags_t out_buf_flags = creq->out_buf_flags;
	mx_segment_t req_seg;
	mx_segment_t ans_seg;
	mx_segment_t reqbuf_seg;
	mx_segment_t ansbuf_seg;
	mx_return_t ret;

	if ( in_buf ) {
	        unsigned int pin_type = 0;
		switch ( in_buf_flags ) {
		case ORFS_BUF_USER_ANY:
			reqbuf_seg.segment_ptr = MX_VA_TO_SEGMENT_PTR(in_buf);
			pin_type = MX_PIN_USER;
			ORFS_DEBUG(MX, "Using MX_PIN_USER to send request buffer\n");
			break;
		case ORFS_BUF_KMALLOC:
			reqbuf_seg.segment_ptr = MX_VA_TO_SEGMENT_PTR(in_buf);
			pin_type = MX_PIN_KERNEL;
			ORFS_DEBUG(MX, "Using MX_PIN_KERNEL to send request buffer\n");
			break;
		case ORFS_BUF_PAGE:
			reqbuf_seg.segment_ptr = MX_PA_TO_SEGMENT_PTR(page_to_phys((struct page *)in_buf)+in_page_off);
			pin_type = MX_PIN_PHYSICAL;
			ORFS_DEBUG(MX, "Using MX_PIN_PHYSICAL to send request buffer\n");
			break;
		default:
			ORFS_ALERT("Out buffer flags incorrect %d\n", in_buf_flags);
			BUG();
			break;
		}
		reqbuf_seg.segment_length = in_buf_len;

		// no rdma for now
		if (0) {
			req->net.gm.rdma_reqbuf = 1;
			// create rdma window
			// set rdma params
			mxreq->to_cancel |= RDMA_WIN_REQBUF;
		} else {
			req->net.gm.rdma_reqbuf = 0;
			/* send the request buffer */
			if ( in_buf_len > ORFS_MX_PREPROVIDED_REQBUF_LEN ) {
				ret = mx_isend(ep,
					       &reqbuf_seg, 1, pin_type,
					       addr,
					       ORFS_MX_LARGE_REQBUF_MATCH(id),
					       creq, &mxreq->reqbuf_req);
				ORFS_DEBUG(MX, "Posted send for large request buffer\n");
			} else {
				ret = mx_isend(ep,
					       &reqbuf_seg, 1, pin_type,
					       addr,
					       ORFS_MX_PREPROVIDED_REQBUF_MATCH(id),
					       creq, &mxreq->reqbuf_req);
				ORFS_DEBUG(MX, "Posted send for preprovided request buffer\n");
			}
			if ( ret != MX_SUCCESS ) {
				ORFS_ERROR("failed to post request buffer send (error %s)\n",
					   mx_strerror(ret));
				return -EIO;
			}
			mxreq->to_cancel |= ISEND_REQBUF;
		}
	}

	if ( out_buf ) {
	        unsigned int pin_type = 0;
		switch ( out_buf_flags ) {
		case ORFS_BUF_USER_ANY:
			ansbuf_seg.segment_ptr = MX_VA_TO_SEGMENT_PTR(out_buf);
			pin_type = MX_PIN_USER;
			ORFS_DEBUG(MX, "Using MX_PIN_USER to receive answer buffer\n");
			break;
		case ORFS_BUF_KMALLOC:
			ansbuf_seg.segment_ptr = MX_VA_TO_SEGMENT_PTR(out_buf);
			pin_type = MX_PIN_KERNEL;
			ORFS_DEBUG(MX, "Using MX_PIN_KERNEL to receive answer buffer\n");
			break;
		case ORFS_BUF_PAGE:
			ansbuf_seg.segment_ptr = MX_PA_TO_SEGMENT_PTR(page_to_phys((struct page *)out_buf)+out_page_off);
			pin_type = MX_PIN_PHYSICAL;
			ORFS_DEBUG(MX, "Using MX_PIN_PHYSICAL to receive answer buffer\n");
			break;
		default:
			ORFS_ALERT("Out buffer flags incorrect %d\n", out_buf_flags);
			BUG();
			break;
		}
		ansbuf_seg.segment_length = out_buf_len;

		// no rdma for now
		if (0) {
			req->net.gm.rdma_ansbuf = 1;
			// create rdma window
			// set rdma params
			mxreq->to_cancel |= RDMA_WIN_ANSBUF;
		} else {
			req->net.gm.rdma_ansbuf = 0;
			/* post a recv for the answer buffer */
			ret = mx_irecv(ep,
				       &ansbuf_seg, 1, pin_type,
				       ORFS_MX_ANSBUF_RECV_MATCH(id), ORFS_MX_ANSBUF_RECV_MASK,
				       creq, &mxreq->ansbuf_req);
			ORFS_DEBUG(MX, "Posted receive for answer buffer\n");
			if ( ret != MX_SUCCESS ) {
				ORFS_ERROR("failed to post answer buffer receive (error %s)\n",
					   mx_strerror(ret));
				return -EIO;
			}
			mxreq->to_cancel |= IRECV_ANSBUF;
		}
	}

	/* prepare answer */
	ans_seg.segment_ptr = MX_VA_TO_SEGMENT_PTR(ans);
	ans_seg.segment_length = sizeof(*ans);
	ret = mx_irecv(ep,
		       &ans_seg, 1, MX_PIN_KERNEL,
		       /* use the same ID to receive the answer */
		       ORFS_MX_ANS_RECV_MATCH(id), ORFS_MX_ANS_RECV_MASK,
		       creq, &mxreq->ans_req);
	ORFS_DEBUG(MX, "Posted receive for answer\n");
	if ( ret != MX_SUCCESS ) {
		ORFS_ERROR("failed to post answer receive (error %s)\n",
			   mx_strerror(ret));
		return -EIO;
	}
	mxreq->to_cancel |= IRECV_ANS;
	
	/* send request */
	req_seg.segment_ptr = MX_VA_TO_SEGMENT_PTR(req);
	req_seg.segment_length = sizeof(*req);
	ret = mx_isend(connection->ep,
		       &req_seg, 1, MX_PIN_KERNEL,
		       addr,
		       ORFS_MX_REQ_MATCH(id),
		       creq, &mxreq->req_req);
	ORFS_DEBUG(MX, "Posted send for request\n");
	if ( ret != MX_SUCCESS ) {
		ORFS_ERROR("failed to post request send (error %s)\n",
			   mx_strerror(ret));
		return -EIO;
	}
	mxreq->to_cancel |= ISEND_REQ;
	
	return 0;
}

static inline int
orfs_mx_wait_requests(struct orfs_client_request *creq,
		      struct orfs_mx_connection *connection,
		      struct orfs_mx_request *mxreq)
{
	mx_endpoint_t ep = connection->ep;
	struct orfs_request *req = ORFS_CREQ2REQ(creq);
	void *in_buf = creq->in_buf;
	char *out_buf = creq->out_buf;
	mx_status_t status;
	uint32_t result;
	mx_return_t ret;

	/* wait for completion of the request sending */
	mxreq->to_cancel &= ~ISEND_REQ;
	ORFS_DEBUG(MX, "Waiting for request send completion...\n");
	ret = mx_wait(ep, &mxreq->req_req, MX_INFINITE, &status, &result);
	if ( ret == MX_SUCCESS ) {
		if ( result == 1 ) {
			ORFS_DEBUG(MX, "Request send completed\n");
		} else {
			ORFS_ERROR("Request send wait woken up\n");
			return -EIO;
		}
	} else {
		ORFS_ERROR("Request send failed (return %s)\n", mx_strerror(ret));
		return -EIO;
	}
	
	/* wait for the answer */
	mxreq->to_cancel &= ~IRECV_ANS;
	ORFS_DEBUG(MX, "Waiting for answer receive completion...\n");
	ret = mx_wait(ep, &mxreq->ans_req, MX_INFINITE, &status, &result);
	if ( ret == MX_SUCCESS ) {
		if ( result == 1 ) {
			ORFS_DEBUG(MX, "Answer receive completed\n");
		} else {
			ORFS_ERROR("Answer receive wait woken up\n");
			return -EIO;
		}
	} else {
		ORFS_ERROR("Answer receive failed (return %s)\n", mx_strerror(ret));
		return -EIO;
	}

	/* wait for the answer buffer */
	if ( out_buf ) {
		if ( req->net.mx.rdma_ansbuf ) {
			/* RDMA has been done before the answer arrived */
			mxreq->to_cancel &= ~RDMA_WIN_ANSBUF;
			// destroy rdma window
		} else {
			/* wait for the answer buffer */
			mxreq->to_cancel &= ~IRECV_ANSBUF;
			ORFS_DEBUG(MX, "Waiting for answer buffer receive completion...\n");
			ret = mx_wait(ep, &mxreq->ansbuf_req, MX_INFINITE, &status, &result);
			if ( ret == MX_SUCCESS ) {
				if ( result == 1 ) {
					ORFS_DEBUG(MX, "Answer buffer receive completed\n");
				} else {
					ORFS_ERROR("Answer buffer receive wait woken up\n");
					return -EIO;
				}
			} else {
				ORFS_ERROR("Answer buffer receive failed (return %s)\n",
					   mx_strerror(ret));
				return -EIO;
			}
		}
	}

	/* wait for the request buffer */
	if ( in_buf ) {
		if ( req->net.mx.rdma_reqbuf ) {
			/* RDMA has been done before the answer arrived */
			mxreq->to_cancel &= ~RDMA_WIN_REQBUF;
			// destroy rdma window
		} else {
			/* wait for the request buffer */
			mxreq->to_cancel &= ~ISEND_REQBUF;
			ORFS_DEBUG(MX, "Waiting for request buffer send completion...\n");
			ret = mx_wait(ep, &mxreq->reqbuf_req, MX_INFINITE, &status, &result);
			if ( ret == MX_SUCCESS ) {
				if ( result == 1 ) {
					ORFS_DEBUG(MX, "Request buffer send completed\n");
				} else {
					ORFS_ERROR("Request buffer send wait woken up\n");
					return -EIO;
				}
			} else {
				ORFS_ERROR("Request buffer send failed (return %s)\n",
					   mx_strerror(ret));
				return -EIO;
			}
		}
	}

	return 0;
}

static inline void
orfs_mx_cancel_requests(struct orfs_mx_connection *connection,
			struct orfs_mx_request *mxreq)
{
	mx_endpoint_t ep = connection->ep;
	mx_status_t status;
	uint32_t result;
	if (mxreq->to_cancel & RDMA_WIN_ANSBUF) {
		// destroy rdma window
	} else if (mxreq->to_cancel & IRECV_ANSBUF) {
//		mx_cancel(ep, mxreq->ansbuf_req);
		mx_wait(ep, &mxreq->ansbuf_req, MX_INFINITE, &status, &result);
	}
	if (mxreq->to_cancel & IRECV_ANS) {
//		mx_cancel(ep, mxreq->ans_req);
		mx_wait(ep, &mxreq->ans_req, MX_INFINITE, &status, &result);
	}
	if (mxreq->to_cancel & RDMA_WIN_REQBUF) {
		// destroy rdma window
	} else if (mxreq->to_cancel & ISEND_REQBUF) {
//		mx_cancel(ep, mxreq->reqbuf_req);
		mx_wait(ep, &mxreq->reqbuf_req, MX_INFINITE, &status, &result);
	}
	if (mxreq->to_cancel & ISEND_REQ) {
//		mx_cancel(ep, mxreq->req_req);
		mx_wait(ep, &mxreq->req_req, MX_INFINITE, &status, &result);
	}
}

static int
orfs_mx_generic_remote_call(struct orfs_sb_info *sbi,
			    struct orfs_client_request *creq,
			    uint32_t id)
{
	struct orfs_mx_connection *connection = CONNECTION(sbi);
	struct orfs_request *req = ORFS_CREQ2REQ(creq);
	struct orfs_answer *ans;
	struct orfs_mx_request mxreq = { NULL, NULL, NULL, NULL, /* NULL, NULL, */ 0 };
	int error;

	if (sbi->connection_status != ORFS_CONNECTION_OK) {
		ORFS_DEBUG(MX, "Process refused\n");
		return -EIO;
	}
	/* The connection status may change after if this point if another
	 * thread encounters an error or if umount -f is called.
	 * This buggy thread will loop in orfs_wakeup_all_requests, waking up
	 * everybody and waiting for everybody to leave this routine.
	 */

	ans = orfs_get_client_answer(sbi);
	if ( !ans )
		return -ENOMEM;
	ORFS_CREQ2ANS(creq) = ans;

	/* no network specific private data */
	creq->net_pd = NULL;

	/* initializes tag to 0 */
	req->tag = 0;

	/* use the same ID to receive the answer */
	error = orfs_mx_post_requests(creq, connection, &mxreq, id);
	if (error < 0)
	  goto error;

	error = orfs_mx_wait_requests(creq, connection, &mxreq);
	if (error < 0)
	  goto error;

	return 0;

      error:
	orfs_mx_cancel_requests(connection, &mxreq);
	orfs_set_connection_error(sbi);
	return error;
}

#define MONO_ID 0x21071980

static int
orfs_mx_mono_remote_call(struct orfs_sb_info *sbi,
			 struct orfs_client_request *creq)
{
	return orfs_mx_generic_remote_call(sbi, creq, MONO_ID);
}

static int
orfs_mx_multi_remote_call(struct orfs_sb_info *sbi,
			  struct orfs_client_request *creq)
{
	return orfs_mx_generic_remote_call(sbi, creq,
					   (uint32_t)(unsigned long) creq);
}

#if ORFS_LINUX_26

static int
orfs_mx_async_remote_call(struct orfs_sb_info *sbi,
			  struct orfs_client_request *creq)
{
	struct orfs_mx_connection *connection = CONNECTION(sbi);
	struct orfs_request *req = ORFS_CREQ2REQ(creq);
	struct orfs_answer *ans;
	struct orfs_mx_request *mxreq;
	int error;

	if (sbi->connection_status != ORFS_CONNECTION_OK) {
		ORFS_DEBUG(MX, "Process refused\n");
		return -EIO;
	}
	/* The connection status may change after if this point if another
	 * thread encounters an error or if umount -f is called.
	 * This buggy thread will loop in orfs_wakeup_all_requests, waking up
	 * everybody and waiting for everybody to leave this routine.
	 */

	/* allocate the mx structure
	 * and put in into the network specific private data*/
	mxreq = kmalloc(sizeof(*mxreq), GFP_KERNEL);
	if (!mxreq) {
		ORFS_ERROR("Failed to allocate mxreq\n");
		return -ENOMEM;
	}
	memset(mxreq, 0, sizeof(*mxreq));
	creq->net_pd = mxreq;

	/* allocate the answer */
	ans = orfs_get_client_answer(sbi);
	if ( !ans ) {
		kfree(mxreq);
		return -ENOMEM;
	}
	ORFS_CREQ2ANS(creq) = ans;

	/* initializes tag to 0 */
	req->tag = 0;

	error = orfs_mx_post_requests(creq, connection, mxreq,
				      (uint32_t)(unsigned long) creq);
	return error;
}

/********************************
 * Async Reqs Processing Thread *
 ********************************/

static void
orfs_mx_kt_loop(struct orfs_sb_info *sbi)
{
	struct orfs_mx_connection *connection = CONNECTION(sbi);	
	mx_endpoint_t ep = connection->ep;
	struct orfs_client_request *creq;
	mx_request_t mx_request_handle;
	struct orfs_mx_request *mxreq;
	struct kiocb *kiocb;
	mx_return_t ret;
	mx_status_t status;
	uint32_t result;

	while (1) {
		ret = mx_peek(ep, MX_INFINITE, &mx_request_handle, &result);
		if (sbi->param.multi.kt_finish)
			break;
		if (ret != MX_SUCCESS || result != 1) {
			// handle error here ?
			break;
		}

		/* get the ORFS creq from the mx request context */
		ret = mx_context(&mx_request_handle, (void**)&creq);
		if (ret != MX_SUCCESS) {
			// handle error here ?
			break;
		}

		/* get the kiocb */
		kiocb = creq->kiocb;
		if (!kiocb) {
			/* no iocb, it's not for us !
			 * schedule() to let the target thread take care of this request */
			schedule();
			continue;
		}

		/* get the structure describing all mx requests for this ORFS creq */
		mxreq = creq->net_pd;
		BUG_ON(!mxreq);

		/* update the creq status */
		if (mx_request_handle == mxreq->req_req) {
			ORFS_DEBUG(MX, "Asynchronously completed request send of creq %p\n", creq);
			mxreq->to_cancel &= ~ISEND_REQ;
		} else if (mx_request_handle == mxreq->reqbuf_req) {
			ORFS_DEBUG(MX, "Asynchronously completed request buffer send of creq %p\n", creq);
			mxreq->to_cancel &= ~ISEND_REQBUF;
		} else if (mx_request_handle == mxreq->ans_req) {
			ORFS_DEBUG(MX, "Asynchronously completed answer receive of creq %p\n", creq);
			mxreq->to_cancel &= ~IRECV_ANS;
		} else if (mx_request_handle == mxreq->ansbuf_req) {
			ORFS_DEBUG(MX, "Asynchronously completed answer buffer receive of creq %p\n", creq);
			mxreq->to_cancel &= ~IRECV_ANSBUF;
		} else {
			BUG();
		}

		/* complete the mx request */
		ORFS_DEBUG(MX, "Completing MX request %p\n", mx_request_handle);
		ret = mx_test(ep, &mx_request_handle, &status, &result);
		BUG_ON(ret != MX_SUCCESS || result != 1);
		// check status

		// mx_request_handle is NULL now

		if (!(mxreq->to_cancel & (ISEND_REQ|ISEND_REQBUF|IRECV_ANS|IRECV_ANSBUF))) {
			/* this request is done */
			long res, res2;
			/* cleanup MX stuff */
			if (mxreq->to_cancel & RDMA_WIN_REQBUF) {
				// destroy rdma window
			}
			if (mxreq->to_cancel & RDMA_WIN_ANSBUF) {
				// destroy rdma window
			}
                        /* call its callback */
			creq->callback(sbi, creq, &res, &res2);
			/* complete the request */
			aio_complete(kiocb, res, res2);
			/* release structures */
			kfree(mxreq);
			orfs_put_client_request(sbi, creq);
		}
	}
}
#endif /* ORFS_LINUX_26 */

/**************
 * Connection *
 **************/

static int
orfs_mx_connect(struct orfs_sb_info *sbi,
		struct orfs_user_mount_options *options)
{
	struct orfs_mx_connection *connection;
	unsigned long server_ep, server_board;
	mx_endpoint_t ep;
	mx_endpoint_addr_t server_addr;
	int len = strlen(options->common.name);
	mx_return_t ret;
	int error = -EIO;
	uint64_t nic_id;

	/* fix server endpoint and board */
	server_ep = options->specific.mask & ORFS_SERVER_EP_MASK
	  ? options->mx.ep : ORFS_MX_SERVER_EP;
	server_board = options->specific.mask & ORFS_SERVER_BOARD_MASK
	  ? options->mx.board : ORFS_MX_SERVER_BOARD;

	ORFS_DEBUG(NET, "Trying to connect to server %s through MX endpoint %ld on board %ld\n",
		   options->common.name, server_ep, server_board);

	/* mx_init unavailable in the kernel */

	/* open endpoint */
	// need to use client_board here if precised
	ret = mx_open_endpoint(MX_ANY_NIC, MX_ANY_ENDPOINT,
			       ORFS_MX_EP_FILTER, NULL, 0, &ep);
	if ( ret )
		goto error;

	/* Get server nics */
	if ( !strcmp(options->common.name, "localhost") ) {
		ret = mx_board_number_to_nic_id(server_board, &nic_id);
		if ( ret != MX_SUCCESS ) {
			ORFS_ERROR("Cannot find MX local nic id (error %s)\n",
				   mx_strerror(ret));
			goto error_with_ep;
		}
	} else {
		uint32_t nics = 1;
		// use server_board
		uint64_t nic_ids[1];
		ret = mx_hostname_to_nic_ids(options->common.name, nic_ids, &nics);
		if ( ret != MX_SUCCESS ) {
			ORFS_ERROR("Cannot find MX nic ids of node %s (error %s)\n",
				   options->common.name, mx_strerror(ret));
			goto error_with_ep;
		}
		nic_id = nic_ids[0];
	}
	
	/* Connect to server's endpoint */
	ret = mx_connect(ep, nic_id, server_ep, ORFS_MX_EP_FILTER,
			 &server_addr);
	if ( ret != MX_SUCCESS ) {
		ORFS_ERROR("Cannot connect to server's endpoint (error %s)\n",
			   mx_strerror(ret));
		goto error_with_ep;
	}

	/* Setup connection */
	error = -ENOMEM;
	connection = kmalloc(sizeof (*connection), GFP_KERNEL);
	if (!connection)
		goto error_with_ep;
	connection->ep = ep;
	connection->server_addr = server_addr;

        SET_CONNECTION(sbi, connection);

	/* Setup path */
	error = -ENOMEM;
	sbi->server_path = kmalloc(len + 4 + 10 + 1 + 10 + 1, GFP_KERNEL);
	if (!sbi->server_path)
		goto error_with_connection;
	sprintf(sbi->server_path, "%s:MX:%ld:%ld", options->common.name,
		(unsigned long) server_board, (unsigned long) server_ep);

	return 0;

      error_with_connection:
	kfree(connection);
      error_with_ep:
	mx_close_endpoint(ep);
      error:
	/* mx_finalize unavailable in the kernel */
	return error;
}

static void
orfs_mx_disconnect(struct orfs_sb_info *sbi)
{
	struct orfs_mx_connection *connection = CONNECTION(sbi);
	mx_endpoint_t ep = connection->ep;

	kfree(sbi->server_path);
	kfree(connection);
	mx_close_endpoint(ep);
	/* mx_finalize unavailable in the kernel */
}

/***********
 * Wake up *
 ***********/

static void
orfs_mx_wake_up(struct orfs_sb_info *sbi)
{
	struct orfs_mx_connection *connection = CONNECTION(sbi);
	mx_endpoint_t ep = connection->ep;
	mx_wakeup(ep);
}

/**********************
 * Network Operations *
 **********************/

static struct orfs_net_operations orfs_mx_nops = {
	.connect = orfs_mx_connect,
	.disconnect = orfs_mx_disconnect,

	.multi_remote_call = orfs_mx_multi_remote_call,
	.mono_remote_call = orfs_mx_mono_remote_call,
	.async_remote_call = orfs_mx_async_remote_call,

#if ORFS_LINUX_26
	.kt_loop = orfs_mx_kt_loop, /* for async request processing */
#else
	.kt_loop = NULL,
#endif

	.wake_up_waiters = orfs_mx_wake_up,
};

/*****************
 * Network Layer *
 *****************/

static DECLARE_ORFS_NET_LAYER(orfs_mx_layer, "mx", &orfs_mx_nops);

int
orfs_mx_layer_init(void)
{
	return orfs_register_net_layer(&orfs_mx_layer);
}

void
orfs_mx_layer_exit(void)
{
	orfs_unregister_net_layer(&orfs_mx_layer);
}

#ifdef CONFIG_ORFS_NETMODULES
MODULE_AUTHOR("Brice Goglin <Brice.Goglin@ens-lyon.fr>");
MODULE_LICENSE("GPL");
module_init(orfs_mx_layer_init);
module_exit(orfs_mx_layer_exit);
#endif				/* CONFIG_ORFS_NETMODULES */
