/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/net.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/in.h>
#include <net/sock.h>
#include <linux/spinlock.h>

#include "orfs_fs.h"
#include "orfs_common_network.h"
#include "orfs_request.h"
#include "orfs_mount.h"
#include "orfs_misc.h"
#include "orfs_messages.h"

#if ORFS_LINUX_26
#define SOCKET_SLEEPQ(s) (s)->sk->sk_sleep
#elif ORFS_LINUX_24
#define SOCKET_SLEEPQ(s) (s)->sk->sleep
#endif

/******************
 * TCP connection *
 ******************/

struct orfs_tcp_connection {
	/* connection socket and status */
	struct socket *socket;
	/* set to 1 to make ORFSD wait before receiving again */
	int client_receiving;
};

#define CONNECTION(sbi) ((struct orfs_tcp_connection*)(sbi)->connection)
#define SET_CONNECTION(sbi, connection) ((sbi)->connection = connection)

/***************************
 * Internal Send Functions *
 ***************************/

static inline int
orfs_tcp_send(struct orfs_sb_info *sbi, void *buffer, size_t len)
{
	struct orfs_tcp_connection *connection = CONNECTION(sbi);
	struct socket *socket = connection->socket;
	struct msghdr msg;
	struct iovec iov;
        DECLARE_WAITQUEUE(send_wq, current);
	int error;

	msg.msg_name = NULL;
	msg.msg_namelen = 0;
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags = MSG_DONTWAIT;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	iov.iov_base = buffer;
	iov.iov_len = len;
	
        add_wait_queue(SOCKET_SLEEPQ(socket), &send_wq);
	for(;;) {
		set_current_state(TASK_INTERRUPTIBLE);
		error = sock_sendmsg(socket, &msg, len);
		/* Warning, msg and iov might have been modified,
		 * do not rely on their contents.
		 */
		if (sbi->connection_status != ORFS_CONNECTION_OK) {
			ORFS_ERROR("Send stopped\n");
			error = -EIO;
			break;
		} else if (error == len) {
			error = 0;
			break;
		} else if (error >= 0) {
			iov.iov_base = (buffer += error);
			iov.iov_len = (len -= error);
			continue;
		} else if (error != -EAGAIN) {
			ORFS_ERROR("Send error %d\n", -error);
			break;
		}
		schedule();
	}
	set_current_state(TASK_RUNNING);
        remove_wait_queue(SOCKET_SLEEPQ(socket), &send_wq);
	return error;
}

static inline int
orfs_tcp_send_whole_request(struct orfs_sb_info *sbi,
			    struct orfs_client_request *creq)
{
	void *in_buf = creq->in_buf;
	int error;

	error = orfs_tcp_send(sbi, ORFS_CREQ2REQ(creq),
			      sizeof (*ORFS_CREQ2REQ(creq)));
	if (error < 0)
		return error;

	if (in_buf) {
		size_t in_buf_len = creq->in_buf_len;
		off_t in_page_off = creq->in_page_off;
		orfs_buffer_flags_t flags = creq->in_buf_flags;
		void *buffer = in_buf;
		void *vaddr;

		switch (flags) {
		case ORFS_BUF_USER_ANY:
		case ORFS_BUF_KMALLOC:
			break;
		case ORFS_BUF_PAGE:
			vaddr = kmap((struct page *) in_buf);
			if (!vaddr) {
				ORFS_ERROR("Input page kmap failed !\n");
				return -ENOMEM;
			}
			buffer = vaddr + in_page_off;
			break;
		default:
			ORFS_ALERT("Input buffer flags incorrect %d\n", flags);
			BUG();
			break;
		}

		error = orfs_tcp_send(sbi, buffer, in_buf_len);
		
		if (in_buf && flags == ORFS_BUF_PAGE)
			kunmap((struct page *) in_buf);
	}

	return error;
}

/***************************
 * Internal Recv Functions *
 ***************************/

static inline int
orfs_tcp_try_recv(struct orfs_sb_info *sbi, void *buffer, size_t len)
{
	struct orfs_tcp_connection *connection = CONNECTION(sbi);
	struct socket *socket = connection->socket;
	struct msghdr msg;
	struct iovec iov;
	int received = 0;
	int error;

	msg.msg_name = NULL;
	msg.msg_namelen = 0;
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags = MSG_DONTWAIT;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	iov.iov_base = buffer;
	iov.iov_len = len;

	for(;;) {
		set_current_state(TASK_INTERRUPTIBLE);
		error = sock_recvmsg(socket, &msg, len, MSG_DONTWAIT);
		/* Warning, msg and iov might have been modified,
		 * do not rely on their contents.
		 */
		if (sbi->connection_status != ORFS_CONNECTION_OK) {
			ORFS_ERROR("Receive stopped\n");
			error = -EIO;
			break;
		} else if (error == len) {
			error = 0;
			break;
		} else if (error >= 0) {
			received = 1;
			iov.iov_base = (buffer += error);
			iov.iov_len = (len -= error);
			continue;
		} else if (error != -EAGAIN) {
			ORFS_ERROR("Receive error %d\n", -error);
			break;
		} else if (!received) {
			/* only block if we already received something */
			break;
		}
		schedule();
	}
	set_current_state(TASK_RUNNING);
	return error;
}

static inline int
orfs_tcp_recv(struct orfs_sb_info *sbi, void *buffer, size_t len)
{
	struct orfs_tcp_connection *connection = CONNECTION(sbi);
	struct socket *socket = connection->socket;
	struct msghdr msg;
	struct iovec iov;
        DECLARE_WAITQUEUE(recv_wq, current);
	int error;

	msg.msg_name = NULL;
	msg.msg_namelen = 0;
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags = MSG_DONTWAIT;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	iov.iov_base = buffer;
	iov.iov_len = len;

        add_wait_queue(SOCKET_SLEEPQ(socket), &recv_wq);
	for(;;) {
		set_current_state(TASK_INTERRUPTIBLE);
		error = sock_recvmsg(socket, &msg, len, MSG_DONTWAIT);
		/* Warning, msg and iov might have been modified,
		 * do not rely on their contents.
		 */
		if (sbi->connection_status != ORFS_CONNECTION_OK) {
			ORFS_ERROR("Receive stopped\n");
			error = -EIO;
			break;
		} else if (error == len) {
			error = 0;
			break;
		} else if (error >= 0) {
			iov.iov_base = (buffer += error);
			iov.iov_len = (len -= error);
			continue;
		} else if (error != -EAGAIN) {
			ORFS_ERROR("Receive error %d\n", -error);
			break;
		}
		schedule();
	}
	set_current_state(TASK_RUNNING);
        remove_wait_queue(SOCKET_SLEEPQ(socket), &recv_wq);
	return error;
}

static inline int
orfs_tcp_recv_buffer(struct orfs_sb_info *sbi,
		     struct orfs_client_request *creq)
{
	void *out_buf = creq->out_buf;
	size_t out_buf_len = creq->out_buf_len;
	off_t out_page_off = creq->out_page_off;
	orfs_buffer_flags_t flags = creq->out_buf_flags;
	void *buffer = out_buf;
	void *vaddr;
	int error;

	switch (flags) {
	case ORFS_BUF_USER_ANY:
	case ORFS_BUF_KMALLOC:
		break;
	case ORFS_BUF_PAGE:
		vaddr = kmap((struct page *) out_buf);
		if (!vaddr) {
			ORFS_ERROR("Output page kmap failed !\n");
			return -ENOMEM;
		}
		buffer = vaddr + out_page_off;
		break;
	default:
		ORFS_ALERT("Output buffer flags incorrect %d\n", flags);
		BUG();
		break;
	}

	error = orfs_tcp_recv(sbi, buffer, out_buf_len);
		
	if (out_buf && flags == ORFS_BUF_PAGE)
		kunmap((struct page *) out_buf);

	return error;
}

/***************************
 * Remote Calls Processing *
 ***************************/

static int
orfs_tcp_mono_remote_call(struct orfs_sb_info *sbi, struct orfs_client_request *creq)
{
	struct orfs_answer *ans;
	int error;
	size_t data_len;
	ORFS_DECLARE_KERNEL_DS();
	ORFS_DECLARE_SIGMASK();
	ORFS_INIT_SIGMASK();

	ans = orfs_get_client_answer(sbi);
	if ( !ans )
		return -ENOMEM;
	ORFS_CREQ2ANS(creq) = ans;

	ORFS_START_KERNEL_DS();
	ORFS_START_SIGMASK();

	if (sbi->connection_status != ORFS_CONNECTION_OK) {
		ORFS_DEBUG(TCP, "Process refused\n");
		error = -EIO;
		goto disconnected;
	}
	/* The connection status may change after if this point if another
	 * thread encounters an error or if umount -f is called.
	 * This buggy thread will loop in orfs_wakeup_all_requests, waking up
	 * everybody and waiting for everybody to leave this routine.
	 */

	ORFS_DEBUG(TCP, "Sending request\n");
	error = orfs_tcp_send_whole_request(sbi, creq);
	if (error < 0)
		goto error;

	ORFS_DEBUG(TCP, "Receiving answer\n");
	error = orfs_tcp_recv(sbi, ans, sizeof(*ans));
	if (error < 0)
		goto error;

	if (get_answer_buffer_len(ans, &data_len)) {
		ORFS_DEBUG(TCP, "Process has %ld data to receive\n", (long) data_len);

		if (data_len > creq->out_buf_len) {
			ORFS_ALERT("Receiving %ld instead of %ld\n",
				   (long) data_len, (long) creq->out_buf_len);
			BUG();
		}
		creq->out_buf_len = data_len;

		ORFS_DEBUG(TCP, "Receiving buffer\n");
		error = orfs_tcp_recv_buffer(sbi, creq);
		if (error < 0)
			goto error;

		ORFS_DEBUG(TCP, "Process received all data\n");
	}

	/* return success */
	ORFS_STOP_KERNEL_DS();
	ORFS_STOP_SIGMASK();
	return 0;

      error:
	orfs_set_connection_error(sbi);
      disconnected:
	ORFS_STOP_KERNEL_DS();
	ORFS_STOP_SIGMASK();
	return error;
}

static int
orfs_tcp_multi_remote_call(struct orfs_sb_info *sbi,
			   struct orfs_client_request *creq)
{
	struct orfs_answer *ans;
	struct orfs_tcp_connection *connection = CONNECTION(sbi);
	int error;
	size_t data_len;
	ORFS_DECLARE_KERNEL_DS();
	ORFS_DECLARE_SIGMASK();
	ORFS_INIT_SIGMASK();

	ORFS_START_KERNEL_DS();

	if (sbi->connection_status != ORFS_CONNECTION_OK) {
		ORFS_DEBUG(TCP, "Process defused\n");
		error = -EIO;
		goto disconnected;
	}
	/* The connection status may change after if this point if another
	 * thread encounters an error or if umount -f is called.
	 * This buggy thread will loop in orfs_wakeup_all_requests, waking up
	 * everybody and waiting for everybody to leave this routine.
	 */

	ORFS_CREQ_NEED_REQ(creq);

	ORFS_DEBUG(TCP, "Sending request\n");
	ORFS_START_SIGMASK();
	error = orfs_tcp_send_whole_request(sbi, creq);
	ORFS_STOP_SIGMASK();
	if (error < 0)
		goto error;

	ORFS_DEBUG(TCP, "Sleeping until answer is received...\n");
	error = wait_event_interruptible(creq->queue,
					 /* wait until request wait is done */
					 !IS_ORFS_CREQ_WAITING(creq)
					 /* or client disconnected */
					 || (sbi->connection_status != ORFS_CONNECTION_OK));
	if (error < 0) {
		ORFS_DEBUG(TCP, "Process woken up by a signal\n");
		ORFS_CREQ_IS_ORPHAN(creq);
		goto interrupted;
	}
	if (sbi->connection_status != ORFS_CONNECTION_OK) {
		ORFS_DEBUG(TCP, "Process woken up by ORFSD network error\n");
		error = -EIO;
		goto disconnected;
	}

	ORFS_DEBUG(TCP, "Process woken up by ORFSD\n");
	/* we can now assume that ORFS_CREQ2ANS(creq) and its eventual
	 * buffer have been filled */
	ans = ORFS_CREQ2ANS(creq);

	if (get_answer_buffer_len(ans, &data_len)) {
		ORFS_DEBUG(TCP, "Process has %ld data to receive\n", (long) data_len);

		if (data_len > creq->out_buf_len) {
			ORFS_ALERT("Receiving %ld instead of %ld\n",
				   (long) data_len, (long) creq->out_buf_len);
			BUG();
		}
		creq->out_buf_len = data_len;

		ORFS_DEBUG(TCP, "Receiving buffer\n");
		ORFS_START_SIGMASK();
		error = orfs_tcp_recv_buffer(sbi, creq);
		ORFS_STOP_SIGMASK();

		/* release ORFSD */
		connection->client_receiving = 0;
		wake_up(SOCKET_SLEEPQ(connection->socket));

		if (error < 0)
			goto error;

		ORFS_DEBUG(TCP, "Process received all data\n");
	}

	/* return success */
	ORFS_STOP_KERNEL_DS();
	return 0;

      error:
	orfs_set_connection_error(sbi);
      interrupted:
      disconnected:
	ORFS_STOP_KERNEL_DS();
	return error;
}

/*********
 * ORFSD *
 *********/

/* This function reads data_len from the sbi socket
 * and discards all.
 * It is used when a request has been abandonned by an
 * interrupted process
 */
static inline int
orfs_tcp_discard_abandonned_data(struct orfs_sb_info *sbi,
				 size_t data_len)
{
	struct page *dummy;
	struct orfs_client_request fake_creq;
	int error;

	ORFS_DEBUG(ORFSD, "ORFSD discards %ld data\n",
		   (unsigned long) data_len);

	/* allocate dummy buffer */
	dummy = alloc_page(GFP_USER);
	if (!dummy) {
		ORFS_ALERT("No memory available for dummy page !\n");
		BUG();
	}
	fake_creq.out_buf = page_address(dummy);
	fake_creq.out_buf_flags = ORFS_BUF_KMALLOC;

	/* receive data */
	while (data_len > 0) {
		fake_creq.in_buf_len =
		    data_len > PAGE_SIZE ? PAGE_SIZE : data_len;
		error = orfs_tcp_recv_buffer(sbi, &fake_creq);
		if (error < 0)
			goto error_on_network;
		data_len -= error;
	}

	/* release dummy page */
	__free_page(dummy);

	return 0;

      error_on_network:
	__free_page(dummy);
	orfs_set_connection_error(sbi);
	return error;
}

/* This function check the validity of the answer,
 * updates its corresponding creq.
 * Then it wakes up its waiting process or discard
 * all.
 */
static inline int
orfs_tcp_process_answer(struct orfs_sb_info *sbi, struct orfs_answer *ans)
{
	struct orfs_tcp_connection *connection = CONNECTION(sbi);
	struct orfs_client_request *creq;
	int error = 0;

	/* get tag and check validity */
	creq = (void*) (unsigned long) ans->tag;

	/* update creq */
	creq->ans = ans;
	ORFS_CREQ_RECV_REQ(creq);

	if (!IS_ORFS_CREQ_ADOPTED(creq)) {
		/* usual request */

		if (creq->out_buf) {
			/* stop orfsd receive loop */
			connection->client_receiving = 1;
		}

		/* wake up process */
		wake_up_interruptible(&creq->queue);
		ORFS_DEBUG(ORFSD, "ORFSD wakes up process\n");

	} else {
		/* abandonned request */
		size_t data_len;

		if (get_answer_buffer_len(ans, &data_len))
			error = orfs_tcp_discard_abandonned_data(sbi, data_len);

		orfs_put_client_request(sbi, creq);
		ORFS_DEBUG(ORFSD, "ORFSD releases orphan request\n");
	}

	return error;
}

static void
orfs_tcp_kt_loop(struct orfs_sb_info *sbi)
{
	struct orfs_tcp_connection *connection = CONNECTION(sbi);
	struct socket *socket = connection->socket;
	struct orfs_answer *ans;
	int error;
	DECLARE_WAITQUEUE(recv_wq, current);

	ORFS_DECLARE_KERNEL_DS();
	ORFS_DECLARE_SIGMASK();
	ORFS_INIT_SIGMASK();

	/* get an answer */
	ans = orfs_get_client_answer(sbi);

	/* when receiving answers,
	 * we have to copy_to_user into kernel space */
	ORFS_START_KERNEL_DS();
	ORFS_START_SIGMASK();

        add_wait_queue(SOCKET_SLEEPQ(socket), &recv_wq);
	for (;;) {
		set_current_state(TASK_INTERRUPTIBLE);

		if (!connection->client_receiving) {
			error = orfs_tcp_try_recv(sbi, ans, sizeof(*ans));
			if (!error) {
				error = orfs_tcp_process_answer(sbi, ans);
				if (error < 0)
					break;
				/* this answer has been filled,
				 * get another one  and try again */
				ans = orfs_get_client_answer(sbi);
				continue;
			}
			if (error != -EAGAIN) {
				/* receive error, abort */
				orfs_set_connection_error(sbi);
				break;
			}
		}

		if (sbi->param.multi.kt_finish)
			/* finish required */
			break;

		/* nothing to do */
		schedule();
	}
	set_current_state(TASK_RUNNING);
        remove_wait_queue(SOCKET_SLEEPQ(socket), &recv_wq);

	/* cleanup */
	ORFS_STOP_SIGMASK();
	ORFS_STOP_KERNEL_DS();

	/* put the answer */
	orfs_put_client_answer(sbi, ans);
}

/**************
 * Connection *
 **************/

static int
orfs_tcp_connect(struct orfs_sb_info *sbi,
		 struct orfs_user_mount_options *options)
{
	struct sockaddr_in *ski = &options->tcp.sockaddr_in;
	struct orfs_tcp_connection *connection;
	struct socket *socket = NULL;
	int len = strlen(options->common.name);
	unsigned long port;
	int error;

	/* fix server port */
	port = options->specific.mask & ORFS_SERVER_PORT_MASK ?
	    options->tcp.port : ORFS_TCP_SERVER_PORT;
	options->tcp.sockaddr_in.sin_port = htons(port);

	ORFS_DEBUG(NET, "Trying to connect to server %s through TCP port %ld\n",
		   options->common.name, port);

	/* check sockaddr_in */
	if ( !(options->specific.mask & ORFS_SERVER_TCP_SOCKADDR_MASK) ) {
		ORFS_ERROR("Cannot find TCP server %s\n", options->common.name);
		error = -EINVAL;
		goto error;
	}

	/* create the socket */
	if ((error = sock_create(PF_INET, SOCK_STREAM, 0, &socket)) < 0)
		goto error;

	/* connect */
	if ((error = socket->ops->connect(socket,
					  (struct sockaddr *) ski,
					  sizeof (*ski), O_RDWR)) < 0) {
		ORFS_ERROR
		    ("TCP connection refused from server %s on port %ld, error %d\n",
		     options->common.name, port, -error);
		goto error_with_socket;
	}

	/* Setup connection */
	connection = kmalloc(sizeof (*connection), GFP_KERNEL);
	if (!connection) {
		error = -ENOMEM;
		goto error_with_socket;
	}
	connection->socket = socket;
	connection->client_receiving = 0;
	SET_CONNECTION(sbi, connection);

	/* Setup path */
	error = -ENOMEM;
	sbi->server_path = kmalloc(len + 5 + 10 + 1, GFP_KERNEL);
	if (!sbi->server_path)
		goto error_with_connection;
	sprintf(sbi->server_path, "%s:TCP:%ld", options->common.name, port);

	return 0;

      error_with_connection:
	kfree(connection);
      error_with_socket:
	sock_release(socket);
      error:
	return error;
}

static void
orfs_tcp_disconnect(struct orfs_sb_info *sbi)
{
	struct orfs_tcp_connection *connection = CONNECTION(sbi);
	struct socket *socket = connection->socket;
	kfree(sbi->server_path);
	kfree(connection);
	sock_release(socket);
}

/***********
 * Wake up *
 ***********/

static void
orfs_tcp_wake_up(struct orfs_sb_info *sbi)
{
	wake_up(SOCKET_SLEEPQ(CONNECTION(sbi)->socket));
}

/**********************
 * Network Operations *
 **********************/

static struct orfs_net_operations orfs_tcp_nops = {
	.connect = orfs_tcp_connect,
	.disconnect = orfs_tcp_disconnect,

	.multi_remote_call = orfs_tcp_multi_remote_call,
	.mono_remote_call = orfs_tcp_mono_remote_call,

	.kt_loop = orfs_tcp_kt_loop,

	.wake_up_waiters = orfs_tcp_wake_up,
};

/*****************
 * Network Layer *
 *****************/

static DECLARE_ORFS_NET_LAYER(orfs_tcp_layer, "tcp", &orfs_tcp_nops);

int
orfs_tcp_layer_init(void)
{
	return orfs_register_net_layer(&orfs_tcp_layer);
}

void
orfs_tcp_layer_exit(void)
{
	orfs_unregister_net_layer(&orfs_tcp_layer);
}

#ifdef CONFIG_ORFS_NETMODULES
MODULE_AUTHOR("Brice Goglin <Brice.Goglin@ens-lyon.fr>");
MODULE_LICENSE("GPL");
module_init(orfs_tcp_layer_init);
module_exit(orfs_tcp_layer_exit);
#endif				/* CONFIG_ORFS_NETMODULES */
