/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/pagemap.h>

#include "orfs_fs.h"
#include "orfs_private_data.h"
#include "orfs_protocol.h"
#include "orfs_request.h"
#include "orfs_messages.h"

/****************************
 * Address Space Operations *
 ****************************/

int
orfs_readpage(struct file *file, struct page *page)
{
	struct inode *inode = page->mapping->host;
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	orfs_off_t offset = page->index << PAGE_CACHE_SHIFT;
	orfs_off_t count;
	ssize_t error;

	ORFS_DEBUG(REQ,
		   "READPAGE called for inode %lld at offset %lld in page %p\n",
		   (long long) inode->i_ino, (long long) offset, page);

	ORFS_UPDATE_REQ_STAT(sbi, readpage);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_PAGE_REQ;
	req->subtype = ORFS_READ_FULL_PAGE_REQ;
	fill_inode_id_from_pd(inode, &req->target.page.inode);
	req->target.page.data.data_access.offset = offset;
	req->target.page.data.data_access.count = PAGE_SIZE;
	creq->out_buf = page;
	creq->out_buf_len = PAGE_SIZE;
	creq->out_page_off = 0;
	creq->out_buf_flags = ORFS_BUF_PAGE;

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR
		    ("Remote call for READ FULL PAGE failed with error %d\n",
		     -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.page.node_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* update node attributes */
			orfs_update_inode_attr(inode, node_attr);
			/* get count and bzero the end of the page */
			count = ans->target.page.data.data_access.count;
			if (count != PAGE_SIZE) {
				char *kaddr = kmap(page);
				memset(kaddr + count, 0, PAGE_SIZE - count);
				kunmap(page);
			}
			flush_dcache_page(page);
			SetPageUptodate(page);
			break;
		case ORFS_SERVER_INVALID_PARAMETER:
			ORFS_NOTICE("READ FULL PAGE returns Server Invalid Parameter\n");
			error = -EIO;
			break;
		case ORFS_BAD_INODE_TYPE:
			// for now the state based protocol make the error impossible
			ORFS_ALERT("Bad inode %lld mode %lld\n",
				   (long long) inode->i_ino, (long long) node_attr->mode);
			BUG();
			break;
		case ORFS_NO_SUCH_INODE:
			// for now the state based protocol make the error impossible
			ORFS_ALERT("Bad inode %lld\n",
				   (long long) inode->i_ino);
			BUG();
			break;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

#if ORFS_LINUX_26
	unlock_page(page);
#elif ORFS_LINUX_24
	UnlockPage(page);
#endif
	orfs_put_client_request(sbi, creq);
	return error;
}

int
#if ORFS_LINUX_26
orfs_writepage(struct page *page, struct writeback_control *wbc)
#elif ORFS_LINUX_24
orfs_writepage(struct page *page)
#endif
{
	struct inode *inode = page->mapping->host;
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	orfs_off_t offset = page->index << PAGE_CACHE_SHIFT;
	orfs_off_t count;
	ssize_t error;

	ORFS_DEBUG(REQ,
		   "WRITEPAGE called for inode %lld at offset %lld in page %p\n",
		   (long long) inode->i_ino, (long long) offset, page);

	ORFS_UPDATE_REQ_STAT(sbi, writepage);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	count = inode->i_size - offset;
	if (count > PAGE_SIZE)
		count = PAGE_SIZE;

	req->type = ORFS_PAGE_REQ;
	req->subtype = ORFS_WRITE_FULL_PAGE_REQ;
	fill_inode_id_from_pd(inode, &req->target.page.inode);
	req->target.page.data.data_access.offset = offset;
	req->target.page.data.data_access.count = count;
	creq->in_buf = page;
	creq->in_buf_len = count;
	creq->in_page_off = 0;
	creq->in_buf_flags = ORFS_BUF_PAGE;

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR
		    ("Remote call for WRITE FULL PAGE failed with error %d\n",
		     -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.page.node_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* update node attributes */
			orfs_update_inode_attr(inode, node_attr);
			SetPageUptodate(page);
			break;
		case ORFS_SERVER_INVALID_PARAMETER:
			ORFS_NOTICE("WRITE FULL PAGE returns Server Invalid Parameter\n");
			error = -EIO;
			break;
		case ORFS_NO_SPACE_LEFT:
			error = -ENOSPC;
			break;
		case ORFS_BAD_INODE_TYPE:
			// for now the state based protocol make the error impossible
			ORFS_ALERT("Bad inode %lld mode %lld\n",
				   (long long) inode->i_ino, (long long) node_attr->mode);
			BUG();
			break;
		case ORFS_NO_SUCH_INODE:
			// for now the state based protocol make the error impossible
			ORFS_ALERT("Bad inode %lld\n",
				   (long long) inode->i_ino);
			BUG();
			break;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

#if ORFS_LINUX_26
	unlock_page(page);
#elif ORFS_LINUX_24
	UnlockPage(page);
#endif
	orfs_put_client_request(sbi, creq);
	return error;
}

int
orfs_prepare_write(struct file *file, struct page *page,
		   unsigned from, unsigned to)
{
	struct inode *inode = page->mapping->host;
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	orfs_off_t offset = page->index << PAGE_CACHE_SHIFT;
	orfs_off_t count = to - from;
	ssize_t error;

	ORFS_DEBUG(REQ,
		   "PREPARE WRITE called for inode %lld at offset %lld in page %p from %d to %d\n",
		   (long long) inode->i_ino, (long long) offset, page, from,
		   to);

	ORFS_UPDATE_REQ_STAT(sbi, prepare_write);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_PAGE_REQ;
	req->subtype = ORFS_PREPARE_WRITE_PAGE_REQ;
	fill_inode_id_from_pd(inode, &req->target.page.inode);
	req->target.page.data.data_access.offset = offset + from;
	req->target.page.data.data_access.count = count;

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR
		    ("Remote call for PREPARE WRITE failed with error %d\n",
		     -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.page.node_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* update node attributes */
			orfs_update_inode_attr(inode, node_attr);
			break;
		case ORFS_SERVER_INVALID_PARAMETER:
			ORFS_NOTICE("PREPARE WRITE returns Server Invalid Parameter\n");
			error = -EIO;
			break;
		case ORFS_NO_SPACE_LEFT:
			error = -ENOSPC;
			break;
		case ORFS_INVALID_PARAMETER:
			ORFS_ALERT("Invalid Parameter\n");
			BUG();
			break;
		case ORFS_BAD_INODE_TYPE:
			// for now the state based protocol make the error impossible
			ORFS_ALERT("Bad inode %lld mode %lld\n",
				   (long long) inode->i_ino, (long long) node_attr->mode);
			BUG();
			break;
		case ORFS_NO_SUCH_INODE:
			// for now the state based protocol make the error impossible
			ORFS_ALERT("Bad inode %lld\n",
				   (long long) inode->i_ino);
			BUG();
			break;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

int
orfs_commit_write(struct file *file, struct page *page,
		  unsigned from, unsigned to)
{
	struct inode *inode = page->mapping->host;
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	orfs_off_t offset = page->index << PAGE_CACHE_SHIFT;
	orfs_off_t count = to - from;
	ssize_t error;

	ORFS_DEBUG(REQ,
		   "COMMIT WRITE called for inode %lld at offset %lld in page %p from %d to %d\n",
		   (long long) inode->i_ino, (long long) offset, page, from,
		   to);

	ORFS_UPDATE_REQ_STAT(sbi, commit_write);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_PAGE_REQ;
	req->subtype = ORFS_COMMIT_WRITE_PAGE_REQ;
	fill_inode_id_from_pd(inode, &req->target.page.inode);
	req->target.page.data.data_access.offset = offset + from;
	req->target.page.data.data_access.count = count;
	creq->in_buf = page;
	creq->in_buf_len = count;
	creq->in_page_off = from;
	creq->in_buf_flags = ORFS_BUF_PAGE;

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR
		    ("Remote call for COMMIT WRITE failed with error %d\n",
		     -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.page.node_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* update node attributes */
			orfs_update_inode_attr(inode,
					       &ans->target.page.node_attr);
			SetPageUptodate(page);
			break;
		case ORFS_SERVER_INVALID_PARAMETER:
			ORFS_NOTICE("COMMIT WRITE returns Server Invalid Parameter\n");
			error = -EIO;
			break;
		case ORFS_INVALID_PARAMETER:
			ORFS_ALERT("Invalid Parameter\n");
			BUG();
			break;
		case ORFS_BAD_INODE_TYPE:
			// for now the state based protocol make the error impossible
			ORFS_ALERT("Bad inode %lld mode %lld\n",
				   (long long) inode->i_ino, (long long) node_attr->mode);
			BUG();
			break;
		case ORFS_NO_SUCH_INODE:
			// for now the state based protocol make the error impossible
			ORFS_ALERT("Bad inode %lld\n",
				   (long long) inode->i_ino);
			BUG();
			break;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

#if ORFS_LINUX_26
static int
orfs_direct_IO(int rw, struct kiocb *iocb,
	       const struct iovec *iov, loff_t offset,
	       unsigned long nr_segs)
{
	return -ENOSYS;
}
#endif

struct address_space_operations orfs_file_aops = {
	.readpage = orfs_readpage,
	.writepage = orfs_writepage,
	.prepare_write = orfs_prepare_write,
	.commit_write = orfs_commit_write,
#if ORFS_LINUX_26
	.direct_IO = orfs_direct_IO,
#endif
};
