/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/dcache.h>

#include "orfs_fs.h"
#include "orfs_private_data.h"
#include "orfs_request.h"
#include "orfs_mount.h"
#include "orfs_messages.h"

/*********************
 * Dentry Operations *
 *********************/

static inline int
__orfs_dentry_revalidate(struct dentry *dentry)
{
	struct dentry *parent = dentry->d_parent;
	struct inode *dir = parent->d_inode;
	struct orfs_sb_info *sbi = ORFS_SBI(dentry->d_sb);
	orfs_name_len_t name_len = dentry->d_name.len;
	struct orfs_client_request *creq;
	struct orfs_request *req;
	int error;

	ORFS_DEBUG(REQ, "Real REVALIDATE called for dentry %s in %lld\n",
		   dentry->d_name.name, (long long) dir->i_ino);

	ORFS_UPDATE_REQ_STAT(sbi, revalidate);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_DIR_REQ;
	req->subtype = ORFS_REVALIDATE_DIR_REQ;
	fill_inode_id_from_pd(dir, &req->target.dir.dir);
	req->target.dir.name_len = name_len;
	memcpy(req->target.dir.name, dentry->d_name.name, name_len + 1);

	error = remote_call(sbi, sbi->nops, creq);

	if (error) {
		ORFS_ERROR("Remote call for REVALIDATE failed with error %d\n",
			   -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.dir.data.revalidate.node_attr;
		struct orfs_getattr_data *father_attr = &ans->target.dir.father_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* update node attributes */
			orfs_update_inode_attr(dentry->d_inode, node_attr);
		      update_dir:
			/* update dir attributes */
			orfs_update_inode_attr(dir, father_attr);
			break;
		case ORFS_NO_SUCH_DIR_ENTRY:
			error = -ENOENT;
			goto update_dir;
		case ORFS_NO_SUCH_INODE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("REVALIDATE returns No Such Inode %lld\n",
					   (long long) dir->i_ino);
		       invalidate:
			/* drop old dir inode */
			orfs_invalidate_dir_inode(dir);
			error = -EIO;
			break;
		case ORFS_BAD_INODE_TYPE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("REVALIDATE returns Bad Inode %lld mode %lld\n",
					   (long long) dir->i_ino, (long long) father_attr->mode);
			goto invalidate;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

int
orfs_dentry_revalidate(struct dentry *dentry
#if ORFS_LINUX_24
		       , int flags
#endif
#if ORFS_LINUX_26
		       , struct nameidata *nd
#endif
    )
{
	struct inode *parent_inode = dentry->d_parent->d_inode;
	struct orfs_sb_info *sbi = ORFS_SBI(parent_inode->i_sb);
	orfs_time_t current_time, parent_time, dentry_time, parent_mtime;
	int res;

	if (ORFS_SBI_IS_NON_CONCURRENT(sbi))
		return 1;

	ORFS_DEBUG(REQ, "REVALIDATE called on dentry %s in %lld\n",
		   dentry->d_name.name, (long long) parent_inode->i_ino);

	/* do not care about "/" (root dentry) */
	if (*dentry->d_name.name == '/')
		return 1;

	ORFS_UPDATE_REVAL_STAT(sbi, dentry, call);

	KERNEL2ORFS_TIME(CURRENT_TIME, current_time);
	parent_time = ORFS_INODE_PD(parent_inode)->last_update;
	dentry_time = ORFS_DENTRY_PD(dentry)->last_update;
	KERNEL2ORFS_TIME(parent_inode->i_mtime, parent_mtime);

	/* A dentry is supposed to be valid when
	 * its parent has been updated recently,
	 * its parent has not been modified since the dentry was updated
	 * and the dentry has been updated recently.
	 */
	if (current_time <= parent_time + sbi->opts.revalidate_delay.inode
	    && parent_mtime < dentry_time
	    && current_time <= dentry_time + sbi->opts.revalidate_delay.dentry)
		return 1;

	ORFS_UPDATE_REVAL_STAT(sbi, dentry, remote);

	res = __orfs_dentry_revalidate(dentry);

	if (res)
		ORFS_UPDATE_REVAL_STAT(sbi, dentry, failed);

	/* returns 1 when revalidating succeeded, 0 else */
	return res ? 0 : 1;
}

static void
orfs_dentry_release(struct dentry *dentry)
{
	struct orfs_sb_info *sbi = ORFS_SBI(dentry->d_sb);

	free_dentry_pd(sbi, dentry);
}

struct dentry_operations orfs_dops = {
	.d_revalidate = orfs_dentry_revalidate,
	.d_hash = NULL,
	.d_compare = NULL,
	.d_delete = NULL,
	.d_release = orfs_dentry_release,
	.d_iput = NULL,
};

/****************
 * Dirent Cache *
 ****************/

struct orfs_dirent_cache {
	int size;
	orfs_doff_t first_off, last_off;
	orfs_bool_t eof_reached;
	int current_pos;
	orfs_doff_t current_off;
	struct orfs_dirent_cache_entry {
		struct dentry *dentry;
		orfs_doff_t next_off;
		orfs_dirent_type_t type;
	} dirents[0];
};

#define PREADDIR_DATA_MAX PAGE_SIZE
#define PREADDIR_NR_MAX (PAGE_SIZE/(sizeof(struct orfs_dirent)+1))

static int
__orfs_preaddir(struct file *file,
		void *dirent,
		orfs_doff_t offset,
		orfs_data_len_t data_max,
		orfs_doff_t * read, orfs_bool_t * eof_reached)
{
	struct dentry *dentry = file->f_dentry;
	struct inode *inode = dentry->d_inode;
	struct super_block *sb = dentry->d_sb;
	struct orfs_sb_info *sbi = ORFS_SBI(sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	int error;

	ORFS_DEBUG(REQ, "Real READDIR called for inode %lld"
		   " at offset %lld with %lld data max\n",
		   (long long) inode->i_ino, (long long) offset,
		   (long long) data_max);

	ORFS_UPDATE_REQ_STAT(sbi, preaddir);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_FILE_REQ;
	req->subtype = ORFS_PREADDIR_FILE_REQ;
	fill_file_id_from_pd(file, &req->target.file.file);
	req->target.file.data.preaddir.offset = offset;
	req->target.file.data.preaddir.data_max = data_max;
	creq->out_buf = dirent;
	creq->out_buf_len = data_max;
	creq->out_buf_flags = ORFS_BUF_KMALLOC;

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR
		    ("Remote call of Real PREADDIR failed with error %d\n",
		     -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.file.node_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* update node attributes */
			orfs_update_inode_attr(inode, node_attr);
			/* get count and update readdir parameters */
			*read = ans->target.file.data.preaddir.count;
			ORFS_DEBUG(FS, "Real READDIR gets %lld dirents\n",
				   ans->target.file.data.preaddir.count);
			*eof_reached =
			    ans->target.file.data.preaddir.eof_reached;
			ORFS_DEBUG(FS, "Real READDIR reached EOF\n");
			break;
		case ORFS_BAD_INODE_TYPE:
			// for now the state based protocol make the error impossible
			ORFS_ALERT("Bad inode %lld mode %lld\n",
				   (long long) inode->i_ino, (long long) node_attr->mode);
			BUG();
			break;
		case ORFS_BAD_FILE:
			// for now the state based protocol make the error impossible
			ORFS_ALERT("Bad file %lld\n",
				   (long long) req->target.file.file.id);
			BUG();
			break;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

/* This function has to be called for real dentry,
 * especially not for . and ..
 */
static inline struct dentry *
orfs_add_dentry(struct dentry *parent, const char *name, unsigned int name_len, struct orfs_getattr_data
		*attr)
{
	struct super_block *sb = parent->d_sb;
	struct dentry *dentry;
	struct inode *inode;
	struct qstr qstr;
	int error;

	qstr.name = name;
	qstr.len = name_len;
	qstr.hash = full_name_hash(name, name_len);

	dentry = d_lookup(parent, &qstr);
	if (!dentry) {
		/* if does not exist, allocate the dentry */
		dentry = d_alloc(parent, &qstr);
		if (!dentry)
			return ERR_PTR(-ENOMEM);

		/* try to alloc private data */
		error = alloc_dentry_pd(ORFS_SBI(dentry->d_sb), dentry);
		if (error < 0)
			return ERR_PTR(error);
		/* we now can safely program d_ops */
		dentry->d_op = parent->d_op;
	}

	if (dentry->d_inode) {
		/* dentry positive, just update attributes */
		orfs_update_inode_attr(dentry->d_inode, attr);
	} else {
		/* dentry negative, create inode with given attributes */
		inode = orfs_iget_with_attr(sb, attr);
		if (!inode || is_bad_inode(inode)) {
			dput(dentry);
			return ERR_PTR(-ENOMEM);
		}
		d_add(dentry, inode);
	}

	ORFS_UPDATE_DENTRY_PD_TIME(dentry);

	return dentry;
}

static void
orfs_drop_dirent_cache(struct file *file)
{
	struct orfs_dirent_cache *cache = ORFS_FILE_PD(file)->dirent_cache;
	orfs_doff_t i;

	for (i = 0; i < cache->size; i++) {
		struct dentry *dentry = cache->dirents[i].dentry;
		/* release locked dentry, the file's dentry for . and .. */
		dput(IS_ERR(dentry) ? file->f_dentry : dentry);
	}
}

void
orfs_kill_dirent_cache(struct file *file)
{
	struct orfs_dirent_cache *cache = ORFS_FILE_PD(file)->dirent_cache;

	if (cache) {
		orfs_drop_dirent_cache(file);
		kfree(cache);
	}
}

static inline int
orfs_fill_dirent_cache(struct file *file, orfs_doff_t offset)
{
	struct orfs_file_private_data *file_pd = ORFS_FILE_PD(file);
	struct orfs_dirent_cache *cache = file_pd->dirent_cache;
	orfs_doff_t read, i;
	orfs_bool_t eof_reached;
	void *buffer;
	int error = -ENOMEM;
	char *dirent_ptr;

	/* get the cache */
	if (cache) {
		orfs_drop_dirent_cache(file);
	} else {
		/* allocate it */
		cache = kmalloc(sizeof (*cache)
				/* main structure */
				+ PREADDIR_NR_MAX * sizeof (*cache->dirents),
				/* majored number of orfs_dirent in the request */
				GFP_KERNEL);
		if (!cache) {
			ORFS_ERROR("Failed to allocate dirent cache\n");
			goto error;
		}
	}

	/* allocating temporary readdir buffer */
	buffer = kmalloc(PREADDIR_DATA_MAX, GFP_KERNEL);
	if (!buffer) {
		ORFS_ERROR("Failed to alloc READDIR buffer\n");
		goto error_with_empty_cache;
	}

	/* filling this buffer */
	error = __orfs_preaddir(file, buffer, offset, PREADDIR_DATA_MAX,
				&read, &eof_reached);
	if (error) {
		ORFS_ERROR("Failed to fill READDIR buffer\n");
		goto error_with_failed_readdir;
	}

	cache->size = read;
	cache->first_off = cache->last_off = offset;
	cache->eof_reached = eof_reached;
	cache->current_pos = 0;
	cache->current_off = offset;

	dirent_ptr = buffer;
	for (i = 0; i < read; i++) {
		struct orfs_dirent *dirent = (struct orfs_dirent *) dirent_ptr;
		orfs_name_len_t name_len = dirent->name_len;
		char *name = dirent->name;
		struct dentry *new;

		if (*name == '.' && name_len < 3
		    && (*(name + 1) == '.' || name_len == 1)) {
			/* handle special case of . and .. */
			new = ERR_PTR(-name_len);
			/* locking the dentry or its parent is equivalent here */
			dget(file->f_dentry);
		} else {
			/* try to add and lock a new dentry */
			new = orfs_add_dentry(file->f_dentry,
					      name, name_len, &dirent->attr);
			if (IS_ERR(new)) {
				ORFS_ERROR("Failed to add dentry %s in %lld\n",
					   name,
					   (long long) file->f_dentry->d_inode->
					   i_ino);
				cache->size = i;
				cache->eof_reached = 0;
				break;
			}
		}

		/* we got the target dentry locked */
		cache->dirents[i].dentry = new;

		/* update dirent fields */
		cache->last_off = cache->dirents[i].next_off = dirent->off;
		cache->dirents[i].type = dirent->type;

		ORFS_DEBUG(FS, "Added entry %s at pos %lld with off %lld\n",
			   name, (long long) i, dirent->off);

		/* next dirent */
		dirent_ptr += sizeof (*dirent) + name_len + 1;
	}

	kfree(buffer);

	file_pd->dirent_cache = cache;

	return 0;

      error_with_failed_readdir:
	kfree(buffer);
      error_with_empty_cache:
	kfree(cache);
	file_pd->dirent_cache = NULL;
      error:
	return error;
}

static inline int
orfs_read_dirent_cache(struct file *file, void *dirent, filldir_t filldir)
{
	orfs_doff_t offset = file->f_pos;
	orfs_doff_t pos;
	struct orfs_dirent_cache *cache = ORFS_FILE_PD(file)->dirent_cache;
	struct orfs_dirent_cache_entry *entry;

	struct dentry *dirent_dentry;
	const unsigned char *dirent_name;
	unsigned int dirent_name_len;
	unsigned long dirent_ino;

	if (offset < cache->first_off || offset >= cache->last_off)
		return 0;

	if (offset != cache->current_off) {
		pos = offset < cache->current_off ? 0 : cache->current_pos;
		for (; cache->dirents[pos].next_off > offset; pos++) ;
	} else
		pos = cache->current_pos;

	entry = &cache->dirents[pos];
	dirent_dentry = entry->dentry;
	if (dirent_dentry == ERR_PTR(-1)) {
		dirent_name = ".";
		dirent_name_len = 1;
		dirent_ino = file->f_dentry->d_inode->i_ino;
	} else if (dirent_dentry == ERR_PTR(-2)) {
		dirent_name = "..";
		dirent_name_len = 2;
		dirent_ino = file->f_dentry->d_parent->d_inode->i_ino;
	} else {
		dirent_name = dirent_dentry->d_name.name;
		dirent_name_len = dirent_dentry->d_name.len;
		dirent_ino = dirent_dentry->d_inode->i_ino;
	}

	ORFS_DEBUG(FS,
		   "READDIR gets entry %s with len %d offset %lld ino %lld\n",
		   dirent_name, dirent_name_len, (long long) entry->next_off,
		   (long long) dirent_ino);
	filldir(dirent, dirent_name, dirent_name_len, entry->next_off,
		dirent_ino, entry->type);

	/* update cache position */
	cache->current_pos++;
	cache->current_off = entry->next_off;
	/* update file position */
	file->f_pos = entry->next_off;

	return 0;
}

/********************************
 * Dir Specific File Operations *
 ********************************/

static int
orfs_preaddir(struct file *file, void *dirent, filldir_t filldir)
{
	orfs_doff_t offset = file->f_pos;
	struct orfs_dirent_cache *cache = ORFS_FILE_PD(file)->dirent_cache;
	int error;

	/* This call is protected by inode->i_sem.
	 * This guarantes that the cache won't change
	 * between filling and reading.
	 */

	if (!cache || offset < cache->first_off
	    || (offset >= cache->last_off && !cache->eof_reached)) {
		error = orfs_fill_dirent_cache(file, offset);
		if (error) {
			ORFS_ERROR("Failed to fill READDIR cache\n");
			return error;
		}
	}

	error = orfs_read_dirent_cache(file, dirent, filldir);

	return error;
}

extern int orfs_open(struct inode *inode, struct file *file);
extern int orfs_release(struct inode *inode, struct file *file);

struct file_operations orfs_dir_fops = {
	.owner = THIS_MODULE,
	.read = generic_read_dir,
	.readdir = orfs_preaddir,
	.open = orfs_open,
	.flush = NULL,
	.release = orfs_release,
	.fsync = NULL,
	.fasync = NULL,
	.lock = NULL,
};
