/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <asm/uaccess.h>

#include "orfs_fs.h"
#include "orfs_private_data.h"
#include "orfs_protocol.h"
#include "orfs_request.h"
#include "orfs_messages.h"

/*************************
 * File Inode Operations *
 *************************/

extern int orfs_setattr(struct dentry *dentry, struct iattr *iattr);
#if ORFS_LINUX_24
extern int orfs_getattr(struct dentry *dentry, struct iattr *iattr);
extern int orfs_inode_revalidate(struct dentry *dentry);
#elif ORFS_LINUX_26
extern int orfs_getattr(struct vfsmount *mnt,
			struct dentry *dentry, struct kstat *kst);
#endif

struct inode_operations orfs_file_iops = {
	.truncate = NULL,
	.permission = NULL,
	.setattr = orfs_setattr,
	.getattr = orfs_getattr,
#if ORFS_LINUX_24
	.revalidate = orfs_inode_revalidate,
#endif
};

/***************************
 * Generic File Operations *
 ***************************/

int
orfs_open(struct inode *inode, struct file *file)
{
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	int error;

	ORFS_DEBUG(REQ, "OPEN called for inode %lld\n",
		   (long long) inode->i_ino);

	if ((error = alloc_file_pd(sbi, file)) < 0)
		return error;

	ORFS_UPDATE_REQ_STAT(sbi, open);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_INODE_REQ;
	req->subtype = ORFS_OPEN_INODE_REQ;
	fill_inode_id_from_pd(inode, &req->target.inode.inode);

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR("Remote call for OPEN failed with error %d\n",
			   -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.inode.node_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			fill_file_pd_from_id(file, 
					     &ans->target.inode.data.open.file);
		      update_inode:
			/* update inode attributes */
			orfs_update_inode_attr(inode, node_attr);
			break;
		case ORFS_NO_FILE_LEFT:
			error = -ENFILE;
			goto update_inode;
		case ORFS_NO_SUCH_INODE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("OPEN returns No Such Inode %lld\n",
					   (long long) inode->i_ino);
			/* drop inode */
			orfs_invalidate_generic_inode(inode);
			error = -EIO;
			break;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

	if (error)
		free_file_pd(sbi, file);

	orfs_put_client_request(sbi, creq);
	return error;
}

int
orfs_release(struct inode *inode, struct file *file)
{
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	struct orfs_file_private_data *file_pd = ORFS_FILE_PD(file);
	int error;

	ORFS_DEBUG(REQ, "CLOSE called for inode %lld\n",
		   (long long) inode->i_ino);

	ORFS_UPDATE_REQ_STAT(sbi, close);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_FILE_REQ;
	req->subtype = ORFS_CLOSE_FILE_REQ;
	fill_file_id_from_pd(file, &req->target.file.file);

	error = remote_call(sbi, sbi->nops, creq);

	if (file_pd) {
		/* release private data */
		free_file_pd(sbi, file);
	}

	if (error < 0) {
		ORFS_ERROR("Remote call for CLOSE failed with error %d\n",
			   -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.file.node_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* update node attributes */
			orfs_update_inode_attr(inode, node_attr);
			break;
		case ORFS_BAD_FILE:
			// for now the state based protocol make the error impossible
			ORFS_ALERT("Bad file %lld\n",
				   (long long) req->target.file.file.id);
			BUG();
			break;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

/*********************************
 * File Specific File Operations *
 *********************************/

extern ssize_t orfs_odirect_pread(struct file *, char *, size_t, loff_t *);
extern ssize_t orfs_odirect_pwrite(struct file *, const char *, size_t,
				   loff_t *);

static ssize_t
orfs_pread(struct file *file, char *buffer, size_t count, loff_t * offset)
{
#ifndef CONFIG_ORFS_FAKEDATA
	struct inode *inode = file->f_dentry->d_inode;
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);

	ORFS_DEBUG(REQ, "PREAD called for inode %lld\n",
		   (long long) inode->i_ino);

	if ((file->f_flags & O_DIRECT) || ORFS_SBI_IS_ALWAYS_O_DIRECT(sbi))
		return orfs_odirect_pread(file, buffer, count, offset);
	else
		return generic_file_read(file, buffer, count, offset);
#else				/* CONFIG_ORFS_FAKEDATA */
	loff_t size = file->f_dentry->d_inode->i_size;
	loff_t off = *offset;
	ssize_t res = off > size ? 0 : off + count > size ? size - off : count;
	*offset += res;
	return res;
#endif				/* CONFIG_ORFS_FAKEDATA */
}

static ssize_t
orfs_pwrite(struct file *file,
	    const char *buffer, size_t count, loff_t * offset)
{
#ifndef CONFIG_ORFS_FAKEDATA
	struct inode *inode = file->f_dentry->d_inode;
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);

	ORFS_DEBUG(REQ, "PWRITE called for inode %lld\n",
		   (long long) inode->i_ino);

	if ((file->f_flags & O_DIRECT) || ORFS_SBI_IS_ALWAYS_O_DIRECT(sbi))
		return orfs_odirect_pwrite(file, buffer, count, offset);
	else
		return generic_file_write(file, buffer, count, offset);
#else				/* CONFIG_ORFS_FAKEDATA */
	*offset += count;
	return count;
#endif				/* CONFIG_ORFS_FAKEDATA */
}

static inline int
orfs_complete_fsync(struct orfs_client_request * creq,
		    struct inode *inode)
{
	struct orfs_request *req = ORFS_CREQ2REQ(creq);
	struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
	struct orfs_getattr_data *node_attr = &ans->target.file.node_attr;
	int error = 0;

	switch (ans->result) {
	case ORFS_SUCCESS:
		/* update node attributes */
		orfs_update_inode_attr(inode, node_attr);
		break;
	case ORFS_BAD_FILE:
		// for now the state based protocol make the error impossible
		ORFS_ALERT("Bad file %lld\n",
			   (long long) req->target.file.file.id);
		BUG();
		break;
	default:
		ORFS_ALERT("Unrecognized error %ld\n", (long) error);
		BUG();
	}

	return error;
}

static int
orfs_fsync(struct file *file, struct dentry *dentry, int datasync)
{
	struct inode *inode = dentry->d_inode;
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	int error;

	ORFS_DEBUG(REQ, "FSYNC called for inode %lld\n",
		   (long long) inode->i_ino);

	ORFS_UPDATE_REQ_STAT(sbi, fsync);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_FILE_REQ;
	req->subtype = ORFS_SYNC_FILE_REQ;
	fill_file_id_from_pd(file, &req->target.file.file);
	req->target.file.data.fsync.datasync = datasync;

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR("Remote call for FSYNC failed with error %d\n",
			   -error);
	} else {
		error = orfs_complete_fsync(creq, inode);
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

#if ORFS_LINUX_26
/*************************************
 * File Specific AIO File Operations *
 *************************************/

#include <linux/aio.h>

extern ssize_t orfs_odirect_aio_read(struct kiocb *, char *, size_t, loff_t);
extern ssize_t orfs_odirect_aio_write(struct kiocb *, const char *, size_t,
				      loff_t);

static ssize_t
orfs_aio_read(struct kiocb *iocb, char * buf, size_t count, loff_t pos)
{
#ifndef CONFIG_ORFS_FAKEDATA
	struct file *file = iocb->ki_filp;
	struct inode *inode = file->f_dentry->d_inode;
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);

	ORFS_DEBUG(REQ, "AIO_PREAD called for inode %lld\n",
		   (long long) inode->i_ino);

	if ((file->f_flags & O_DIRECT) || ORFS_SBI_IS_ALWAYS_O_DIRECT(sbi))
		return orfs_odirect_aio_read(iocb, buf, count, pos);
	else
		return -ENOSYS;
//		return generic_file_aio_read(iocb, buf, count, pos);
#else				/* CONFIG_ORFS_FAKEDATA */
	return count;
#endif				/* CONFIG_ORFS_FAKEDATA */
}

static ssize_t
orfs_aio_write(struct kiocb *iocb, const char * buf, size_t count, loff_t pos)
{
#ifndef CONFIG_ORFS_FAKEDATA
	struct file *file = iocb->ki_filp;
	struct inode *inode = file->f_dentry->d_inode;
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);

	ORFS_DEBUG(REQ, "AIO_PWRITE called for inode %lld\n",
		   (long long) inode->i_ino);

	if ((file->f_flags & O_DIRECT) || ORFS_SBI_IS_ALWAYS_O_DIRECT(sbi))
		return orfs_odirect_aio_write(iocb, buf, count, pos);
	else
		return -ENOSYS;
//		return generic_file_aio_write(iocb, buf, count, pos);
#else				/* CONFIG_ORFS_FAKEDATA */
	loff_t size = file->f_dentry->d_inode->i_size;
	ssize_t res = pos > size ? 0 : pos + count > size ? size - pos : count;
	return res;
#endif				/* CONFIG_ORFS_FAKEDATA */
}

static void
orfs_aio_fsync_callback(struct orfs_sb_info *sbi,
			struct orfs_client_request *creq,
			long *res, long *res2)
{
	struct inode *inode = creq->kiocb->ki_filp->f_dentry->d_inode;
	*res = orfs_complete_fsync(creq, inode);
	*res2 = 0;
}

static int
orfs_aio_fsync(struct kiocb *iocb, int datasync)
{
	struct file *file = iocb->ki_filp;
	struct dentry *dentry = file->f_dentry;
	struct inode *inode = dentry->d_inode;
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	int error;

	ORFS_DEBUG(REQ, "AIO FSYNC called for inode %lld\n",
		   (long long) inode->i_ino);

	ORFS_UPDATE_REQ_STAT(sbi, aiofsync);

	if (!ORFS_SBI_IS_MULTICLIENT(sbi)
	    || !sbi->nops->async_remote_call) {
		/* revert to synchronous */
		return orfs_fsync(file, dentry, datasync);
	}

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_FILE_REQ;
	req->subtype = ORFS_SYNC_FILE_REQ;
	fill_file_id_from_pd(file, &req->target.file.file);
	req->target.file.data.fsync.datasync = datasync;

	creq->kiocb = iocb;
	creq->callback = orfs_aio_fsync_callback;

	ORFS_KIOCB_SBI(iocb) = sbi;
	ORFS_KIOCB_CREQ(iocb) = creq;

	error = sbi->nops->async_remote_call(sbi, creq);
	if (error < 0)
		ORFS_ERROR("Asynchronous remote call for FSYNC failed with error %d\n",
			   -error);

	return -EIOCBQUEUED;
}
#endif				/* ORFS_LINUX_26 */

/*******************
 * File Operations *
 *******************/

struct file_operations orfs_file_fops = {
	.owner = THIS_MODULE,
	.llseek = NULL,
	.read = orfs_pread,
	.write = orfs_pwrite,
	.ioctl = NULL,
	.mmap = NULL,
	.open = orfs_open,
	.flush = NULL,
	.release = orfs_release,
	.fsync = orfs_fsync,
	.fasync = NULL,
	.lock = NULL,
	.readv = NULL,
	.writev = NULL,
	.sendpage = NULL,
#if ORFS_LINUX_26
	.aio_read = orfs_aio_read,
	.aio_write = orfs_aio_write,
	.aio_fsync = orfs_aio_fsync,
#endif
};
