/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/list.h>
#include <linux/spinlock.h>

#include "orfs_fs.h"
#include "orfs_request.h"
#include "orfs_private_data.h"
#include "orfs_proc.h"

/*********************************
 * Linux 2.4 Filesystem mounting *
 *********************************/
#if ORFS_LINUX_24

DECLARE_FSTYPE(orfs_fs_type, "orfs", orfs_read_super, 0 && FS_ODD_RENAME);

/*********************************
 * Linux 2.6 Filesystem mounting *
 *********************************/
#elif ORFS_LINUX_26

static struct super_block *
orfs_get_sb(struct file_system_type *fs_type,
	    int flags, const char *dev_name, void *data)
{
	return get_sb_nodev(fs_type, flags, data, orfs_fill_super);
}

struct file_system_type orfs_fs_type = {
	.name = "orfs",
	.fs_flags = 0 && FS_ODD_RENAME,
	.get_sb = orfs_get_sb,
	.kill_sb = generic_shutdown_super,
	.owner = THIS_MODULE,
};

#endif				/* ORFS_LINUX_?? */

/*****************************
 * Filesystem initialization *
 *****************************/

/* ORFS sbi list */
struct list_head orfs_sbi_list = LIST_HEAD_INIT(orfs_sbi_list);
spinlock_t orfs_sbi_list_lock = SPIN_LOCK_UNLOCKED;

/* private data caches */
kmem_cache_t *orfs_file_cachep = NULL;
kmem_cache_t *orfs_dentry_cachep = NULL;
kmem_cache_t *orfs_inode_cachep = NULL;

/* client request and answer caches */
kmem_cache_t *orfs_creq_cachep;
kmem_cache_t *orfs_ans_cachep;

int
orfs_init_fs(void)
{
	int err = -ENOMEM;

	orfs_file_cachep = kmem_cache_create("orfs_file",
					     sizeof (struct
						     orfs_file_private_data),
					     0, 0, NULL, NULL);
	if (!orfs_file_cachep)
		goto out;

	orfs_dentry_cachep = kmem_cache_create("orfs_dentry",
					       sizeof (struct
						       orfs_dentry_private_data),
					       0, 0, NULL, NULL);
	if (!orfs_dentry_cachep)
		goto out_with_file_cachep;

	orfs_inode_cachep = kmem_cache_create("orfs_inode",
					      sizeof (struct
						      orfs_inode_private_data),
					      0, 0, NULL, NULL);
	if (!orfs_inode_cachep)
		goto out_with_dentry_cachep;

	err = register_filesystem(&orfs_fs_type);
	if (err)
		goto out_with_inode_cachep;

	orfs_creq_cachep = kmem_cache_create("orfs_creq",
					     sizeof(struct orfs_client_request), 0,
					     SLAB_HWCACHE_ALIGN, NULL, NULL);
	if (!orfs_creq_cachep)
		goto out_with_fs;
	orfs_ans_cachep = kmem_cache_create("orfs_ans",
					    sizeof(struct orfs_answer), 0,
					    SLAB_HWCACHE_ALIGN, NULL, NULL);
	if (!orfs_ans_cachep)
		goto out_with_creq_cachep;

#ifndef CONFIG_ORFS_NETMODULES
#ifdef CONFIG_ORFS_TCP
	orfs_tcp_layer_init();
#endif
#ifdef CONFIG_ORFS_GM
	orfs_gm_layer_init();
#endif
#ifdef CONFIG_ORFS_MX
	orfs_mx_layer_init();
#endif
#endif				/* CONFIG_ORFS_NETMODULES */

#ifdef CONFIG_ORFS_PROC
	orfs_proc_init();	/* do not care if it fails */
#endif				/* CONFIG_ORFS_PROC */

	return 0;

      out_with_creq_cachep:
	kmem_cache_destroy(orfs_creq_cachep);
      out_with_fs:
	unregister_filesystem(&orfs_fs_type);
      out_with_inode_cachep:
	kmem_cache_destroy(orfs_inode_cachep);
      out_with_dentry_cachep:
	kmem_cache_destroy(orfs_dentry_cachep);
      out_with_file_cachep:
	kmem_cache_destroy(orfs_file_cachep);
      out:
	return err;
}

void
orfs_exit_fs(void)
{
#ifdef CONFIG_ORFS_PROC
	orfs_proc_exit();
#endif				/* CONFIG_ORFS_PROC */

#ifndef CONFIG_ORFS_NETMODULES
#ifdef CONFIG_ORFS_TCP
	orfs_tcp_layer_exit();
#endif
#ifdef CONFIG_ORFS_GM
	orfs_gm_layer_exit();
#endif
#ifdef CONFIG_ORFS_MX
	orfs_mx_layer_exit();
#endif
#endif				/* CONFIG_ORFS_NETMODULES */

	kmem_cache_destroy(orfs_ans_cachep);
	kmem_cache_destroy(orfs_creq_cachep);

	unregister_filesystem(&orfs_fs_type);

	kmem_cache_destroy(orfs_inode_cachep);
	kmem_cache_destroy(orfs_dentry_cachep);
	kmem_cache_destroy(orfs_file_cachep);
}
