/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __ORFS_FS_H__
#define __ORFS_FS_H__

#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/list.h>
#include <linux/spinlock.h>
#include <asm/semaphore.h>

#include "orfs_mount.h"
#include "orfs_protocol.h"
#include "orfs_network.h"
#include "orfs_stats.h"

extern struct file_system_type orfs_fs_type;

/************************
 * super block handling *
 ************************/

/* file system magic */
#define ORFS_SUPER_MAGIC 0x26111978

/* Needed to get size of buffers.
 * Actually no used since we do not use the page cache.
 */
#define ORFS_FAKE_BLOCKSIZE_BITS 12

/* This blocksize is the prefered size for I/O.
 * The bigger it is, the bigger request will be,
 * the better performance will be.
 * Could be replaced by server's blocksize if defined.
 */
#define ORFS_DEFAULT_BLOCKSIZE 16384

/* fill superblock structures */
#if ORFS_LINUX_24
extern struct super_block *orfs_read_super(struct super_block *, void *, int);
#elif ORFS_LINUX_26
extern int orfs_fill_super(struct super_block *, void *, int);
#endif

/* superblock operations */
extern struct super_operations orfs_sops;

/* ORFS super block list */
extern struct list_head orfs_sbi_list;
extern spinlock_t orfs_sbi_list_lock;

/*****************************
 * super block specific info *
 *****************************/

#define ORFS_SBI_IS_MULTICLIENT(sbi) ((sbi)->opts.specific_flags & ORFS_MULTICLIENT_MASK)
#define ORFS_SBI_IS_NON_CONCURRENT(sbi) ((sbi)->opts.specific_flags & ORFS_NON_CONCURRENT_MASK)
#define ORFS_SBI_IS_ALWAYS_O_DIRECT(sbi) ((sbi)->opts.specific_flags & ORFS_ALWAYS_O_DIRECT_MASK)

#define ORFS_SBI_HAS_CAP(sbi, cap) HAS_ORFS_CAP((sbi)->opts.capabilities, ORFS_ ## cap ## _CAP)

#define ORFS_INIT_SPECIFIC_FLAGS (ORFS_MULTICLIENT_MASK)

struct orfs_sbi_opts {
	/* specific mount options flags */
	unsigned long specific_flags;
	/* specific mount options */
	struct {
		unsigned long dentry;
		unsigned long inode;
		unsigned long page;
	} revalidate_delay;

	/* server specific options */
	orfs_blksize_t blksize;
	orfs_path_len_t symlen;
	orfs_cap_t capabilities;
};

struct orfs_sb_info {
	/* options */
	struct orfs_sbi_opts opts;

#ifdef CONFIG_ORFS_STATS
	/* statistics */
	struct orfs_sbi_stat stat;
#endif				/* CONFIG_ORFS_STATS */

	/* real connection, its operations and layer */
	void *connection;
	connection_status_t connection_status;
	struct orfs_net_operations *nops;
	struct orfs_net_layer *layer;

	/* contains a string describing the server, its port, ... */
	char *server_path;

	union {
		struct _monops_sbi {
			/* semaphore */
			struct semaphore creq_sem;
		} mono;
		struct _multips_sbi {
			/* requests */
        		struct list_head creq_list;
	        	spinlock_t creq_lock;

			/* orfsd kernel thread pid */
			int kt_pid;
			/* completion queue for launch/termination control */
			struct completion kt_done;
			/* set to 1 when the thread has to finish */
			int kt_finish;
		} multi;
	} param;

	/* sbi list */
	struct list_head list;
};

#if ORFS_LINUX_24
#define ORFS_SBI(sb) ((struct orfs_sb_info*)((sb)->u.generic_sbp))
#define SET_ORFS_SBI(sb, sbi) ((sb)->u.generic_sbp = sbi)
#elif ORFS_LINUX_26
#define ORFS_SBI(sb) ((struct orfs_sb_info*)((sb)->s_fs_info))
#define SET_ORFS_SBI(sb, sbi) ((sb)->s_fs_info = sbi)
#endif

/******************************
 * file system initialization *
 ******************************/

extern int orfs_init_fs(void);
extern void orfs_exit_fs(void);

extern int orfs_process_init_answer(struct orfs_sb_info *sbi,
				    struct orfs_answer *ans);

/*************************************
 * inode, dentry and file operations *
 *************************************/

extern struct inode_operations orfs_file_iops;
extern struct file_operations orfs_file_fops;
extern struct address_space_operations orfs_file_aops;

extern struct inode_operations orfs_dir_iops;
extern struct file_operations orfs_dir_fops;

extern struct inode_operations orfs_symlink_iops;

extern struct dentry_operations orfs_dops;

/**********************
 * Specific Functions *
 **********************/

extern struct inode *orfs_iget_with_attr(struct super_block *sb,
					 struct orfs_getattr_data *data);
extern struct inode *orfs_iget_root(struct super_block *sb);

extern int orfs_update_inode_attr(struct inode *inode,
				  struct orfs_getattr_data *attr);

extern int orfs_autofill_symlink_cache(struct inode *inode,
				       const void *old_buffer, size_t path_len);

extern void orfs_kill_dirent_cache(struct file *file);
extern void orfs_kill_symlink_cache(struct inode *inode);

extern void orfs_invalidate_generic_inode(struct inode *inode);
extern void orfs_invalidate_reg_inode(struct inode *inode);
extern void orfs_invalidate_dir_inode(struct inode *inode);
extern void orfs_invalidate_lnk_inode(struct inode *inode);
extern void orfs_invalidate_special_inode(struct inode *inode);
extern void orfs_invalidate_empty_inode(struct inode *inode);

#endif				/* __ORFS_FS_H__ */
