/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __ORFS_NETWORK_H__
#define __ORFS_NETWORK_H__

#include <linux/module.h>
#include <linux/fs.h>
#include <asm/types.h>

struct orfs_client_request;
struct orfs_sb_info;
struct orfs_user_mount_options;

/*********************
 * Connection Status *
 *********************/

typedef enum {
	ORFS_CONNECTION_OK,
	ORFS_CONNECTION_ERROR,
} connection_status_t;

/**********************
 * Network Operations *
 **********************/

struct orfs_net_operations {
	/* Connection to the server, required. */
	int (*connect) (struct orfs_sb_info * sbi,
			struct orfs_user_mount_options * opts);
	/* Disconnection from the server, required. */
	void (*disconnect) (struct orfs_sb_info * sbi);

	/* Remote call, required for multips mount. */
	int (*multi_remote_call) (struct orfs_sb_info * sbi,
				  struct orfs_client_request * creq);
	/* Remote call, required for monops mount. */
	int (*mono_remote_call) (struct orfs_sb_info * sbi,
				 struct orfs_client_request * creq);
	/* Asynchronous remote call. */
	int (*async_remote_call) (struct orfs_sb_info * sbi,
				  struct orfs_client_request * creq);

	/* Kernel thread main loop, required only if
	 * you need a kernel thread to centralize events. */
	void (*kt_loop) (struct orfs_sb_info * sbi);

	/* Wake up network waiters, required. */
	void (*wake_up_waiters) (struct orfs_sb_info * sbi);
};

#define remote_call(sbi, nops, creq) \
ORFS_SBI_IS_MULTICLIENT(sbi)         \
? nops->multi_remote_call(sbi, creq) \
: nops->mono_remote_call(sbi, creq)

/******************************
 * Network Layer Registration *
 ******************************/

struct orfs_net_layer {
	char *name;
	struct orfs_net_operations *nops;
	atomic_t count;
	struct module *owner;
	struct list_head layers;
};

#define DECLARE_ORFS_NET_LAYER(var,var_name,var_nops)	\
struct orfs_net_layer var = {				\
	.name = var_name,				\
	.nops = var_nops,				\
	.owner = THIS_MODULE,				\
}

extern int orfs_register_net_layer(struct orfs_net_layer *layer);
extern int orfs_unregister_net_layer(struct orfs_net_layer *layer);
extern int orfs_get_net_layers_list(char *page);

/******************
 * Builtin Layers *
 ******************/

#ifndef CONFIG_ORFS_NETMODULES
#ifdef CONFIG_ORFS_TCP
extern int orfs_tcp_layer_init(void);
extern void orfs_tcp_layer_exit(void);
#endif
#ifdef CONFIG_ORFS_GM
extern int orfs_gm_layer_init(void);
extern void orfs_gm_layer_exit(void);
#endif
#ifdef CONFIG_ORFS_MX
extern int orfs_mx_layer_init(void);
extern void orfs_mx_layer_exit(void);
#endif
#endif				/* CONFIG_ORFS_NETMODULES */


#endif				/* __ORFS_NETWORK_H__ */
