/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __ORFS_PRIVATE_DATA_H__
#define __ORFS_PRIVATE_DATA_H__

#include <linux/fs.h>

#include "orfs_messages.h"
#include "orfs_portable.h"
#include "orfs_stats.h"

#define ORFS_UPDATE_PD_TIME(pd) KERNEL2ORFS_TIME(CURRENT_TIME, (pd)->last_update);

/***********
 * File PD *
 ***********/

struct orfs_file_private_data {
	orfs_file_id_t file_id;
	void *dirent_cache;
};

#define ORFS_FILE_PD(file) ((struct orfs_file_private_data*)(file)->private_data)
#define SET_ORFS_FILE_PD(file, pd) ((file)->private_data = pd)

extern kmem_cache_t *orfs_file_cachep;

extern inline int
alloc_file_pd(struct orfs_sb_info *sbi, struct file *file)
{
	struct orfs_file_private_data *file_pd;
	file_pd = kmem_cache_alloc(orfs_file_cachep, SLAB_KERNEL);
	if (!file_pd) {
		ORFS_DEBUG(PD, "Failed to allocate file private data\n");
		return -ENOMEM;
	}
	ORFS_INC_ALLOC_STAT(sbi, file);

	file_pd->dirent_cache = NULL;
	SET_ORFS_FILE_PD(file, file_pd);
	return 0;
}

extern inline void
free_file_pd(struct orfs_sb_info *sbi, struct file *file)
{
	struct orfs_file_private_data *file_pd = ORFS_FILE_PD(file);
	if (!file_pd) {
		ORFS_NOTICE("No private data in file to release\n");
	} else {
		orfs_kill_dirent_cache(file);
		ORFS_DEC_ALLOC_STAT(sbi, file);
		kmem_cache_free(orfs_file_cachep, file_pd);
	}
}

extern inline void
fill_file_pd_from_id(struct file *file, struct orfs_file_id *open_file)
{
	struct orfs_file_private_data *file_pd = ORFS_FILE_PD(file);
	if (!file_pd) {
		ORFS_ALERT("No private data in file\n");
		BUG();
	}
	file_pd->file_id = open_file->id;
}

extern inline void
fill_file_id_from_pd(struct file *file, struct orfs_file_id *file_id)
{
	struct orfs_file_private_data *file_pd = ORFS_FILE_PD(file);
	if (!file_pd) {
		ORFS_ALERT("No private data in file\n");
		BUG();
	}
	file_id->id = file_pd->file_id;
}

/*************
 * Dentry PD *
 *************/

struct orfs_dentry_private_data {
	orfs_time_t last_update;
};

#define ORFS_DENTRY_PD(dentry) ((struct orfs_dentry_private_data*)(dentry)->d_fsdata)
#define SET_ORFS_DENTRY_PD(dentry, pd) ((dentry)->d_fsdata = pd)

#define ORFS_UPDATE_DENTRY_PD_TIME(inode) ORFS_UPDATE_PD_TIME(ORFS_DENTRY_PD(inode))

extern kmem_cache_t *orfs_dentry_cachep;

extern inline int
alloc_dentry_pd(struct orfs_sb_info *sbi, struct dentry *dentry)
{
	struct orfs_dentry_private_data *dentry_pd;
	dentry_pd = kmem_cache_alloc(orfs_dentry_cachep, SLAB_KERNEL);
	if (!dentry_pd) {
		ORFS_DEBUG(PD, "Failed to allocate dentry private data\n");
		return -ENOMEM;
	}
	ORFS_INC_ALLOC_STAT(sbi, dentry);

	ORFS_UPDATE_PD_TIME(dentry_pd);
	SET_ORFS_DENTRY_PD(dentry, dentry_pd);
	return 0;
}

extern inline void
free_dentry_pd(struct orfs_sb_info *sbi, struct dentry *dentry)
{
	struct orfs_dentry_private_data *dentry_pd = ORFS_DENTRY_PD(dentry);
	if (!dentry_pd) {
		ORFS_NOTICE("No private data in dentry to release\n");
	} else {
		ORFS_DEC_ALLOC_STAT(sbi, dentry);
		kmem_cache_free(orfs_dentry_cachep, dentry_pd);
	}
}

/************
 * Inode PD *
 ************/

struct orfs_inode_private_data {
	orfs_time_t last_update;
	orfs_sb_t sb_id;
	void *symlink_cache;
};

#define ORFS_INODE_PD(inode) ((struct orfs_inode_private_data*)(inode)->u.generic_ip)
#define SET_ORFS_INODE_PD(inode, pd) ((inode)->u.generic_ip = pd)

#define ORFS_UPDATE_INODE_PD_TIME(inode) ORFS_UPDATE_PD_TIME(ORFS_INODE_PD(inode))

extern kmem_cache_t *orfs_inode_cachep;

extern inline int
alloc_inode_pd(struct orfs_sb_info *sbi, struct inode *inode)
{
	struct orfs_inode_private_data *inode_pd;
	
	inode_pd = kmem_cache_alloc(orfs_inode_cachep, SLAB_KERNEL);
	if (!inode_pd) {
		ORFS_DEBUG(PD, "Failed to allocate inode private data\n");
		return -ENOMEM;
	}
	ORFS_INC_ALLOC_STAT(sbi, inode);

	ORFS_UPDATE_PD_TIME(inode_pd);
	inode_pd->symlink_cache = NULL;
	SET_ORFS_INODE_PD(inode, inode_pd);
	return 0;
}

extern inline void
free_inode_pd(struct orfs_sb_info *sbi, struct inode *inode)
{
	struct orfs_inode_private_data *inode_pd = ORFS_INODE_PD(inode);
	if (!inode_pd) {
		ORFS_NOTICE("No private data in inode to release\n");
	} else {
		orfs_kill_symlink_cache(inode);
		ORFS_DEC_ALLOC_STAT(sbi, inode);
		kmem_cache_free(orfs_inode_cachep, inode_pd);
	}
}

extern inline void
fill_inode_pd_from_getattr(struct inode *inode,
			   struct orfs_getattr_data *getattr)
{
	struct orfs_inode_private_data *inode_pd = ORFS_INODE_PD(inode);
	if (!inode_pd) {
		ORFS_ALERT("No private data in inode\n");
		BUG();
	}
	inode->i_ino = getattr->id.ino;
	inode_pd->sb_id = getattr->id.sb;
}

extern inline void
fill_inode_id_from_pd(struct inode *inode, struct orfs_inode_id *inode_id)
{
	struct orfs_inode_private_data *inode_pd;

	inode_id->ino = inode->i_ino;

	if (!inode->i_mode)
		/* read_inode marks inode as initializing through mode = 0 */
		return;

	inode_pd = ORFS_INODE_PD(inode);
	if (!inode_pd) {
		ORFS_ALERT("No private data in inode\n");
		BUG();
	}
	inode_id->sb = inode_pd->sb_id;
}

#endif				/* __ORFS_PRIVATE_DATA_H__ */
