/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __ORFS_REQUEST_H__
#define __ORFS_REQUEST_H__

#include <linux/sched.h>
#include <linux/list.h>
#include <linux/spinlock.h>
#if ORFS_LINUX_26
#include <linux/aio.h>
#endif

/* Request Status */
enum {
	ORFS_CREQ_REQ2RECV_BIT = 0,
	ORFS_CREQ_BUF2RECV_BIT,
	ORFS_CREQ_ORPHAN_BIT,
	ORFS_CREQ_ADOPTED_BIT,
};
#define ORFS_CREQ_REQ2RECV (1<<ORFS_CREQ_REQ2RECV_BIT)
#define ORFS_CREQ_BUF2RECV (1<<ORFS_CREQ_BUF2RECV_BIT)
#define ORFS_CREQ_ORPHAN (1<<ORFS_CREQ_ORPHAN_BIT)
#define ORFS_CREQ_ADOPTED (1<<ORFS_CREQ_ADOPTED_BIT)
typedef unsigned char orfs_request_status_t;

#define ORFS_CREQ_NEED_REQ(creq) ((creq)->status |= ORFS_CREQ_REQ2RECV)
#define ORFS_CREQ_NEED_BUF(creq) ((creq)->status |= ORFS_CREQ_BUF2RECV)
#define ORFS_CREQ_RECV_REQ(creq) ((creq)->status ^= ORFS_CREQ_REQ2RECV)
#define ORFS_CREQ_RECV_BUF(creq) ((creq)->status ^= ORFS_CREQ_BUF2RECV)
#define ORFS_CREQ_IS_ORPHAN(creq) ((creq)->status |= ORFS_CREQ_ORPHAN)
#define ORFS_CREQ_IS_ADOPTED(creq) ((creq)->status ^= (ORFS_CREQ_ADOPTED|ORFS_CREQ_ORPHAN))

#define IS_ORFS_CREQ_WAITING(creq) (((creq)->status) & (ORFS_CREQ_REQ2RECV|ORFS_CREQ_BUF2RECV))
#define IS_ORFS_CREQ_ORPHAN(creq) (((creq)->status) & ORFS_CREQ_ORPHAN)
#define IS_ORFS_CREQ_ADOPTED(creq) (((creq)->status) & ORFS_CREQ_ADOPTED)

/* Buffer flags */
enum {
	ORFS_BUF_PAGE,		/* a struct page * */
	ORFS_BUF_KMALLOC,	/* a kernel virtual address that can be translated through virt_to_phys */
	ORFS_BUF_USER_ANY,	/* any user virtual address */
};
typedef unsigned char orfs_buffer_flags_t;

/*******************
 * Request/Answers *
 *******************/

#include <linux/slab.h>

extern kmem_cache_t *orfs_creq_cachep;
extern kmem_cache_t *orfs_ans_cachep;

typedef void (*orfs_creq_callback_t)(struct orfs_sb_info *, struct orfs_client_request *, long *, long *);

struct orfs_client_request {
	/* request */
	struct orfs_request req;
	void *in_buf;
	size_t in_buf_len;
	off_t in_page_off;
	orfs_buffer_flags_t in_buf_flags;

	/* answer */
	struct orfs_answer *ans;
	void *out_buf;
	size_t out_buf_len;
	off_t out_page_off;
	orfs_buffer_flags_t out_buf_flags;
	
#if ORFS_LINUX_26
	/* kiocb for async request */
	struct kiocb *kiocb;
	/* callback */
	orfs_creq_callback_t callback;
#endif
	/* request status */
	orfs_request_status_t status;
	/* waitqueue_head */
	wait_queue_head_t queue;
	/* sbi list */
	struct list_head list;
	/* netlayer private data */
	void *net_pd;
};

#define ORFS_CREQ2REQ(creq) (&(creq)->req)
#define ORFS_CREQ2ANS(creq) ((creq)->ans)

#if ORFS_LINUX_26
struct orfs_kiocb_private {
	struct orfs_sb_info *sbi;
	struct orfs_client_request *creq;
};

static inline struct orfs_kiocb_private *
orfs_kiocb_to_private(struct kiocb *kiocb)
{
	/* dirty... but that's what they do in include/net/sock.h */
	BUG_ON(sizeof(struct orfs_kiocb_private) > KIOCB_PRIVATE_SIZE);
	return (struct orfs_kiocb_private *) kiocb->private;
}

#define ORFS_KIOCB_SBI(kiocb) orfs_kiocb_to_private(kiocb)->sbi
#define ORFS_KIOCB_CREQ(kiocb) orfs_kiocb_to_private(kiocb)->creq
#endif /* ORFS_LINUX_26 */

/**************************
 * Externalized Functions *
 **************************/

extern int orfs_set_connection_error(struct orfs_sb_info *sbi);
extern void orfs_killall_client_requests(struct orfs_sb_info *sbi);

extern struct orfs_client_request *orfs_get_client_request(struct orfs_sb_info *sbi);
extern void orfs_put_client_request(struct orfs_sb_info *sbi,
				    struct orfs_client_request *creq);

extern struct orfs_answer *orfs_get_client_answer(struct orfs_sb_info
						  *sbi);
extern void orfs_put_client_answer(struct orfs_sb_info *sbi,
				   struct orfs_answer *ans);

extern int orfs_launch_orfsd(struct orfs_sb_info *sbi);
extern void orfs_finish_orfsd(struct orfs_sb_info *sbi);

#endif				/* __ORFS_REQUEST_H__ */
