/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/mm.h>

#include "orfs_fs.h"
#include "orfs_portable.h"
#include "orfs_private_data.h"
#include "orfs_mount.h"
#include "orfs_request.h"
#include "orfs_messages.h"

/****************************************
 * Updating and initializing attributes *
 ****************************************/

static inline int
__orfs_init_attr(struct inode *inode, struct orfs_getattr_data *attr)
{
	int error;
	umode_t mode = attr->mode;

	ORFS_DEBUG(FS, "Init Attr of inode %lld\n", (long long) inode->i_ino);

	error = alloc_inode_pd(ORFS_SBI(inode->i_sb), inode);
	if ( error < 0 ) {
		ORFS_ERROR("Inode %lld initializing failed\n",
			   (long long) inode->i_ino);
		orfs_invalidate_empty_inode(inode);
		return error;
	}

	/* fill private data */
	fill_inode_pd_from_getattr(inode, attr);

	/* fix operations and rdev for special inode */
	if (S_ISREG(mode)) {
		inode->i_op = &orfs_file_iops;
		inode->i_fop = &orfs_file_fops;
		inode->i_mapping->a_ops = &orfs_file_aops;
	} else if (S_ISDIR(mode)) {
		inode->i_op = &orfs_dir_iops;
		inode->i_fop = &orfs_dir_fops;
	} else if (S_ISLNK(mode)) {
		inode->i_op = &orfs_symlink_iops;
	} else {
		init_special_inode(inode, mode, ORFS2KERNEL_DEV(attr->rdev));
	}

	return 0;
}

static inline int
__orfs_check_attr(struct inode *inode, struct orfs_getattr_data *attr)
{
	umode_t new_mode = attr->mode;

	ORFS_DEBUG(FS, "Check Attr of inode %lld\n", (long long) inode->i_ino);

	if ( is_bad_inode(inode) )
		return 0;

	/* check mode */
	if ((S_IFMT & inode->i_mode) != (S_IFMT & new_mode)) {
		ORFS_DEBUG(FS, "Inode type changed\n");
		// invalidate/drop it
		return -EIO;
	}

	/* check type specific fields */

	if (S_ISCHR(inode->i_mode) || S_ISBLK(inode->i_mode)) {
		/* check rdev of chr/blk nodes */
		orfs_dev_t old_rdev = KERNEL2ORFS_DEV(inode->i_rdev);
		if (old_rdev != attr->rdev) {
			ORFS_DEBUG(FS, "Inode rdev changed\n");
			// invalidate/drop it
			return -EIO;
		}

	} else if (S_ISLNK(inode->i_mode)) {
		/* check mtime of symlink to know whether it was overwritten */
		orfs_time_t old_mtime;
		KERNEL2ORFS_TIME(inode->i_mtime, old_mtime);
		if (attr->mtime != old_mtime) {
			ORFS_DEBUG(FS, "Inode symlink path changed\n");
			// invalidate/drop it
			return -EIO;
		}
	}
	return 0;
}

static inline void
__orfs_update_attr(struct inode *inode, struct orfs_getattr_data *attr)
{
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	orfs_time_t old_mtime;
	mode_t mode;

	inode->i_mode = mode = attr->mode;
	inode->i_uid = attr->uid;
	inode->i_gid = attr->gid;
	inode->i_nlink = attr->nlink;

	if (!ORFS_SBI_IS_NON_CONCURRENT(sbi)) {
		/* If the file was reduced, vmtruncate it */
		if (S_ISREG(mode) && attr->size < inode->i_size) {
			ORFS_DEBUG(FS, "vmtruncating of inode %lld at %lld\n",
				   (long long) inode->i_ino,
				   (long long) attr->size);
			vmtruncate(inode, (loff_t) attr->size);
		}
		/* If the file was modified, invalidate its pages */
		KERNEL2ORFS_TIME(inode->i_mtime, old_mtime);
		if (S_ISREG(mode) && attr->mtime > old_mtime) {
			ORFS_DEBUG(FS, "Invalidating inode %lld pages\n",
				   (long long) inode->i_ino);
#if ORFS_LINUX_26
			invalidate_inode_pages(inode->i_mapping);
#elif ORFS_LINUX_24
			invalidate_inode_pages(inode);
#endif
		}
	}

	/* And now update checked fields */
	inode->i_size = attr->size;
	inode->i_blocks = attr->blocks;
	ORFS2KERNEL_TIME(attr->atime, inode->i_atime);
	ORFS2KERNEL_TIME(attr->mtime, inode->i_mtime);
	ORFS2KERNEL_TIME(attr->ctime, inode->i_ctime);

	// use version for anything ??

	ORFS_UPDATE_INODE_PD_TIME(inode);
}

int
orfs_update_inode_attr(struct inode *inode, struct orfs_getattr_data *attr)
{
	int error;

	/* protect private data allocating */
	up(&inode->i_sem);

	if (!inode->i_mode) {
		/* read_inode sets mode to 0 to mark it as "initializing".
		 * so we initialize it.
		 */
		if ((error =__orfs_init_attr(inode, attr)) < 0)
			/* what to do ? */
			goto error;
	} else {
		/* already initialized inode.
		 * check that it did not change.
		 */
		if ((error = __orfs_check_attr(inode, attr)) < 0)
			goto error;
	}

	/* update usual attributes */
	__orfs_update_attr(inode, attr);

      error:
	/* release protection */
	down(&inode->i_sem);
	return 0;
}

/****************************
 * Generic Inode Operations *
 ****************************/

static inline int
__orfs_inode_getattr(struct inode *inode)
{
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	int error;

	ORFS_DEBUG(REQ, "Real GETATTR called for inode %lld\n",
		   (long long) inode->i_ino);

	ORFS_UPDATE_REQ_STAT(sbi, getattr);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_INODE_REQ;
	req->subtype = ORFS_GETATTR_INODE_REQ;
	fill_inode_id_from_pd(inode, &req->target.inode.inode);

	error = remote_call(sbi, sbi->nops, creq);

	if (error) {
		ORFS_ERROR
		    ("Remote call for Real GETATTR failed with error %d\n",
		     -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.inode.node_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* update whole inode fields */
			orfs_update_inode_attr(inode, node_attr);
			break;
		case ORFS_NO_SUCH_INODE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("GETATTR returns No Such Inode %lld\n",
					   (long long) inode->i_ino);
			/* drop inode */
			orfs_invalidate_generic_inode(inode);
			error = -EIO;
			break;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

int
orfs_inode_revalidate(struct dentry *dentry)
{
	struct inode *inode = dentry->d_inode;
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	orfs_time_t current_time, inode_time;
	int res;

	if (ORFS_SBI_IS_NON_CONCURRENT(sbi))
		return 0;

	ORFS_DEBUG(REQ, "REVALIDATE called on inode %lld\n",
		   (long long) inode->i_ino);

	ORFS_UPDATE_REVAL_STAT(sbi, inode, call);

	KERNEL2ORFS_TIME(CURRENT_TIME, current_time);
	inode_time = ORFS_INODE_PD(inode)->last_update;

	/* An inode is supposed to be valid when
	 * it has been updated recently
	 */
	if (current_time <= inode_time + sbi->opts.revalidate_delay.inode)
		return 0;

	ORFS_UPDATE_REVAL_STAT(sbi, inode, remote);

	res = __orfs_inode_getattr(inode);

	if (res)
		ORFS_UPDATE_REVAL_STAT(sbi, inode, failed);

	/* returns 0 when revalidating succeeded, -error else */
	return res;
}

#if ORFS_LINUX_26

int
orfs_getattr(struct vfsmount *mnt, struct dentry *dentry, struct kstat *kst)
{
	struct inode *inode = dentry->d_inode;
	int res;

	ORFS_DEBUG(REQ, "GETATTR called on inode %lld\n",
		   (long long) inode->i_ino);

	res = orfs_inode_revalidate(dentry);
	if (res < 0)
		return res;

	generic_fillattr(inode, kst);

	return 0;
}

#elif ORFS_LINUX_24

int
orfs_getattr(struct dentry *dentry, struct iattr *iattr)
{
	ORFS_ERROR("Getattr currently not supported for 2.4\n");

	return -ENOSYS;
}

#endif				/* ORFS_LINUX_?? */

int
orfs_setattr(struct dentry *dentry, struct iattr *iattr)
{
	struct inode *inode = dentry->d_inode;
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	int error;

	if (!(iattr->ia_valid & ~ATTR_CTIME))
		return 0;

	ORFS_DEBUG(REQ, "SETATTR called for inode %lld with valid flags %d\n",
		   (long long) inode->i_ino, iattr->ia_valid);

	error = inode_change_ok(inode, iattr);
	if (error)
		return error;

	ORFS_UPDATE_REQ_STAT(sbi, setattr);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_INODE_REQ;
	req->subtype = ORFS_SETATTR_INODE_REQ;
	fill_inode_id_from_pd(inode, &req->target.inode.inode);

	req->target.inode.data.setattr.new_attr.valid = iattr->ia_valid;
	req->target.inode.data.setattr.new_attr.mode = iattr->ia_mode;
	req->target.inode.data.setattr.new_attr.uid = iattr->ia_uid;
	req->target.inode.data.setattr.new_attr.gid = iattr->ia_gid;
	req->target.inode.data.setattr.new_attr.size = iattr->ia_size;
	KERNEL2ORFS_TIME(iattr->ia_atime,
			 req->target.inode.data.setattr.new_attr.atime);
	KERNEL2ORFS_TIME(iattr->ia_mtime,
			 req->target.inode.data.setattr.new_attr.mtime);

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR("Remote call for SETATTR failed with error %d\n",
			   -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.inode.node_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			// is that really needed ? following update will do the same ?
			error = inode_setattr(inode, iattr);
			error = 0;
			// only truncate can return EFBIG, should be
			// caught by server first ?
			orfs_update_inode_attr(inode, node_attr);
			break;
		case ORFS_NO_SUCH_INODE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("SETATTR returns No Such Inode %lld\n",
					   (long long) inode->i_ino);
		       invalidate:
			/* drop inode */
			orfs_invalidate_generic_inode(inode);
			error = -EIO;
			break;
		case ORFS_BAD_INODE_TYPE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("SETATTR returns Bad Inode %lld mode %lld\n",
					   (long long) inode->i_ino, (long long) node_attr->mode);
			goto invalidate;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

/********************
 * Creating objects *
 ********************/

struct inode *
orfs_iget_with_attr(struct super_block *sb, struct orfs_getattr_data *attr)
{
	struct inode *inode;
	unsigned long ino = attr->id.ino;

	ORFS_DEBUG(FS, "IGET WITH ATTR called for ino %ld\n", ino);

	inode = iget(sb, ino);
	if (!inode)
		return NULL;

	if (is_bad_inode(inode))
		goto error_with_inode;

	if (orfs_update_inode_attr(inode, attr))
		goto error_with_inode;

	return inode;

      error_with_inode:
	iput(inode);
	return NULL;
}

struct inode *
orfs_iget_root(struct super_block *sb)
{
	struct inode *inode;

	ORFS_DEBUG(FS, "IGET ROOT called\n");

	inode = iget(sb, ORFS_ROOT_INO);
	if (!inode)
		return NULL;

	if (is_bad_inode(inode))
		goto error_with_inode;

	if (__orfs_inode_getattr(inode))
		goto error_with_inode;

	return inode;

      error_with_inode:
	iput(inode);
	return NULL;
}

/************************
 * Invalidating objects *
 ************************/

void
orfs_invalidate_empty_inode(struct inode *inode)
{
	ORFS_DEBUG(FS, "invalidate empty inode %lld\n", (long long) inode->i_ino);

	make_bad_inode(inode);
#if ORFS_LINUX_24
	remove_inode_hash(inode);
#endif
}

void
orfs_invalidate_reg_inode(struct inode *inode)
{
	ORFS_DEBUG(FS, "invalidate reg inode %lld\n", (long long) inode->i_ino);

	if (!S_ISREG(inode->i_mode)) {
		ORFS_ALERT("bad dir inode type %ld\n", (long) inode->i_mode);
		BUG();
	}

	// invalidate regular inode specific things: page cache ?

	orfs_invalidate_empty_inode(inode);
}

void
orfs_invalidate_dir_inode(struct inode *inode)
{
	ORFS_DEBUG(FS, "invalidate dir inode %lld\n", (long long) inode->i_ino);

	if (!S_ISDIR(inode->i_mode)) {
		ORFS_ALERT("bad reg inode type %ld\n", (long) inode->i_mode);
		BUG();
	}

	// invalidate directory specific things: dirent cache ?

	orfs_invalidate_empty_inode(inode);
}

void
orfs_invalidate_lnk_inode(struct inode *inode)
{
	ORFS_DEBUG(FS, "invalidate lnk inode %lld\n", (long long) inode->i_ino);

	if (!S_ISLNK(inode->i_mode)) {
		ORFS_ALERT("bad lnk inode type %ld\n", (long) inode->i_mode);
		BUG();
	}

	// invalidate symlink specific things: symlink cache ?

	orfs_invalidate_empty_inode(inode);
}

void
orfs_invalidate_special_inode(struct inode *inode)
{
	ORFS_DEBUG(FS, "invalidate special inode %lld\n",
		   (long long) inode->i_ino);

	if (!S_ISCHR(inode->i_mode) && !S_ISBLK(inode->i_mode)
	    && !S_ISFIFO(inode->i_mode) && !S_ISSOCK(inode->i_mode)) {
		ORFS_ALERT("bad special inode type %ld\n",
			   (long) inode->i_mode);
		BUG();
	}

	// nothing specific to invalidate ?

	orfs_invalidate_empty_inode(inode);
}

void
orfs_invalidate_generic_inode(struct inode *inode)
{
	ORFS_DEBUG(FS, "invalidate generic inode %lld\n",
		   (long long) inode->i_ino);

	switch (inode->i_mode & S_IFMT) {
	case S_IFDIR:
		orfs_invalidate_dir_inode(inode);
		break;
	case S_IFREG:
		orfs_invalidate_reg_inode(inode);
		break;
	case S_IFLNK:
		orfs_invalidate_lnk_inode(inode);
		break;
	case S_IFCHR:
	case S_IFBLK:
	case S_IFIFO:
	case S_IFSOCK:
		orfs_invalidate_special_inode(inode);
		break;
	default:
		ORFS_ERROR("trying to invalidate non-orfs inode %lld\n",
			   (long long) inode->i_ino);
	}

}
