/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/fs.h>

#include "orfs_fs.h"
#include "orfs_private_data.h"
#include "orfs_protocol.h"
#include "orfs_request.h"
#include "orfs_messages.h"

/************************
 * Dir Inode Operations *
 ************************/

struct dentry *
orfs_lookup(struct inode *dir, struct dentry *dentry
#if ORFS_LINUX_26
	    , struct nameidata *nd
#endif
    )
{
	struct super_block *sb = dir->i_sb;
	struct orfs_sb_info *sbi = ORFS_SBI(sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	orfs_name_len_t name_len = dentry->d_name.len;
	int error;
	struct dentry *res = NULL;
	struct inode *inode;

	ORFS_DEBUG(REQ, "LOOKUP called for entry %s in dir inode %lld\n",
		   dentry->d_name.name, (long long) dir->i_ino);

	/* try to alloc private data */
	error = alloc_dentry_pd(sbi, dentry);
	if (error < 0)
		return ERR_PTR(error);
	/* we now can safely program d_ops */
	dentry->d_op = dentry->d_parent->d_op;

	if (dentry->d_name.len > ORFS_NAME_LEN)
		return ERR_PTR(-ENAMETOOLONG);

	ORFS_UPDATE_REQ_STAT(sbi, lookup);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_DIR_REQ;
	req->subtype = ORFS_LOOKUP_DIR_REQ;
	fill_inode_id_from_pd(dir, &req->target.dir.dir);
	req->target.dir.name_len = name_len;
	memcpy(req->target.dir.name, dentry->d_name.name, name_len + 1);

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR("Remote call for LOOKUP failed with error %d\n",
			   -error);
		res = ERR_PTR(error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.dir.data.lookup.node_attr;
		struct orfs_getattr_data *father_attr = &ans->target.dir.father_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* create child */
			inode = orfs_iget_with_attr(dir->i_sb, node_attr);
			if (!inode || is_bad_inode(inode))
				error = -EACCES;
			else
				d_add(dentry, inode);
		      update_dir:
			/* update dentry */
			ORFS_UPDATE_DENTRY_PD_TIME(dentry);
			/* update dir attributes */
			orfs_update_inode_attr(dir, father_attr);
			break;
		case ORFS_NO_SUCH_DIR_ENTRY:
			/* make dentry negative */
			d_add(dentry, NULL);
			goto update_dir;
		case ORFS_NO_SUCH_INODE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("LOOKUP returns No Such Inode %lld\n",
					   (long long) dir->i_ino);
		       invalidate:
			/* drop dir inode */
			orfs_invalidate_dir_inode(dir);
			error = -EIO;
			break;
		case ORFS_BAD_INODE_TYPE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("LOOKUP returns Bad Inode %lld mode %lld\n",
					   (long long) dir->i_ino, (long long) father_attr->mode);
			goto invalidate;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
		res = ERR_PTR(error);
	}

	orfs_put_client_request(sbi, creq);
	return res;
}

int
orfs_mkdir(struct inode *dir, struct dentry *dentry, int mode)
{
	struct orfs_sb_info *sbi = ORFS_SBI(dir->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	orfs_name_len_t name_len = dentry->d_name.len;
	orfs_uid_t uid;
	orfs_gid_t gid;
	int error;
	struct inode *inode;

	ORFS_DEBUG(REQ,
		   "MKDIR called for entry %s in dir inode %lld with mode %x\n",
		   dentry->d_name.name, (long long) dir->i_ino, mode);

	ORFS_UPDATE_REQ_STAT(sbi, mkdir);

	/* deal with uid */
	uid = current->fsuid;
	/* deal with gid */
	if (dir->i_mode & S_ISGID) {
		mode |= S_ISGID;
		gid = dir->i_gid;
	} else
		gid = current->fsgid;

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_DIR_REQ;
	req->subtype = ORFS_MKDIR_DIR_REQ;
	fill_inode_id_from_pd(dir, &req->target.dir.dir);
	req->target.dir.name_len = name_len;
	memcpy(req->target.dir.name, dentry->d_name.name, name_len + 1);
	req->target.dir.data.create.mode = mode;
	req->target.dir.data.create.uid = uid;
	req->target.dir.data.create.gid = gid;

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR("Remote call for MKDIR failed with error %d\n",
			   -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.dir.data.create.new_node_attr;
		struct orfs_getattr_data *father_attr = &ans->target.dir.father_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* create child */
			inode = orfs_iget_with_attr(dir->i_sb, node_attr);
			if (!inode || is_bad_inode(inode)) {
				// ugh... server still has it...
				d_drop(dentry);
				error = -ENOMEM;
			} else {
				d_instantiate(dentry, inode);
				// mark inode dirty ???

				ORFS_DEBUG(FS, "MKDIR returns new inode %lld\n",
					   (long long) node_attr->id.ino);
			}
		      update_dir:
			/* update dentry */
			ORFS_UPDATE_DENTRY_PD_TIME(dentry);
			/* update dir attributes */
			orfs_update_inode_attr(dir, father_attr);
			break;
		case ORFS_NO_SPACE_LEFT:
			error = -ENOMEM;
			goto update_dir;
		case ORFS_EXISTING_DIR_ENTRY:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("MKDIR returns Existing Dir %lld Entry %s\n",
					   (long long) dir->i_ino, dentry->d_name.name);
			// update dirent/dentry cache of new dir
			error = -EEXIST;
			goto update_dir;
		case ORFS_NO_SUCH_INODE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("MKDIR returns No Such Inode %lld\n",
					   (long long) dir->i_ino);
		       invalidate:
			/* drop old dir inode */
			orfs_invalidate_dir_inode(dir);
			error = -EIO;
			break;
		case ORFS_BAD_INODE_TYPE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("MKDIR returns Bad Inode %lld mode %lld\n",
					   (long long) dir->i_ino, (long long) father_attr->mode);
			goto invalidate;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

int
orfs_mknod(struct inode *dir, struct dentry *dentry, int mode,
#if ORFS_LINUX_24
	   int rdev
#elif ORFS_LINUX_26
	   dev_t rdev
#endif
    )
{
	struct orfs_sb_info *sbi = ORFS_SBI(dir->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	orfs_name_len_t name_len = dentry->d_name.len;
	orfs_uid_t uid;
	orfs_gid_t gid;
	orfs_dev_t dev;
	int error;
	struct inode *inode;

	/* no need to contact the server if it doesn't
	 * have the needed capability
	 */
	if (!ORFS_SBI_HAS_CAP(sbi, MKNOD))
		return -EPERM;

	ORFS_DEBUG(REQ,
		   "MKDIR called for entry %s in dir inode %lld with mode %x\n",
		   dentry->d_name.name, (long long) dir->i_ino, mode);

	ORFS_UPDATE_REQ_STAT(sbi, mkdir);

	/* deal with dev */
	dev = KERNEL2ORFS_DEV(rdev);

	/* deal with uid */
	uid = current->fsuid;
	/* deal with gid */
	if (dir->i_mode & S_ISGID) {
		mode |= S_ISGID;
		gid = dir->i_gid;
	} else
		gid = current->fsgid;

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_DIR_REQ;
	req->subtype = ORFS_MKNOD_DIR_REQ;
	fill_inode_id_from_pd(dir, &req->target.dir.dir);
	req->target.dir.name_len = name_len;
	memcpy(req->target.dir.name, dentry->d_name.name, name_len + 1);
	req->target.dir.data.create.mode = mode;
	req->target.dir.data.create.uid = uid;
	req->target.dir.data.create.gid = gid;
	req->target.dir.data.create.dev = dev;

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR("Remote call for MKNOD failed with error %d\n",
			   -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.dir.data.create.new_node_attr;
		struct orfs_getattr_data *father_attr = &ans->target.dir.father_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* create child */
			inode = orfs_iget_with_attr(dir->i_sb, node_attr);
			if (!inode || is_bad_inode(inode)) {
				// ugh... server still has it...
				d_drop(dentry);
				error = -ENOMEM;
			} else {
				d_instantiate(dentry, inode);
				// mark inode dirty ???

				ORFS_DEBUG(FS, "MKNOD returns new inode %lld\n",
					   (long long) node_attr->id.ino);
			}
		      update_dir:
			/* update dentry */
			ORFS_UPDATE_DENTRY_PD_TIME(dentry);
			/* update dir attributes */
			orfs_update_inode_attr(dir, father_attr);
			break;
		case ORFS_NO_SPACE_LEFT:
			error = -ENOMEM;
			goto update_dir;
		case ORFS_EXISTING_DIR_ENTRY:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("MKNOD returns Existing Dir %lld Entry %s\n",
					   (long long) dir->i_ino, dentry->d_name.name);
			// update dirent/dentry cache of new dir
			error = -EEXIST;
			goto update_dir;
		case ORFS_NO_SUCH_INODE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("MKNOD returns No Such Inode %lld\n",
					   (long long) dir->i_ino);
		       invalidate:
			/* drop old dir inode */
			orfs_invalidate_dir_inode(dir);
			error = -EIO;
			break;
		case ORFS_BAD_INODE_TYPE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("MKNOD returns Bad Inode %lld mode %lld\n",
					   (long long) dir->i_ino, (long long) father_attr->mode);
			goto invalidate;
		case ORFS_INVALID_PARAMETER:
			ORFS_ALERT("Invalid parameter\n");
			BUG();
			break;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

int
orfs_create(struct inode *dir, struct dentry *dentry, int mode
#if ORFS_LINUX_26
	    , struct nameidata *nd
#endif
    )
{
	struct orfs_sb_info *sbi = ORFS_SBI(dir->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	orfs_name_len_t name_len = dentry->d_name.len;
	orfs_uid_t uid;
	orfs_gid_t gid;
	int error;
	struct inode *inode;

	ORFS_DEBUG(REQ,
		   "MKREG called for entry %s in dir inode %lld with mode %x\n",
		   dentry->d_name.name, (long long) dir->i_ino, mode);

	ORFS_UPDATE_REQ_STAT(sbi, mkreg);

	/* deal with uid */
	uid = current->fsuid;
	/* deal with gid */
	if (dir->i_mode & S_ISGID)
		gid = dir->i_gid;
	else
		gid = current->fsgid;

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_DIR_REQ;
	req->subtype = ORFS_MKREG_DIR_REQ;
	fill_inode_id_from_pd(dir, &req->target.dir.dir);
	req->target.dir.name_len = name_len;
	memcpy(req->target.dir.name, dentry->d_name.name, name_len + 1);
	req->target.dir.data.create.mode = mode;
	req->target.dir.data.create.uid = uid;
	req->target.dir.data.create.gid = gid;

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR("Remote call for MKREG failed with error %d\n",
			   -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.dir.data.create.new_node_attr;
		struct orfs_getattr_data *father_attr = &ans->target.dir.father_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* create child */
			inode = orfs_iget_with_attr(dir->i_sb, node_attr);
			if (!inode || is_bad_inode(inode)) {
				// ugh... server still has it...
				d_drop(dentry);
				error = -ENOMEM;
			} else {
				d_instantiate(dentry, inode);
				// mark inode dirty ???

				ORFS_DEBUG(FS, "MKREG returns new inode %lld\n",
					   (long long) node_attr->id.ino);
			}
		      update_dir:
			/* update dentry */
			ORFS_UPDATE_DENTRY_PD_TIME(dentry);
			/* update dir attributes */
			orfs_update_inode_attr(dir, father_attr);
			break;
		case ORFS_NO_SPACE_LEFT:
			error = -ENOMEM;
			goto update_dir;
		case ORFS_EXISTING_DIR_ENTRY:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("MKREG returns Existing Dir %lld Entry %s\n",
					   (long long) dir->i_ino, dentry->d_name.name);
			// update dirent/dentry cache of new dir
			error = -EEXIST;
			goto update_dir;
		case ORFS_NO_SUCH_INODE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("MKREG returns No Such Inode %lld\n",
					   (long long) dir->i_ino);
		       invalidate:
			/* drop old dir inode */
			orfs_invalidate_dir_inode(dir);
			error = -EIO;
			break;
		case ORFS_BAD_INODE_TYPE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("MKREG returns Bad Inode %lld mode %lld\n",
					   (long long) dir->i_ino, (long long) father_attr->mode);
			goto invalidate;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

int
orfs_symlink(struct inode *dir, struct dentry *dentry, const char *symname)
{
	struct orfs_sb_info *sbi = ORFS_SBI(dir->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	struct inode *inode;
	orfs_name_len_t name_len = dentry->d_name.len;
	orfs_path_len_t path_len = strlen(symname);
	orfs_uid_t uid;
	orfs_gid_t gid;
	int error;

	/* no need to contact the server if it doesn't
	 * have the needed capability
	 */
	if (!ORFS_SBI_HAS_CAP(sbi, SYMLINK))
		return -EPERM;

	ORFS_DEBUG(REQ, "SYMLINK called for entry %s in dir inode %lld\n",
		   dentry->d_name.name, (long long) dir->i_ino);

	/* check symlen */
	if (path_len > sbi->opts.symlen)
		return -ENAMETOOLONG;

	ORFS_UPDATE_REQ_STAT(sbi, symlink);

	/* deal with uid */
	uid = current->fsuid;
	/* deal with gid */
	if (dir->i_mode & S_ISGID)
		gid = dir->i_gid;
	else
		gid = current->fsgid;

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_DIR_REQ;
	req->subtype = ORFS_SYMLINK_DIR_REQ;
	fill_inode_id_from_pd(dir, &req->target.dir.dir);
	req->target.dir.name_len = name_len;
	req->target.dir.data.symlink.path_len = path_len;
	req->target.dir.data.symlink.uid = uid;
	req->target.dir.data.symlink.gid = gid;
	memcpy(req->target.dir.name, dentry->d_name.name, name_len + 1);
	creq->in_buf = (char *) symname;
	creq->in_buf_len = path_len;
	creq->in_buf_flags = ORFS_BUF_KMALLOC;

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR("Remote call for SYMLINK failed with error %d\n",
			   -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.dir.data.create.new_node_attr;
		struct orfs_getattr_data *father_attr = &ans->target.dir.father_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* create child */
			inode = orfs_iget_with_attr(dir->i_sb, node_attr);
			if (!inode || is_bad_inode(inode)) {
				// ugh... server still has it...
				d_drop(dentry);
				error = -ENOMEM;
			} else {
				d_instantiate(dentry, inode);
				// mark inode dirty ???

				orfs_autofill_symlink_cache(inode, symname,
							    path_len);

				ORFS_DEBUG(FS,
					   "SYMLINK returns new inode %lld\n",
					   (long long) node_attr->id.ino);
			}
		      update_dir:
			/* update dentry */
			ORFS_UPDATE_DENTRY_PD_TIME(dentry);
			/* update dir attributes */
			orfs_update_inode_attr(dir, father_attr);
			break;
		case ORFS_NO_SPACE_LEFT:
			error = -ENOMEM;
			goto update_dir;
		case ORFS_EXISTING_DIR_ENTRY:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("SYMLINK returns Existing Dir %lld Entry %s\n",
					   (long long) dir->i_ino, dentry->d_name.name);
			// update dirent/dentry cache of new dir
			error = -EEXIST;
			goto update_dir;
		case ORFS_NO_SUCH_INODE:
		case ORFS_BAD_INODE_TYPE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("SYMLINK returns No Such Inode %lld\n",
					   (long long) dir->i_ino);
		       invalidate:
			/* drop old dir inode */
			orfs_invalidate_dir_inode(dir);
			error = -EIO;
			break;
		case ORFS_INVALID_PARAMETER:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("SYMLINK returns Bad Inode %lld mode %lld\n",
					   (long long) dir->i_ino, (long long) father_attr->mode);
			goto invalidate;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

int
orfs_unlink(struct inode *dir, struct dentry *dentry)
{
	struct orfs_sb_info *sbi = ORFS_SBI(dir->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	orfs_name_len_t name_len = dentry->d_name.len;
	int error;

	ORFS_DEBUG(REQ, "UNLINK called for entry %s in dir inode %lld\n",
		   dentry->d_name.name, (long long) dir->i_ino);

	ORFS_UPDATE_REQ_STAT(sbi, unlink);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_DIR_REQ;
	req->subtype = ORFS_UNLINK_DIR_REQ;
	fill_inode_id_from_pd(dir, &req->target.dir.dir);
	req->target.dir.name_len = name_len;
	memcpy(req->target.dir.name, dentry->d_name.name, name_len + 1);

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR("Remote call for UNLINK failed with error %d\n",
			   -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *father_attr = &ans->target.dir.father_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* just update dir */
		      update_dir:
			/* update dentry */
			ORFS_UPDATE_DENTRY_PD_TIME(dentry);
			/* update dir attributes */
			orfs_update_inode_attr(dir, father_attr);
			break;
		case ORFS_NO_SUCH_DIR_ENTRY:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("UNLINK returns No Such Dir %lld Entry %s\n",
					   (long long) dir->i_ino, dentry->d_name.name);
			// update dirent/dentry cache of new dir
			error = -ENOENT;
			goto update_dir;
		case ORFS_NO_SUCH_INODE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("UNLINK returns No Such Inode %lld\n",
					   (long long) dir->i_ino);
		       invalidate:
			/* drop old dir inode */
			orfs_invalidate_dir_inode(dir);
			error = -EIO;
			break;
		case ORFS_BAD_INODE_TYPE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("UNLINK returns Bad Inode %lld mode %lld\n",
					   (long long) dir->i_ino, (long long) father_attr->mode);
			goto invalidate;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

int
orfs_rmdir(struct inode *dir, struct dentry *dentry)
{
	struct orfs_sb_info *sbi = ORFS_SBI(dir->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	orfs_name_len_t name_len = dentry->d_name.len;
	int error;

	ORFS_DEBUG(REQ, "RMDIR called for entry %s in dir inode %lld\n",
		   dentry->d_name.name, (long long) dir->i_ino);

	ORFS_UPDATE_REQ_STAT(sbi, rmdir);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_DIR_REQ;
	req->subtype = ORFS_RMDIR_DIR_REQ;
	fill_inode_id_from_pd(dir, &req->target.dir.dir);
	req->target.dir.name_len = name_len;
	memcpy(req->target.dir.name, dentry->d_name.name, name_len + 1);

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR("Remote call for RMDIR failed with error %d\n",
			   -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *father_attr = &ans->target.dir.father_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* just update dir */
		      update_dir:
			/* update dentry */
			ORFS_UPDATE_DENTRY_PD_TIME(dentry);
			/* update dir attributes */
			orfs_update_inode_attr(dir, father_attr);
			break;
		case ORFS_NOT_EMPTY_DIR:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("RMDIR returns Not Empty Dir %lld\n",
					   (long long) dir->i_ino);
			error = -ENOTEMPTY;
			goto update_dir;
		case ORFS_NO_SUCH_DIR_ENTRY:
			// update dirent/dentry cache of new dir
			error = -ENOENT;
			goto update_dir;
		case ORFS_NO_SUCH_INODE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("RMDIR returns No Such Inode %lld\n",
					   (long long) dir->i_ino);
		       invalidate:
			/* drop old dir inode */
			orfs_invalidate_dir_inode(dir);
			error = -EIO;
			break;
		case ORFS_BAD_INODE_TYPE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("RMDIR returns Bad Inode %lld mode %lld\n",
					   (long long) dir->i_ino, (long long) father_attr->mode);
			goto invalidate;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

int
orfs_rename(struct inode *old_dir,
	    struct dentry *old_dentry,
	    struct inode *new_dir, struct dentry *new_dentry)
{
	struct orfs_sb_info *sbi = ORFS_SBI(old_dentry->d_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	orfs_name_len_t old_name_len = old_dentry->d_name.len;
	orfs_name_len_t new_name_len = new_dentry->d_name.len;
	int error;

	ORFS_DEBUG(REQ, "MOVE called from (%lld,%s) to (%lld,%s)\n",
		   (long long) old_dir->i_ino, old_dentry->d_name.name,
		   (long long) new_dir->i_ino, new_dentry->d_name.name);

	ORFS_UPDATE_REQ_STAT(sbi, move);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_2DIR_REQ;
	req->subtype = ORFS_MOVE_2DIR_REQ;
	fill_inode_id_from_pd(old_dir, &req->target.twodir.old_dir);
	req->target.twodir.old_name_len = old_name_len;
	memcpy(req->target.twodir.old_name, old_dentry->d_name.name,
	       old_name_len + 1);
	fill_inode_id_from_pd(new_dir, &req->target.twodir.new_dir);
	req->target.twodir.new_name_len = new_name_len;
	memcpy(req->target.twodir.new_name, new_dentry->d_name.name,
	       new_name_len + 1);

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR("Remote call for MOVE failed with error %d\n",
			   -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.twodir.node_attr;
		struct orfs_getattr_data *old_father_attr = &ans->target.twodir.old_father_attr;
		struct orfs_getattr_data *new_father_attr = &ans->target.twodir.new_father_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* just update all nodes */
		      update_all:
			/* update dentries */
			ORFS_UPDATE_DENTRY_PD_TIME(old_dentry);
			ORFS_UPDATE_DENTRY_PD_TIME(new_dentry);
			/* update node attributes */
			orfs_update_inode_attr(old_dentry->d_inode, node_attr);
		      update_dirs:
			/* update new dir attributes */
			orfs_update_inode_attr(new_dir, new_father_attr);
		      update_old_dir:
			/* update old dir attributes */
			orfs_update_inode_attr(old_dir, old_father_attr);
			break;
		case ORFS_NO_SPACE_LEFT:
			error = -ENOMEM;
			goto update_all;
		case ORFS_EXISTING_DIR_ENTRY:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("MOVE returns Existing Dir %lld Entry %s\n",
					   (long long) new_dir->i_ino, new_dentry->d_name.name);
			// update dirent/dentry cache of new dir
			error = -EEXIST;
			goto update_all;
		case ORFS_NO_SUCH_DIR_ENTRY:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("MOVE returns No Such Dir %lld Entry %s\n",
					   (long long) old_dir->i_ino, old_dentry->d_name.name);
			/* drop non existing inode */
			orfs_invalidate_generic_inode(old_dentry->d_inode);
			// update dirent/dentry cache of old dir
			error = -ENOENT;
			goto update_dirs;
		case ORFS_NO_SUCH_INODE_2:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("MOVE returns No Such Inode %lld\n",
					   (long long) new_dir->i_ino);
		       invalidate_new_dir:
			/* drop new dir inode */
			orfs_invalidate_dir_inode(new_dir);
			error = -EIO;
			goto update_old_dir;
		case ORFS_BAD_INODE_TYPE_2:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("MOVE returns Bad Inode %lld mode %lld\n",
					   (long long) new_dir->i_ino, (long long) new_father_attr->mode);
			goto invalidate_new_dir;
		case ORFS_NO_SUCH_INODE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("MOVE returns No Such Inode %lld\n",
					   (long long) old_dir->i_ino);
		       invalidate_old_dir:
			/* drop old dir inode */
			orfs_invalidate_dir_inode(old_dir);
			error = -EIO;
			break;
		case ORFS_BAD_INODE_TYPE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("MOVE returns Bad Inode %lld mode %lld\n",
					   (long long) old_dir->i_ino, (long long) old_father_attr->mode);
			goto invalidate_old_dir;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

int
orfs_link(struct dentry *old_dentry,
	  struct inode *new_dir, struct dentry *new_dentry)
{
	struct orfs_sb_info *sbi = ORFS_SBI(old_dentry->d_sb);
	struct inode *old_dir = old_dentry->d_parent->d_inode;
	struct inode *inode = old_dentry->d_inode;
	struct orfs_client_request *creq;
	struct orfs_request *req;
	orfs_name_len_t old_name_len = old_dentry->d_name.len;
	orfs_name_len_t new_name_len = new_dentry->d_name.len;
	int error;

#ifdef ORFS_LINUX_24
	if (S_ISDIR(inode->i_mode))
		return -EPERM;
#endif

	/* no need to contact the server if it doesn't
	 * have the needed capability
	 */
	if (!ORFS_SBI_HAS_CAP(sbi, LINK))
		return -EPERM;

	ORFS_DEBUG(REQ, "LINK called from (%lld,%s) to (%lld,%s)\n",
		   (long long) old_dir->i_ino, old_dentry->d_name.name,
		   (long long) new_dir->i_ino, new_dentry->d_name.name);

	ORFS_UPDATE_REQ_STAT(sbi, link);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_2DIR_REQ;
	req->subtype = ORFS_LINK_2DIR_REQ;
	fill_inode_id_from_pd(old_dir, &req->target.twodir.old_dir);
	req->target.twodir.old_name_len = old_name_len;
	memcpy(req->target.twodir.old_name, old_dentry->d_name.name,
	       old_name_len + 1);
	fill_inode_id_from_pd(new_dir, &req->target.twodir.new_dir);
	req->target.twodir.new_name_len = new_name_len;
	memcpy(req->target.twodir.new_name, new_dentry->d_name.name,
	       new_name_len + 1);

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR("Remote call for LINK failed with error %d\n",
			   -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.twodir.node_attr;
		struct orfs_getattr_data *old_father_attr = &ans->target.twodir.old_father_attr;
		struct orfs_getattr_data *new_father_attr = &ans->target.twodir.new_father_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* update links */
			atomic_inc(&old_dentry->d_inode->i_count);
			d_instantiate(new_dentry, inode);
		      update_all:
			/* update dentries */
			ORFS_UPDATE_DENTRY_PD_TIME(old_dentry);
			ORFS_UPDATE_DENTRY_PD_TIME(new_dentry);
			/* update node attributes */
			orfs_update_inode_attr(old_dentry->d_inode, node_attr);
		      update_dirs:
			/* update new dir attributes */
			orfs_update_inode_attr(new_dir, new_father_attr);
		      update_old_dir:
			/* update old dir attributes */
			orfs_update_inode_attr(old_dir, old_father_attr);
			break;
		case ORFS_NO_SPACE_LEFT:
			error = -ENOMEM;
			goto update_all;
		case ORFS_EXISTING_DIR_ENTRY:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("LINK returns Existing Dir %lld Entry %s\n",
					   (long long) new_dir->i_ino, new_dentry->d_name.name);
			// update dirent/dentry cache of new dir
			error = -EEXIST;
			goto update_all;
		case ORFS_NO_SUCH_DIR_ENTRY:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("LINK returns No Such Dir %lld Entry %s\n",
					   (long long) old_dir->i_ino, old_dentry->d_name.name);
			/* drop non existing inode */
			orfs_invalidate_generic_inode(old_dentry->d_inode);
			// update dirent/dentry cache of old dir
			error = -ENOENT;
			goto update_dirs;
		case ORFS_BAD_INODE_TYPE_3:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("LINK returns Bad Inode %lld mode %lld\n",
					   (long long) new_dir->i_ino, (long long) node_attr->mode);
			/* drop bad type inode */
			orfs_invalidate_dir_inode(old_dentry->d_inode);
			// update dirent/dentry cache of old dir
			error = -EISDIR;
			goto update_dirs;			
		case ORFS_NO_SUCH_INODE_2:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("LINK returns No Such Inode %lld\n",
					   (long long) new_dir->i_ino);
		       invalidate_new_dir:
			/* drop new dir inode */
			orfs_invalidate_dir_inode(new_dir);
			error = -EIO;
			goto update_old_dir;
		case ORFS_BAD_INODE_TYPE_2:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("LINK returns Bad Inode %lld mode %lld\n",
					   (long long) new_dir->i_ino, (long long) new_father_attr->mode);
			goto invalidate_new_dir;
		case ORFS_NO_SUCH_INODE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("LINK returns No Such Inode %lld\n",
					   (long long) old_dir->i_ino);
		       invalidate_old_dir:
			/* drop old dir inode */
			orfs_invalidate_dir_inode(old_dir);
			error = -EIO;
			break;
		case ORFS_BAD_INODE_TYPE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("LINK returns Bad Inode %lld mode %lld\n",
					   (long long) old_dir->i_ino, (long long) old_father_attr->mode);
			goto invalidate_old_dir;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

extern int orfs_setattr(struct dentry *dentry, struct iattr *iattr);
#if ORFS_LINUX_24
extern int orfs_getattr(struct dentry *dentry, struct iattr *iattr);
extern int orfs_inode_revalidate(struct dentry *dentry);
#elif ORFS_LINUX_26
extern int orfs_getattr(struct vfsmount *mnt,
			struct dentry *dentry, struct kstat *kst);
#endif

struct inode_operations orfs_dir_iops = {
	.create = orfs_create,
	.lookup = orfs_lookup,
	.link = orfs_link,
	.unlink = orfs_unlink,
	.symlink = orfs_symlink,
	.mkdir = orfs_mkdir,
	.rmdir = orfs_rmdir,
	.mknod = orfs_mknod,
	.rename = orfs_rename,
	.permission = NULL,
	.setattr = orfs_setattr,
	.getattr = orfs_getattr,
#if ORFS_LINUX_24
	.revalidate = orfs_inode_revalidate,
#endif
};
