/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <asm/uaccess.h>

#include "orfs_fs.h"
#include "orfs_private_data.h"
#include "orfs_protocol.h"
#include "orfs_request.h"
#include "orfs_messages.h"

/*************************************
 * O_DIRECT Specific File Operations *
 *************************************/

static inline ssize_t
orfs_complete_odirect_pread(struct orfs_client_request * creq,
			    struct orfs_sb_info * sbi,
			    struct inode * inode,
			    loff_t * offset)
{
	struct orfs_request *req = ORFS_CREQ2REQ(creq);
	struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
	struct orfs_getattr_data *node_attr = &ans->target.file.node_attr;
	ssize_t error = 0;
	
	switch (ans->result) {
	case ORFS_SUCCESS:
		/* update node attributes */
		orfs_update_inode_attr(inode, node_attr);
		/* get count and update file pointer */
		error = ans->target.file.data.data_access.count;
		if (offset) /* NULL for aio_pread */
			*offset += error;
		break;
	case ORFS_SERVER_INVALID_PARAMETER:
		ORFS_NOTICE("PREAD returns Server Invalid Parameter\n");
		error = -EIO;
		break;
	case ORFS_BAD_INODE_TYPE:
		if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
			ORFS_ERROR("PREAD returns Bad Inode %lld mode %lld\n",
				   (long long) inode->i_ino, (long long) node_attr->mode);
		/* drop inode */
		orfs_invalidate_reg_inode(inode);
		error = -EINVAL;
		break;
	case ORFS_BAD_FILE:
		// for now the state based protocol make the error impossible
		ORFS_ALERT("Bad File %lld\n",
			   (long long) req->target.file.file.id);
		BUG();
		break;
	default:
		ORFS_ALERT("Unrecognized error %ld\n", (long) error);
		BUG();
	}

	return error;
}

ssize_t
orfs_odirect_pread(struct file *file,
		   char *buffer, size_t count, loff_t * offset)
{
	struct inode *inode = file->f_dentry->d_inode;
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	ssize_t error;

	ORFS_DEBUG(REQ, "O_DIRECT PREAD called for inode %lld\n",
		   (long long) inode->i_ino);

	ORFS_UPDATE_REQ_STAT(sbi, o_pread);

	if (!ORFS_SBI_IS_ALWAYS_O_DIRECT(sbi)) {
		/* flush inode pages */
#if ORFS_LINUX_26
		error = filemap_fdatawrite(inode->i_mapping);
#elif ORFS_LINUX_24
		error = filemap_fdatasync(inode->i_mapping);
#endif
		if (error) {
			ORFS_ERROR("filemap_fdatasync failed for inode %lld\n",
				   (long long) inode->i_ino);
			goto out;
		}
	}

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_FILE_REQ;
	req->subtype = ORFS_PREAD_FILE_REQ;
	fill_file_id_from_pd(file, &req->target.file.file);
	req->target.file.data.data_access.offset = *offset;
	req->target.file.data.data_access.count = count;
	creq->out_buf = buffer;
	creq->out_buf_len = count;
	creq->out_buf_flags = ORFS_BUF_USER_ANY;

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR
		    ("Remote call for O_DIRECT PREAD failed with error %d\n",
		     -error);
	} else {
		error = orfs_complete_odirect_pread(creq, sbi, inode, offset);
	}

	orfs_put_client_request(sbi, creq);
      out:
	return error;
}

static inline ssize_t
orfs_complete_odirect_pwrite(struct orfs_client_request * creq,
			     struct orfs_sb_info * sbi,
			     struct inode * inode,
			     loff_t * offset)
{
	struct orfs_request *req = ORFS_CREQ2REQ(creq);
	struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
	struct orfs_getattr_data *node_attr = &ans->target.file.node_attr;
	ssize_t error = 0;

	switch (ans->result) {
	case ORFS_SUCCESS:
		/* update node attributes */
		orfs_update_inode_attr(inode, node_attr);
		/* get count and update file pointer */
		error = ans->target.file.data.data_access.count;
		if (offset) /* NULL for aio_pwrite */
			*offset += error;
		break;
	case ORFS_NO_SPACE_LEFT:
		error = -ENOSPC;
		break;
	case ORFS_SERVER_INVALID_PARAMETER:
		ORFS_NOTICE("PWRITE returns Server Invalid Parameter\n");
		error = -EIO;
		break;
	case ORFS_BAD_INODE_TYPE:
		if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
			ORFS_ERROR("PWRITE returns Bad Inode %lld mode %lld\n",
				   (long long) inode->i_ino, (long long) node_attr->mode);
		/* drop inode */
		orfs_invalidate_reg_inode(inode);
		error = -EINVAL;
		break;
	case ORFS_BAD_FILE:
		// for now the state based protocol make the error impossible
		ORFS_ALERT("Bad File %lld\n",
			   (long long) req->target.file.file.id);
		BUG();
		break;
	default:
		ORFS_ALERT("Unrecognized error %ld\n", (long) error);
		BUG();
	}
	
	return error;
}

ssize_t
orfs_odirect_pwrite(struct file * file,
		    const char *buffer, size_t count, loff_t * offset)
{
	struct inode *inode = file->f_dentry->d_inode;
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	ssize_t error;

	ORFS_DEBUG(REQ, "O_DIRECT PWRITE called for inode %lld\n",
		   (long long) inode->i_ino);

	ORFS_UPDATE_REQ_STAT(sbi, o_pwrite);

	if (!ORFS_SBI_IS_ALWAYS_O_DIRECT(sbi)) {
		/* flush inode pages */
#if ORFS_LINUX_26
		error = filemap_fdatawrite(inode->i_mapping);
#elif ORFS_LINUX_24
		error = filemap_fdatasync(inode->i_mapping);
#endif
		if (error) {
			ORFS_ERROR("filemap_fdatasync failed for inode %lld\n",
				   (long long) inode->i_ino);
			goto out;
		}
	}

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_FILE_REQ;
	req->subtype = ORFS_PWRITE_FILE_REQ;
	fill_file_id_from_pd(file, &req->target.file.file);
	req->target.file.data.data_access.offset = *offset;
	req->target.file.data.data_access.count = count;
	creq->in_buf = (char *) buffer;
	creq->in_buf_len = count;
	creq->in_buf_flags = ORFS_BUF_USER_ANY;

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR
		    ("Remote call for O_DIRECT PWRITE failed with error %d\n",
		     -error);
	} else {
		error = orfs_complete_odirect_pwrite(creq, sbi, inode, offset);
	}

	orfs_put_client_request(sbi, creq);
      out:
	return error;
}

#if ORFS_LINUX_26
#include <linux/aio.h>

static void
orfs_odirect_aio_read_callback(struct orfs_sb_info *sbi,
			       struct orfs_client_request * creq,
			       long *res, long *res2)
{
	struct inode *inode = creq->kiocb->ki_filp->f_dentry->d_inode;
	*res = orfs_complete_odirect_pread(creq, sbi, inode, NULL);
	*res2 = 0;
}

ssize_t
orfs_odirect_aio_read(struct kiocb * iocb, char * buffer, size_t count, loff_t offset)
{
	struct file * file = iocb->ki_filp;
	struct inode *inode = file->f_dentry->d_inode;
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	ssize_t error;

	ORFS_DEBUG(REQ, "O_DIRECT AIO READ called for inode %lld\n",
		   (long long) inode->i_ino);

	ORFS_UPDATE_REQ_STAT(sbi, o_aioread);

	if (!ORFS_SBI_IS_ALWAYS_O_DIRECT(sbi)) {
		/* flush inode pages */
		error = filemap_fdatawrite(inode->i_mapping);
		if (error) {
			ORFS_ERROR("filemap_fdatasync failed for inode %lld\n",
				   (long long) inode->i_ino);
			return error;
		}
	}

	if (!ORFS_SBI_IS_MULTICLIENT(sbi)
	    || !sbi->nops->async_remote_call) {
		/* revert to synchronous */
		return orfs_odirect_pread(file, buffer, count, &offset);
	}
	
	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_FILE_REQ;
	req->subtype = ORFS_PREAD_FILE_REQ;
	fill_file_id_from_pd(file, &req->target.file.file);
	req->target.file.data.data_access.offset = offset;
	req->target.file.data.data_access.count = count;
	creq->out_buf = buffer;
	creq->out_buf_len = count;
	creq->out_buf_flags = ORFS_BUF_USER_ANY;

	creq->kiocb = iocb;
	creq->callback = orfs_odirect_aio_read_callback;

	ORFS_KIOCB_SBI(iocb) = sbi;
	ORFS_KIOCB_CREQ(iocb) = creq;

	error = sbi->nops->async_remote_call(sbi, creq);
	if (error < 0)
		ORFS_ERROR("Asynchronous remote call for PREAD failed with error %d\n",
			   -error);

	return -EIOCBQUEUED;
}

static void
orfs_odirect_aio_write_callback(struct orfs_sb_info *sbi,
				struct orfs_client_request * creq,
				long *res, long *res2)
{
	struct inode *inode = creq->kiocb->ki_filp->f_dentry->d_inode;
	*res = orfs_complete_odirect_pwrite(creq, sbi, inode, NULL);
	*res2 = 0;
}

ssize_t
orfs_odirect_aio_write(struct kiocb * iocb, const char * buffer, size_t count, loff_t offset)
{
	struct file * file = iocb->ki_filp;
	struct inode *inode = file->f_dentry->d_inode;
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	ssize_t error;

	ORFS_DEBUG(REQ, "O_DIRECT AIO WRITE called for inode %lld\n",
		   (long long) inode->i_ino);

	ORFS_UPDATE_REQ_STAT(sbi, o_aiowrite);

	if (!ORFS_SBI_IS_ALWAYS_O_DIRECT(sbi)) {
		/* flush inode pages */
		error = filemap_fdatawrite(inode->i_mapping);
		if (error) {
			ORFS_ERROR("filemap_fdatasync failed for inode %lld\n",
				   (long long) inode->i_ino);
			return error;
		}
	}

	if (!ORFS_SBI_IS_MULTICLIENT(sbi)
	    || !sbi->nops->async_remote_call) {
		/* revert to synchronous */
		return orfs_odirect_pwrite(file, buffer, count, &offset);
	}

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_FILE_REQ;
	req->subtype = ORFS_PWRITE_FILE_REQ;
	fill_file_id_from_pd(file, &req->target.file.file);
	req->target.file.data.data_access.offset = offset;
	req->target.file.data.data_access.count = count;
	creq->in_buf = (char *) buffer;
	creq->in_buf_len = count;
	creq->in_buf_flags = ORFS_BUF_USER_ANY;

	creq->kiocb = iocb;
	creq->callback = orfs_odirect_aio_write_callback;

	ORFS_KIOCB_SBI(iocb) = sbi;
	ORFS_KIOCB_CREQ(iocb) = creq;

	error = sbi->nops->async_remote_call(sbi, creq);
	if (error < 0)
		ORFS_ERROR("Asynchronous remote call for PWRITE failed with error %d\n",
			   -error);

	return -EIOCBQUEUED;
}
#endif
