/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/fs.h>
#include <linux/sysctl.h>
#include <asm/uaccess.h>

#include "orfs_fs.h"
#include "orfs_messages.h"
#include "orfs_stats.h"

/***************************
 * Proc NetLayers Function *
 ***************************/

static int
orfs_proc_netlayers_func(char *page, char **start, off_t off,
			 int count, int *eof, void *data)
{
	int len = orfs_get_net_layers_list(page);
	/* this is a copy of proc_calc_metrics (fs/proc/proc_misc.c) */
	if (len <= off + count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;
	return len;
}

/***********************
 * Proc Reset Function *
 ***********************/

#define RESET_STRING_MAX 64

static int
orfs_proc_reset_func(struct file *file, const char *buffer,
		     unsigned long count, void *data)
{
#ifdef CONFIG_ORFS_STATS
	char kbuffer[RESET_STRING_MAX + 1];
	char *end;
	struct list_head *list_elt;
	int all = 0;

	/* check count */
	if (!count)
		return -EINVAL;
	if (count > RESET_STRING_MAX)
		count = RESET_STRING_MAX;

	/* copy and fix buffer */
	if (copy_from_user(kbuffer, buffer, count))
		return -EFAULT;
	kbuffer[RESET_STRING_MAX] = '\0';

	/* remove '\n' */
	end = strchr(kbuffer, '\n');
	if (end)
		*end = '\0';

	/* handle "*" case */
	if (!strcmp(kbuffer, "*")) {
		ORFS_DEBUG(PROC, "Resetting all superblocks statistics\n");
		all = 1;
	} else {
		ORFS_DEBUG(PROC,
			   "Resetting statistics of superblocks matching %s\n",
			   kbuffer);
	}

	spin_lock(&orfs_sbi_list_lock);

	list_for_each(list_elt, &orfs_sbi_list) {
		struct orfs_sb_info *sbi =
		    list_entry(list_elt, struct orfs_sb_info, list);
		if (all || !strcmp(kbuffer, sbi->server_path))
			orfs_sbi_reset_stats(sbi);
	}

	spin_unlock(&orfs_sbi_list_lock);

#endif				/* ~CONFIG_ORFS_STATS */

	return count;
}

/**********************
 * Seq Show Functions *
 **********************/

static int
orfs_proc_alloc_seq_show(struct seq_file *m, void *v)
{
	if (v) {
		struct list_head *list_elt;

		seq_printf(m, "server\t\t\tfile\tdentry\tinode\n");

		spin_lock(&orfs_sbi_list_lock);

		list_for_each(list_elt, &orfs_sbi_list) {
			struct orfs_sb_info *sbi =
			    list_entry(list_elt, struct orfs_sb_info,
				       list);
			seq_printf(m, "%s\t%ld\t%ld\t%ld\n",
				   sbi->server_path,
				   ORFS_GET_ALLOC_STAT(sbi, file),
				   ORFS_GET_ALLOC_STAT(sbi, dentry),
				   ORFS_GET_ALLOC_STAT(sbi, inode));
		}

		spin_unlock(&orfs_sbi_list_lock);

	}
	return 0;
}

static int
orfs_proc_reval_seq_show(struct seq_file *m, void *v)
{
	if (v) {
		struct list_head *list_elt;

		seq_printf(m,
			   "server\t\t\t"
			   "inode\tremote\tfailed\t"
			   "dentry\tremote\tfailed\n");

		spin_lock(&orfs_sbi_list_lock);

		list_for_each(list_elt, &orfs_sbi_list) {
			struct orfs_sb_info *sbi =
			    list_entry(list_elt, struct orfs_sb_info,
				       list);
			seq_printf(m, "%s\t"	/* path */
				   "%ld\t%ld\t%ld\t"	/* inode revalidating */
				   "%ld\t%ld\t%ld\n",	/* dentry revalidating */
				   sbi->server_path,
				   ORFS_GET_REVAL_STAT(sbi, inode, call),
				   ORFS_GET_REVAL_STAT(sbi, inode, remote),
				   ORFS_GET_REVAL_STAT(sbi, inode, failed),
				   ORFS_GET_REVAL_STAT(sbi, dentry, call),
				   ORFS_GET_REVAL_STAT(sbi, dentry, remote),
				   ORFS_GET_REVAL_STAT(sbi, dentry, failed));
		}

		spin_unlock(&orfs_sbi_list_lock);

	}
	return 0;
}

static int
orfs_proc_req_seq_show(struct seq_file *m, void *v)
{
	if (v) {
		struct list_head *list_elt;

		seq_printf(m,
			   "server\t\t\t"
			   "sattr\tgattr\trdlk\topen\tclose\t"
			   "ord\towr\t"
#if ORFS_LINUX_26
			   "aor\taow\taos\t"
#endif
			   "rdp\twrp\tpwr\tcwr\trddir\tfsync\t"
			   "lkup\tmkreg\tmkdir\tmknod\tunlk\trmdir\treval\t"
			   "move\tlink\tsymlk\tstatfs\n");

		spin_lock(&orfs_sbi_list_lock);

		list_for_each(list_elt, &orfs_sbi_list) {
			struct orfs_sb_info *sbi =
			    list_entry(list_elt, struct orfs_sb_info,
				       list);
			seq_printf(m,
				   "%s\t" "%ld\t%ld\t%ld\t%ld\t%ld\t"
				   "%ld\t%ld\t%ld\t%ld\t%ld\t"
#if ORFS_LINUX_26
				   "%ld\t%ld\t%ld\t"
#endif
				   "%ld\t%ld\t%ld\t%ld\t%ld\t"
				   "%ld\t%ld\t%ld\t%ld\t%ld\t"
				   "%ld\t%ld\t%ld\t%ld\n",
				   sbi->server_path,
				   ORFS_GET_REQ_STAT(sbi, setattr),
				   ORFS_GET_REQ_STAT(sbi, getattr),
				   ORFS_GET_REQ_STAT(sbi, readlink),
				   ORFS_GET_REQ_STAT(sbi, open),
				   ORFS_GET_REQ_STAT(sbi, close),
				   ORFS_GET_REQ_STAT(sbi, o_pread),
				   ORFS_GET_REQ_STAT(sbi, o_pwrite),
#if ORFS_LINUX_26
				   ORFS_GET_REQ_STAT(sbi, o_aioread),
				   ORFS_GET_REQ_STAT(sbi, o_aiowrite),
				   ORFS_GET_REQ_STAT(sbi, aiofsync),
#endif
				   ORFS_GET_REQ_STAT(sbi, readpage),
				   ORFS_GET_REQ_STAT(sbi, writepage),
				   ORFS_GET_REQ_STAT(sbi, prepare_write),
				   ORFS_GET_REQ_STAT(sbi, commit_write),
				   ORFS_GET_REQ_STAT(sbi, preaddir),
				   ORFS_GET_REQ_STAT(sbi, fsync),
				   ORFS_GET_REQ_STAT(sbi, lookup),
				   ORFS_GET_REQ_STAT(sbi, mkreg),
				   ORFS_GET_REQ_STAT(sbi, mkdir),
				   ORFS_GET_REQ_STAT(sbi, mknod),
				   ORFS_GET_REQ_STAT(sbi, unlink),
				   ORFS_GET_REQ_STAT(sbi, rmdir),
				   ORFS_GET_REQ_STAT(sbi, revalidate),
				   ORFS_GET_REQ_STAT(sbi, move),
				   ORFS_GET_REQ_STAT(sbi, link),
				   ORFS_GET_REQ_STAT(sbi, symlink),
				   ORFS_GET_REQ_STAT(sbi, statfs));
		}

		spin_unlock(&orfs_sbi_list_lock);

	}
	return 0;
}

static int
orfs_proc_server_seq_show(struct seq_file *m, void *v)
{
	if (v) {
		struct list_head *list_elt;

		seq_printf(m, "server\t\t\tcapabilies\tblksize\torfsd\n");

		spin_lock(&orfs_sbi_list_lock);

		list_for_each(list_elt, &orfs_sbi_list) {
			struct orfs_sb_info *sbi =
			    list_entry(list_elt, struct orfs_sb_info,
				       list);
			int first = 1;

			/* server path */
			seq_printf(m, "%s\t", sbi->server_path);

			/* capabilities */
			if (ORFS_SBI_HAS_CAP(sbi, LINK)) {
				if (!first)
					seq_putc(m, ',');
				first = 0;
				seq_printf(m, "link");
			}
			if (ORFS_SBI_HAS_CAP(sbi, SYMLINK)) {
				if (!first)
					seq_putc(m, ',');
				first = 0;
				seq_printf(m, "symlink");
			}
			if (ORFS_SBI_HAS_CAP(sbi, MKNOD)) {
				if (!first)
					seq_putc(m, ',');
				first = 0;
				seq_printf(m, "mknod");
			}

			/* blksize */
			seq_printf(m, "\t%ld\t",
				   (unsigned long) sbi->opts.blksize);

			/* orfsd pid */
			if (ORFS_SBI_IS_MULTICLIENT(sbi))
				seq_printf(m, "%ld\n",
					   (unsigned long) sbi->param.multi.kt_pid);
			else
				seq_printf(m, "none\n");
		}

		spin_unlock(&orfs_sbi_list_lock);

	}
	return 0;
}

/***********************
 * Single Declarations *
 ***********************/

/* alloc */
static int
orfs_proc_alloc_open(struct inode *inode, struct file *file)
{
	return single_open(file, orfs_proc_alloc_seq_show, NULL);
}

static struct file_operations orfs_proc_alloc_ops = {
	open:orfs_proc_alloc_open,
	read:seq_read,
	llseek:seq_lseek,
	release:single_release,
};

/* reval */
static int
orfs_proc_reval_open(struct inode *inode, struct file *file)
{
	return single_open(file, orfs_proc_reval_seq_show, NULL);
}

static struct file_operations orfs_proc_reval_ops = {
	open:orfs_proc_reval_open,
	read:seq_read,
	llseek:seq_lseek,
	release:single_release,
};

/* req */
static int
orfs_proc_req_open(struct inode *inode, struct file *file)
{
	return single_open(file, orfs_proc_req_seq_show, NULL);
}

static struct file_operations orfs_proc_req_ops = {
	open:orfs_proc_req_open,
	read:seq_read,
	llseek:seq_lseek,
	release:single_release,
};

/* server */
static int
orfs_proc_server_open(struct inode *inode, struct file *file)
{
	return single_open(file, orfs_proc_server_seq_show, NULL);
}

static struct file_operations orfs_proc_server_ops = {
	open:orfs_proc_server_open,
	read:seq_read,
	llseek:seq_lseek,
	release:single_release,
};

/******************
 * Initialization *
 ******************/

#ifdef CONFIG_ORFS_DEBUG
/* /proc/sys/debug/orfs node */

#define CTL_DEBUG_ORFS 434	/* random number */

static struct ctl_table_header *orfs_ctl_header;

static struct ctl_table orfs_debug_ctl_table[] = {
	{CTL_DEBUG_ORFS, "orfs", &orfs_debug, sizeof (orfs_debug),
	 0644,
	 NULL, &proc_dointvec}
	,
	{0}
};

static struct ctl_table orfs_root_ctl_table[] = {
	{CTL_DEBUG, "debug", NULL, 0, 0555, orfs_debug_ctl_table},
	{0},
};

#endif				/* CONFIG_ORFS_DEBUG */

/* main /proc/fs/orfs/ node */

static struct proc_dir_entry *orfs_proc_root;

/* init/exit */

int
orfs_proc_init(void)
{
	struct proc_dir_entry *orfs_proc_netlayers;
	struct proc_dir_entry *orfs_proc_alloc;
	struct proc_dir_entry *orfs_proc_reval;
	struct proc_dir_entry *orfs_proc_req;
	struct proc_dir_entry *orfs_proc_server;
	struct proc_dir_entry *orfs_proc_reset;

	/* root node */
	orfs_proc_root = proc_mkdir("orfs", proc_root_fs);
	if (!orfs_proc_root) {
		ORFS_ERROR("Cannot create /proc/fs/orfs/\n");
		goto error;
	}

	/* net layers node */
	orfs_proc_netlayers =
	    create_proc_entry("netlayers", 0444, orfs_proc_root);
	if (!orfs_proc_netlayers) {
		ORFS_ERROR("Cannot create /proc/fs/orfs/netlayers\n");
		goto error_with_root;
	}
	orfs_proc_netlayers->read_proc = orfs_proc_netlayers_func;
	orfs_proc_netlayers->data = NULL;

	/* alloc node */
	orfs_proc_alloc = create_proc_entry("alloc", 0444, orfs_proc_root);
	if (!orfs_proc_alloc) {
		ORFS_ERROR("Cannot create /proc/fs/orfs/alloc\n");
		goto error_with_netlayers;
	}
	orfs_proc_alloc->proc_fops = &orfs_proc_alloc_ops;

	/* reval node */
	orfs_proc_reval = create_proc_entry("reval", 0444, orfs_proc_root);
	if (!orfs_proc_reval) {
		ORFS_ERROR("Cannot create /proc/fs/orfs/reval\n");
		goto error_with_alloc;
	}
	orfs_proc_reval->proc_fops = &orfs_proc_reval_ops;

	/* req node */
	orfs_proc_req = create_proc_entry("req", 0444, orfs_proc_root);
	if (!orfs_proc_req) {
		ORFS_ERROR("Cannot create /proc/fs/orfs/req\n");
		goto error_with_reval;
	}
	orfs_proc_req->proc_fops = &orfs_proc_req_ops;

	/* server node */
	orfs_proc_server = create_proc_entry("server", 0444, orfs_proc_root);
	if (!orfs_proc_server) {
		ORFS_ERROR("Cannot create /proc/fs/orfs/server\n");
		goto error_with_req;
	}
	orfs_proc_server->proc_fops = &orfs_proc_server_ops;

	/* reset node */
	orfs_proc_reset = create_proc_entry("reset", 0200, orfs_proc_root);
	if (!orfs_proc_reset) {
		ORFS_ERROR("Cannot create /proc/fs/orfs/reset\n");
		goto error_with_server;
	}
	orfs_proc_reset->write_proc = orfs_proc_reset_func;

#ifdef CONFIG_ORFS_DEBUG
	/* debug node */
	orfs_ctl_header = register_sysctl_table(orfs_root_ctl_table, 0);
	if (!orfs_ctl_header) {
		ORFS_ERROR("Cannot register sysctl table\n");
		goto error_with_reset;
	}
#endif				/* CONFIG_ORFS_DEBUG */

	return 0;

#ifdef CONFIG_ORFS_DEBUG
      error_with_reset:
	remove_proc_entry("reset", orfs_proc_root);
#endif				/* CONFIG_ORFS_DEBUG */
      error_with_server:
	remove_proc_entry("server", orfs_proc_root);
      error_with_req:
	remove_proc_entry("req", orfs_proc_root);
      error_with_reval:
	remove_proc_entry("reval", orfs_proc_root);
      error_with_alloc:
	remove_proc_entry("alloc", orfs_proc_root);
      error_with_netlayers:
	remove_proc_entry("netlayers", orfs_proc_root);
      error_with_root:
	remove_proc_entry("orfs", proc_root_fs);
      error:
	return -1;
}

void
orfs_proc_exit(void)
{
#ifdef CONFIG_ORFS_DEBUG
	unregister_sysctl_table(orfs_ctl_header);
#endif				/* CONFIG_ORFS_DEBUG */
	remove_proc_entry("reset", orfs_proc_root);
	remove_proc_entry("req", orfs_proc_root);
	remove_proc_entry("server", orfs_proc_root);
	remove_proc_entry("reval", orfs_proc_root);
	remove_proc_entry("alloc", orfs_proc_root);
	remove_proc_entry("netlayers", orfs_proc_root);
	remove_proc_entry("orfs", proc_root_fs);
}
