/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <asm/semaphore.h>

#include "orfs_fs.h"
#include "orfs_request.h"
#include "orfs_mount.h"
#include "orfs_messages.h"

/***********
 * Killall * 
 ***********/

int
orfs_set_connection_error(struct orfs_sb_info *sbi)
{
	static spinlock_t lock = SPIN_LOCK_UNLOCKED;
	if ( !spin_trylock(&lock) )
		return -EBUSY;

	if ( sbi->connection_status != ORFS_CONNECTION_OK ) {
		spin_unlock(&lock);
		return 1;
	}

	ORFS_DEBUG(ORFSD, "Setting connection status to error\n");

	sbi->connection_status = ORFS_CONNECTION_ERROR;
	spin_unlock(&lock);
	return 0;
}

void
orfs_killall_client_requests(struct orfs_sb_info *sbi)
{
	struct orfs_client_request *creq;

	ORFS_DEBUG(ORFSD, "Killall requests\n");

	if ( orfs_set_connection_error(sbi) )
		return;

	if ( ORFS_SBI_IS_MULTICLIENT(sbi) ) {
		if ( sbi->nops->kt_loop ) {
			/* When there's a kernel thread,
			 * requests might have their own wait queue.
			 * Wake up all of them.
			 */	
			spin_lock(&sbi->param.multi.creq_lock);
			list_for_each_entry(creq, &sbi->param.multi.creq_list, list)
				wake_up(&creq->queue);
			spin_unlock(&sbi->param.multi.creq_lock);
		}
		/* Requests might also be waiting on
		 * some network specific wait queues.
		 * Use the network specific wakeup routine
		 * until there are no pending requests.
		 */
		do {
		  sbi->nops->wake_up_waiters(sbi);
		  schedule();
		} while ( !list_empty(&sbi->param.multi.creq_list) );
	} else {
		/* Only one request might be waiting, and it holds the semaphore.
		 * Use the network specific wakeup routine until nobody holds the
		 * semaphore anymore.
		 */
		do {
			sbi->nops->wake_up_waiters(sbi);
			schedule ();
		} while ( !atomic_read(&sbi->param.mono.creq_sem.count) );
	}
}

/************
 * Requests * 
 ************/

struct orfs_client_request *
orfs_get_client_request(struct orfs_sb_info *sbi)
{
	struct orfs_client_request *creq;

	creq = kmem_cache_alloc(orfs_creq_cachep, GFP_KERNEL);
	if ( !creq ) {
		ORFS_ERROR("creq allocation failed\n");
		return NULL;
	}

	init_waitqueue_head(&creq->queue);

	if ( ORFS_SBI_IS_MULTICLIENT(sbi) ) {
		spin_lock(&sbi->param.multi.creq_lock);
		list_add_tail(&creq->list, &sbi->param.multi.creq_list);
		spin_unlock(&sbi->param.multi.creq_lock);
	} else
		down(&sbi->param.mono.creq_sem);

	ORFS_CREQ2REQ(creq)->tag = (orfs_tag_t)(unsigned long) creq;
#if ORFS_LINUX_26
	creq->kiocb = NULL;
#endif
	creq->status = 0;
	creq->ans = NULL;
	creq->in_buf = NULL;
	creq->out_buf = NULL;

	ORFS_DEBUG(CREQ, "Allocated client request %p\n", creq);

	return creq;
}

void
orfs_put_client_request(struct orfs_sb_info *sbi, struct orfs_client_request *creq)
{
	ORFS_DEBUG(CREQ, "Releasing client request %p\n", creq);
	
	if (IS_ORFS_CREQ_ORPHAN(creq)) {
		ORFS_DEBUG(CREQ, "Adopting orphan request %p\n", creq);
		ORFS_CREQ_IS_ADOPTED(creq);
	} else {
		if (IS_ORFS_CREQ_ADOPTED(creq))
			ORFS_DEBUG(CREQ, "Destroying adopted request %p\n", creq);
		else
			ORFS_DEBUG(CREQ, "Destroying request %p\n", creq);
			
		/* release the answer if it exists */
		if (creq->ans)
			orfs_put_client_answer(sbi, creq->ans);

		if ( ORFS_SBI_IS_MULTICLIENT(sbi) ) {
			spin_lock(&sbi->param.multi.creq_lock);
			list_del(&creq->list);
			spin_unlock(&sbi->param.multi.creq_lock);
		} else
			up(&sbi->param.mono.creq_sem);

		kmem_cache_free(orfs_creq_cachep, creq);
	} 
}

/***********
 * Answers *
 ***********/

struct orfs_answer *
orfs_get_client_answer(struct orfs_sb_info *sbi)
{
	struct orfs_answer *ans;

	ans = kmem_cache_alloc(orfs_ans_cachep, GFP_KERNEL);
	if ( !ans ) {
		ORFS_ERROR("answer allocation failed\n");
		return NULL;
	}

	ORFS_DEBUG(CREQ, "Allocated answer %p\n", ans);

	return ans;
}

void
orfs_put_client_answer(struct orfs_sb_info *sbi, struct orfs_answer *ans)
{
	ORFS_DEBUG(CREQ, "Releasing answer %p\n", ans);

	kmem_cache_free(orfs_ans_cachep, ans);
}

/*
 * ORFSD is used to centralize receiving answers ans
 * giving them back to different clients.
 * ORFS tries to receive an answer. When succeded, it
 * reads its tag and finds the corresponding request.
 * It wakes up corresponding process after giving it
 * the answer.
 */

/*********
 * ORFSD *
 *********/

static int
orfs_receive_kernel_thread(struct orfs_sb_info *sbi)
{
#if ORFS_LINUX_24
	daemonize();
	strcpy(current->comm, ORFS_DAEMON_NAME);
#elif ORFS_LINUX_26
	daemonize(ORFS_DAEMON_NAME);
#endif

	ORFS_NOTICE("ORFSD launched\n");

	/* loop until &param->wakeup_q is woken up */
	sbi->nops->kt_loop(sbi);

	ORFS_NOTICE("ORFSD finished\n");

	/* tell we are ready to finish */
	complete_and_exit(&sbi->param.multi.kt_done, 0);
}

int
orfs_launch_orfsd(struct orfs_sb_info *sbi)
{
	int error;

	/* initialize kernel thread */
	init_completion(&sbi->param.multi.kt_done);
	sbi->param.multi.kt_finish = 0;

	/* launch kernel thread */
	error =
	    kernel_thread((int (*)(void *)) orfs_receive_kernel_thread,
			  sbi, CLONE_FS | CLONE_FILES);
	if (error < 0)
		goto error;

	sbi->param.multi.kt_pid = error;
	return 0;

      error:
	return error;
}

void
orfs_finish_orfsd(struct orfs_sb_info *sbi)
{
	/* ask kernel thread termination */
	sbi->param.multi.kt_finish = 1;
	sbi->nops->wake_up_waiters(sbi);

	/* wait for its completion */
	wait_for_completion(&sbi->param.multi.kt_done);
}
