/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/vfs.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/seq_file.h>
#include <linux/mount.h>
#include <linux/list.h>
#include <linux/pagemap.h>
#include <asm/atomic.h>
#include <linux/spinlock.h>

#include "orfs_fs.h"
#include "orfs_private_data.h"
#include "orfs_mount.h"
#include "orfs_request.h"
#include "orfs_messages.h"

#ifdef CONFIG_ORFS_STATS

/**************
 * Statistics *
 **************/

void
orfs_sbi_reset_stats(struct orfs_sb_info *sbi)
{
	memset(&sbi->stat.reval, 0, sizeof (sbi->stat.reval));
	memset(&sbi->stat.req, 0, sizeof (sbi->stat.req));
}

static inline void
orfs_sbi_full_reset_stats(struct orfs_sb_info *sbi)
{
	memset(&sbi->stat.alloc, 0, sizeof (sbi->stat.alloc));
	orfs_sbi_reset_stats(sbi);
}

#endif				/* CONFIG_ORFS_STATS */

/**************************
 * Init Answer Processing *
 **************************/

int
orfs_process_init_answer(struct orfs_sb_info *sbi, struct orfs_answer *ans)
{
	if (ans->target.fs.data.init.data.magic != ORFS_PROTOCOL_MAGIC) {
		ORFS_ERROR("Bad Magic received from server %s\n",
			   sbi->server_path);
		return -EIO;
	}

	// nothing to do with ans->target.fs.data.init.data.protocol

	if ( !(ans->target.fs.data.init.data.blksize%PAGE_SIZE) ) {
		ORFS_DEBUG(FS, "Got blksize %ld\n",
			   (long) ans->target.fs.data.init.data.blksize);
		sbi->opts.blksize = ans->target.fs.data.init.data.blksize;
	} else {
		ORFS_ERROR("Cannot size blksize %lld (must be a multiple of PAGE_SIZE=%ld)\n",
			   (long long)ans->target.fs.data.init.data.blksize, (long)PAGE_SIZE);
		return -EIO;
	}

	sbi->opts.symlen = ans->target.fs.data.init.data.symlen;
	ORFS_DEBUG(FS, "Got symlink maximum length %ld\n",
		   (long) ans->target.fs.data.init.data.symlen);

	sbi->opts.capabilities = ans->target.fs.data.init.data.capabilities;
	ORFS_DEBUG(FS, "Got capabilities %ld\n",
		   (long) ans->target.fs.data.init.data.capabilities);

	return 0;
}

/******************
 * Network Layers *
 ******************/

static LIST_HEAD(orfs_net_layers);
static spinlock_t orfs_net_layers_lock = SPIN_LOCK_UNLOCKED;

/* called with lock held */
static inline struct orfs_net_layer *
orfs_find_net_layer(char *name)
{
	struct orfs_net_layer *layer;

	list_for_each_entry(layer, &orfs_net_layers, layers) {
		if (!strcmp(name, layer->name))
			return layer;
	}
	return NULL;
}

int
orfs_register_net_layer(struct orfs_net_layer *new_layer)
{
	char *name = new_layer->name;

	ORFS_DEBUG(NET, "Registering network layer %s\n", name);

	spin_lock(&orfs_net_layers_lock);
	if (orfs_find_net_layer(name)) {
		ORFS_ERROR("Network layer %s already registered\n", name);
		spin_unlock(&orfs_net_layers_lock);
		return -EEXIST;
	}

	list_add_tail(&new_layer->layers, &orfs_net_layers);
	atomic_set(&new_layer->count, 0);
	spin_unlock(&orfs_net_layers_lock);

	return 0;
}

int
orfs_unregister_net_layer(struct orfs_net_layer *old_layer)
{
	char *name = old_layer->name;
	struct orfs_net_layer *layer;
	struct list_head *head;

	ORFS_DEBUG(NET, "Unregistering network layer %s\n", name);

	spin_lock(&orfs_net_layers_lock);

	layer = orfs_find_net_layer(name);
	if (layer != old_layer) {
		ORFS_ERROR("Network layer %s not registered\n", name);
		spin_unlock(&orfs_net_layers_lock);
		return -ENOENT;
	}

	if (atomic_read(&layer->count)) {
		ORFS_ERROR("Network layer %s still used\n", name);
		spin_unlock(&orfs_net_layers_lock);
		return -EBUSY;
	}

	head = &layer->layers;
	__list_del(head->prev, head->next); /* no need to reinit */

	spin_unlock(&orfs_net_layers_lock);

	return 0;
}

int
orfs_get_net_layers_list(char *buf)
{
	int len = 0;
	struct list_head *tmp;

	/* this code is inspired from get_filesystem_list */
	spin_lock(&orfs_net_layers_lock);

	tmp = orfs_net_layers.next;
	while (tmp != &orfs_net_layers && len < PAGE_SIZE - 80) {
		struct orfs_net_layer *layer;
		layer = list_entry(tmp, struct orfs_net_layer, layers);
		len += sprintf(buf + len, "%s\t%d\t[%s]\n",
			       layer->name, atomic_read(&layer->count),
			       layer->owner->name);
		tmp = tmp->next;
	}

	spin_unlock(&orfs_net_layers_lock);

	return len;
}

static struct orfs_net_layer *
orfs_get_net_layer(char *name)
{
	struct orfs_net_layer *layer;

	if (!*name) {
		/* if no layer asked, use the first one */
		ORFS_DEBUG(NET, "Using default network layer\n");

		spin_lock(&orfs_net_layers_lock);
		if (list_empty(&orfs_net_layers)) {
			ORFS_DEBUG(NET, "No layer registered\n");
			goto error;
		}
		layer = list_entry(orfs_net_layers.next,
				   struct orfs_net_layer, layers);
		ORFS_DEBUG(NET, "Default network layer is %s\n", layer->name);
	} else {
		/* looking for asked network layer */
		ORFS_DEBUG(NET, "Looking for network layer %s\n", name);

		spin_lock(&orfs_net_layers_lock);
		layer = orfs_find_net_layer(name);
		if (!layer) {
			ORFS_ERROR("Cannot find unexisting network layer %s\n",
				   name);
			goto error;
		}
	}

	if (
#if ORFS_LINUX_24
		   !try_inc_mod_count(layer->owner)
#elif ORFS_LINUX_26
		   !try_module_get(layer->owner)
#endif
	    ) {
		ORFS_ERROR("Cannot get layer %s module\n", name);
		goto error;
	}
	atomic_inc(&layer->count);
	spin_unlock(&orfs_net_layers_lock);
	return layer;

      error:
	spin_unlock(&orfs_net_layers_lock);
	return NULL;
}

static void
orfs_put_net_layer(struct orfs_net_layer *layer)
{
	atomic_dec(&layer->count);
#if ORFS_LINUX_24
	__MOD_DEC_USE_COUNT(layer->owner);
#elif ORFS_LINUX_26
	module_put(layer->owner);
#endif
}

/************
 * Mounting *
 ************/

static inline void
orfs_init_sbi(struct orfs_sb_info *sbi)
{
	memset(sbi, 0, sizeof (*sbi));
	sbi->opts.specific_flags = ORFS_INIT_SPECIFIC_FLAGS;
	sbi->opts.revalidate_delay.dentry = ORFS_DEFAULT_DENTRY_REVALIDATE_DELAY;
	sbi->opts.revalidate_delay.inode = ORFS_DEFAULT_INODE_REVALIDATE_DELAY;
	sbi->opts.revalidate_delay.page = ORFS_DEFAULT_PAGE_REVALIDATE_DELAY;
}

static void
orfs_update_sbi_opts(struct orfs_sb_info *sbi,
		     struct orfs_mount_flagmask *specific,
		     struct orfs_mount_revalidate_delay *delay)
{
	sbi->opts.specific_flags
	    = (sbi->opts.specific_flags & ~specific->mask) | specific->flags;

	if (specific->mask & ORFS_DENTRY_REVALIDATE_DELAY_MASK)
		sbi->opts.revalidate_delay.dentry = delay->dentry;
	if (specific->mask & ORFS_INODE_REVALIDATE_DELAY_MASK)
		sbi->opts.revalidate_delay.inode = delay->inode;
	if (specific->mask & ORFS_PAGE_REVALIDATE_DELAY_MASK)
		sbi->opts.revalidate_delay.page = delay->page;
}

#if ORFS_LINUX_24
struct super_block *
orfs_read_super(struct super_block *sb, void *raw_data, int silent)
#elif ORFS_LINUX_26
int
orfs_fill_super(struct super_block *sb, void *raw_data, int silent)
#endif
{
	struct orfs_user_mount_options *options
	    = (struct orfs_user_mount_options *) raw_data;
	struct orfs_net_layer *layer;
	struct orfs_net_operations *nops;
	struct orfs_sb_info *sbi;
	struct orfs_client_request *creq;
	struct dentry *root;
	int error = 0;

	ORFS_DEBUG(FS, "READ SUPER called\n");

	sb->s_magic = ORFS_SUPER_MAGIC;
	sb->s_op = &orfs_sops;
	sb->s_maxbytes = MAX_LFS_FILESIZE;

	/* This field is only used for statfs and stat
	 * to get the "prefered size" for I/O.
	 */
	sb->s_blocksize = ORFS_DEFAULT_BLOCKSIZE;

	/* This field is given to each newly allocated inode.
	 * Inode will only use it to allocate buffers for
	 * the page cache. We are not using the page cache.
	 */
	sb->s_blocksize_bits = ORFS_FAKE_BLOCKSIZE_BITS;

	/* initialize sbi */
	sbi = kmalloc(sizeof (*sbi), GFP_KERNEL);
	orfs_init_sbi(sbi);
	orfs_update_sbi_opts(sbi, &options->specific,
			     &options->revalidate_delay);
#ifdef CONFIG_ORFS_STATS
	orfs_sbi_full_reset_stats(sbi);
#endif				/* CONFIG_ORFS_STATS */
	SET_ORFS_SBI(sb, sbi);

	ORFS_DEBUG(FS, "SBI filled\n");

	/* get network layer and ops */
	error = -EINVAL;
	layer =
	    orfs_get_net_layer(options->specific.
			       mask & ORFS_SERVER_TYPE_MASK
			       ? options->common.layer
			       : "");
	if (!layer)
		goto error;
	sbi->layer = layer;
	nops = sbi->nops = layer->nops;

	/* check nops */
	if (ORFS_SBI_IS_MULTICLIENT(sbi)) {
	  if (!nops->multi_remote_call) {
	    ORFS_ERROR("Cannot mount multips on %s layer"
		       " without a multi_remote_call routine\n",
		       layer->name);
	    goto error_with_layer;
	  }
	} else {
	  if (!nops->mono_remote_call) {
	    ORFS_ERROR("Cannot mount monops on %s layer"
		       " without a mono_remote_call routine\n",
		       layer->name);
	    goto error_with_layer;
	  }
	}

	/* connecting to server */
	if ((error = nops->connect(sbi, options)) < 0)
		goto error_with_layer;

	ORFS_DEBUG(NET, "Connected to server %s\n", sbi->server_path);

	/* init creqs */
	if ( ORFS_SBI_IS_MULTICLIENT(sbi) ) {
	        INIT_LIST_HEAD(&sbi->param.multi.creq_list);
        	spin_lock_init(&sbi->param.multi.creq_lock);
	} else
		sema_init(&sbi->param.mono.creq_sem, 1);

	/* launch ORFSD */
	if (ORFS_SBI_IS_MULTICLIENT(sbi) && nops->kt_loop)
		if ((error = orfs_launch_orfsd(sbi)) < 0)
			goto error_with_connection;

	/* Init Request */
	creq = orfs_get_client_request(sbi);
	ORFS_CREQ2REQ(creq)->type = ORFS_FS_REQ;
	ORFS_CREQ2REQ(creq)->subtype = ORFS_INIT_FS_REQ;
	if ((error = remote_call(sbi, nops, creq)) < 0) {
		ORFS_ERROR("Network Error during Init Request Processing\n");
		goto error_with_creq;
	}
	if ((error = orfs_process_init_answer(sbi, ORFS_CREQ2ANS(creq))) < 0) {
		goto error_with_creq;
	}
	orfs_put_client_request(sbi, creq);

	/* Finalize connection */
	sbi->connection_status = ORFS_CONNECTION_OK;

	ORFS_DEBUG(NET, "Connection initialized\n");

	/* Connection is ready, we now can get fs root */
	root = d_alloc_root(orfs_iget_root(sb));
	if (!root) {
		ORFS_ERROR("Cannot allocate root dentry\n");
		error = -ENOMEM;
		goto error_with_orfsd;
	}

	/* try to alloc private data */
	error = alloc_dentry_pd(sbi, root);
	if (error < 0) {
		ORFS_ERROR("Cannot allocate root dentry private data\n");
		goto error_with_empty_root;
	}
	root->d_op = &orfs_dops;
	sb->s_root = root;

	/* insert sbi in list */
	spin_lock(&orfs_sbi_list_lock);
	list_add_tail(&sbi->list, &orfs_sbi_list);
	spin_unlock(&orfs_sbi_list_lock);

#if ORFS_LINUX_24
	return sb;
#elif ORFS_LINUX_26
	return 0;
#endif

      error_with_empty_root:
	dput(root);
	goto error_with_orfsd;
      error_with_creq:
	orfs_put_client_request(sbi, creq);
      error_with_orfsd:
	if (ORFS_SBI_IS_MULTICLIENT(sbi) && nops->kt_loop)
		orfs_finish_orfsd(sbi);
      error_with_connection:
	nops->disconnect(sbi);
      error_with_layer:
	orfs_put_net_layer(sbi->layer);
      error:
	kfree(sbi);
#if ORFS_LINUX_24
	return NULL;
#elif ORFS_LINUX_26
	return error;
#endif
}

static void
orfs_put_super(struct super_block *super)
{
	struct orfs_sb_info *sbi = ORFS_SBI(super);
	struct orfs_net_operations *nops  = sbi->nops;
	struct list_head *head;

	ORFS_DEBUG(FS, "PUT SUPER called\n");

	/* if the connection is still ready, send a FINI request to the server */
	if ( sbi->connection_status == ORFS_CONNECTION_OK ) {
		struct orfs_client_request *creq = orfs_get_client_request(sbi);
		ORFS_CREQ2REQ(creq)->type = ORFS_FS_REQ;
		ORFS_CREQ2REQ(creq)->subtype = ORFS_FINI_FS_REQ;
		if (remote_call(sbi, nops, creq) < 0) {
			ORFS_ERROR("Network Error during Fini Request Processing\n");
			/* too late to handle this error */
		}
	        orfs_put_client_request(sbi, creq);

		sbi->connection_status = ORFS_CONNECTION_ERROR;
	}

	/* remove sbi from list */
	spin_lock(&orfs_sbi_list_lock);
	head = &sbi->list;
	__list_del(head->prev, head->next); /* no need to reinit */
	spin_unlock(&orfs_sbi_list_lock);

	/* exit orfsd */
	if (ORFS_SBI_IS_MULTICLIENT(sbi) && nops->kt_loop)
		orfs_finish_orfsd(sbi);

	/* disconnect */
	sbi->nops->disconnect(sbi);

	/* release network layer */
	orfs_put_net_layer(sbi->layer);

	kfree(sbi);
}

static void
orfs_umount_begin(struct super_block *super)
{
	struct orfs_sb_info *sbi = ORFS_SBI(super);

	ORFS_DEBUG(FS, "UMOUNT BEGIN called\n");

	orfs_killall_client_requests(sbi);
}

static int
orfs_remount_fs(struct super_block *super, int *flags, char *data)
{
	struct orfs_user_remount_options *options
	    = (struct orfs_user_remount_options *) data;
	struct orfs_sb_info *sbi = ORFS_SBI(super);

	ORFS_DEBUG(FS, "REMOUNT FS called\n");

	/* update generic flags */
	*flags =
	    (super->s_flags & ~options->generic.mask) | options->generic.flags;

	/* select only flags that are available under remount */
	if (options->specific.mask & ~ORFS_REMOUNT_SPECIFIC_MASK)
		ORFS_NOTICE("Several remount flags were discarded\n");
	options->specific.flags &= ORFS_REMOUNT_SPECIFIC_MASK;
	options->specific.mask &= ORFS_REMOUNT_SPECIFIC_MASK;

	/* update specific flags */
	orfs_update_sbi_opts(sbi, &options->specific,
			     &options->revalidate_delay);

	return 0;
}

static int
orfs_show_options(struct seq_file *seq_file, struct vfsmount *vfsmount)
{
	struct super_block *sb = vfsmount->mnt_sb;
	struct orfs_sb_info *sbi = ORFS_SBI(sb);

	ORFS_DEBUG(FS, "SHOW OPTIONS called\n");

	seq_printf(seq_file, ",server=%s", sbi->server_path);

	if (ORFS_SBI_IS_NON_CONCURRENT(sbi)) {
		seq_puts(seq_file, ",nonconc");
	} else {
		seq_puts(seq_file, ",conc");
		seq_printf(seq_file, ",dentry=%ld",
			   sbi->opts.revalidate_delay.dentry);
		seq_printf(seq_file, ",inode=%ld",
			   sbi->opts.revalidate_delay.inode);
		seq_printf(seq_file, ",page=%ld",
			   sbi->opts.revalidate_delay.page);
	}

	if (!ORFS_SBI_IS_MULTICLIENT(sbi))
		seq_printf(seq_file, ",monops");

	if (ORFS_SBI_IS_ALWAYS_O_DIRECT(sbi))
		seq_printf(seq_file, ",odirect");

	return 0;
}

/**********************
 * Inode Manipulation *
 **********************/

static void
orfs_read_inode(struct inode *inode)
{
	ORFS_DEBUG(FS, "READ INODE called for inode %lld\n",
		   (long long) inode->i_ino);

	/* set mode to 0 to mark the inode as "initializing".
	 * This will provoque inode_pd allocation when
	 * update_attr will be called for the first time.
	 */
	inode->i_mode = 0;

	/* set default rdev value */
	inode->i_rdev = 0;

 	/* blksize is the "prefered size for I/O" */
	inode->i_blksize = ORFS_SBI(inode->i_sb)->opts.blksize;

	/* Inode are set to be NOATIME:
	 * The real atime is kept by the server which updates
	 * it at each read request.
	 * ATIME[_SET] is only used through utime[s] syscall
	 * which will imply an explicit remote setattr.
	 * This flag disable marking inode dirty due to atime
	 * update (in read requests), and also disable atime
	 * updating. But atime will be updated from the server
	 * after all requests, especially read requests.
	 */
	inode->i_flags |= S_NOATIME;
}

static void
orfs_clear_inode(struct inode *inode)
{
	ORFS_DEBUG(FS, "CLEAR INODE Called for inode %lld\n",
		   (long long) inode->i_ino);
	free_inode_pd(ORFS_SBI(inode->i_sb), inode);
}

/**************************
 * Other super operations *
 **************************/

static int
orfs_statfs(struct super_block *super,
#if ORFS_LINUX_24
	    struct statfs *buf
#elif ORFS_LINUX_26
	    struct kstatfs *buf
#endif
    )
{
	struct orfs_sb_info *sbi = ORFS_SBI(super);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	int error;

	ORFS_DEBUG(REQ, "STATFS called\n");

	ORFS_UPDATE_REQ_STAT(sbi, statfs);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_FS_REQ;
	req->subtype = ORFS_STATFS_FS_REQ;

	error = remote_call(sbi, sbi->nops, creq);

	if (error) {
		ORFS_ERROR("Remote call for STATFS failed with error %d\n",
			   -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);

		if (ans->result == ORFS_SUCCESS) {
			buf->f_type = ORFS_SUPER_MAGIC;
			buf->f_bsize = ORFS_SBI(super)->opts.blksize;
			buf->f_blocks =
			    ans->target.fs.data.statfs.fsattr.blocks;
			buf->f_bfree = ans->target.fs.data.statfs.fsattr.bfree;
			buf->f_bavail =
			    ans->target.fs.data.statfs.fsattr.bavail;
			buf->f_files = ans->target.fs.data.statfs.fsattr.files;
			buf->f_ffree = ans->target.fs.data.statfs.fsattr.ffree;
			// long f_sid = ????
			buf->f_namelen = ORFS_NAME_LEN;
#if ORFS_LINUX_26
			// long f_frsize = ????
#endif
			// long f_spare[5] = ????
		} else {
			// impossible for now
			error = -EIO;
		}
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

struct super_operations orfs_sops = {
	.write_super = NULL,
	.write_super_lockfs = NULL,
	.unlockfs = NULL,

	.put_super = orfs_put_super,
	.remount_fs = orfs_remount_fs,
	.umount_begin = orfs_umount_begin,
	.show_options = orfs_show_options,

	.statfs = orfs_statfs,

	.read_inode = orfs_read_inode,
	.dirty_inode = NULL,	/* inodes always are synched */
	.write_inode = NULL,	/* nothing to sync in inodes */
	.put_inode = NULL,	/* nothing to free here when releasing an inode */
	.delete_inode = NULL,	/* nothing to free when deleting an inode */
	.clear_inode = orfs_clear_inode,	/* need to clear inode data */
#if ORFS_LINUX_26
	.alloc_inode = NULL,
	.destroy_inode = NULL,
	.drop_inode = NULL,	/* nothing special to free here */
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,21)
	.sync_fs = NULL,	/* nothing to sync in fs */
#endif
};
