/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/fs.h>

#include "orfs_fs.h"
#include "orfs_private_data.h"
#include "orfs_request.h"

/*************************************
 * Symlink Specific Inode Operations *
 *************************************/

extern int orfs_setattr(struct dentry *dentry, struct iattr *iattr);
#if ORFS_LINUX_24
extern int orfs_getattr(struct dentry *dentry, struct iattr *iattr);
extern int orfs_inode_revalidate(struct dentry *dentry);
#elif ORFS_LINUX_26
extern int orfs_getattr(struct vfsmount *mnt,
			struct dentry *dentry, struct kstat *kst);
#endif

/* Used by symlink to fill the inode symlink cache
 * after symlink request succeeded on server.
 * Need to copy the user old_buffer in a kernel one.
 * Choosed to use network buffers as readlink does.
 */
int
orfs_autofill_symlink_cache(struct inode *inode,
			    const void *old_buffer, size_t path_len)
{
	char *buffer;
	int error = -ENOMEM;

	buffer = kmalloc(path_len + 1, GFP_KERNEL);
	if (!buffer) {
		ORFS_ERROR("Failed to alloc symlink cache\n");
		goto error;
	}

	memcpy(buffer, old_buffer, path_len + 1);

	ORFS_INODE_PD(inode)->symlink_cache = buffer;

	return 0;

      error:
	return error;
}

/* Real remote readlink request.
 */
static inline int
__orfs_readlink(struct inode *inode, char *buffer)
{
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	struct orfs_client_request *creq;
	struct orfs_request *req;
	int error;

	ORFS_DEBUG(REQ, "Real READLINK called for inode %lld\n",
		   (long long) inode->i_ino);

	ORFS_UPDATE_REQ_STAT(sbi, readlink);

	creq = orfs_get_client_request(sbi);
	req = ORFS_CREQ2REQ(creq);

	req->type = ORFS_INODE_REQ;
	req->subtype = ORFS_READLINK_INODE_REQ;
	fill_inode_id_from_pd(inode, &req->target.inode.inode);
	creq->out_buf = buffer;
	creq->out_buf_len = sbi->opts.symlen + 1;
	creq->out_buf_flags = ORFS_BUF_USER_ANY;

	error = remote_call(sbi, sbi->nops, creq);

	if (error < 0) {
		ORFS_ERROR
		    ("Remote call of Real READLINK failed with error %d\n",
		     -error);
	} else {
		struct orfs_answer *ans = ORFS_CREQ2ANS(creq);
		struct orfs_getattr_data *node_attr = &ans->target.inode.node_attr;

		switch (ans->result) {
		case ORFS_SUCCESS:
			/* update node attributes */
			orfs_update_inode_attr(inode, node_attr);
			/* add zero at the end */
			buffer[ans->target.inode.data.readlink.count] = '\0';
			break;
		case ORFS_BAD_INODE_TYPE:
			if ( ORFS_SBI_IS_NON_CONCURRENT(sbi) )
				ORFS_ERROR("READLINK returns bad inode %lld mode %lld\n",
					   (long long) inode->i_ino, (long long) node_attr->mode);
			/* drop inode */
			orfs_invalidate_lnk_inode(inode);
			error = -EINVAL;
			break;
		default:
			ORFS_ALERT("Unrecognized error %ld\n", (long) error);
			BUG();
		}
	}

	orfs_put_client_request(sbi, creq);
	return error;
}

/* Fills inode symlink cache with
 * server's one.
 */
static inline int
orfs_fill_symlink_cache(struct inode *inode)
{
	struct orfs_sb_info *sbi = ORFS_SBI(inode->i_sb);
	char *buffer;
	int error = -ENOMEM;

	buffer = kmalloc(sbi->opts.symlen + 1, GFP_KERNEL);
	if (!buffer) {
		ORFS_ERROR("Failed to alloc symlink cache\n");
		goto error;
	}

	error = __orfs_readlink(inode, buffer);
	if (error)
		goto error_with_buffer;

	ORFS_INODE_PD(inode)->symlink_cache = buffer;

	return 0;

      error_with_buffer:
	kfree(buffer);
      error:
	return error;
}

/* Releases symlink cache.
 */
void
orfs_kill_symlink_cache(struct inode *inode)
{
	char *symlink = ORFS_INODE_PD(inode)->symlink_cache;

	if (symlink)
		kfree(symlink);
}

static int
orfs_readlink(struct dentry *dentry, char *buffer, int buflen)
{
	struct inode *inode = dentry->d_inode;
	char *symlink;
	int error;

	if (!ORFS_INODE_PD(inode)->symlink_cache
	    && (error = orfs_fill_symlink_cache(inode)) < 0)
		return error;

	symlink = ORFS_INODE_PD(inode)->symlink_cache;

	return vfs_readlink(dentry, buffer, buflen, symlink);
}

static int
orfs_follow_link(struct dentry *dentry, struct nameidata *nd)
{
	struct inode *inode = dentry->d_inode;
	char *symlink;
	int error;

	if (!ORFS_INODE_PD(inode)->symlink_cache
	    && (error = orfs_fill_symlink_cache(inode)) < 0)
		return error;

	symlink = ORFS_INODE_PD(inode)->symlink_cache;

	return vfs_follow_link(nd, symlink);
}

struct inode_operations orfs_symlink_iops = {
	.readlink = orfs_readlink,
	.follow_link = orfs_follow_link,
	.permission = NULL,
	.setattr = orfs_setattr,	// useless ?
	.getattr = orfs_getattr,
#if ORFS_LINUX_24
	.revalidate = orfs_inode_revalidate,
#endif
};
