/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <sys/mount.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "orfs_mount.h"
#include "orfs_common_network.h"
#include "mount_options.h"


/*******************
 * Cmdline Parsing *
 *******************/

/* need two macro to stringified the argument */
#define xstr(x) str(x)
#define str(x) #x

enum { MOUNT, UMOUNT, REMOUNT, MOVE_BIND };

static char *callname;
static int verbose = 0;
static int type;
static char *stropts = "";

#define MOVE_BIND_MASK (MS_MOVE|MS_BIND|MS_REC)

static int special_flags = MS_MGC_VAL;
static int umount_flags = 0;


static
void exit_with_usage()
{
  fprintf(stderr,
	  "Usage : %s [-murBMflRv] [-o options] [server|old_mnt_path] mnt_path\n"
	  "  --mnt|-m     mount (default when " xstr(MOUNT_BINARY) " is called)\n"
          "            " xstr(MOUNT_BINARY) " [-v] [-o options] server mnt_path\n"
	  "  --umnt|-u    umount (default when " xstr(UMOUNT_BINARY) " is called)\n"
          "            " xstr(UMOUNT_BINARY) " [-v] [-l] [-f] mnt_path\n"
	  "  --rmnt|-r    remount (default when " xstr(REMOUNT_BINARY) " is called)\n"
          "            " xstr(REMOUNT_BINARY) " [-v] [-o options] mnt_path\n"
	  "  --move|-M    atomically move a mount point\n"
          "            " xstr(MOUNT_BINARY) " --move [-v] old_mnt_path mnt_path\n"
	  "  --bind|-B    bind a new mount point to an old one\n"
          "            " xstr(MOUNT_BINARY) " --bind [-v] old_mnt_path mnt_path\n"
	  "  --recbind|-R recursively bind a new mount point to an old one\n"
          "            " xstr(MOUNT_BINARY) " --recbind [-v] old_mnt_path mnt_path\n"
	  "\n"
	  "  --force|-f   force umount\n"
	  "  --lazy|-l    lazy umount\n"
	  "  --verbose|-v verbose mode\n"
	  "\n"
	  "  -o    set options, from a comma separated string of\n"
	  "          net=tcp/gm/mx            network layer\n"
	  "          port=<int>               server TCP/GM port\n"
	  "          ep=<int>                 server MX endpoint\n"
	  "          board=<int>              server GM/MX board\n"
	  "          client_board=<int>       server GM/MX client_board\n"
	  "          [non]conc                concurrent mode\n"
	  "          dentry/inode/page=<int>  revalidate delay\n"
	  "          monops/multips           allow only one or multiple processes simultaneously\n"
	  "          odirect                  always use O_DIRECT\n"
	  "          nodirect                 use O_DIRECT only when passed to open\n"
	  "\n",
	  callname);
  exit(-1);
}


static
void parse_cmdline(int argc, char *const argv[])
{
  char c;
  static struct option long_opts[] = {
    { "mnt", 0, NULL, 'm' },
    { "umnt", 0, NULL, 'u' },
    { "rmnt", 0, NULL, 'r' },
    { "move", 0, NULL, 'M' },
    { "bind", 0, NULL, 'B' },
    { "recbind", 0, NULL, 'R' },
    { "force", 0, NULL, 'f' },
    { "lazy", 0, NULL, 'l' },
    { "verbose", 0, NULL, 'v' },
    { NULL, 0, NULL, 0 },
  };
  
  callname = strrchr(argv[0], '/')+1;

  if ( !strcmp(callname, xstr(UMOUNT_BINARY)) )
    type = UMOUNT;
  else if ( !strcmp(callname, xstr(REMOUNT_BINARY)) )
    type = REMOUNT;
  else
    type = MOUNT;

  while ( (c=getopt_long(argc, argv, "murBMRflvo:", long_opts, NULL))
	  != EOF ) {
    switch ( c ) {
    case 'm':
      type = MOUNT;
      special_flags = special_flags & (~MOVE_BIND_MASK);
      break;
    case 'u':
      type = UMOUNT;
      special_flags = special_flags & (~MOVE_BIND_MASK);
      break;
    case 'r':
      type = REMOUNT;
      special_flags = special_flags & (~MOVE_BIND_MASK);
      break;
    case 'M':
      type = MOVE_BIND;
      special_flags = (special_flags & (~MOVE_BIND_MASK)) | MS_MOVE;
      break;
    case 'B':
      type = MOVE_BIND;
      special_flags = (special_flags & (~MOVE_BIND_MASK)) | MS_BIND;
     break;
    case 'R':
      type = MOVE_BIND;
      special_flags = (special_flags & (~MOVE_BIND_MASK)) | MS_BIND | MS_REC;
      break;
    case 'f':
      umount_flags |= MNT_FORCE;
      break;
    case 'l':
      umount_flags |= MNT_DETACH;
      break;
    case 'v':
      special_flags |= MS_VERBOSE;
      verbose = 1;
      break;
    case 'o':
      stropts = optarg;
      break;

    case '?':
    default:
      exit_with_usage(argv);
    }
  }
}


/********************
 * Syscall wrappers *
 ********************/

static
int try_mount(char *server, char *mnt_path, char *stropts)
{
  int res=0;
  struct orfs_parsed_mount_options parsed_opts = {{0}};
  struct orfs_user_mount_options mount_opts = {{0}};

  parsed_opts.generic.flags = special_flags;

  if ( verbose )
    fprintf(stderr, "Mounting server %s on %s with options %s\n",
	    server, mnt_path, stropts);
  
  if ( parse_options(stropts, server, &parsed_opts) < 0 )
    goto out;
  
  memcpy(&mount_opts.specific, &parsed_opts.specific,
	 sizeof(mount_opts.specific));
  memcpy(&mount_opts.revalidate_delay, &parsed_opts.revalidate_delay,
	 sizeof(mount_opts.revalidate_delay));
  memcpy(&mount_opts.common, &parsed_opts.common,
	 sizeof(mount_opts.common));
  memcpy(&mount_opts.tcp, &parsed_opts.tcp,
	 sizeof(mount_opts.tcp));
  memcpy(&mount_opts.gm, &parsed_opts.gm,
	 sizeof(mount_opts.gm));
  memcpy(&mount_opts.mx, &parsed_opts.mx,
	 sizeof(mount_opts.mx));

  res = mount("", mnt_path, "orfs", parsed_opts.generic.flags, &mount_opts);

 out:
  if ( res < 0 )
    printf("ORFS mount: mount failed (errno %m)\n");

  return res;
}


static
int try_remount(char *mnt_path, char *stropts)
{
  int res=0;
  struct orfs_parsed_mount_options parsed_opts = {{0}};
  struct orfs_user_remount_options remount_opts = {{0}};

  parsed_opts.generic.flags = special_flags | MS_REMOUNT;

  if ( verbose )
    fprintf(stderr, "Remounting %s with options %s\n",
	    mnt_path, stropts);

  if ( parse_options(stropts, NULL, &parsed_opts) < 0 )
    goto out;

  memcpy(&remount_opts.generic, &parsed_opts.generic,
	 sizeof(remount_opts.generic));
  memcpy(&remount_opts.specific, &parsed_opts.specific,
	 sizeof(remount_opts.specific));
  memcpy(&remount_opts.revalidate_delay, &parsed_opts.revalidate_delay,
	 sizeof(remount_opts.revalidate_delay));

  res = mount("", mnt_path, "orfs", parsed_opts.generic.flags, &remount_opts);
  
 out:
  if ( res < 0 )
    printf("ORFS remount: remount failed (errno %m)\n");

  return res;
}


static
int try_umount(char *mnt_path)
{
  int res;

  if ( verbose )
    fprintf(stderr, "Umounting %s\n",
	    mnt_path);

  res = umount2(mnt_path, umount_flags);

  if ( res < 0 )
    printf("ORFS umount: umount failed (errno %m)\n");

  return res;
}


static
int try_move_bind(char *old_mnt_path, char *mnt_path)
{
  int res=0;
  
  if ( verbose )
    fprintf(stderr, "%s %s to %s\n",
	    special_flags & MS_BIND ? special_flags & MS_REC ?
	    "Recursively binding" : "Binding" : "Moving",
	    old_mnt_path, mnt_path);
  
  res = mount(old_mnt_path, mnt_path, NULL, special_flags, NULL);

  if ( res < 0 )
    printf("ORFS move/bind mount: mount failed (errno %m)\n");

  return res;
}


/****************
 * Main Program *
 ****************/

int main(int argc, char *argv[])
{
  int res = -1;

  parse_cmdline(argc, argv);

  argc -= optind;
  argv += optind;

  /* mount, remount or umount ? */
  switch ( type ) {
  case MOUNT:
    if ( argc < 2 )
      exit_with_usage(argv);
    res = try_mount(argv[0], argv[1], stropts);
    break;
  case UMOUNT:
    if ( argc < 1 )
      exit_with_usage(argv);
    res = try_umount(argv[0]);
    break;
  case REMOUNT:
    if ( argc < 1 )
      exit_with_usage(argv);
    res = try_remount(argv[0], stropts);
    break;
  case MOVE_BIND:
    if ( argc < 2 )
      exit_with_usage(argv);
    res = try_move_bind(argv[0], argv[1]);
  }

  return res;
}
