/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <sys/mount.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>

#include "orfs_common_network.h"
#include "orfs_mount.h"
#include "mount_options.h"


/*************************
 * Generic Mount Options *
 *************************/

struct generic_option {
  const char *name;
  unsigned long flag;
  unsigned long mask;
};


static 
struct generic_option generic_opts[] = {
  { "dev", 0, MS_NODEV }, // useless with ORFS ?
  { "nodev", MS_NODEV, MS_NODEV }, // useless with ORFS ?
  { "exec", 0, MS_NOEXEC },
  { "noexec", MS_NOEXEC, MS_NOEXEC },
  { "ro", MS_RDONLY, MS_RDONLY },
  { "rw", 0, MS_RDONLY },
  { "sync", MS_SYNCHRONOUS, MS_SYNCHRONOUS }, // with ORFS ?
  { "async", 0, MS_SYNCHRONOUS }, // with ORFS ?
  { "suid", 0, MS_NOSUID },
  { "nosuid", MS_NOSUID, MS_NOSUID },
  { "remount", MS_REMOUNT, 0 },
  // mandlock MS_MANDLOCK difficult ?
  // noatime MS_NOATIME useless ?
  // nodiratime MS_NODIRATIME useless ?
  { NULL, 0, 0 },
  // "default" = "rw" "suid" "dev" "exec" "auto" "nouser" "async";
  // remount allow ro/rw + sync + mandlock + noatime + nodiratime
};


static
int parse_generic_option(char *name, char *value,
			 struct orfs_parsed_mount_options *opts)
{
  struct generic_option *opt;

  for(opt=&generic_opts[0];opt->name;opt++) {
    if ( strcmp(name, opt->name) )
      continue;
    opts->generic.flags = (opts->generic.flags&~opt->mask)|opt->flag;
    opts->generic.mask |= opt->mask;
    return 0;
  }
  return -1;
}


/*******************************
 * ORFS Specific Mount Options *
 *******************************/

static
int parse_specific_option(char *opt_name, char *value,
			  struct orfs_parsed_mount_options *opts)
{
  /******************************
   * handle communication layer *
   ******************************/
  if ( !strcmp(opt_name, "net") ) {
    int len;
    if (!value || !*value) {
      fprintf (stderr,
	       "ORFS mount error: "
	       "net option requires an argument\n");
      return -1;
    }
    len = strlen(value);
    if ( len > ORFS_MAX_SERVER_TYPE_LEN ) {
      fprintf(stderr,
	      "ORFS mount error; "
	      "net option %s to long (max is %d)\n",
	      value, ORFS_MAX_SERVER_TYPE_LEN);
      return -1;
    }
    strcpy(opts->common.layer, value);
    opts->specific.mask |= ORFS_SERVER_TYPE_MASK;
  }

  /********************************
   * handle TCP or GM server port *
   ********************************/
  else if ( !strcmp(opt_name, "port") ) {
    unsigned long port;
    if (!value || !*value) {
      fprintf(stderr,
	      "ORFS mount error: "
	      "port option requires an argument\n");
      return -1;
    }
    port = strtol(value, &value, 10);
    if ( *value ) {
      fprintf(stderr,
	      "ORFS mount error: "
	      "Invalid port option: %s\n",
	      value);
      return -1;
    }
    opts->tcp.port = port;
    opts->gm.port = port;
    opts->specific.mask |= ORFS_SERVER_PORT_MASK;
  }

  /*****************************
   * handle MX server endpoint *
   *****************************/
  else if ( !strcmp(opt_name, "ep") ) {
    unsigned long ep;
    if (!value || !*value) {
      fprintf(stderr,
	      "ORFS mount error: "
	      "ep option requires an argument\n");
      return -1;
    }
    ep = strtol(value, &value, 10);
    if ( *value ) {
      fprintf(stderr,
	      "ORFS mount error: "
	      "Invalid ep option: %s\n",
	      value);
      return -1;
    }
    opts->mx.ep = ep;
    opts->specific.mask |= ORFS_SERVER_EP_MASK;
  }

  /********************************
   * handle GM or MX server board *
   ********************************/
  else if ( !strcmp(opt_name, "board") ) {
    unsigned long board;
    if (!value || !*value) {
      fprintf(stderr,
	      "ORFS mount error: "
	      "board option requires an argument\n");
      return -1;
    }
    board = strtol(value, &value, 10);
    if ( *value ) {
      fprintf(stderr,
	      "ORFS mount error: "
	      "Invalid board option: %s\n",
	      value);
      return -1;
    }
    opts->gm.board = board;
    opts->mx.board = board;
    opts->specific.mask |= ORFS_SERVER_BOARD_MASK;
  }

  /********************************
   * handle GM or MX client board *
   ********************************/
  else if ( !strcmp(opt_name, "client_board") ) {
    unsigned long board;
    if (!value || !*value) {
      fprintf(stderr,
	      "ORFS mount error: "
	      "client_board option requires an argument\n");
      return -1;
    }
    board = strtol(value, &value, 10);
    if ( *value ) {
      fprintf(stderr,
	      "ORFS mount error: "
	      "Invalid client_board option: %s\n",
	      value);
      return -1;
    }
    opts->gm.client_board = board;
    opts->mx.client_board = board;
    opts->specific.mask |= ORFS_CLIENT_BOARD_MASK;
  }

  /*****************************
   * handle multiclient or not *
   *****************************/
  else if ( !strcmp(opt_name, "multips") ) {
    opts->specific.flags |= ORFS_MULTICLIENT_MASK;
    opts->specific.mask |= ORFS_MULTICLIENT_MASK;
  }
  else if ( !strcmp(opt_name, "monops") ) {
    opts->specific.flags &= ~ORFS_MULTICLIENT_MASK;
    opts->specific.mask |= ORFS_MULTICLIENT_MASK;
  }
  
  /*********************************
   * handle concurrent mode or not *
   *********************************/
  else if ( !strcmp(opt_name, "nonconc") ) {
    opts->specific.flags |= ORFS_NON_CONCURRENT_MASK;
    opts->specific.mask |= ORFS_NON_CONCURRENT_MASK;
  }
  else if ( !strcmp(opt_name, "conc") ) {
    opts->specific.flags &= ~ORFS_NON_CONCURRENT_MASK;
    opts->specific.mask |= ORFS_NON_CONCURRENT_MASK;
  }
  
  /*****************************
   * handle multiclient or not *
   *****************************/
  else if ( !strcmp(opt_name, "odirect") ) {
    opts->specific.flags |= ORFS_ALWAYS_O_DIRECT_MASK;
    opts->specific.mask |= ORFS_ALWAYS_O_DIRECT_MASK;
  }
  else if ( !strcmp(opt_name, "nodirect") ) {
    opts->specific.flags &= ~ORFS_ALWAYS_O_DIRECT_MASK;
    opts->specific.mask |= ORFS_ALWAYS_O_DIRECT_MASK;
  }
  
  /****************************
   * handle revalidate delays *
   ****************************/

  /* dentry revalidate delay */
  else if ( !strcmp(opt_name, "dentry") ) {
    long delay;
    if (!value || !*value) {
      fprintf (stderr,
	       "ORFS mount error: "
	       "dentry option requires an argument\n");
      return -1;
    }
    delay = strtol (value, &value, 10);
    if (*value ) {
      fprintf (stderr,
	       "ORFS mount error: "
	       "Invalid dentry option: %s\n",
	       value);
      return -1;
    }
    if ( delay < 0 ) {
      delay = 0;
    }
    opts->revalidate_delay.dentry = delay;
    opts->specific.mask |= ORFS_DENTRY_REVALIDATE_DELAY_MASK;
  }   

  /* inode revalidate delay */
  else if ( !strcmp(opt_name, "inode") ) {
    long delay;
    if (!value || !*value) {
      fprintf (stderr,
	       "ORFS mount error: "
	       "inode option requires an argument\n");
      return -1;
    }
    delay = strtol (value, &value, 10);
    if (*value ) {
      fprintf (stderr,
	       "ORFS mount error: "
	       "Invalid inode option: %s\n",
	       value);
      return -1;
    }
    if ( delay < 0 ) {
      delay = 0;
    }
    opts->revalidate_delay.inode = delay;
    opts->specific.mask |= ORFS_INODE_REVALIDATE_DELAY_MASK;
  }   

  /* page revalidate delay */
  else if ( !strcmp(opt_name, "page") ) {
    long delay;
    if (!value || !*value) {
      fprintf (stderr,
	       "ORFS mount error: "
	       "page option requires an argument\n");
      return -1;
    }
    delay = strtol (value, &value, 10);
    if (*value ) {
      fprintf (stderr,
	       "ORFS mount error: "
	       "Invalid page option: %s\n",
	       value);
      return -1;
    }
    if ( delay < 0 ) {
      delay = 0;
    }
    opts->revalidate_delay.page = delay;
    opts->specific.mask |= ORFS_PAGE_REVALIDATE_DELAY_MASK;
  }   

  /* unrecognized option */
  else {
    fprintf (stderr,
	     "ORFS mount error: "
	     "Invalid option: %s\n",
	     opt_name);
    return -1;
  }
  
  return 0;
}


/********************
 * Deal with Server *
 ********************/

int parse_server(char *server,
		 struct orfs_parsed_mount_options *opts)
{
  /* TCP specific work */
  struct hostent *server_hostent;
  struct sockaddr_in *ski = &opts->tcp.sockaddr_in;
  /* find server address */
  if ( (server_hostent = gethostbyname(server)) != NULL ) {
    /* fill the sockaddr_in structure */
    memset(ski, 0, sizeof(*ski));
    ski->sin_family = server_hostent->h_addrtype;
    memcpy(&ski->sin_addr,
	   server_hostent->h_addr,
	   server_hostent->h_length);

    /* the sockaddr_in is valid */
    opts->specific.mask |= ORFS_SERVER_TCP_SOCKADDR_MASK;
  }

  /* copy server name */
  strcpy(opts->common.name, server);

  return 0;
}


/***********************
 * Main Option Parsing *
 ***********************/

int parse_options(char *opt_data, char *server,
		  struct orfs_parsed_mount_options *opts)
{
  char *this_char;
  char *value;

  while ((this_char = strsep (&opt_data, ",")) != NULL) {

    if (!*this_char)
      continue;

    if ((value = strchr (this_char, '=')) != NULL)
      *value++ = 0;

    if ( parse_generic_option(this_char, value, opts) < 0
	 &&  parse_specific_option(this_char, value, opts) < 0 ) {
      return -1;
    }
  }

  if ( server )
    if ( parse_server(server, opts) < 0 )
      return -1;

  return 0;
}
