/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __ORFS_COMMON_NETWORK_H__
#define __ORFS_COMMON_NETWORK_H__


#ifndef __KERNEL__
#include <stdint.h>
#endif /* __KERNEL__ */


/* These definitions can be used by
 * some network independant builds.
 * Their definitions cannot depend on NETWORK value
 */


/*******************
 * TCP definitions *
 *******************/

/*
 * ORFS TCP protocol:
 *
 * The server opens a socket and binds it to port ORFS_TCP_SERVER_PORT.
 * The server accepts client connections on this initial socket.
 * It thus gets a pool a sockets which are connected to one client each.
 * Then, the server polls on all these sockets to handle incoming
 * requests, and polls on the initial socket too to handle new client
 * connections.
 *
 * The client sends the request followed by its eventual additional
 * request data. The server sends the answer followed by its eventual
 * additional answer data back to the client.
 */

#define ORFS_TCP_SERVER_PORT 8256

struct orfs_tcp_request_param {
};
struct orfs_tcp_answer_param {
};


/******************
 * GM definitions *
 ******************/

/*
 * ORFS GM protocol:
 *
 * The server opens GM port ORFS_GM_SERVER_PORT.
 * Clients open whatever port they want.
 * There's no connections, but the server stores incoming
 * request port and node id and sends the answer back to it.
 *
 * One priority is reserved for requests/answers, the other
 * one for additional data. Using high priority for
 * requests/answers ensures that they arrive before their
 * additional data.
 *
 * Incoming requests are received in a pool of provided requests.
 * The request describes the length of the eventual additional
 * buffer so that the server knows what data message is coming
 * after each request it receives.
 *
 * The server preprovide a pool a receive buffers for request
 * additional data. Their length is ORFS_GM_PREPROVIDED_REQBUF_LEN.
 * When a client additional data is smaller, its uses the
 * ORFS_GM_PREPROVIDED_REQBUF_SIZE size for the GM message.
 * When larger, it uses the macro ORFS_GM_REQBUF_SIZE(len)
 * to determine the message GM size.
 * At the same time, the server knew this message won't fit
 * in its preprovided buffers. Thus it provided a specific
 * larger one.
 *
 * The client request <rdv_size> field also tells the server
 * what GM size it must use to send the answer additional data
 * back to the client.
 *
 * The client request <rdma_*buf> also allow the client to
 * ask the server to get or put additional buffers through RDMA.
 * Neither a GM size nor a provided buffer are then required.
 */

#define ORFS_GM_SERVER_BOARD 0
#define ORFS_GM_SERVER_PORT 2

/* priority */
#define ORFS_GM_REQ_PRIO GM_HIGH_PRIORITY
#define ORFS_GM_REQBUF_PRIO GM_LOW_PRIORITY
#define ORFS_GM_ANS_PRIO GM_HIGH_PRIORITY
#define ORFS_GM_ANSBUF_PRIO GM_LOW_PRIORITY

/* associated recv event */
#ifdef ORFS_SERVER
#define ORFS_GM_REQ_RECV_EV GM_HIGH_RECV_EVENT
#define ORFS_GM_REQBUF_RECV_EV GM_RECV_EVENT
#endif
#ifdef ORFS_CLIENT
#define ORFS_GM_ANS_RECV_EV GM_HIGH_RECV_EVENT
#define ORFS_GM_ANSBUF_RECV_EV GM_RECV_EVENT
#endif

/* preprovided buffer size */
#define ORFS_GM_PREPROVIDED_REQBUF_LEN 65536

/* size tags for requests and their data */
#define ORFS_GM_REQ_SIZE 31
#define ORFS_GM_PREPROVIDED_REQBUF_SIZE 16
#define ORFS_GM_REQBUF_SIZE(len)		\
	( len <= ORFS_GM_PREPROVIDED_REQBUF_LEN	\
	? ORFS_GM_PREPROVIDED_REQBUF_SIZE	\
	: gm_min_size_for_length(len)		\
	)
#define ORFS_GM_REQBUF_SIZE_MAX 32

/* size tags for answers */
#define ORFS_GM_ANS_SIZE 31

struct orfs_gm_request_param {
  unsigned char rdma_reqbuf;
  unsigned char rdma_ansbuf;
  union {
    uint64_t rdma_ptr;
  } reqbuf;
  union {
    uint8_t rdv_size;
    uint64_t rdma_ptr;
  } ansbuf;
};
struct orfs_gm_answer_param {
};


/******************
 * MX definitions *
 ******************/

/*
 * ORFS MX protocol:
 *
 * The server opens endpoint ORFS_MX_SERVER_EP on board
 * ORFS_MX_SERVER_BOARD with filter ORFS_MX_EP_FILTER.
 * There's no connections, but the server stores incoming
 * request endpoint addr and sends the answer back to it.
 *
 * Maskable bits are used to identify requests and answers
 * Unmaskable bits are used to differenciate the type of
 * messages.
 *
 * Incoming requests are received in a pool of preprovided
 * requests, with a blank mask.
 * The client chooses a 32bit identifier that will be used
 * as a maskable for the request message and its buffer, and
 * the answer and its buffer too.
 * Small request buffers are received in a pool of provided
 * buffer with a blank mask.
 * Big request buffers are received in a specifically provided
 * buffer with full match for the request identifier.
 *
 * Unmaskable bits are used to describe message type:
 * request, small request buffer (preprovided), large request buffer,
 * answer, answer buffer.
 *
 * The client receives its answer and buffer with its full match.
 */

#define ORFS_MX_SERVER_BOARD 0
#define ORFS_MX_SERVER_EP 0
#define ORFS_MX_EP_FILTER 0x26111978UL

#define ORFS_MX_PREPROVIDED_REQBUF_LEN 32768

/* unmaskable bits are used to describe the type of message */
#define ORFS_MX_REQ_TYPE			0x00000001
#define ORFS_MX_PREPROVIDED_REQBUF_TYPE		0x00000002
#define ORFS_MX_LARGE_REQBUF_TYPE		0x00000004
#define ORFS_MX_RDMA_REQBUF_TYPE		0x00000008
#define ORFS_MX_ANS_TYPE			0x00000010
#define ORFS_MX_ANSBUF_TYPE			0x00000020
#define ORFS_MX_RDMA_ANSBUF_TYPE		0x00000040

/* for all message type, we provide the matching info of the message,
 * the mask used by the receiver, and the matching info of the receiver.
 */

/* the request goes in whatever receive buffer in the server */
#define ORFS_MX_REQ_MATCH(id)			((uint64_t)ORFS_MX_REQ_TYPE << 32 | (id))
#define ORFS_MX_REQ_RECV_MATCH			((uint64_t)ORFS_MX_REQ_TYPE << 32)
#define ORFS_MX_REQ_RECV_MASK			MX_MATCH_MASK_BC
/* if not big, the request buffer goes in whatever preprovided receive buffer */
#define ORFS_MX_PREPROVIDED_REQBUF_MATCH(id) 	((uint64_t)ORFS_MX_PREPROVIDED_REQBUF_TYPE << 32 | (id))
#define ORFS_MX_PREPROVIDED_REQBUF_RECV_MATCH	((uint64_t)ORFS_MX_PREPROVIDED_REQBUF_TYPE << 32)
#define ORFS_MX_PREPROVIDED_REQBUF_RECV_MASK	MX_MATCH_MASK_BC
/* if big, the request buffer requires a dedicated receive buffer */
#define ORFS_MX_LARGE_REQBUF_MATCH(id)		((uint64_t)ORFS_MX_LARGE_REQBUF_TYPE << 32 | (id))
#define ORFS_MX_LARGE_REQBUF_RECV_MATCH(id)	((uint64_t)ORFS_MX_LARGE_REQBUF_TYPE << 32 | (id))
#define ORFS_MX_LARGE_REQBUF_RECV_MASK		MX_MATCH_MASK_NONE

/* the answer goes in its dedicated receive buffer */
#define ORFS_MX_ANS_MATCH(id)			((uint64_t)ORFS_MX_ANS_TYPE << 32 | (id))
#define ORFS_MX_ANS_RECV_MATCH(id)		((uint64_t)ORFS_MX_ANS_TYPE << 32 | (id))
#define ORFS_MX_ANS_RECV_MASK			MX_MATCH_MASK_NONE
/* the answer buffer goes in its dedicated receive buffer */
#define ORFS_MX_ANSBUF_MATCH(id)		((uint64_t)ORFS_MX_ANSBUF_TYPE << 32 | (id))
#define ORFS_MX_ANSBUF_RECV_MATCH(id)		((uint64_t)ORFS_MX_ANSBUF_TYPE << 32 | (id))
#define ORFS_MX_ANSBUF_RECV_MASK		MX_MATCH_MASK_NONE

/* extract from matching info */
#define ORFS_MX_MATCH_TYPE(match)			((uint32_t)((match) >> 32))
#define ORFS_MX_MATCH_ID(match)			((uint32_t)(match) & 0xffffffff)

struct orfs_mx_request_param {
  unsigned char rdma_reqbuf;
  unsigned char rdma_ansbuf;
  union {
    uint32_t window_handle;
    uint32_t offset;
  } reqbuf;
  union {
    uint32_t window_handle;
    uint32_t offset;
  } ansbuf;
};
struct orfs_mx_answer_param {
};


#endif /* __ORFS_COMMON_NETWORK_H__ */
