/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __ORFS_PROTOCOL_H__
#define __ORFS_PROTOCOL_H__


#include "orfs_types.h"
#include "orfs_common_network.h"


/****************************
 * Some constants and types *
 ****************************/

/* various contants */
#define ORFS_NAME_LEN 255 /* hardcoded because used in message structure */
#define ORFS_ROOT_INO 1
#define ORFS_PROTOCOL_MAGIC 0x21071980


/**********
 * Errors *
 **********/

enum {
  /* success */
  ORFS_SUCCESS=0,

  /* unrecognized request */
  ORFS_BAD_REQUEST,
  /* unrecognized file descriptor */
  ORFS_BAD_FILE,
  /* invalid parameter */
  ORFS_INVALID_PARAMETER,

  /* not implemented */
  ORFS_FUNCTION_NOT_IMPLEMENTED,

  /* server space overloaded */
  ORFS_NO_FILE_LEFT,
  ORFS_NO_SPACE_LEFT,
  /* invalid parameter during the processing in the server */
  ORFS_SERVER_INVALID_PARAMETER,
  /* server hadn't enough memory to process the request */
  ORFS_SERVER_NOMEM,

  /* trying to remove a non-empty directory */
  ORFS_NOT_EMPTY_DIR,
  /* not uptodate client cache */
  ORFS_EXISTING_DIR_ENTRY,
  ORFS_NO_SUCH_DIR_ENTRY,
  ORFS_BAD_INODE_TYPE,
  ORFS_NO_SUCH_INODE,
  /* not uptodate client cache, for second inode in 2dir req */
  ORFS_BAD_INODE_TYPE_2,
  ORFS_NO_SUCH_INODE_2,
  /* not uptodate client cache, for third inode in link req */
  ORFS_BAD_INODE_TYPE_3,
  
  /* error max */
  ORFS_ERROR_MAX,
};


/*****************
 * Request Types *
 *****************/

/* general request type */
enum {
  ORFS_FILE_REQ=0,
  ORFS_PAGE_REQ,
  ORFS_INODE_REQ,
  ORFS_DIR_REQ,
  ORFS_2DIR_REQ,
  ORFS_FS_REQ,
};

/* file request type */
enum {
  ORFS_CLOSE_FILE_REQ=0,
  ORFS_PREAD_FILE_REQ,
  ORFS_PWRITE_FILE_REQ,
  ORFS_PREADDIR_FILE_REQ,
  ORFS_SYNC_FILE_REQ,
};

/* page request type */
enum {
  ORFS_READ_FULL_PAGE_REQ=0,
  ORFS_WRITE_FULL_PAGE_REQ,
  ORFS_PREPARE_WRITE_PAGE_REQ,
  ORFS_COMMIT_WRITE_PAGE_REQ,
};

/* inode request type */
enum {
  ORFS_SETATTR_INODE_REQ=0,
  ORFS_GETATTR_INODE_REQ,
  ORFS_READLINK_INODE_REQ,
  ORFS_OPEN_INODE_REQ,
};

/* dir request type */
enum {
  ORFS_LOOKUP_DIR_REQ=0,
  ORFS_MKREG_DIR_REQ,
  ORFS_MKDIR_DIR_REQ,
  ORFS_MKNOD_DIR_REQ,
  ORFS_SYMLINK_DIR_REQ,
  ORFS_UNLINK_DIR_REQ,
  ORFS_RMDIR_DIR_REQ,
  ORFS_REVALIDATE_DIR_REQ,
};

/* 2dir request type */
enum {
  ORFS_MOVE_2DIR_REQ=0,
  ORFS_LINK_2DIR_REQ,
};

/* fs request type */
enum {
  ORFS_INIT_FS_REQ=0,
  ORFS_FINI_FS_REQ,
  ORFS_PING_FS_REQ,
  ORFS_STATFS_FS_REQ,
};


/********************************
 * Request Type's Specific Data *
 ********************************/

struct orfs_file_request_data {
  struct orfs_file_id file;
  union {
    struct {
      orfs_off_t offset;
      orfs_off_t count;
    } data_access;
    struct {
      orfs_doff_t offset;
      orfs_data_len_t data_max;
    } preaddir;
    struct {
      uint8_t datasync;
    } fsync;
  } data;
};

struct orfs_page_request_data {
  struct orfs_inode_id inode;
  union {
    struct {
      orfs_off_t offset;
      orfs_off_t count;      
    } data_access;
  } data;
};

struct orfs_inode_request_data {
  struct orfs_inode_id inode;
  union {
    struct {
      struct orfs_setattr_data new_attr;
    } setattr;
  } data;
};

struct orfs_dir_request_data {
  struct orfs_inode_id dir;
  char name[ORFS_NAME_LEN+1];
  orfs_name_len_t name_len;
  union {
    struct {
      orfs_mode_t mode;
      orfs_uid_t uid;
      orfs_gid_t gid;
      orfs_dev_t dev;
    } create;
    struct {
      orfs_path_len_t path_len;
      orfs_uid_t uid;
      orfs_gid_t gid;  
    } symlink;
  } data;
};

struct orfs_2dir_request_data {
  struct orfs_inode_id old_dir;
  char old_name[ORFS_NAME_LEN+1];
  orfs_name_len_t old_name_len;
  struct orfs_inode_id new_dir;
  char new_name[ORFS_NAME_LEN+1];
  orfs_name_len_t new_name_len;
};

struct orfs_fs_request_data {
};


/*******************************
 * Answer Type's Specific Data *
 *******************************/

struct orfs_file_answer_data {
  struct orfs_getattr_data node_attr;
  union {
    struct {
      orfs_off_t count;
    } data_access;
    struct {
      orfs_doff_t count;
      orfs_bool_t eof_reached;
      orfs_data_len_t data_len;
    } preaddir;
  } data;
};

struct orfs_page_answer_data {
  struct orfs_getattr_data node_attr;
  union {
    struct {
      orfs_off_t count;
    } data_access;
  } data;
};

struct orfs_inode_answer_data {
  struct orfs_getattr_data node_attr;
  union {
    struct {
      orfs_path_len_t count;
    } readlink;
    struct {
      struct orfs_file_id file;
    } open;
  } data;
};

struct orfs_dir_answer_data {
  struct orfs_getattr_data father_attr;
  union {
    struct {
      struct orfs_getattr_data node_attr;
    } lookup;
    struct {
      struct orfs_getattr_data new_node_attr;
    } create;
    struct {
      struct orfs_getattr_data node_attr;
    } revalidate;
  } data;
};

struct orfs_2dir_answer_data {
  struct orfs_getattr_data old_father_attr;
  struct orfs_getattr_data new_father_attr;
  struct orfs_getattr_data node_attr;
};

struct orfs_fs_answer_data {
  union {
    struct {
      struct orfs_init_data data;
    } init;
    struct {
      struct orfs_fsattr_data fsattr;
    } statfs;
  } data;
};


/**********************
 * Message Structures *
 **********************/

struct orfs_request {
  orfs_req_type_t type;
  orfs_req_subtype_t subtype;
  orfs_tag_t tag;
  union {
    struct orfs_file_request_data file;
    struct orfs_page_request_data page;
    struct orfs_inode_request_data inode;
    struct orfs_dir_request_data dir;
    struct orfs_2dir_request_data twodir;
    struct orfs_fs_request_data fs;
  } target;
  union {
    struct orfs_tcp_request_param tcp;
    struct orfs_gm_request_param gm;
    struct orfs_mx_request_param mx;
  } net;
};

struct orfs_answer {
  orfs_result_t result;
  orfs_req_type_t type;
  orfs_req_subtype_t subtype;
  orfs_tag_t tag;
  union {
    struct orfs_file_answer_data file;
    struct orfs_page_answer_data page;
    struct orfs_inode_answer_data inode;
    struct orfs_dir_answer_data dir;
    struct orfs_2dir_answer_data twodir;
    struct orfs_fs_answer_data fs;
  } target;    
  union {
    struct orfs_tcp_answer_param tcp;
    struct orfs_gm_answer_param gm;
    struct orfs_mx_answer_param mx;
  } net;
};


/******************************
 * Get Optional Buffer Length *
 ******************************/

extern inline
int get_request_buffer_len(struct orfs_request *req, size_t *len)
{
  orfs_req_type_t type = req->type;
  orfs_req_subtype_t subtype = req->subtype;

  if ( type == ORFS_FILE_REQ ) {
    if ( subtype == ORFS_PWRITE_FILE_REQ ) {
      *len = req->target.file.data.data_access.count;
      return 1;
    }
  } else if ( type == ORFS_PAGE_REQ ) {
    if ( subtype == ORFS_WRITE_FULL_PAGE_REQ
	 || subtype == ORFS_COMMIT_WRITE_PAGE_REQ ) {
      *len = req->target.page.data.data_access.count;
      return 1;
    }
  } else if ( type == ORFS_DIR_REQ ) {
    if ( subtype == ORFS_SYMLINK_DIR_REQ ) {
      *len = req->target.dir.data.symlink.path_len;
      return 1;
    }
  }
  return 0;
}


extern inline
int get_answer_buffer_len(struct orfs_answer *ans, size_t *len)
{
  orfs_req_type_t type = ans->type;
  orfs_req_subtype_t subtype = ans->subtype;

  if ( type == ORFS_FILE_REQ ) {
    if ( subtype == ORFS_PREAD_FILE_REQ ) {
      *len = (ans->result == ORFS_SUCCESS) ?
      ans->target.file.data.data_access.count : 0;
      return 1;
    } else if ( subtype == ORFS_PREADDIR_FILE_REQ ) {
      *len = (ans->result == ORFS_SUCCESS) ?
      ans->target.file.data.preaddir.data_len : 0;
      return 1;
    }
  } else if ( type == ORFS_PAGE_REQ ) {
   if ( subtype == ORFS_READ_FULL_PAGE_REQ ) {
      *len = (ans->result == ORFS_SUCCESS) ?
      ans->target.page.data.data_access.count : 0;
      return 1;
    }
  } else if ( type == ORFS_INODE_REQ ) {
    if ( subtype == ORFS_READLINK_INODE_REQ ) {
      *len = (ans->result == ORFS_SUCCESS) ?
	ans->target.inode.data.readlink.count : 0;
      return 1;
    }
  }
  return 0;
}


/****************
 * Capabilities *
 ****************/

enum {
  ORFS_LINK_CAP=0,
  ORFS_SYMLINK_CAP,
  ORFS_MKNOD_CAP
};

#define ADD_ORFS_CAP(x, cap) do { x |= 1UL << cap; } while (0)
#define HAS_ORFS_CAP(x, cap) (x & (1UL << cap))


#endif /* __ORFS_PROTOCOL_H__ */
