/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <gm.h>
#include <unistd.h>
#include <errno.h>

#include "orfs_protocol.h"
#include "orfs_common_network.h"
#include "ramfs_network.h"
#include "ramfs_fs.h"
#include "ramfs_io.h"
#include "ramfs_server.h"
#include "ramfs_queue.h"
#include "ramfs_messages.h"


static INLINE void process_request(struct server_request *req);

/**********
 * Queues *
 **********/

/* Declare a queue to store requests which
 * we are waiting to receive the data buffer for.
 */
DECLARE_REQUEST_QUEUE(partial_request_queue);

/* A queue for requests which need a token to
 * provide a buffer to receive data.
 */
DECLARE_REQUEST_QUEUE(recv_token_request_queue);

/* Queue for requests which need a second send token.
 */
DECLARE_REQUEST_QUEUE(second_send_token_request_queue);

/* Queue for requests which need a send token
 * to get the reqbuf through a RDMA read
 */
DECLARE_REQUEST_QUEUE(rdma_reqbuf_token_request_queue);

/*******************
 * GM declarations *
 *******************/

/* port */
static struct gm_port *port;

unsigned long gm_server_port = ORFS_GM_SERVER_PORT;
unsigned long gm_server_board = ORFS_GM_SERVER_BOARD;


/**********
 * Malloc *
 **********/

void *GM_malloc(size_t size)
{
  return gm_dma_malloc(port, size);
}


void GM_free(void *ptr)
{
  gm_dma_free(port, ptr);
}


/***************
 * Send Tokens *
 ***************/

/* total tokens */
static unsigned int send_tokens;


static INLINE
int try_get_send_token()
{
  if ( send_tokens ) {
    send_tokens--;
    return 0;
  } else
    return -1;
}


static void auto_send();

static INLINE
void put_send_token()
{
  send_tokens++;
  auto_send();
}


/***********
 * Sending *
 ***********/

/* Called when sending is resumed */
static
void resume_sending_callback(struct gm_port *port,
			     void *context,
			     gm_status_t status)
{
  put_send_token(); 
}

/* Answer sending */
static
void answer_send_callback(struct gm_port *port,
			  void *context,
			  gm_status_t status)
{
  struct server_request *req = context;

  if ( status != GM_SUCCESS ) {
    ERROR("Answer send callback got status %s\n", gm_strerror(status));
    gm_resume_sending(port,
		      ORFS_GM_ANS_PRIO,
		      req->connection.gm.node, req->connection.gm.port,
		      resume_sending_callback, NULL);
    /* keep this token for the resume sending event */
    // this send is trashed
    // we may try again for several errors
  } else {
    put_send_token();
  }

  if ( !--req->count ) {
    put_free_request(req);
    counters.sent++;
  }
}

static INLINE
void answer_send(gm_u16_t client_node,
		 gm_u8_t client_port,
		 struct server_request *req)
{
  gm_send_with_callback(port,
			&req->ans,
			ORFS_GM_ANS_SIZE, sizeof(req->ans),
			ORFS_GM_ANS_PRIO,
			client_node, client_port,
			answer_send_callback, req);
  req->count++;
}

/* Buffer sending */
static
void buffer_send_callback(struct gm_port *port,
			  void *context,
			  gm_status_t status)
{
  struct server_request *req = context;

  if ( status != GM_SUCCESS ) {
    ERROR("Buffer send callback got status %s\n", gm_strerror(status));
    gm_resume_sending(port,
		      ORFS_GM_ANSBUF_PRIO,
		      req->connection.gm.node, req->connection.gm.port,
		      resume_sending_callback, NULL);
    /* keep this token for the resume sending event */
    // this send is trashed
    // we may try again for several errors
  } else {
    put_send_token();
  }

  put_io_vector(req->out_vect);

  if ( !--req->count ) {
    put_free_request(req);
    counters.sent++;
  }
}

static INLINE
void buffer_send(gm_u16_t client_node,
		 gm_u8_t client_port,
		 struct server_request *req,
		 size_t data_len)
{
  void *buffer;

  /* If buffer is empty, send one fake byte
   * to use the client receive buffer
   */
  if ( data_len ) {
    req->out_vect = merge_io_vector(req->out_vect);
    buffer = VECT_PTR(req->out_vect, 0);
  } else {
    buffer = &req->ans;
    data_len = 1;
  }
  
  /* Send the buffer */
  gm_send_with_callback(port,
			buffer,
			req->req.net.gm.ansbuf.rdv_size, data_len,
			ORFS_GM_ANSBUF_PRIO,
			client_node, client_port,
			buffer_send_callback, req);
  req->count++;
}

/* Buffer RDMA write */
static
void buffer_rdma_write_callback(struct gm_port *port,
				void *context,
				gm_status_t status)
{
  struct server_request *req = context;

  if ( status != GM_SUCCESS ) {
    ERROR("Buffer rdma write callback got status %s\n", gm_strerror(status));
    gm_resume_sending(port,
		      /* ANS priority (see later) */
		      ORFS_GM_ANS_PRIO,
		      req->connection.gm.node, req->connection.gm.port,
		      resume_sending_callback, NULL);
    /* keep this token for the resume sending event */
    // this RDMA write is trashed
    // we may try again for several errors
  } else {
    put_send_token();
  }

  put_io_vector(req->out_vect);

  if ( !--req->count ) {
    put_free_request(req);
    counters.sent++;
  }
}

static INLINE
void buffer_rdma_write(gm_u16_t client_node,
		       gm_u8_t client_port,
		       struct server_request *req,
		       size_t data_len)
{
  req->out_vect = merge_io_vector(req->out_vect);
  gm_put(port,
	 VECT_PTR(req->out_vect, 0),
	 req->req.net.gm.ansbuf.rdma_ptr,
	 data_len,
	 /* use ANS priority to maintain order to ensure
	  * the client will be notified by te answer after
	  * the RDMA is received.
	  */
	 ORFS_GM_ANS_PRIO,
	 client_node, client_port,
	 buffer_rdma_write_callback, req);
  req->count++;
}

/* Buffer RDMA Read */
static
void buffer_rdma_read_callback(struct gm_port *port,
			       void *context,
			       gm_status_t status)
{
  struct server_request *req = context;

  if ( status != GM_SUCCESS ) {
    ERROR("Buffer rdma read callback got status %s\n", gm_strerror(status));
    gm_resume_sending(port,
		      ORFS_GM_REQBUF_PRIO,
		      req->connection.gm.node, req->connection.gm.port,
		      resume_sending_callback, NULL);
    /* keep this token for the resume sending event */
    // this RDMA read is trashed
    // this will store fake data in the server
    // we may try again for several errors
  } else {
    put_send_token();
  }
  req->count--;

  /* process the request */
  process_request(req);
  counters.received++;
}

static INLINE
void buffer_rdma_read(gm_u16_t client_node,
		      gm_u8_t client_port,
		      struct server_request *req,
		      size_t data_len)
{
  gm_get(port,
	 req->req.net.gm.reqbuf.rdma_ptr,
	 VECT_PTR(req->in_vect, 0),
	 data_len,
	 ORFS_GM_REQBUF_PRIO, /* don't care */
	 client_node, client_port,
	 buffer_rdma_read_callback, req);
  req->count++;
}

/*
 * Answer sending routines. Three cases:
 * 1. Just an answer. Need 1 send token.
 * 2. Answer and sent Buffer. Need 2 send tokens.
 *    Have to send a fake byte in case of empty
 *    buffer since the client requires a notification
 *    for both buffer and answer.
 * 3. Answer and RDMA Buffer. Need 2 send tokens.
 *    No need to send empty buffer since there's no
 *    RDMA notification on the remote side.
 *    However, need to send the answer at last to
 *    notify the client when both RDMA and answer 
 *    have been received.
 *    Thus the RDMA uses the same priority than the
 *    answer to ensure it is received first.
 *
 * Both cases 2 and 3 may fail if only one send token
 * is available. In this case, the request is queued
 * back. It is then dequeued and processed first when
 * this routine is called again.
 */

static
void auto_send()
{
  struct server_request *req;
  gm_u16_t client_node;
  gm_u8_t client_port;
  size_t data_len;
  int has_buffer;


  /*
   * Finish Sending of Request.
   * This Queue should not contain multiple requests.
   */
  if ( !QUEUE_EMPTY(&second_send_token_request_queue) ) {
    if ( try_get_send_token() < 0 )
      return;

    req = DEQUEUE_REQUEST(&second_send_token_request_queue);
    client_node = req->connection.gm.node;
    client_port = req->connection.gm.port;

    if ( req->req.net.gm.rdma_ansbuf ) {
      /* RDMA case: need to send the answer */
      answer_send(client_node, client_port, req);
    } else {
      /* Usual case: need to send the buffer */
      get_answer_buffer_len(&req->ans, &data_len);
      buffer_send(client_node, client_port, req, data_len);
    }
  }


  /* Finish getting reqbuf through rdma read */
  while ( !QUEUE_EMPTY(&rdma_reqbuf_token_request_queue) ) {
    if ( try_get_send_token() < 0 )
      return;

    req = DEQUEUE_REQUEST(&rdma_reqbuf_token_request_queue);
    client_node = req->connection.gm.node;
    client_port = req->connection.gm.port;

    get_request_buffer_len(&req->req, &data_len);
    buffer_rdma_read(client_node, client_port, req, data_len);
  }  


  /* Send other Requests */ 
  while ( !QUEUE_EMPTY(&done_request_queue) ) {
    if ( try_get_send_token() < 0 )
      return;

    req = DEQUEUE_REQUEST(&done_request_queue);
    client_node = req->connection.gm.node;
    client_port = req->connection.gm.port;

    has_buffer = get_answer_buffer_len(&req->ans, &data_len);

    /* RDMA case launches the RDMA first */
    if ( has_buffer && req->req.net.gm.rdma_ansbuf && data_len ) {
      buffer_rdma_write(client_node, client_port, req, data_len);

      /* Try to get a send token for the answer */
      if ( try_get_send_token() < 0 ) {
	QUEUE_REQUEST(&second_send_token_request_queue, req);
	return;
      }
    }
    
    /* Send the answer now (first for usual case, second for RDMA case) */
    answer_send(client_node, client_port, req);

    /* Now send the buffer */
    if ( has_buffer && !req->req.net.gm.rdma_ansbuf ) {
      if ( try_get_send_token() < 0 ) {
	QUEUE_REQUEST(&second_send_token_request_queue, req);
	return;
      }

      buffer_send(client_node, client_port, req, data_len);
    }
  }
}


/***************
 * Recv Tokens *
 ***************/

/* total tokens */
static unsigned int recv_tokens;

/* dedicated tokens */
static unsigned int req_recv_tokens, reqbuf_recv_tokens, other_recv_tokens;


static INLINE
int provide_buffer(size_t len)
{
  int size = gm_min_size_for_length(len);
  size_t maxlen = gm_max_length_for_size(size);
  void *buffer = GM_malloc(maxlen);
  if ( !buffer ) {
    ERROR("Failed to allocate buffer of len %d\n", len);
    return -1;
  }
  gm_provide_receive_buffer(port,
			    buffer,
			    size,
			    ORFS_GM_REQBUF_PRIO);
  other_recv_tokens--;
  return 0;
}


static INLINE
int try_provide_buffer(size_t len)
{
  if ( other_recv_tokens )
    return provide_buffer(len);
  else
    return -1;
}


static INLINE
void auto_provide_buffers()
{
  struct server_request *req;

  /* provide required buffers */
  while ( other_recv_tokens
	  && (req=DEQUEUE_REQUEST(&recv_token_request_queue)) != NULL ) {
    size_t len;
    get_request_buffer_len(&req->req, &len);
    if ( provide_buffer(len) < 0 ) {
      QUEUE_REQUEST(&recv_token_request_queue, req);
      break;
    }
    QUEUE_REQUEST(&partial_request_queue, req);
  }
}

static INLINE
void auto_preprovide_requests()
{
  struct server_request *req;

  /* preprovide requests */
  while ( req_recv_tokens ) {
    if ( (req=get_free_request()) != NULL ) {
      req_recv_tokens--;
      gm_provide_receive_buffer(port,
				&req->req,
				ORFS_GM_REQ_SIZE,
				ORFS_GM_REQ_PRIO);
      req->count++;
    } else {
      ERROR("Failed to get a new request\n");
      break;
    }
  }
}

static INLINE
void auto_preprovide_buffers()
{
  void *buffer;

 /* preprovide buffers */
  while ( reqbuf_recv_tokens ) {
    if ( (buffer=GM_malloc(ORFS_GM_PREPROVIDED_REQBUF_LEN)) != NULL ) {
      reqbuf_recv_tokens--;
      gm_provide_receive_buffer(port,
				buffer,
				ORFS_GM_PREPROVIDED_REQBUF_SIZE,
				ORFS_GM_REQBUF_PRIO);
    } else {
      ERROR("Failed to preallocate a buffer\n");
      break;
    }
  }
}

static INLINE
void unpreprovide_request()
{
  req_recv_tokens++;
  auto_preprovide_requests();
}


static INLINE
void unpreprovide_buffer()
{
  reqbuf_recv_tokens++;
  auto_preprovide_buffers();
}


static INLINE
void unprovide_buffer()
{
  other_recv_tokens++;
  auto_provide_buffers();
}


/*************
 * Receiving *
 *************/

static inline
int GM_connection_compare(struct orfs_gm_connection *c1,
			  struct orfs_gm_connection *c2)
{
  return ( c1->port == c2->port && c1->node == c2->node );
}

static
int GM_server_request_compare(struct server_request *req,
			      void *data)
{
  struct orfs_gm_connection *connection = data;
  return GM_connection_compare(connection, &req->connection.gm);
}


static INLINE
void process_request_recv_event(gm_recv_event_t *ev)
{
  struct orfs_request *oreq = gm_ntohp(ev->recv.buffer);
  struct server_request *req = REQ2SREQ(oreq);
  gm_u16_t client_node;
  gm_u8_t client_port;
  size_t data_len;

  /* get connection */
  client_node = req->connection.gm.node = gm_ntoh_u16(ev->recv.sender_node_id);
  client_port = req->connection.gm.port = gm_ntoh_u8(ev->recv.sender_port_id);

  assert(!--req->count);

  if ( get_request_buffer_len(&req->req, &data_len) ) {
    /* this request needs to receive some data */
    if ( req->req.net.gm.rdma_reqbuf ) {
      req->in_vect = get_io_monovector_alloc(data_len);
      if ( try_get_send_token() < 0 ) {
	QUEUE_REQUEST(&rdma_reqbuf_token_request_queue, req);
      } else {
	buffer_rdma_read(client_node, client_port, req, data_len);
      }
    } else {
      /* provide a dedicated buffer if needed */
      if ( data_len > ORFS_GM_PREPROVIDED_REQBUF_LEN 
	   &&  try_provide_buffer(data_len) < 0 )
	QUEUE_REQUEST(&recv_token_request_queue, req);
      else
	QUEUE_REQUEST(&partial_request_queue, req);
    }
  } else {
    /* this request has nothing to receive, it's ready */
    process_request(req);
    counters.received++;
  }
}


static INLINE
void process_buffer_recv_event(gm_recv_event_t *ev)
{
  void *buf = gm_ntohp(ev->recv.buffer);
  size_t len = gm_ntoh_u32(ev->recv.length);
  struct server_request *req;
  struct orfs_gm_connection connection;
  
  /* get connection */
  connection.node = gm_ntoh_u16(ev->recv.sender_node_id);
  connection.port = gm_ntoh_u8(ev->recv.sender_port_id);
  
  /* find the partial request */
  req = DEQUEUE_SPECIFIC_REQUEST(&partial_request_queue,
				 GM_server_request_compare,
				 &connection);
  /* fill the request */
  req->in_vect = get_io_monovector(buf, len, 0);
  /* process it */
  process_request(req);
  counters.received++;
}


/**********************
 * Processing Request *
 **********************/

static INLINE
void process_request(struct server_request *req)
{
  QUEUE_REQUEST(&ready_request_queue, req);
  ramfs_process_queue();
  auto_send();
}


/*************
 * Main Loop *
 *************/

void GM_process(int *exiting)
{
  int slept = 1;

  while ( !*exiting ) {
    gm_recv_event_t *ev;

    /* update current time if we slept */
    if ( slept ) {
      update_current_time();
      slept = 0;
    }
    
    /* receive events */
    ev = gm_blocking_receive(port);

    switch ( gm_ntoh_u8(ev->recv.type) ) {
    case ORFS_GM_REQ_RECV_EV:
      process_request_recv_event(ev);
      unpreprovide_request();
      break;
    case ORFS_GM_REQBUF_RECV_EV:
      process_buffer_recv_event(ev);
      if ( gm_ntoh_u8(ev->recv.size) == ORFS_GM_PREPROVIDED_REQBUF_SIZE )
	unpreprovide_buffer();
      else
	unprovide_buffer();
      break;
    case _GM_SLEEP_EVENT:
      slept = 1;
      /* fallthough */
    default:
      gm_unknown(port, ev);
    }
  }
}


/****************
 * Initializing *
 ****************/

int GM_init(char *name)
{
  gm_status_t status;

  /* Initialize GM */
  status = gm_init();
  if ( status != GM_SUCCESS ) {
    ERROR("GM INIT failed with error %s\n",
	  gm_strerror(status));
    goto error;
  }

  /* Open the port */
  status = gm_open(&port,
		   gm_server_board, gm_server_port,
		   "ramfs", GM_API_VERSION);
  if ( status != GM_SUCCESS ) {
    ERROR("GM cannot open port (error %s)\n",
	  gm_strerror(status));
    goto error_with_init;
  }

  /* Deal with tokens */
  send_tokens = gm_num_send_tokens(port);

  recv_tokens = gm_num_receive_tokens(port);

  req_recv_tokens = recv_tokens*GM_PREPROVIDED_REQ_RATIO;
  reqbuf_recv_tokens = recv_tokens*GM_PREPROVIDED_REQBUF_RATIO;
  other_recv_tokens = recv_tokens-req_recv_tokens-reqbuf_recv_tokens;

  /* Initialize Queues */
  INIT_REQUEST_QUEUE(partial_request_queue);
  INIT_REQUEST_QUEUE(recv_token_request_queue);
  INIT_REQUEST_QUEUE(second_send_token_request_queue);
  INIT_REQUEST_QUEUE(rdma_reqbuf_token_request_queue);

  /* Preprovide Requests and Buffers */
  auto_preprovide_requests();
  auto_preprovide_buffers();

  /* check everything has been allocated and provided */
  if ( req_recv_tokens || reqbuf_recv_tokens )
    goto error_with_port;

  return 0;

 error_with_port:
  gm_close(port);
 error_with_init:
  gm_finalize();
 error:
  return -1;
}


void GM_exit()
{
  /* lots of buffer to unprovide, currently impossible */
  gm_close(port);
  gm_finalize();
}
