/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <myriexpress.h>
#include <unistd.h>
#include <errno.h>

#include "orfs_protocol.h"
#include "orfs_common_network.h"
#include "ramfs_network.h"
#include "ramfs_fs.h"
#include "ramfs_io.h"
#include "ramfs_server.h"
#include "ramfs_queue.h"
#include "ramfs_messages.h"


/*******************
 * MX declarations *
 *******************/

static mx_endpoint_t ep;

unsigned long mx_server_ep = ORFS_MX_SERVER_EP;
unsigned long mx_server_board = ORFS_MX_SERVER_BOARD;

/* These routines compare connections to retrieve a
 * queued request or buffer.
 */

static inline
int MX_connection_compare(struct orfs_mx_connection *c1,
			  struct orfs_mx_connection *c2)
{
  return !memcmp(c1, c2, sizeof(*c1));
}

static
int MX_server_request_compare(struct server_request *req,
			      void *data)
{
  struct orfs_mx_connection *connection = data;
  return MX_connection_compare(connection, &req->connection.mx);
}

/**********
 * Queues *
 **********/

/* Declare a queue to store requests which
 * we are waiting to receive the data buffer for.
 */
DECLARE_REQUEST_QUEUE(partial_request_queue);

/* Declare a list to store buffers whose
 * requests has not arrived yet.
 */
struct list_head orphan_buffers;

struct orphan_buffer {
  struct list_head list;
  void *ptr;
  struct orfs_mx_connection connection;
};

static inline int
queue_orphan_buffer(void *ptr, struct orfs_mx_connection *connection)
{
  struct orphan_buffer *buf = malloc(sizeof(*buf));
  if ( !buf ) {
    ERROR("Failed to allocate orphan buffer\n");
    return -ENOMEM;
  }
  buf->ptr = ptr;
  memcpy(&buf->connection, connection, sizeof(*connection));
#if DEBUG_LEVEL >= 0
  bzero(&buf->list, sizeof(buf->list));
#endif
  LIST_ADD_TAIL(&buf->list, &orphan_buffers);
  return 0;
}

static inline void *
dequeue_orphan_buffer(struct orfs_mx_connection *connection)
{
  struct orphan_buffer *buf;
  struct list_head *head;
  void *ptr;
  LIST_FOR_EACH(buf, head, list, &orphan_buffers)
    if ( MX_connection_compare(connection, &buf->connection) ) {
      ptr = buf->ptr;
      LIST_DEL(head);
      free(buf);
      return ptr;
    }

  return NULL;
}

/***********
 * Context *
 ***********/

struct context {
  uint32_t type;
  void *ptr;
};

/****************
 * Post receive *
 ****************/

static inline int
post_request_receive(void)
{
  struct server_request *req;
  mx_segment_t seg;
  mx_request_t mx_req;
  mx_return_t ret;
  struct context *ctx;

  ctx = malloc(sizeof(*ctx));
  if (!ctx) {
    ERROR("Cannot allocate context\n");
    return -ENOMEM;
  }

  req = get_free_request();
  if ( !req ) {
    ERROR("Cannot get new free request\n");
    free(ctx);
    return -ENOMEM;
  }  

  LOG(1, "Posting a receive for request %p...\n", req);

  seg.segment_ptr = &req->req;
  seg.segment_length = sizeof(req->req);
  ctx->type = ORFS_MX_REQ_TYPE;
  ctx->ptr = req;
  ret = mx_irecv(ep, &seg, 1,
		 ORFS_MX_REQ_RECV_MATCH,
		 ORFS_MX_REQ_RECV_MASK,
		 ctx, &mx_req);
  if ( ret != MX_SUCCESS ) {
    ERROR("Failed to post MX request receive (error %s)\n",
	  mx_strerror(ret));
    free(ctx);
    put_free_request(req);
    return -EIO;
  }

  return 0;
}

static inline int
post_small_buffer_receive(void)
{
  void *buffer;
  mx_segment_t seg;
  mx_request_t mx_req;
  mx_return_t ret;
  struct context *ctx;

  ctx = malloc(sizeof(*ctx));
  if (!ctx) {
    ERROR("Cannot allocate context\n");
    return -ENOMEM;
  }

  buffer = malloc(ORFS_MX_PREPROVIDED_REQBUF_LEN);
  if ( !buffer ) {
    ERROR("Cannot allocate new preprovided buffer\n");
    free(ctx);
    return -ENOMEM;
  }

  LOG(1, "Posting receive for small buffer %p...\n", buffer);

  seg.segment_ptr = buffer;
  seg.segment_length = ORFS_MX_PREPROVIDED_REQBUF_LEN;
  ctx->type = ORFS_MX_PREPROVIDED_REQBUF_TYPE;
  ctx->ptr = buffer;
  ret = mx_irecv(ep, &seg, 1,
		 ORFS_MX_PREPROVIDED_REQBUF_RECV_MATCH,
		 ORFS_MX_PREPROVIDED_REQBUF_RECV_MASK,
		 ctx, &mx_req);
  if ( ret != MX_SUCCESS ) {
    ERROR("Failed to post MX preprovided buffer receive (error %s)\n",
	  mx_strerror(ret));
    free(ctx);
    free(buffer);
    return -EIO;
  }

  return 0;
}

static inline int
post_large_buffer_receive(struct server_request *req)
{
  mx_segment_t seg;
  mx_request_t mx_req;
  mx_return_t ret;
  uint32_t id = ORFS_MX_MATCH_ID(req->connection.mx.match_id);
  struct context *ctx;

  ctx = malloc(sizeof(*ctx));
  if (!ctx) {
    ERROR("Cannot allocate context\n");
    return -ENOMEM;
  }

  LOG(1, "Posting receive for large buffer %p size %lld for request %p...\n",
      VECT_PTR(req->in_vect, 0),
      (unsigned long long) VECT_SIZE(req->in_vect, 0),
      req);

  seg.segment_ptr = VECT_PTR(req->in_vect, 0);
  seg.segment_length = VECT_SIZE(req->in_vect, 0);
  ctx->type = ORFS_MX_LARGE_REQBUF_TYPE;
  ctx->ptr = req;
  ret = mx_irecv(ep, &seg, 1,
		 ORFS_MX_LARGE_REQBUF_RECV_MATCH(id),
		 ORFS_MX_LARGE_REQBUF_RECV_MASK,
		 ctx, &mx_req);
  if ( ret != MX_SUCCESS ) {
    ERROR("Failed to post MX large buffer receive (error %s)\n",
	  mx_strerror(ret));
    free(ctx);
    return -EIO;
  }  

  return 0;
}

static inline int
post_rdma_read(struct server_request *req)
{
  mx_segment_t seg;
  mx_request_t mx_req;
  mx_return_t ret;
  struct context *ctx;

  LOG(1, "Getting request %p buffer through RDMA...\n", req);

  ctx = malloc(sizeof(*ctx));
  if (!ctx) {
    ERROR("Cannot allocate context\n");
    return -ENOMEM;
  }

  seg.segment_ptr = VECT_PTR(req->in_vect, 0);
  seg.segment_length = VECT_SIZE(req->in_vect, 0);
  ctx->type = ORFS_MX_RDMA_REQBUF_TYPE;
  ctx->ptr = req;
  ret = mx_rdma_read(ep,
		     &seg, 1,
		     req->connection.mx.addr,
		     req->req.net.mx.reqbuf.window_handle,
		     req->req.net.mx.reqbuf.offset,
		     ctx, &mx_req);
  if ( ret != MX_SUCCESS ) {
    ERROR("Failed to post MX rdma read (error %s)\n",
	  mx_strerror(ret));
    free(ctx);
    return -EIO;
  }

  return 0;
}

/***************
 * Peek events *
 ***************/

static inline void
receive()
{
  mx_request_t mx_req;
  mx_return_t ret;
  mx_status_t status;
  uint32_t result;
  int got_full_request = 0;
  struct context * context;

  /* loop until nothing more to receive */
  while (1) {
    if ( got_full_request ) {
      LOG(2, "Polling for a new completed request...\n");
      /* process other events */
      ret = mx_ipeek(ep, &mx_req, &result);
      if ( ret != MX_SUCCESS ) {
	ERROR("MX ipeek failed (error %s)\n",
	      mx_strerror(ret));
	// do something
      }
      if ( result != 1 ) {
	/* nothing more ? exit the loop and process the requests */
	break;
      }
    } else {
      LOG(2, "Blocking until next request completes...\n");
      /* block if no request are ready to be processed */
      ret = mx_peek(ep, MX_INFINITE, &mx_req, &result);
      if ( ret != MX_SUCCESS ) {
	ERROR("MX peek failed (error %s)\n", mx_strerror(ret));
	// do something
      }
      if ( result != 1 ) {
	ERROR("MX peek return without request ????\n");
	// do something
      }
    }
    ret = mx_test(ep, &mx_req, &status, &result);
    if ( ret != MX_SUCCESS ) {
      ERROR("MX test failed (error %s)\n", mx_strerror(ret));
      // do something
    }
    if ( result != 1 ) {
      ERROR("MX request not ready ????\n");
      // do something
    }

    // what if bad status ? depends on the type...

    context = status.context;

    switch ( context->type ) {

    case ORFS_MX_REQ_TYPE: {
      struct server_request * req = context->ptr;
      size_t in_buf_len;
      
      // what if xfer_length != msg_length ?

      LOG(1, "Req %p filled\n", req);

      /* store connection */
      memcpy(&req->connection.mx.addr,
	     &status.source, sizeof(mx_endpoint_addr_t));
      req->connection.mx.match_id = ORFS_MX_MATCH_ID(status.match_info);

      /* get input buffer lenght */
      if ( get_request_buffer_len(&req->req, &in_buf_len) ) {
	if ( req->req.net.mx.rdma_reqbuf ) {
	  /* rdma read if necessary */
	  req->in_vect = get_io_monovector_alloc(in_buf_len);
	  post_rdma_read(req);
	  // do something on error
	} else {
	  if ( in_buf_len > ORFS_MX_PREPROVIDED_REQBUF_LEN ) {
	    /* large buffers need to be posted here */
	    req->in_vect = get_io_monovector_alloc(in_buf_len);
	    post_large_buffer_receive(req);
	    // do something on error
	    /* No need to queue the request here,
	     * we'll found it back from the large receive context. */
	  } else {
	    /* small buffers are already posted, and may be even already received */
	    void *ptr = dequeue_orphan_buffer(&req->connection.mx);
	    if ( ptr ) {
	      req->in_vect = get_io_monovector(ptr, in_buf_len, 0);
	      LOG(2, "Found buffer for request %p\n", req);
	      QUEUE_REQUEST(&ready_request_queue, req);
	      got_full_request = 1;
	      counters.received++;
	    } else {
	      QUEUE_REQUEST(&partial_request_queue, req);
	    }
	  }
	}
      } else {
	LOG(2, "Request %p doesn't need any buffer\n", req);
	QUEUE_REQUEST(&ready_request_queue, req);
	got_full_request = 1;
	counters.received++;
      }
      post_request_receive();
      // do something on error
      break;
    }

    case ORFS_MX_PREPROVIDED_REQBUF_TYPE: {
      void * buffer = context->ptr;
      struct server_request *req;
      struct orfs_mx_connection connection;
      
      // what if xfer_length != msg_length ?

      LOG(1, "Buffer %p filled\n", buffer);

      /* store connection */
      memcpy(&connection, &status.source, sizeof(mx_endpoint_addr_t));
      connection.match_id = ORFS_MX_MATCH_ID(status.match_info);

      req = DEQUEUE_SPECIFIC_REQUEST(&partial_request_queue,
				     MX_server_request_compare,
				     &connection);
      if ( req ) {
	LOG(2, "Found request %p for buffer\n", req);
	req->in_vect = get_io_monovector(buffer, status.xfer_length, 0);
	QUEUE_REQUEST(&ready_request_queue, req);
	got_full_request = 1;
	counters.received++;
      } else {
	queue_orphan_buffer(buffer, &connection);
      }

      post_small_buffer_receive();
      // do something on error
      break;
    }

    case ORFS_MX_RDMA_REQBUF_TYPE:
    case ORFS_MX_LARGE_REQBUF_TYPE: {
      struct server_request * req = context->ptr;

      // what if xfer_length != msg_length ?

      LOG(1, "Buffer for request %p filled\n", req);
      QUEUE_REQUEST(&ready_request_queue, req);
      got_full_request = 1;
      counters.received++;
      break;
    }

    case ORFS_MX_ANS_TYPE: {
      struct server_request * req = context->ptr;

      LOG(1, "Request %p sent\n", req);
      if ( !--req->count ) {
	if ( req->req.type == ORFS_FS_REQ && req->req.subtype == ORFS_FINI_FS_REQ ) {
	  // HACK to reset connection so that the next client using this endpoint will be able to talk to us
	  // hope the mx_isend is really done and not simply buffered
	  extern void mx__partner_reset(mx_endpoint_t ep, mx_endpoint_addr_t addr);
	  mx__partner_reset(ep, req->connection.mx.addr);
	}
	put_free_request(req);
	counters.sent++;
      }
      break;
    }
      
    case ORFS_MX_ANSBUF_TYPE:
    case ORFS_MX_RDMA_ANSBUF_TYPE: {
      struct server_request * req = context->ptr;
      
      LOG(1, "Buffer of request %p sent\n", req);
      put_io_vector(req->out_vect);
      if ( !--req->count ) {
	put_free_request(req);
	counters.sent++;
      }
      break;
    }

    default:
      ERROR("Got request with context type %d ???\n",
	    context->type);
      assert(0);
      break;
    }

    /* the MX request has been processed, release its context */
    free(context);
  }
}

/********
 * Send *
 ********/

static inline int
post_answer_send(struct server_request *req, uint32_t id)
{
  mx_request_t ans_req;
  mx_segment_t ans_seg;
  mx_return_t ret;
  struct context *ctx;

  LOG(1, "Sending request %p...\n", req);
  
  ctx = malloc(sizeof(*ctx));
  if (!ctx) {
    ERROR("Cannot allocate context\n");
    return -ENOMEM;
  }

  ans_seg.segment_ptr = &req->ans;
  ans_seg.segment_length = sizeof(req->ans);
  ctx->type = ORFS_MX_ANS_TYPE;
  ctx->ptr = req;
  ret = mx_isend(ep,
		 &ans_seg, 1,
		 req->connection.mx.addr,
		 ORFS_MX_ANS_MATCH(id),
		 ctx, &ans_req);
  if ( ret != MX_SUCCESS ) {
    ERROR("Failed to post request send (error %s)\n",
	  mx_strerror(ret));
    free(ctx);
    return -EIO;
  }

  req->count++;
  return 0;
}

static inline int
post_buffer_send(struct server_request *req, size_t len, uint32_t id)
{
  mx_request_t ansbuf_req;
  mx_segment_t ansbuf_seg;
  mx_return_t ret;
  struct context *ctx;

  LOG(1, "Sending buffer of request %p...\n", req);
  
  ctx = malloc(sizeof(*ctx));
  if (!ctx) {
    ERROR("Cannot allocate context\n");
    return -ENOMEM;
  }

  if ( len ) {
    ansbuf_seg.segment_ptr = VECT_PTR(req->out_vect, 0);
    ansbuf_seg.segment_length = VECT_SIZE(req->out_vect, 0);
    // vectorial here !
  } else {
    ansbuf_seg.segment_ptr = NULL;
    ansbuf_seg.segment_length = 0;
    // vectorial here !
  }
  ctx->type = ORFS_MX_ANSBUF_TYPE;
  ctx->ptr = req;
  ret = mx_isend(ep,
		 &ansbuf_seg, 1,
		 req->connection.mx.addr,
		 ORFS_MX_ANSBUF_MATCH(id),
		 ctx, &ansbuf_req);
  if ( ret != MX_SUCCESS ) {
    ERROR("Failed to post buffer send (error %s)\n",
	  mx_strerror(ret));
    free(ctx);
    return -EIO;
  }

  req->count++;
  return 0;
}

static inline int
post_rdma_write(struct server_request *req, uint32_t id)
{
  mx_request_t ansbuf_req;
  mx_segment_t ansbuf_seg;
  mx_return_t ret;
  struct context *ctx;

  LOG(1, "Putting buffer of request %p through RDMA...\n", req);
  
  ctx = malloc(sizeof(*ctx));
  if (!ctx) {
    ERROR("Cannot allocate context\n");
    return -ENOMEM;
  }

  ansbuf_seg.segment_ptr = VECT_PTR(req->out_vect, 0);
  ansbuf_seg.segment_length = VECT_SIZE(req->out_vect, 0);
  // vectorial here !
  ctx->type = ORFS_MX_RDMA_ANSBUF_TYPE;
  ctx->ptr = req;
  ret = mx_rdma_write(ep,
		      &ansbuf_seg, 1,
		      req->connection.mx.addr,
		      req->req.net.mx.ansbuf.window_handle,
		      req->req.net.mx.ansbuf.offset,
		      ctx, &ansbuf_req);
  if ( ret != MX_SUCCESS ) {
    ERROR("Failed to post rdma write (error %s)\n",
	  mx_strerror(ret));
    free(ctx);
    return -EIO;
  }

  return 0;
}

static inline void
send()
{
  struct server_request *req;
  size_t out_buf_len;
  
  while ((req = DEQUEUE_REQUEST(&done_request_queue)) != NULL) {
    uint32_t match_id = ORFS_MX_MATCH_ID(req->connection.mx.match_id);

    if ( get_answer_buffer_len(&req->ans, &out_buf_len)
         && out_buf_len
         && req->req.net.mx.rdma_ansbuf ) {
      /* rdma write ansbuf if not empty */
      post_rdma_write(req, match_id);
      // do something on error
    }

    /* send answer */
    post_answer_send(req, match_id);
    // do something on error

    if ( get_answer_buffer_len(&req->ans, &out_buf_len)
         && !req->req.net.mx.rdma_ansbuf ) {
      /* send answer buffer, even if empty */
      post_buffer_send(req, out_buf_len, match_id);
      // do something on error
    }
  }
}

/*************
 * Main Loop *
 *************/

void MX_process(int *exiting)
{
  while ( !*exiting ) {
    receive();
    ramfs_process_queue();
    send();
  }
}


/****************
 * Initializing *
 ****************/

int MX_init(char *name)
{
  mx_return_t ret;
  uint64_t nic_id;
  uint32_t native_req;
  int i;

  /* Initialize MX */
  ret = mx_init();
  if ( ret != MX_SUCCESS ) {
    ERROR("MX INIT failed with error %s\n",
	  mx_strerror(ret));
    goto error;
  }

  /* Get NIC id */
  ret = mx_board_number_to_nic_id(mx_server_board, &nic_id);
  if ( ret != MX_SUCCESS ) {
    fprintf(stderr, "Cannot get MX nic id (error %s)\n", mx_strerror(ret));
    goto error_with_init;
  }

  /* Open the endpoint */
  ret = mx_open_endpoint(nic_id, mx_server_ep, ORFS_MX_EP_FILTER,
			 NULL, 0, &ep);
  if ( ret != MX_SUCCESS ) {
    ERROR("MX cannot open endpoint (error %s)\n",
	  mx_strerror(ret));
    goto error_with_init;
  }

  printf("open endpoint on nic %llx ep %ld filter %lx\n",
	 nic_id, mx_server_ep, ORFS_MX_EP_FILTER);

  /* Retrieve native request amount */
  ret = mx_get_info(ep, MX_NATIVE_REQUESTS, &native_req, sizeof(native_req));
  if ( ret != MX_SUCCESS ) {
    ERROR("MX cannor retrive native request amount (error %s)\n",
	  mx_strerror(ret));
    goto error_with_ep;
  }

  LOG(2, "%d native requests available\n", native_req);

  /* Post several request and buffer receive */
  for(i=0; i<native_req/3; i++) {
    post_request_receive();
    // do something on error
  }
  for(i=0; i<native_req/3; i++) {
    post_small_buffer_receive();
    // do something on error
  }

  /* Initialize Queues */
  INIT_REQUEST_QUEUE(partial_request_queue);
  LIST_INIT(&orphan_buffers);

  return 0;

 error_with_ep:
  mx_close_endpoint(ep);
 error_with_init:
  mx_finalize();
 error:
  return -1;
}


void MX_exit()
{
  mx_close_endpoint(ep);
  mx_finalize();
}
