/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/tcp.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/poll.h>
#include <errno.h>

#if SERVER_TCP_POLL == EPOLL
#include <epoll.h>
#endif /* SERVER_TCP_POLL */

#include "orfs_protocol.h"
#include "orfs_common_network.h"
#include "ramfs_network.h"
#include "ramfs_fs.h"
#include "ramfs_io.h"
#include "ramfs_server.h"
#include "ramfs_queue.h"
#include "ramfs_messages.h"


/********************
 * TCP declarations *
 ********************/

#define SET_TCP_NODELAY(s)   \
do {                         \
 int status = 1;             \
 setsockopt(s,               \
            IPPROTO_TCP,     \
            TCP_NODELAY,     \
            &status,         \
            sizeof(status)); \
} while ( 0 )


unsigned long tcp_server_port = ORFS_TCP_SERVER_PORT;
static int main_socket = -1;


/***************
 * Connections *
 ***************/

#if SERVER_TCP_POLL == EPOLL

static int epoll_fd = -1;

static INLINE
int epoll_add_connection(int fd)
{
  int res;
  struct epoll_event epoll_event = { 0 };
  epoll_event.events = EPOLLIN|EPOLLET;
  epoll_event.data.fd = fd;
  res = epoll_ctl(epoll_fd, EPOLL_CTL_ADD, fd, &epoll_event);
  if ( res < 0 ) {
    ERROR("epoll add error %m\n");
    return -1;
  }
  LOG(4, "Added connection with socket %d\n", fd);
  return res;
}

static INLINE
void epoll_remove_connection(int fd)
{
  int res;
  struct epoll_event epoll_event = { 0 };
  res = epoll_ctl(epoll_fd, EPOLL_CTL_DEL, fd, &epoll_event);
  close(fd);
  if ( res < 0 )
    ERROR("epoll remove error %m\n");
  else
    LOG(4, "Removed connection with socket %d\n", fd);  
}

static INLINE
int epoll_init()
{
  int res;

  res = epoll_create(TCP_SERVER_EVENT_QUEUE_SIZE);
  if ( res < 0 ) {
    ERROR("epoll create error %m\n");
    return -1;
  }
  epoll_fd = res;

  /* add the main connection to epoll fd */
  if ( epoll_add_connection(main_socket) ) {
    ERROR("epoll add failed error %m\n");
    close(epoll_fd);
    return -1;
  }

  return 0;
}

static INLINE
void epoll_finish()
{
  close(epoll_fd);
}

#elif SERVER_TCP_POLL == POLL

static int pollfd_nr;

static struct pollfd *pollfd;

static INLINE
int poll_add_connection(int fd)
{
  if ( pollfd_nr == TCP_SERVER_EVENT_QUEUE_SIZE ) {
    ERROR("Cannot add new pollfd structure\n");
    return -1;
  }
  pollfd[pollfd_nr].fd = fd;
  pollfd[pollfd_nr].events = POLLIN;
  pollfd_nr++;
  LOG(4, "Added connection with socket %d\n", fd);
  return 0;
}

static INLINE
void poll_remove_connection(int i)
{
  close(pollfd[i].fd);
  if ( i != pollfd_nr-1 )
    memcpy(&pollfd[i], &pollfd[pollfd_nr-1], sizeof(*pollfd));
  pollfd_nr--;
}

static INLINE
int poll_init()
{
  pollfd = malloc(TCP_SERVER_EVENT_QUEUE_SIZE*sizeof(*pollfd));
  if ( !pollfd ) {
    ERROR("Cannot allocated pollfd structures\n");
    return -1;
  }
  pollfd_nr = 0;
  poll_add_connection(main_socket);
  return 0;
}

static INLINE
void poll_finish()
{
  free(pollfd);
}

#endif /* SERVER_TCP_POLL */


/******************
 * Initialization *
 ******************/

int TCP_init()
{
  int res = 0;
  struct sockaddr_in server_address = { 0 };

  /* open an internet tcp socket */
  res = socket(PF_INET, SOCK_STREAM, 0);
  if ( res < 0 ) {
    ERROR("socket creation error %m\n");
    goto error;
  }
  main_socket = res;

  /* set up Internet address structure for the server */
  server_address.sin_family = AF_INET;
  server_address.sin_port = htons((short)tcp_server_port);
  server_address.sin_addr.s_addr = htonl(INADDR_ANY);
  LOG(1, "Port %d\n", ntohs(server_address.sin_port));

  /* bind the socket to the server's Internet address */
  res = bind(main_socket,
	     (struct sockaddr *)&server_address,
	     sizeof(server_address));
  if ( res < 0 ) {
    ERROR("socket binding error %m\n");
    goto error_with_socket;
  }

  /* set up listening backlog for connect requests from clients */
  res =  listen(main_socket, TCP_LISTENING_DEPTH);
  if ( res < 0 ) {
    ERROR("socket listening error %m\n");
    goto error_with_socket;
  }

  LOG(2, "Now listening on socket %d\n", main_socket);

  /* set non-blocking to accept multiple connexion easily */
  fcntl(main_socket, F_SETFL, fcntl(main_socket, F_GETFL) | O_NONBLOCK);

#if SERVER_TCP_POLL == EPOLL
  res = epoll_init();
#elif SERVER_TCP_POLL == POLL
  res = poll_init();
#endif /* SERVER_TCP_POLL */
  if ( res < 0 )
    goto error_with_socket;

  return 0;

 error_with_socket:
  close(main_socket);
 error:
  return -1;
}


void TCP_exit()
{
#if SERVER_TCP_POLL == EPOLL
  epoll_finish();
#elif SERVER_TCP_POLL == POLL
  poll_finish();
#endif /* SERVER_TCP_POLL */
  close(main_socket);
}


/*************
 * Accepting *
 *************/

static INLINE
void accept_new_connection()
{
  int new_socket;
  int res;

  while ( 1 ) {
    new_socket = accept(main_socket, NULL, 0);
    if ( new_socket < 0 ) {
      if ( errno == EAGAIN  ||  errno == EWOULDBLOCK )
	break;
      else {
	ERROR("accept failed on error %m\n");
	break;
      }
    }

    SET_TCP_NODELAY(new_socket);
#if SERVER_TCP_POLL == EPOLL
    res = epoll_add_connection(new_socket);
#elif SERVER_TCP_POLL == POLL
    res = poll_add_connection(new_socket);
#endif /* SERVER_TCP_POLL */
  }
}


/*********************
 * Receiving Request *
 *********************/

static INLINE
int receive_request(int socket)
{
  struct server_request *req = get_free_request();
  void *ptr;
  size_t to_recv;
  ssize_t received;

  ptr = &req->req;
  to_recv = sizeof(req->req);
  while ( to_recv ) {
    received = recv(socket, ptr, to_recv, 0);
    if ( received <= 0 )
      goto connection_closed;
    to_recv -= received;
    ptr += received;
  }

  if ( get_request_buffer_len(&req->req, &to_recv) ) {
    /* assume there's only one buffer */
    req->in_vect = get_io_monovector_alloc(to_recv);
    ptr = VECT_PTR(req->in_vect, 0);
    while ( to_recv ) {
      received = recv(socket, ptr, to_recv, 0);
      if ( received <= 0 )
	goto connection_closed;
      to_recv -= received;
      ptr += received;
    }
  }

  req->connection.tcp.socket = socket;
  QUEUE_REQUEST(&ready_request_queue, req);
  counters.received++;
  return 0;

 connection_closed:
  put_free_request(req);
  return -1;
}


#if SERVER_TCP_POLL == EPOLL

static INLINE
void TCP_recv()
{
  int res = 0;
  int timeout = -1;
  struct epoll_event ev[TCP_SERVER_EVENT_GET_MAX];
  int i;

  do {
    res = epoll_wait(epoll_fd,
		     ev, TCP_SERVER_EVENT_GET_MAX,
		     timeout);
    LOG(3, "Epoll gets %d active sockets\n", res);
    for(i=0; i<res; i++) {
      int fd = ev[i].data.fd;
      int events = ev[i].events;
      LOG(4, "Socket %d gets events %d\n", fd, events);
      if ( fd == main_socket ) {
	if ( events != EPOLLIN )
	  ERROR("poll on main socket received event %d\n", events);
	accept_new_connection();
      } else {
	if ( events != EPOLLIN  ||  receive_request(fd) < 0 )
	  epoll_remove_connection(fd);
      }
    }
    timeout = 0;
  } while ( res == TCP_SERVER_EVENT_GET_MAX );
}

#elif SERVER_TCP_POLL == POLL

static INLINE
void TCP_recv()
{
  int res;
  int timeout = -1;
  int i=0;
  int indix;

  res = poll(pollfd, pollfd_nr, timeout);
  LOG(3, "Poll gets %d active sockets\n", res);

  for(indix=0; i<res; indix++) {
    int fd = pollfd[indix].fd;
    int events = pollfd[indix].revents;
    
    if ( !events )
      continue;

    LOG(4, "Socket %d gets events %d\n", fd, events);
    if ( indix == 0 ) {
      if ( events != POLLIN )
	ERROR("poll on main socket received event %d\n", events);
      accept_new_connection();
    } else {
      if ( events != POLLIN  ||  receive_request(fd) < 0 ) {
	poll_remove_connection(indix);
	fd--;
      }
    }

    i++;
  }
}

#endif /* SERVER_TCP_POLL */


/*******************
 * Sending Answers *
 *******************/

static INLINE
void TCP_send()
{
  struct server_request *req;

  while ( (req = DEQUEUE_REQUEST(&done_request_queue)) != NULL ) {
    int socket = req->connection.tcp.socket;
    int res;
    size_t data_len;

    res = write(socket, &req->ans, sizeof(req->ans));
    if ( res != sizeof(req->ans) )
      goto write_failed;
    if ( get_answer_buffer_len(&req->ans, &data_len) && data_len > 0 ) {
      res = writev(socket, req->out_vect->buf, req->out_vect->nr);
      if ( res != data_len )
	goto write_failed;
    }
    put_io_vector(req->out_vect);
    put_free_request(req);
    counters.sent++;
    continue;

write_failed:
    ERROR("Writing on socket %d failed\n", socket);    
    if ( res  < 0 )
      ERROR("Writing returned errno %m\n");

    put_io_vector(req->out_vect);
    put_free_request(req);
  }
}


/*************
 * Main Loop *
 *************/

void TCP_process(int *exiting)
{
  while ( !*exiting ) {
    /* Recv all */
    TCP_recv();

    /* Global time is updated each time we empty in_queue.
     * This should be enough for millisecond granularity,
     * even if we are using second granularity.
     */
    update_current_time();
    
    /* Process all */
    ramfs_process_queue();

    /* Send all */
    TCP_send();
  }
}
