/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __RAMFS_IO_H__
#define __RAMFS_IO_H__


#include <sys/uio.h>
#include <stdlib.h>
#include <string.h>

#include "ramfs_messages.h"


struct io_vector {
  int nr;
  int do_not_free;
  struct iovec buf[0];
};


#define VECT_NR(vect) ((vect)->nr)
#define VECT_BUF(vect, i) (&((vect)->buf[i]))
#define BUF_PTR(buf) ((buf)->iov_base)
#define BUF_SIZE(buf) ((buf)->iov_len)
#define VECT_PTR(vect, i) ((vect)->buf[i].iov_base)
#define VECT_SIZE(vect, i) ((vect)->buf[i].iov_len)



extern inline
struct io_vector *get_io_vector(int nr,
				int do_not_free)
{
  struct io_vector *vect = malloc(sizeof(*vect)+nr*sizeof(*vect->buf));

  vect->nr = nr;
  vect->do_not_free = do_not_free;

  return vect;
}


extern inline
struct io_vector *get_io_monovector(void *ptr,
				    size_t size,
				    int do_not_free)
{
  struct io_vector *vect = get_io_vector(1, do_not_free);

  vect->buf[0].iov_base = ptr;
  vect->buf[0].iov_len = size;

  return vect;
}


extern inline
struct io_vector *get_io_monovector_alloc(size_t size)
{
  return get_io_monovector(network_malloc(size),
			   size,
			   0);
}


extern inline
void put_io_vector(struct io_vector *vect)
{
  if ( vect ) {
    if ( !vect->do_not_free ) {
      int i;
      for(i=0; i<vect->nr; i++)
	network_free(vect->buf[i].iov_base);
    }
    free(vect);
  }
}


extern inline
void resize_io_monovector(struct io_vector *vect,
			  size_t size)
{
  DEBUG_ASSERT( size<=vect->buf[0].iov_len );
  vect->buf[0].iov_len = size;
}


extern inline
struct io_vector *merge_io_vector(struct io_vector *src)
{
  size_t size = 0;
  struct io_vector *new;
  void *current_ptr;
  int i;

  if ( src == NULL )
    return NULL;

  if ( src->nr == 1 )
    return src;

  for(i=0; i<src->nr; i++)
    size += src->buf[i].iov_len;
  
  new = get_io_monovector_alloc(size);
  
  current_ptr = new->buf[0].iov_base;

  for(i=0; i<src->nr; i++) {
    int to_copy = src->buf[i].iov_len;
    memcpy(current_ptr, src->buf[i].iov_base, to_copy);
    current_ptr += to_copy;
  }

  put_io_vector(src);

  return new;
}


#endif /* __RAMFS_IO_H__ */

