/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __RAMFS_LIST_H__
#define __RAMFS_LIST_H__


#if DEBUG_LEVEL >= 0
#include <assert.h>
#endif


struct list_head {
  struct list_head *prev, *next;
};

/* initialize */
#define LIST_INIT(list) (list)->next = (list)->prev = (list)

/* return 1 when list is empty */
#define LIST_EMPTY(ptr) ( (ptr)->next == ptr )

/* add at the begining */
#if DEBUG_LEVEL < 0
#define LIST_ADD(ptr, list)   \
do {                          \
  (list)->prev->next = (ptr); \
  (ptr)->prev = (list)->prev; \
  (ptr)->next = (list);       \
  (list)->prev = (ptr);       \
} while ( 0 )
#else
#define LIST_ADD(ptr, list)   \
do {                          \
  assert(!(ptr)->prev);       \
  assert(!(ptr)->next);       \
  (list)->prev->next = (ptr); \
  (ptr)->prev = (list)->prev; \
  (ptr)->next = (list);       \
  (list)->prev = (ptr);       \
} while ( 0 )
#endif

/* add at the end */
#define LIST_ADD_TAIL(ptr, list) LIST_ADD(ptr, (list)->prev)

/* remove element */
#if DEBUG_LEVEL < 0
#define LIST_DEL(ptr)              \
do {                               \
  (ptr)->next->prev = (ptr)->prev; \
  (ptr)->prev->next = (ptr)->next; \
} while ( 0 )
#else
#define LIST_DEL(ptr)              \
do {                               \
  (ptr)->next->prev = (ptr)->prev; \
  (ptr)->prev->next = (ptr)->next; \
  (ptr)->prev = NULL;              \
  (ptr)->next = NULL;              \
} while ( 0 )
#endif

/* returns real element from list element */
#define LIST_HEAD_ELEMENT(ptr, type, member) \
((type *)((char *)(ptr)-(unsigned long)(&((type *)0)->member)))

/**
 * for each element of the list
 * @elt:     a pointer to the user element
 * @ptr:     a pointer to the list_head field in @elt
 * @member:  the name of the list_head field in @elt
 * @head:    a pointer to the list_head begining he list
 */
#define LIST_FOR_EACH(elt, ptr, member, head)           \
for(ptr = (head)->next,                                 \
    elt = LIST_HEAD_ELEMENT(ptr, typeof(*elt), member); \
    ptr != head;                                        \
    ptr = ptr->next,                                    \
    elt = LIST_HEAD_ELEMENT(ptr, typeof(*elt), member))


/* get and remove first element of an assumed-nonempty list */
extern inline
struct list_head *LIST_GET_FIRST(struct list_head *list)
{
  struct list_head *elt = list->next;
  LIST_DEL(elt);
  return elt;
}

/* get and remove last element of an assumed-nonempty list */
extern inline
struct list_head *LIST_GET_LAST(struct list_head *list)
{
  struct list_head *elt = list->prev;
  LIST_DEL(elt);
  return elt;
}


#endif /* __RAMFS_LIST_H__ */
