/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __RAMFS_QUEUE_H__
#define __RAMFS_QUEUE_H__


#if DEBUG_LEVEL < 0
#include <assert.h>
#endif

#include "ramfs_server.h"
#include "ramfs_list.h"


#if DEBUG_LEVEL < 0
#define ramfs_queue list_head
#define Q2L(x) (x)
#else
struct ramfs_queue { struct list_head list; unsigned int count; };
#define Q2L(x) (&(x)->list)
#endif


/**********************
 * Free Request Queue *
 **********************/

extern void queue_init();

extern struct server_request *get_free_request();
extern void put_free_request(struct server_request *);


/***************************
 * Ready and Done Requests *
 ***************************/

extern struct ramfs_queue ready_request_queue, done_request_queue;


/*****************************************
 * Macros to extract Request from Queues *
 *****************************************/

/* Getting Request from its List Head field */
#define LIST_HEAD_REQ(ptr) \
LIST_HEAD_ELEMENT(ptr, struct server_request, queue)


/*******************************************
 * Generic Way to Declare a Specific Queue *
 *******************************************/

/* Declaration */
#define DECLARE_REQUEST_QUEUE(q) \
struct ramfs_queue q

/* Initialization */
#if DEBUG_LEVEL < 0
#define INIT_REQUEST_QUEUE(q) \
LIST_INIT(&(q))
#else
#define INIT_REQUEST_QUEUE(q) \
do { LIST_INIT(Q2L(&(q))); (q).count = 0; } while (0)
#endif

/* Is Empty ? */
#if DEBUG_LEVEL < 0
#define QUEUE_EMPTY(q) \
LIST_EMPTY(q)
#else
#define QUEUE_EMPTY(q) \
({ assert( !LIST_EMPTY(Q2L(q)) != !(q)->count ); LIST_EMPTY(Q2L(q)); })
#endif

/* Queueing */
#if DEBUG_LEVEL < 0
#define QUEUE_REQUEST(q, req) \
LIST_ADD_TAIL(&(req)->queue, q)
#else
#define QUEUE_REQUEST(q, req) \
do { LIST_ADD_TAIL(&(req)->queue, Q2L(q)); (q)->count++; } while (0)
#endif

/* Queueing at the head */
#if DEBUG_LEVEL < 0
#define QUEUE_REQUEST_AT_HEAD(q, req) \
LIST_ADD(&(req)->queue, q)
#else
#define QUEUE_REQUEST_AT_HEAD(q, req) \
do { LIST_ADD(&(req)->queue, Q2L(q)); (q)->count++; } while (0)
#endif

/* Dequeueing */
#if DEBUG_LEVEL < 0
#define DEQUEUE_REQUEST(q) \
(QUEUE_EMPTY(q) ? NULL : LIST_HEAD_REQ(LIST_GET_FIRST(q)))
#else
#define DEQUEUE_REQUEST(q)  \
(QUEUE_EMPTY(q) ? NULL : ({ (q)->count--; LIST_HEAD_REQ(LIST_GET_FIRST(Q2L(q))); }))
#endif

/* Dequeueing at the end */
#if DEBUG_LEVEL < 0
#define DEQUEUE_REQUEST_AT_TAIL(q) \
(QUEUE_EMPTY(q) ? NULL : LIST_HEAD_REQ(LIST_GET_LAST(q)))
#else
#define DEQUEUE_REQUEST_AT_TAIL(q) \
(QUEUE_EMPTY(q) ? NULL : ({ (q)->count--; LIST_HEAD_REQ(LIST_GET_LAST(Q2L(q))); }))
#endif


/* Type of Function Comparing a Request and a Data */
typedef int (*server_request_compare_func_t)(struct server_request*, void*);

/* Dequeueing a specific request which matches data
 * through the compare function.
 */
extern inline
struct server_request *
DEQUEUE_SPECIFIC_REQUEST(struct ramfs_queue *queue,
			 server_request_compare_func_t compare,
			 void *data)
{
  struct list_head *head;
  struct server_request *req;

  LIST_FOR_EACH(req, head, queue, Q2L(queue))
    if ( compare(req, data) ) {
      LIST_DEL(head);
#if DEBUG_LEVEL >= 0
      assert(queue->count > 0);
      queue->count--;
#endif
      return req;
    }

  return NULL;
}


#endif /* __RAMFS_QUEUE_H__ */
