/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sys/time.h>
#include <utime.h>
#include <assert.h>

#include "orfs_protocol.h"
#include "ramfs_network.h"
#include "ramfs_server.h"
#include "ramfs_io.h"
#include "ramfs_list.h"
#include "ramfs_queue.h"
#include "ramfs_messages.h"


/***************
 * Fake Buffer *
 ***************/

/* this fake buffer will be used to read in sparse files */
static char *fake_buffer;


/**********
 * Timing *
 **********/

static orfs_time_t current_time;

void update_current_time() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  current_time = tv.tv_sec;
}

#define GET_CURRENT_TIME() current_time
#define INIT_NODE_TIMES(node) node->atime = node->ctime = node->mtime = current_time
#define NOTIFY_WRITE_ACCESS(node) node->ctime = node->mtime = current_time
#define NOTIFY_READ_ACCESS(node) node->atime = current_time
#define NOTIFY_CHANGE_ACCESS(node) node->ctime = current_time


/******************
 * Various Macros *
 ******************/

#define OFFSET_IN_BLOCK(x) ((x)&(MEMORY_BLOCK_SIZE-1))
#define BLOCK_FROM_OFFSET(x) (orfs_off_t)((x)>>MEMORY_BLOCK_BITS)

#define LIST_HEAD_DENTRY(ptr) \
LIST_HEAD_ELEMENT(ptr, struct ramfs_dentry, name_hash)

#define LIST_HEAD_NODE(ptr) \
LIST_HEAD_ELEMENT(ptr, struct ramfs_node, ino_hash)


#define RAMFS_REQUEST_BLOCKED ORFS_ERROR_MAX


/*******************
 * Data structures *
 *******************/

typedef unsigned int ramfs_hash_t;
typedef unsigned int ramfs_count_t;


struct ramfs_file_block {
  void *ptr;
};


struct ramfs_dentry {
  /* name hashing */
  struct list_head name_hash;

  /* node */
  struct ramfs_node *node;
  struct ramfs_node *parent;
  orfs_doff_t indix_for_parent;

  /* dentry data */
  orfs_doff_t off;
  orfs_name_len_t name_len;
  char name[0];
};


struct ramfs_node {
  /* ino hashing */
  struct list_head ino_hash;

  /* ino */
  orfs_ino_t ino;
  /* mode */
  orfs_mode_t mode;
  /* owner */
  orfs_uid_t uid;
  orfs_gid_t gid;
  /* times */
  orfs_time_t atime;
  orfs_time_t mtime;
  orfs_time_t ctime;
  /* link */
  orfs_nlink_t nlink;

  /* specific data */
  ramfs_count_t count;
  union {
    /* regular file */
    struct ramfs_file_data {
      struct ramfs_file_block *block;
      orfs_off_t block_max;
      orfs_off_t block_nr;
      orfs_off_t length;
      /* writers and waiters */
      unsigned long writers;
      struct ramfs_queue waiters;
    } file;
    /* directory */
    struct ramfs_dir_data {
      struct ramfs_dentry **entry;
      orfs_doff_t entry_max;
      orfs_doff_t entry_nr;
    } dir;
    /* symlink */
    struct ramfs_lnk_data {
      char *path;
      orfs_off_t length;
    } lnk;
    /* dev */
    struct ramfs_dev_data {
      orfs_dev_t dev;
    } dev;
  } data;
};


/*******************
 * Our file system *
 *******************/

static
struct ramfs_fs {
  struct list_head ino_table[HASH_TABLE_SIZE];
  struct list_head name_table[HASH_TABLE_SIZE];
  struct ramfs_node *root;
  orfs_ino_t ctr;
} fs;


/*************************
 * Ino Hashing functions *
 *************************/

static INLINE
ramfs_hash_t ino_hash_func(orfs_ino_t ino)
{
  return ino % HASH_VALUE;
}


static INLINE
void ino_hash_node(struct ramfs_node *node)
{
  ramfs_hash_t ino_hash_key = ino_hash_func(node->ino);
  LIST_ADD(&node->ino_hash, &fs.ino_table[ino_hash_key]);
}


static INLINE
void ino_unhash_node(struct ramfs_node *node)
{
  LIST_DEL(&node->ino_hash);
}


/**************************
 * Name Hashing functions *
 **************************/

static INLINE
ramfs_hash_t __name_hash_func(const char *name)
{
  ramfs_hash_t hash = *name;

  if (hash)
    for(name++; *name != '\0'; name++)
      hash = (hash<<5) - hash + *name;

  return hash;
}


static INLINE
ramfs_hash_t name_hash_func(const char *name,
			    struct ramfs_node *parent)
{
  return ino_hash_func(__name_hash_func(name)+parent->ino);
}


static INLINE
void name_hash_dentry(struct ramfs_dentry *dentry)
{
  ramfs_hash_t name_hash_key = name_hash_func(dentry->name, dentry->parent);
  LIST_ADD(&dentry->name_hash, &fs.name_table[name_hash_key]);
}
 

static INLINE
void name_unhash_dentry(struct ramfs_dentry *dentry)
{
  LIST_DEL(&dentry->name_hash);
}


/*******************
 * Waiting Writers *
 *******************/

static
void enqueue_waiter(struct ramfs_node *node,
		    struct server_request *req)
{
  QUEUE_REQUEST(&node->data.file.waiters, req);
}


static
void wake_up_waiters(struct ramfs_node *node)
{
  struct server_request *req;

  /* We want to enqueue at the head.
   * But we also need to keep the initial order.
   */
  while ( (req=DEQUEUE_REQUEST_AT_TAIL(&node->data.file.waiters)) != NULL )
    QUEUE_REQUEST_AT_HEAD(&ready_request_queue, req);
}


/**********
 * Blocks *
 **********/

static
int extend_block_list(struct ramfs_node *node,
		      orfs_off_t end_nr)
{
  struct ramfs_file_data *data = &node->data.file;
  orfs_off_t new_max = data->block_max * 2;
  void *new_list;

  while ( end_nr > new_max )
    new_max <<= 1;
  
  LOG(5, "Increasing node %lld block_max from %lld to %lld\n",
      (long long) node->ino,
      (long long) data->block_max, (long long) new_max);
  
  new_list = realloc(data->block, new_max*sizeof(*data->block));
  if ( !new_list ) {
    ERROR("Cannot reallocate RAMFS file block list\n");
    return ORFS_NO_SPACE_LEFT;
  }
  /* initialize to 0 */
  memset(new_list + data->block_max*sizeof(*data->block),
         0,
         (new_max-data->block_max)*sizeof(*data->block));
  
  data->block_max = new_max;
  data->block = (struct ramfs_file_block*) new_list;
  
  return ORFS_SUCCESS;
}

static INLINE
int alloc_blocks(struct ramfs_node *node,
		 orfs_off_t begin_bl, orfs_off_t end_bl)
{
  struct ramfs_file_data *data = &node->data.file;
  orfs_off_t nr = data->block_nr;
  orfs_off_t i;
  orfs_result_t res;
  void *block;

  /* extend block list */
  if ( end_bl+1 > data->block_max ) {
     res = extend_block_list(node, end_bl+1);
     if ( res != ORFS_SUCCESS )
       return res;
  }

  /* malloc new blocks */
  for(i=begin_bl; i<=end_bl; i++)
    if ( data->block[i].ptr == NULL ) {
      LOG(5, "Allocated node %lld block %lld\n",
	  (long long) node->ino, (long long) i);
      block = network_malloc(MEMORY_BLOCK_SIZE);
      if ( !block ) {
	ERROR("Cannot allocate RAMFS file block\n");
	return ORFS_NO_SPACE_LEFT;
      }
      data->block[i].ptr = block;
      nr++;
    }

  /* adjust block_nr */
  if ( data->block_nr != nr ) {
    NOTIFY_CHANGE_ACCESS(node);
    data->block_nr = nr;
  }

  return ORFS_SUCCESS;
}


static INLINE
void remove_last_blocks(struct ramfs_node *node,
			orfs_off_t end_nr)
{
  struct ramfs_file_data *data = &node->data.file;
  orfs_off_t nr = data->block_nr;
  orfs_off_t i;

  LOG(5, "Truncating node %lld to %lld blocks\n",
      (long long) node->ino, (long long) end_nr);
  
  /* free blocks */
  for(i=end_nr; i<data->block_max; i++) 
    if ( data->block[i].ptr != NULL ) {
      network_free(data->block[i].ptr);
      data->block[i].ptr = NULL;
      nr--;
    }
  
  /* adjust block_nr */
  if ( data->block_nr != nr ) {
    NOTIFY_CHANGE_ACCESS(node);
    data->block_nr = nr;
  }
}


/********
 * Node *
 ********/

static INLINE
orfs_ino_t get_new_ino()
{
  static orfs_ino_t next_ino = ORFS_ROOT_INO;
  return next_ino++;
}


static INLINE
struct ramfs_node *get_new_node(orfs_mode_t mode,
				orfs_uid_t uid,
				orfs_gid_t gid)
{
  struct ramfs_node *node = malloc(sizeof(*node));

  if ( !node ) {
    ERROR("Cannot allocate RAMFS node\n");
    return NULL;
  }

  node->ino = get_new_ino();
  node->mode = mode;
  node->uid = uid;
  node->gid = gid;
  node->nlink = 0;
  node->count = 0;

  INIT_NODE_TIMES(node);

#if DEBUG_LEVEL >= 0
  node->ino_hash.prev = NULL;
  node->ino_hash.next = NULL;
#endif

  ino_hash_node(node);

  return node;
}


static INLINE
void free_unlinked_node(struct ramfs_node *node)
{
  /* free specific data */
  switch ( ORFS_ISFMT(node->mode) ) {
  case ORFS_IFREG:
    remove_last_blocks(node, 0);
    assert(node->data.file.block_nr == 0);
    free(node->data.file.block);
    break;
  case ORFS_IFDIR:
    assert(node->data.dir.entry_nr == 0);
    free(node->data.dir.entry);
    break;
  case ORFS_IFLNK:
    network_free(node->data.lnk.path);
    break;
  case ORFS_IFCHR:
  case ORFS_IFBLK:
  case ORFS_IFIFO:
  case ORFS_IFSOCK:
    break;
  default:
    assert(0);
  }

  /* unhash node */
  ino_unhash_node(node);

  /* free */
  free(node);
}


/**********
 * Dentry *
 **********/

static INLINE
struct ramfs_dentry *get_new_dentry(struct ramfs_node *parent,
				    struct ramfs_node *node,
				    const char *name,
				    orfs_name_len_t name_len)
{
  struct ramfs_dentry *dentry;

  /* allocate */
  dentry = malloc(sizeof(*dentry) + name_len+1);
  if ( !dentry ) {
    ERROR("Cannod allocate RAMFS dir entry\n");
    return NULL;
  }

#if DEBUG_LEVEL >= 0
  dentry->name_hash.prev = NULL;
  dentry->name_hash.next = NULL;
#endif

  /* fix name */
  dentry->name_len = name_len;
  memcpy(dentry->name, name, name_len+1);

  /* fix nodes */
  dentry->node = node;
  dentry->parent = parent;

  return dentry;
}


static INLINE
void put_old_dentry(struct ramfs_dentry *dentry)
{
  free(dentry);
}


static
int link_dentry(struct ramfs_dentry *dentry)
{
  struct ramfs_node *node = dentry->node;
  struct ramfs_node *parent = dentry->parent;
  struct ramfs_dir_data *data = &parent->data.dir;
  orfs_doff_t indix;
  orfs_doff_t off = 0;

  /* find an empty place */
  if ( data->entry_nr == data->entry_max ) {
    void *new_list;
    /* increase list size if necessary */
    new_list = realloc(data->entry,
		       data->entry_max*2*sizeof(*data->entry));
    if ( !new_list ) {
      ERROR("Cannot reallocate RAMFS dir entry list\n");
      return ORFS_NO_SPACE_LEFT;
    } 
    data->entry = (struct ramfs_dentry**)new_list; 
    /* initialize it */
    memset(&data->entry[data->entry_max],
	   0, data->entry_max*sizeof(*data->entry));
    
    /* fix its size */
    data->entry_max *= 2;
    
    /* let's take the first new place */
    indix = data->entry_nr;
  } else {
    /* search the first empty case, we know there's one */
    for(indix=0; data->entry[indix]!=NULL; indix++);
  }

  /* link to parent */
  dentry->indix_for_parent = indix;
  data->entry[indix] = dentry;
  data->entry_nr++;
  /* link to node */
  node->nlink++;

  /* compute off */
  if ( indix ) {
    off = data->entry[indix-1]->off;
    data->entry[indix-1]->off = indix;
  }
  if ( off == indix )
    off = indix + 1;
  dentry->off = off;

  /* hash dentry */
  name_hash_dentry(dentry);

  /* update node times */
  NOTIFY_WRITE_ACCESS(dentry->parent);

  return ORFS_SUCCESS;
}


static INLINE
void unlink_dentry(struct ramfs_dentry *dentry)
{
  struct ramfs_node *parent = dentry->parent;
  struct ramfs_dir_data *data = &parent->data.dir;
  orfs_doff_t indix = dentry->indix_for_parent;
  struct ramfs_node *node = dentry->node;
  orfs_doff_t i;

  if ( indix > 0 ) {
    /* update off of previous dentry */
    for(i=indix-1; !data->entry[i]; i--);
    data->entry[i]->off = data->entry[indix]->off;
  } else {
    
  }

  /* unlink from parent */
  data->entry[indix] =  NULL;
  data->entry_nr--;

  /* unlink from node */
  node->nlink--;

  /* unhash */
  name_unhash_dentry(dentry);

  /* update node times */
  NOTIFY_WRITE_ACCESS(dentry->parent);
}


static INLINE
int move_dentry(struct ramfs_dentry *old_dentry,
		 struct ramfs_node *new_parent,
		 const char *new_name,
		 orfs_name_len_t new_name_len)
{
  struct ramfs_node *node = old_dentry->node;
  struct ramfs_dentry *new_dentry;

  new_dentry = get_new_dentry(new_parent, node, new_name, new_name_len);
  if ( !new_dentry )
    return ORFS_NO_SPACE_LEFT;

  /* unlink from parent */
  unlink_dentry(old_dentry);
  put_old_dentry(old_dentry);

  /* link to new parent */
  link_dentry(new_dentry);

  return ORFS_SUCCESS;
}


/**********
 * Lookup *
 **********/

static
struct ramfs_node *lookup_ino(orfs_ino_t ino)
{
  ramfs_hash_t hash_key;
  struct list_head *head, *elt;

  hash_key = ino_hash_func(ino);
  head = &fs.ino_table[hash_key];

  for(elt=head->next; elt!=head; elt=elt->next) {
    struct ramfs_node *node = LIST_HEAD_NODE(elt);
    if ( node->ino == ino )
      return node;
  }
  return NULL;
}


static
struct ramfs_dentry *lookup_name(const char *name,
				 struct ramfs_node *parent)
{
  ramfs_hash_t hash_key;
  struct list_head *head, *elt;

  hash_key = name_hash_func(name, parent);
  head = &fs.name_table[hash_key];

  for(elt=head->next; elt!=head; elt=elt->next) {
    struct ramfs_dentry *dentry = LIST_HEAD_DENTRY(elt);
    if ( dentry->parent == parent ) {
      const char *dentry_name = dentry->name;
      if ( !strcmp(dentry_name, name) )
	return dentry;
    }
  }
  return NULL;
}


/************************
 * Attributes Functions *
 ************************/

static /* no inline, because it is often used */
void getattr_node(struct ramfs_node *node,
		  struct orfs_getattr_data *attr)
{
  LOG(4, "GETATTR of node %lld\n", (long long)node->ino);

  attr->id.ino = node->ino;
  attr->id.sb = 0;
  attr->mode = node->mode;
  attr->uid = node->uid;
  attr->gid = node->gid;
  attr->atime = node->atime;
  attr->mtime = node->mtime;
  attr->ctime = node->ctime;
  attr->nlink = node->nlink;
  //  attr->version = 0; // TOFIX ??

  switch ( ORFS_ISFMT(node->mode) ) {
  case ORFS_IFREG :
    attr->size = node->data.file.length;
    attr->blocks = node->data.file.block_nr;
    attr->rdev = 0;
    break;
  case ORFS_IFDIR :
    attr->size = node->data.dir.entry_nr;
    attr->blocks = node->data.dir.entry_nr;
    attr->rdev = 0;
    break;
  case ORFS_IFLNK:
    attr->size = node->data.lnk.length;
    attr->blocks = 1;
    attr->rdev = 0;
    break;
  case ORFS_IFBLK:
  case ORFS_IFCHR:
    attr->rdev = node->data.dev.dev;
  case ORFS_IFIFO:
  case ORFS_IFSOCK:
    attr->size = 0;
    attr->blocks = 0;
    break;
    
  default:
    assert(0);
  }
}


static INLINE
void truncate_node(struct ramfs_node *node,
		   orfs_off_t len)
{
  struct ramfs_file_data *data = &node->data.file;
  orfs_off_t end_nr = BLOCK_FROM_OFFSET(len+MEMORY_BLOCK_SIZE-1);

  if ( len > data->length )
    extend_block_list(node, end_nr);
  else if ( len < data->length )
    remove_last_blocks(node, end_nr);

  data->length = len;
}


static INLINE
orfs_result_t setattr_node(struct ramfs_node *node,
			   struct orfs_setattr_data *data)
{
  orfs_valid_attr_t valid = data->valid;

  LOG(4, "SETATTR of node %lld\n", (long long)node->ino);

  /* Cannot change symlink mode */
  if ( (valid & ORFS_ATTR_MODE) && ORFS_ISLNK(node->mode) )
    return ORFS_BAD_INODE_TYPE;

  /* Cannot change size of non-regular files */
  if ( (valid & ORFS_ATTR_SIZE) && !ORFS_ISREG(node->mode) )
    return ORFS_BAD_INODE_TYPE;

  /* Check whether truncate must block */
  if ( (valid & ORFS_ATTR_SIZE) && node->data.file.writers
       && data->size < node->data.file.length )
    return RAMFS_REQUEST_BLOCKED;

  /* Now really setattr */

  if ( valid & ORFS_ATTR_MODE ) {
    node->mode = (node->mode & (~ORFS_ACCESS_PERMS))
      | (data->mode & ORFS_ACCESS_PERMS);
  }
  if ( valid & ORFS_ATTR_UID ) {
    node->uid = data->uid;
  }
  if ( valid & ORFS_ATTR_GID ) {
    node->gid = data->gid;
  }
  if ( valid & ORFS_ATTR_SIZE ) {
    truncate_node(node, data->size);
  }
  if ( valid & ORFS_ATTR_FORCE ) {
    /* nothing for FORCE */    
  }
  if ( valid & ORFS_ATTR_ATTR_FLAG ) {
    /* nothing for ATTR_FLAG */
  }

  /* Protocol Note:
   * Access and Modify Time are updated with
   * their corresponding read/write/readdir request.
   * Thus clients do never update these times,
   * except when utime[s] is called.
   * Change time is maintained by the server.
   */
  if ( valid & ORFS_ATTR_ATIME ) {
    node->atime = valid & ORFS_ATTR_ATIME_SET ?
      GET_CURRENT_TIME() : data->atime;
  }
  if ( valid & ORFS_ATTR_MTIME ) {
    node->mtime = valid & ORFS_ATTR_MTIME_SET ?
      GET_CURRENT_TIME() : data->mtime;
  }

  NOTIFY_CHANGE_ACCESS(node);

  return ORFS_SUCCESS;
}


static INLINE
orfs_result_t readlink_node(struct ramfs_node *node,
			    struct io_vector **vect,
			    orfs_path_len_t *count)
{
  orfs_path_len_t length = node->data.lnk.length;

  LOG(4, "READLINK in node %lld\n", (long long)node->ino);

  if ( !ORFS_ISLNK(node->mode) )
    return ORFS_BAD_INODE_TYPE;
  
  *vect = get_io_monovector(node->data.lnk.path, length, 1);
  *count = length;

  /* update node times */
  NOTIFY_READ_ACCESS(node);

  return ORFS_SUCCESS;
}


/*****************
 * FILE requests *
 *****************/

typedef uint32_t ramfs_fd_t;

struct ramfs_fd {
  struct ramfs_node *node;
  ramfs_fd_t next_free;
} fd_table[DESCRIPTOR_MAX];

static ramfs_fd_t first_free_fd;


static INLINE
orfs_result_t open_node(struct ramfs_node *node,
			ramfs_fd_t *fd)
{
  ramfs_fd_t new_fd = first_free_fd;

  if ( new_fd == DESCRIPTOR_MAX )
    return ORFS_NO_FILE_LEFT;

  first_free_fd = fd_table[new_fd].next_free;
  fd_table[new_fd].node = node;

  node->count++;
  *fd = new_fd;

  LOG(4, "OPEN inode %lld (count %d, nlink %d) into fd %lld\n",
      (long long)node->ino, node->count, node->nlink, (long long)new_fd);

  return ORFS_SUCCESS;
}


static INLINE
orfs_result_t release_file(struct ramfs_node *node,
			   ramfs_fd_t old_fd)
{
  LOG(4, "RELEASE inode %lld (count %d, nlink %d) from fd %lld\n",
      (long long)node->ino, node->count, node->nlink, (long long)old_fd);

  if ( !fd_table[old_fd].node ) {
    ERROR("Put unused returns error\n");
    return ORFS_BAD_FILE;
  }

  fd_table[old_fd].node = NULL;
  fd_table[old_fd].next_free = first_free_fd;
  first_free_fd = old_fd;

  if ( !--node->count  &&  !node->nlink ) {
    /* real remove */
    free_unlinked_node(node);
    LOG(5, "Really unlinking inode %lld\n",
	(long long)node->ino);
  }

  return ORFS_SUCCESS;
}


static INLINE
orfs_result_t preaddir_node(struct ramfs_node *node,
			    struct io_vector **vect,
			    orfs_doff_t off,
			    orfs_data_len_t data_max,
			    orfs_doff_t *count,
			    orfs_bool_t *eof_reached,
			    orfs_data_len_t *data_len)
{
  struct ramfs_dir_data *data = &node->data.dir;
  struct orfs_dirent *dirent;
  char *dirent_ptr;
  orfs_data_len_t total_len;
  orfs_doff_t current;
  orfs_doff_t read;
  orfs_doff_t i;

  LOG(4, "PREADDIR in node %lld at %lld for %lld max\n",
      (long long)node->ino, (long long)off, (long long)data_max);

  NOTIFY_READ_ACCESS(node);

  *eof_reached = 0;

  /* compute data length and number of dirent to return */
  total_len = 0;
  read = 0;
  current = off;
  while ( 1 ) {
    orfs_name_len_t name_len;
    orfs_data_len_t new_len;

    /* check eof */
    if ( current >= data->entry_max  ||  !data->entry[current] ) {
      *eof_reached = 1;
      break;
    }

    name_len = data->entry[current]->name_len;
    new_len = sizeof(*dirent) + name_len+1;

    /* update data len */
    if ( data_max-total_len < new_len )
      break;
    total_len += new_len;

    /* next dirent */
    read++;
    current = data->entry[current]->off;
  }

  /* EOF (or not enough place to get one dirent) */
  if ( !read )
    goto out;

  *vect = get_io_monovector_alloc(total_len);
  if ( !*vect )
    return ORFS_SERVER_NOMEM;

  dirent_ptr = VECT_PTR(*vect, 0);
  current = off;
  for(i=0; i<read; i++) {
    orfs_name_len_t name_len = data->entry[current]->name_len;
    dirent = (struct orfs_dirent *) dirent_ptr;
    getattr_node(data->entry[current]->node, &dirent->attr);
    dirent->off = data->entry[current]->off;
    dirent->type = 0;
    dirent->name_len = name_len;
    memcpy(dirent->name, data->entry[current]->name, name_len+1);
    /* next dirent */
    dirent_ptr += sizeof(*dirent) + name_len+1;
    current = data->entry[current]->off;
  }

 out:
  *count = read;
  *data_len = total_len;
  return ORFS_SUCCESS;
}


static INLINE
orfs_result_t pread_node(struct ramfs_node *node,
			 struct io_vector **vect,
			 orfs_off_t nbyte,
			 orfs_off_t offset,
			 orfs_off_t *read)
{
  struct ramfs_file_data *data = &node->data.file;
  orfs_off_t begin;
  orfs_off_t end;
  orfs_off_t begin_block;
  orfs_off_t end_block;
  orfs_off_t begin_offset;

  LOG(4, "PREAD in node %lld at %lld for %lld\n",
      (long long)node->ino, (long long)offset, (long long)nbyte);

  NOTIFY_READ_ACCESS(node);

  if ( offset >= data->length ) {
    *read = 0;
    goto out_with_success;
  }

  begin = offset;
  end = offset+nbyte-1;
  if ( end > data->length-1 ) {
    end = data->length-1;
    nbyte = end-begin+1;
  }

  if ( end < begin )
    return ORFS_SERVER_INVALID_PARAMETER;

  begin_block = BLOCK_FROM_OFFSET(begin);
  end_block = BLOCK_FROM_OFFSET(end);
  begin_offset = OFFSET_IN_BLOCK(begin);

  if ( begin_block == end_block ) {
    void *ptr = data->block[begin_block].ptr + begin_offset;
    if ( !data->block[begin_block].ptr ) {
      LOG(5, "Using fake buffer instead of block %lld of node %lld\n",
	  (long long) begin_block, (long long) node->ino);
      ptr = fake_buffer;
    }
    *vect = get_io_monovector(ptr, nbyte, 1);
    *read = nbyte;
    goto out_with_success;
  } else {
    struct ramfs_file_block *current_block = &data->block[begin_block];
    struct iovec *current_buffer;
    orfs_off_t remaining_bytes;
    void *ptr;

    /* allocate an output vector */
    *vect = get_io_vector(end_block-begin_block+1, 1);
    if ( !*vect )
      return ORFS_SERVER_NOMEM;
    
    /* first block ptr */
    ptr = current_block->ptr + begin_offset;
    if ( !current_block->ptr ) {
      LOG(5, "Using fake buffer instead of block %lld of node %lld\n",
	  (long long) (current_block-&data->block[0]),
	  (long long) node->ino);
      ptr = fake_buffer;
    }

    /* insert first block */
    current_buffer = VECT_BUF(*vect, 0);
    BUF_PTR(current_buffer) = ptr;
    BUF_SIZE(current_buffer) = MEMORY_BLOCK_SIZE-begin_offset;
    
    /* prepare following copy */
    current_block++;
    current_buffer++;
    remaining_bytes = nbyte - (MEMORY_BLOCK_SIZE-begin_offset);
    do {
      orfs_off_t to_copy;

      to_copy = remaining_bytes > MEMORY_BLOCK_SIZE ?
	MEMORY_BLOCK_SIZE : remaining_bytes;
      
      /* insert new buffer */
      ptr = current_block->ptr;
      if ( !ptr ) {
      LOG(5, "Using fake buffer instead of block %lld of node %lld\n",
	  (long long) begin_block, (long long) node->ino);
	ptr = fake_buffer;
      }
      BUF_PTR(current_buffer) = ptr;
      BUF_SIZE(current_buffer) = to_copy;
      
      /* update parameters */
      current_block++;
      current_buffer++;
      remaining_bytes -= to_copy;
    } while ( remaining_bytes > 0 );
    
    *read = nbyte;
    goto out_with_success;
  }

 out_with_success:
  return ORFS_SUCCESS;
} 


static INLINE
orfs_result_t pwrite_node(struct ramfs_node *node,
			  const char *buf,
			  orfs_off_t nbyte,
			  orfs_off_t offset,
			  orfs_off_t *written)
{
  struct ramfs_file_data *data = &node->data.file;
  orfs_off_t begin;
  orfs_off_t end;
  orfs_off_t begin_block;
  orfs_off_t end_block;
  orfs_off_t begin_offset;
  orfs_result_t res;

  LOG(4, "PWRITE in node %lld at %lld for %lld\n",
      (long long)node->ino, (long long)offset, (long long)nbyte);

  begin = offset;
  end = offset+nbyte-1;

  if ( end < begin )
    return ORFS_SERVER_INVALID_PARAMETER;

  begin_block = BLOCK_FROM_OFFSET(begin);
  end_block = BLOCK_FROM_OFFSET(end);
  begin_offset = OFFSET_IN_BLOCK(begin);

  res = alloc_blocks(node, begin_block, end_block);
  if ( res != ORFS_SUCCESS )
    return res;

  if ( end+1 > data->length ) {
    LOG(5, "Extending regular node %lld from %lld to %lld\n",
	(long long) node->ino,
	(long long) data->length, (long long) end+1);
    data->length = end+1;
  }

  NOTIFY_WRITE_ACCESS(node);

  if ( begin_block == end_block )
    {
      memcpy(data->block[begin_block].ptr + begin_offset,
	     buf,
	     nbyte);
      *written = nbyte;
      return ORFS_SUCCESS;
    }
  else
    {
      struct ramfs_file_block *current_block = &data->block[begin_block];
      char *current_bufptr;
      orfs_off_t remaining_bytes;

      /* copy the end of the first block */
      memcpy(current_block->ptr + begin_offset,
	     buf,
	     MEMORY_BLOCK_SIZE-begin_offset);

      /* prepare following copy */
      current_block++;
      remaining_bytes = nbyte - (MEMORY_BLOCK_SIZE-begin_offset);
      current_bufptr = (char*)buf + (MEMORY_BLOCK_SIZE-begin_offset);

      do {
	orfs_off_t to_copy = remaining_bytes < MEMORY_BLOCK_SIZE
	  ? remaining_bytes
	  : MEMORY_BLOCK_SIZE;
	
	/* copy data */
	memcpy(current_block->ptr,
	       buf,
	       to_copy);
	/* update parameters */
	current_bufptr += to_copy;
	remaining_bytes -= to_copy;
	current_block++;
      } while ( remaining_bytes > 0 );
     
      *written = nbyte;
      return ORFS_SUCCESS;
    }
} 


static INLINE
orfs_result_t prepare_write_node(struct ramfs_node *node,
				 orfs_off_t nbyte,
				 orfs_off_t offset)
{
  struct ramfs_file_data *data = &node->data.file;
  orfs_off_t begin;
  orfs_off_t end;
  orfs_off_t block;
  orfs_result_t res;

  LOG(4, "PREPARE WRITE in node %lld at %lld for %lld\n",
      (long long)node->ino, (long long)offset, (long long)nbyte);

  begin = offset;
  end = offset+nbyte-1;

  if ( end < begin )
    return ORFS_SERVER_INVALID_PARAMETER;

  block = BLOCK_FROM_OFFSET(begin);
  if ( block != BLOCK_FROM_OFFSET(end) )
    return ORFS_INVALID_PARAMETER;

  res = alloc_blocks(node, block, block);
  if ( res != ORFS_SUCCESS )
    return res;

  if ( end+1 > data->length )
    data->length = end+1;

  data->writers++;

  return ORFS_SUCCESS;
} 


static INLINE
orfs_result_t commit_write_node(struct ramfs_node *node,
				const char *buf,
				orfs_off_t nbyte,
				orfs_off_t offset)
{
  struct ramfs_file_data *data = &node->data.file;
  orfs_off_t begin;
  orfs_off_t end;
  orfs_off_t block;
  orfs_off_t begin_offset;

  LOG(4, "COMMIT WRITE in node %lld at %lld for %lld\n",
      (long long)node->ino, (long long)offset, (long long)nbyte);

  begin = offset;
  end = offset+nbyte-1;

  if ( end < begin )
    return ORFS_SERVER_INVALID_PARAMETER;

  block = BLOCK_FROM_OFFSET(begin);
  if ( block != BLOCK_FROM_OFFSET(end) )
    return ORFS_INVALID_PARAMETER;

  if ( end+1 > data->length )
    return ORFS_INVALID_PARAMETER;

  if ( !data->writers )
    return ORFS_INVALID_PARAMETER;

  NOTIFY_WRITE_ACCESS(node);

  begin_offset = OFFSET_IN_BLOCK(begin);
  memcpy(data->block[block].ptr + begin_offset, buf, nbyte);

  if ( !--data->writers )
    wake_up_waiters(node);
  
  return ORFS_SUCCESS;
} 


static INLINE
orfs_result_t process_file_request(struct server_request *req)
{
  orfs_req_subtype_t subtype = req->req.subtype;
  ramfs_fd_t fd = (ramfs_fd_t)req->req.target.file.file.id;
  struct ramfs_node *node;
  orfs_result_t res;

  res = ORFS_BAD_FILE;
  if ( !(node=fd_table[fd].node) )
    goto out;

  switch ( subtype ) {
  case ORFS_CLOSE_FILE_REQ: {
    LOG(3, "Request subtype is CLOSE\n");
    res = release_file(node, fd);
    break;
  }
  case ORFS_PREAD_FILE_REQ: {
    orfs_off_t offset = req->req.target.file.data.data_access.offset;
    orfs_off_t count = req->req.target.file.data.data_access.count;
    LOG(3, "Request subtype is PREAD\n");
    if ( ORFS_ISREG(node->mode) )
      res = pread_node(node,
		       &req->out_vect, count, offset,
		       &req->ans.target.file.data.data_access.count);
    else
      res = ORFS_BAD_INODE_TYPE;
    break;
  }
  case ORFS_PWRITE_FILE_REQ: {
    orfs_off_t offset = req->req.target.file.data.data_access.offset;
    orfs_off_t count = req->req.target.file.data.data_access.count;
    LOG(3, "Request subtype is PWRITE\n");
    if ( ORFS_ISREG(node->mode) )
      res = pwrite_node(node,
			VECT_PTR(req->in_vect, 0), count, offset,
			&req->ans.target.file.data.data_access.count);
    else
      res = ORFS_BAD_INODE_TYPE;
    break;
  }
  case ORFS_PREADDIR_FILE_REQ: {
    orfs_doff_t offset = req->req.target.file.data.preaddir.offset;
    orfs_data_len_t data_max = req->req.target.file.data.preaddir.data_max;
    LOG(3, "Request subtype is PREADDIR\n");
    if ( ORFS_ISDIR(node->mode) )
      res = preaddir_node(node,
			  &req->out_vect, offset, data_max,
			  &req->ans.target.file.data.preaddir.count,
			  &req->ans.target.file.data.preaddir.eof_reached,
			  &req->ans.target.file.data.preaddir.data_len);
    else
      res = ORFS_BAD_INODE_TYPE;
    break;
  }
  case ORFS_SYNC_FILE_REQ: {
    LOG(3, "Request subtype is SYNC\n");
    /* nothing to do */
    break;
  }
  default:
    ERROR("FILE request subtype %d is UNKNOWN\n", (int)subtype);
    res = ORFS_BAD_REQUEST;
  }

  /* update attributes */
  getattr_node(node, &req->ans.target.file.node_attr);

 out:
  return res;
}


/****************
 * Page request *
 ****************/

static INLINE
orfs_result_t process_page_request(struct server_request *req)
{
  orfs_req_subtype_t subtype = req->req.subtype;
  orfs_ino_t ino = req->req.target.page.inode.ino;
  struct ramfs_node *node = lookup_ino(ino);
  orfs_result_t res;

  res = ORFS_NO_SUCH_INODE;
  if ( !node )
    goto out;

  res = ORFS_BAD_INODE_TYPE;
  if ( !ORFS_ISREG(node->mode) )
    goto out;

  switch ( subtype ) {
  case ORFS_READ_FULL_PAGE_REQ: {
    orfs_off_t offset = req->req.target.page.data.data_access.offset;
    orfs_off_t count = req->req.target.page.data.data_access.count;
    LOG(3, "Request subtype is READ FULL PAGE\n");
    res = pread_node(node, &req->out_vect, count, offset,
		     &req->ans.target.page.data.data_access.count);
    break;
  }
  case ORFS_WRITE_FULL_PAGE_REQ: {
    orfs_off_t offset = req->req.target.page.data.data_access.offset;
    orfs_off_t count = req->req.target.page.data.data_access.count;
    LOG(3, "Request subtype is WRITE FULL PAGE\n");
    res = pwrite_node(node, VECT_PTR(req->in_vect, 0), count, offset,
		      &req->ans.target.page.data.data_access.count);
    break;
  }
  case ORFS_PREPARE_WRITE_PAGE_REQ: {
    orfs_off_t offset = req->req.target.page.data.data_access.offset;
    orfs_off_t count = req->req.target.page.data.data_access.count;
    LOG(3, "Request subtype is PREPARE WRITE\n");
    res = prepare_write_node(node, count, offset);
    break;
  }
  case ORFS_COMMIT_WRITE_PAGE_REQ: {
    orfs_off_t offset = req->req.target.page.data.data_access.offset;
    orfs_off_t count = req->req.target.page.data.data_access.count;
    LOG(3, "Request subtype is COMMIT WRITE\n");
    res = commit_write_node(node, VECT_PTR(req->in_vect, 0), count, offset);
    break;
  }
  default:
    ERROR("PAGE request subtype %d is UNKNOWN\n", (int)subtype);
    res = ORFS_BAD_REQUEST;
  }

  /* update attributes */
  getattr_node(node, &req->ans.target.page.node_attr);

 out:
  return res;
}


/******************
 * INODE requests *
 ******************/

static INLINE
orfs_result_t process_inode_request(struct server_request *req)
{
  orfs_req_subtype_t subtype = req->req.subtype;
  orfs_ino_t ino = req->req.target.inode.inode.ino;
  struct ramfs_node *node = lookup_ino(ino);
  orfs_result_t res;

  res = ORFS_NO_SUCH_INODE;
  if ( !node )
    goto out;

  switch ( subtype ) {
  case ORFS_SETATTR_INODE_REQ: {
    LOG(3, "Request subtype is SETATTR\n");
    res = setattr_node(node,
		       &req->req.target.inode.data.setattr.new_attr);
    break;
  }
  case ORFS_GETATTR_INODE_REQ: {
    LOG(3, "Request subtype is GETATTR\n");
    getattr_node(node, &req->ans.target.inode.node_attr);
    res = ORFS_SUCCESS;
    /* no need to update same attributes */
    goto no_need_to_update;
  }
  case ORFS_READLINK_INODE_REQ: {
    LOG(3, "Request subtype is READLINK\n");
    res = readlink_node(node, 
			&req->out_vect,
			&req->ans.target.inode.data.readlink.count);
    break;
  }
  case ORFS_OPEN_INODE_REQ: {
    LOG(3, "Request subtype is OPEN\n");
    res = open_node(node, (ramfs_fd_t*)&req->ans.target.inode.data.open.file.id);
    break;
  }
  default:
    ERROR("INODE request subtype %d is UNKNOWN\n", (int)subtype);
    res = ORFS_BAD_REQUEST;
  }

  if ( res == RAMFS_REQUEST_BLOCKED ) {
    enqueue_waiter(node, req);
    return RAMFS_REQUEST_BLOCKED;
  }

  /* update node attributes */
  getattr_node(node, &req->ans.target.inode.node_attr);

 no_need_to_update:
 out:
  return res;
}



/****************
 * DIR requests *
 ****************/

static INLINE
orfs_result_t lookup_node(struct ramfs_node *dir_node,
			  const char *name,
			  struct ramfs_node **nodep)
{
  struct ramfs_dentry *dentry;

  LOG(4, "LOOKUP of node %s if dir %lld\n",
      name, (long long)dir_node->ino);

  dentry = lookup_name(name, dir_node);
  if ( !dentry )
    return ORFS_NO_SUCH_DIR_ENTRY;
  
  *nodep = dentry->node;
  return ORFS_SUCCESS;
}


static INLINE
orfs_result_t unlink_node(struct ramfs_node *dir_node,
			  const char *name)
{
  struct ramfs_dentry *dentry;
  struct ramfs_node *node;

  dentry = lookup_name(name, dir_node);
  if ( !dentry )
    return ORFS_NO_SUCH_DIR_ENTRY;

  node = dentry->node;
  if ( !ORFS_ISREG(node->mode)
       && !ORFS_ISLNK(node->mode)
       && !ORFS_ISCHR(node->mode)
       && !ORFS_ISBLK(node->mode) )
    return ORFS_BAD_INODE_TYPE;	

  /* remove from parent dentry list */
  unlink_dentry(dentry);
  put_old_dentry(dentry);

  /* update ctime */
  NOTIFY_CHANGE_ACCESS(node);

  LOG(4, "UNLINK inode %lld, still has count %d and nlink %d\n",
      (long long)node->ino, node->count, node->nlink);

  if ( !node->count  &&  !node->nlink ) {
    /* real remove */
    free_unlinked_node(node);
    LOG(5, "Really unlinking inode %lld\n",
	(long long)node->ino);
  }
  return ORFS_SUCCESS;
}


static INLINE
orfs_result_t rmdir_node(struct ramfs_node *dir_node,
			 const char *name)
{
  struct ramfs_dentry *dentry;
  struct ramfs_node *node;
  struct ramfs_dentry *dot_dentry;
  struct ramfs_dentry *dotdot_dentry;

  dentry = lookup_name(name, dir_node);
  if ( !dentry )
    return ORFS_NO_SUCH_DIR_ENTRY;

  node = dentry->node;
  if ( !ORFS_ISDIR(node->mode) )
    return ORFS_BAD_INODE_TYPE;

  if ( node->data.dir.entry_nr > 2 ) /* keep . and .. */
    return ORFS_NOT_EMPTY_DIR;

  if ( node == fs.root ) {
    ERROR("Tried to remove root directory\n");
    return ORFS_INVALID_PARAMETER;
  }

  /* remove from parent dentry list */
  unlink_dentry(dentry);
  put_old_dentry(dentry);

  /* remove link to parent */
  dotdot_dentry = node->data.dir.entry[1];
  unlink_dentry(dotdot_dentry);
  put_old_dentry(dotdot_dentry);

  /* remove link to us */
  dot_dentry = node->data.dir.entry[0];
  unlink_dentry(dot_dentry);
  put_old_dentry(dot_dentry);

  LOG(4, "RMDIR inode %lld, still has count %d and nlink %d\n",
      (long long)node->ino, node->count, node->nlink);

  if ( !node->count ) {
    /* real remove */
    free_unlinked_node(node);
    LOG(5, "Really unlinking inode %lld\n",
	(long long)node->ino);
  }  
  return ORFS_SUCCESS;
}


static INLINE
orfs_result_t mkreg_node(struct ramfs_node *dir_node,
			 const char *name,
			 orfs_name_len_t name_len,
			 orfs_uid_t uid,
			 orfs_gid_t gid,
			 orfs_mode_t mode,
			 struct ramfs_node **newp)
{
  struct ramfs_dentry *dentry;
  struct ramfs_node *node;
  struct ramfs_file_data *data;
  orfs_result_t res;

  LOG(4, "MKREG node %s in dir %lld with mode %lld\n",
      name, (long long)dir_node->ino, (long long)mode);

  /* check whether such dentry already exist */
  res = ORFS_EXISTING_DIR_ENTRY;
  if ( (dentry=lookup_name(name, dir_node)) != NULL )
    goto error;
  
  res = ORFS_NO_SPACE_LEFT;

  /* alloc node */
  node = get_new_node(mode | ORFS_IFREG, uid, gid);
  if ( !node )
    goto error;
  
  /* alloc file data */
  data = &node->data.file;
  data->block = calloc(BLOCKS_DEFAULT_MAX, sizeof(*data->block));
  if ( !data->block ) {
    ERROR("Cannot allocate RAMFS file block list\n");
    goto error_with_empty_node;
  }
  data->block_max = BLOCKS_DEFAULT_MAX;
  data->block_nr = 0;
  data->length = 0;
  data->writers = 0;
  INIT_REQUEST_QUEUE(data->waiters);

  /* alloc dentry */
  dentry = get_new_dentry(dir_node, node, name, name_len);
  if ( !dentry )
    goto error_with_node;

  /* link */
  if ( link_dentry(dentry) != ORFS_SUCCESS )
    goto error_with_dentry;

  LOG(4, "Allocated regular inode %lld with name %s at %lld in dir %lld\n",
      (long long)node->ino, dentry->name,
      (long long)dentry->indix_for_parent, (long long)dir_node->ino);

  *newp = node;
  return ORFS_SUCCESS;

 error_with_dentry:
  put_old_dentry(dentry);
 error_with_node:
  free(data->block);
 error_with_empty_node:
  free(node);
 error:
  return res;
}


static INLINE
orfs_result_t mkdir_node(struct ramfs_node *dir_node,
			 const char *name,
			 orfs_name_len_t name_len,
			 orfs_uid_t uid,
			 orfs_gid_t gid,
			 orfs_mode_t mode,
			 struct ramfs_node **newp)
{
  struct ramfs_dentry *dentry;
  struct ramfs_node *node;
  struct ramfs_dir_data *data;
  struct ramfs_dentry *dot_dentry, *dotdot_dentry;
  orfs_result_t res;

  LOG(4, "MKDIR node %s in dir %lld with mode %lld\n",
      name, (long long)dir_node->ino, (long long)mode);

  res = ORFS_EXISTING_DIR_ENTRY;
  if ( (dentry=lookup_name(name, dir_node)) != NULL )
    goto error;
  
  res = ORFS_NO_SPACE_LEFT;

  /* alloc_node */
  node = get_new_node(mode | ORFS_IFDIR, uid, gid);
  if ( !node )
    goto error;

  /* alloc dir data */
  data = &node->data.dir;
  data->entry = calloc(DIRENTS_DEFAULT_MAX, sizeof(*data->entry));
  if ( !data->entry ) {
    ERROR("Cannot allocate RAMFS dir entry list\n");
    goto error_with_empty_node;
  }
  data->entry_max = DIRENTS_DEFAULT_MAX;
  data->entry_nr = 0;

  /* fill trivial dirents */
  dot_dentry = get_new_dentry(node, node, ".", 1);
  if ( !dot_dentry )
    goto error_with_node;
  dotdot_dentry = get_new_dentry(node, dir_node, "..", 2);
  if ( !dotdot_dentry )
    goto error_with_dot;
  /* these link cannot fail */
  link_dentry(dot_dentry);
  link_dentry(dotdot_dentry);

  /* alloc dentry (except for root node) */
  dentry = get_new_dentry(dir_node, node, name, name_len);
  if ( !dentry )
    goto error_with_dotdot;
  if ( link_dentry(dentry) != ORFS_SUCCESS )
    goto error_with_dentry;
  
  LOG(4, "Allocated directory inode %lld with name %s at %lld in dir %lld\n",
      (long long)node->ino, dentry->name,
      (long long)dentry->indix_for_parent, (long long)dir_node->ino);

  *newp = node;
  return ORFS_SUCCESS;

 error_with_dentry:
  put_old_dentry(dentry);
 error_with_dotdot:
  put_old_dentry(dotdot_dentry);
 error_with_dot:
  put_old_dentry(dot_dentry);
 error_with_node:
  free(data->entry);
 error_with_empty_node:
  free(node);
 error:
  return res;
}


static INLINE
orfs_result_t mknod_node(struct ramfs_node *dir_node,
			 const char *name,
			 orfs_name_len_t name_len,
			 orfs_uid_t uid,
			 orfs_gid_t gid,
			 orfs_dev_t dev,
			 orfs_mode_t mode,
			 struct ramfs_node **newp)
{
  struct ramfs_dentry *dentry;
  struct ramfs_node *node;
  orfs_result_t res;

  LOG(4, "MKNOD node %s in dir %lld with mode %lld dev %lld\n",
      name, (long long)dir_node->ino, (long long)mode, (long long)dev);

  res = ORFS_EXISTING_DIR_ENTRY;
  if ( (dentry=lookup_name(name, dir_node)) != NULL )
    goto error;
  
  /* alloc_node */
  res = ORFS_NO_SPACE_LEFT;
  node = get_new_node(mode, uid, gid);
  if ( !node )
    goto error;

  /* fill special fields */
  switch ( ORFS_ISFMT(mode) ) {
  case ORFS_IFCHR:
  case ORFS_IFBLK:
    node->data.dev.dev = dev;
  case ORFS_IFIFO:
  case ORFS_IFSOCK:
    break;
  default:
    res = ORFS_INVALID_PARAMETER;
    ERROR("Bad mode %ld passed to mknod\n", (long)mode);
    goto error_with_node;
  }

  /* alloc dentry */
  res = ORFS_NO_SPACE_LEFT;
  dentry = get_new_dentry(dir_node, node, name, name_len);
  if ( !dentry )
    goto error_with_node;

  /* link */
  if ( link_dentry(dentry) != ORFS_SUCCESS )
    goto error_with_dentry;

  LOG(4, "Allocated regular inode %lld with name %s at %lld in dir %lld\n",
      (long long)node->ino, dentry->name,
      (long long)dentry->indix_for_parent, (long long)dir_node->ino);

  *newp = node;
  return ORFS_SUCCESS;

 error_with_dentry:
  put_old_dentry(dentry);
 error_with_node:
  free(node);
 error:
  return res;
}


static INLINE
orfs_result_t symlink_node(struct ramfs_node *dir_node,
			   const char *name,
			   orfs_name_len_t name_len,
			   struct io_vector *in_vect,
			   orfs_path_len_t path_len,
			   orfs_uid_t uid,
			   orfs_gid_t gid,
			   struct ramfs_node **newp)
{
  struct ramfs_node *node;
  struct ramfs_dentry *dentry;
  orfs_result_t res;

  LOG(4, "SYMLINK node %s in dir %lld as %s\n",
      name, (long long)dir_node->ino, (char*) VECT_PTR(in_vect, 0));

  res = ORFS_EXISTING_DIR_ENTRY;
  dentry = lookup_name(name, dir_node);
  if ( dentry )
    goto error;

  /* check symlen */
  res = ORFS_INVALID_PARAMETER;
  if ( path_len > MEMORY_BLOCK_SIZE-1 )
    // impossible with uint16 and 65535
    goto error;

  res = ORFS_NO_SPACE_LEFT;

  /* alloc_node */
  node = get_new_node(ORFS_IFLNK | ORFS_IRWXUGO, uid, gid);
  if ( !node )
    goto error;

  /* alloc dentry */
  dentry = get_new_dentry(dir_node, node, name, name_len);
  if ( !dentry )
    goto error_with_node;

  /* link */
  if ( link_dentry(dentry) != ORFS_SUCCESS )
    goto error_with_dentry;

  /* get symlink */
  node->data.lnk.length = path_len;
  node->data.lnk.path = VECT_PTR(in_vect, 0);
  node->data.lnk.path[path_len] = '\0';
  in_vect->do_not_free = 1;

  LOG(4, "Allocated symlink inode %lld with name %s at %lld in dir %lld\n",
      (long long)node->ino, dentry->name,
      (long long)dentry->indix_for_parent, (long long)dir_node->ino);

  *newp = node;
  return ORFS_SUCCESS;

 error_with_dentry:
  put_old_dentry(dentry);
 error_with_node:
  free(node);
 error:
  return res;
}


static INLINE
orfs_result_t process_dir_request(struct server_request *req)
{
  orfs_req_subtype_t subtype = req->req.subtype;
  orfs_ino_t dir_ino = req->req.target.dir.dir.ino;
  struct ramfs_node *dir_node = lookup_ino(dir_ino);
  const char *name = req->req.target.dir.name;
  orfs_name_len_t name_len = req->req.target.dir.name_len;
  orfs_result_t res;

  LOG(2, "Request on inode %lld for child %s\n",
      (long long)dir_ino, name)

  res = ORFS_NO_SUCH_INODE;
  if ( !dir_node )
    goto out_without_father;

  res = ORFS_BAD_INODE_TYPE;
  if ( !ORFS_ISDIR(dir_node->mode) )
    goto out_with_father;

  switch ( subtype ) {
  case ORFS_LOOKUP_DIR_REQ: {
    struct ramfs_node *node;
    LOG(3, "Request subtype is LOOKUP\n");
    res = lookup_node(dir_node, name, &node);
    if ( res == ORFS_SUCCESS )
      getattr_node(node, &req->ans.target.dir.data.lookup.node_attr);
    break;
  }
  case ORFS_REVALIDATE_DIR_REQ: {
    struct ramfs_node *node;
    LOG(3, "Request subtype is REVALIDATE\n");
    res = lookup_node(dir_node, name, &node);
    if ( res == ORFS_SUCCESS )
      getattr_node(node, &req->ans.target.dir.data.revalidate.node_attr);
    break;
  }
  case ORFS_UNLINK_DIR_REQ: {
    LOG(3, "Request subtype is UNLINK\n");
    res = unlink_node(dir_node, name);
    break;
  }
  case ORFS_RMDIR_DIR_REQ: {
    LOG(3, "Request subtype is RMDIR\n");
    res = rmdir_node(dir_node, name);
    break;
  }
  case ORFS_MKREG_DIR_REQ: {
    orfs_uid_t uid = req->req.target.dir.data.create.uid;
    orfs_gid_t gid = req->req.target.dir.data.create.gid;
    orfs_mode_t mode = req->req.target.dir.data.create.mode;
    struct ramfs_node *new;
    LOG(3, "Request subtype is MKREG\n");
    res = mkreg_node(dir_node, name, name_len, uid, gid, mode, &new);
    if ( res == ORFS_SUCCESS )
      getattr_node(new, &req->ans.target.dir.data.create.new_node_attr);
    break;
  }
  case ORFS_MKDIR_DIR_REQ: {
    orfs_uid_t uid = req->req.target.dir.data.create.uid;
    orfs_gid_t gid = req->req.target.dir.data.create.gid;
    orfs_mode_t mode = req->req.target.dir.data.create.mode;
    struct ramfs_node *new;
    LOG(3, "Request subtype is MKDIR\n");
    res = mkdir_node(dir_node, name, name_len, uid, gid, mode, &new);
    if ( res == ORFS_SUCCESS )
      getattr_node(new, &req->ans.target.dir.data.create.new_node_attr);
    break;
  }
  case ORFS_MKNOD_DIR_REQ: {
    orfs_uid_t uid = req->req.target.dir.data.create.uid;
    orfs_gid_t gid = req->req.target.dir.data.create.gid;
    orfs_dev_t dev = req->req.target.dir.data.create.dev;
    orfs_mode_t mode = req->req.target.dir.data.create.mode;
    struct ramfs_node *new;
    LOG(3, "Request subtype is MKNOD\n");
    res = mknod_node(dir_node, name, name_len, uid, gid, dev, mode, &new);
    if ( res == ORFS_SUCCESS )
      getattr_node(new, &req->ans.target.dir.data.create.new_node_attr);
    break;
  }
  case ORFS_SYMLINK_DIR_REQ: {
    orfs_path_len_t path_len = req->req.target.dir.data.symlink.path_len;
    orfs_uid_t uid = req->req.target.dir.data.symlink.uid;
    orfs_gid_t gid = req->req.target.dir.data.symlink.gid;
    struct ramfs_node *new;
    LOG(3, "Request subtype is SYMLINK\n");
    res = symlink_node(dir_node, name, name_len,
		       req->in_vect, path_len,
		       uid, gid, &new);
    if ( res == ORFS_SUCCESS )
      getattr_node(new, &req->ans.target.dir.data.create.new_node_attr);
    break;
  }
  default:
    ERROR("DIR request subtype %d is UNKNOWN\n", (int)subtype);
    res = ORFS_BAD_REQUEST;
  }  

 out_with_father:
  /* update father attributes */
  getattr_node(dir_node, &req->ans.target.dir.father_attr);

 out_without_father:
  return res;
}


/*****************
 * 2DIR requests *
 *****************/

static INLINE
orfs_result_t move_node(struct ramfs_dentry *dentry,
			struct ramfs_node *new_parent,
			const char *new_name,
			orfs_name_len_t new_name_len)
{
  struct ramfs_node *node = dentry->node;

  LOG(4, "MOVE inode %lld from dir %lld with name %s in %lld\n",
      (long long)node->ino, (long long)dentry->parent->ino,
      dentry->name, (long long)dentry->indix_for_parent);

  if ( move_dentry(dentry, new_parent, new_name, new_name_len) == ORFS_NO_SPACE_LEFT )
    return ORFS_NO_SPACE_LEFT;

  if ( ORFS_ISDIR(node->mode) ) {
    struct ramfs_dentry *dotdot_dentry = node->data.dir.entry[1];
    dotdot_dentry->node->nlink--;
    dotdot_dentry->node = new_parent;
    dotdot_dentry->node->nlink++;    
  }

  LOG(4, "Moved inode %lld to dir %lld with name %s\n",
      (long long)node->ino, (long long)new_parent->ino, new_name);

  return ORFS_SUCCESS;
}


static INLINE
orfs_result_t link_node(struct ramfs_dentry *dentry,
			struct ramfs_node *new_parent,
			const char *new_name,
			orfs_name_len_t new_name_len)
{
  struct ramfs_node *node = dentry->node;
  struct ramfs_dentry *new_dentry;

  LOG(4, "LINK inode %lld from dir %lld with name %s in %lld\n",
      (long long)node->ino, (long long)dentry->parent->ino,
      dentry->name, (long long)dentry->indix_for_parent);

  if ( ORFS_ISDIR(node->mode) )
    return ORFS_BAD_INODE_TYPE_3;

  new_dentry = get_new_dentry(new_parent, node, new_name, new_name_len);
  if ( !new_dentry )
    return ORFS_NO_SPACE_LEFT;

  /* link to new parent */
  link_dentry(new_dentry);

  /* update ctime */
  NOTIFY_CHANGE_ACCESS(node);

  LOG(4, "Linked inode %lld to dir %lld with name %s\n",
      (long long)node->ino, (long long)new_parent->ino, new_name);

  return ORFS_SUCCESS;
}


static INLINE
orfs_result_t process_2dir_request(struct server_request *req)
{
  orfs_req_subtype_t subtype = req->req.subtype;
  orfs_ino_t old_dir_ino = req->req.target.twodir.old_dir.ino;
  const char *old_name = req->req.target.twodir.old_name;
  orfs_ino_t new_dir_ino = req->req.target.twodir.new_dir.ino;
  const char *new_name = req->req.target.twodir.new_name;
  orfs_name_len_t new_name_len = req->req.target.twodir.new_name_len;

  orfs_result_t res;
  struct ramfs_node *old_dir_node, *new_dir_node, *node;
  struct ramfs_dentry *dentry;

  /* check old_dir existency */
  res = ORFS_NO_SUCH_INODE;
  old_dir_node = lookup_ino(old_dir_ino);
  if ( !old_dir_node )
    goto out;
  /* check that it's a directory */
  res = ORFS_BAD_INODE_TYPE;
  if ( !ORFS_ISDIR(old_dir_node->mode) )
    goto out;

  /* check new_dir existency */
  res = ORFS_NO_SUCH_INODE_2;
  new_dir_node = lookup_ino(new_dir_ino);
  if ( !new_dir_node )
    goto out_with_good_old_dir;
  /* check that it's a directory */
  res = ORFS_BAD_INODE_TYPE_2;
  if ( !ORFS_ISDIR(new_dir_node->mode) )
    goto out_with_good_old_dir;

  /* check that new dentry does not exist.
   * this _MUST_ be done before next test
   * to prevent dentry overriding.
   */
  res = ORFS_EXISTING_DIR_ENTRY;
  dentry = lookup_name(new_name, new_dir_node);
  if ( dentry )
    goto out_with_good_dirs;

  /* check that old dentry exists */
  res = ORFS_NO_SUCH_DIR_ENTRY;
  dentry = lookup_name(old_name, old_dir_node);
  if ( !dentry )
    goto out_with_good_dirs;
  
  /* get node before it moves */
  node = dentry->node;

  switch ( subtype ) {
  case ORFS_MOVE_2DIR_REQ:
    LOG(3, "Request subtype is MOVE\n");
    res = move_node(dentry, new_dir_node, new_name, new_name_len);
    break;
  case ORFS_LINK_2DIR_REQ:
    LOG(3, "Request subtype is LINK\n");
    res = link_node(dentry, new_dir_node, new_name, new_name_len);
    break;
  default:
    ERROR("2DIR request subtype %d is UNKNOWN\n", (int)subtype);
    res = ORFS_BAD_REQUEST;
  }  
  
  /* get node attributes, even if move failed */
  getattr_node(node, &req->ans.target.twodir.node_attr);

 out_with_good_dirs:
  /* get new dir attributes */
  getattr_node(new_dir_node, &req->ans.target.twodir.new_father_attr);

 out_with_good_old_dir:
  /* get old dir attributes */
  getattr_node(old_dir_node, &req->ans.target.twodir.old_father_attr);

 out:
  return res;
}


/******************
 * Statfs Request *
 ******************/

static INLINE
orfs_result_t init_req(struct orfs_init_data *data)
{
  data->magic = ORFS_PROTOCOL_MAGIC;
  data->protocol = 0;
  data->blksize = MEMORY_BLOCK_SIZE;
  data->symlen = MEMORY_BLOCK_SIZE-1;

  data->capabilities = 0;
  ADD_ORFS_CAP(data->capabilities, ORFS_LINK_CAP);
  ADD_ORFS_CAP(data->capabilities, ORFS_SYMLINK_CAP);
  ADD_ORFS_CAP(data->capabilities, ORFS_MKNOD_CAP);

  return ORFS_SUCCESS;
}


static INLINE
orfs_result_t statfs_req(struct orfs_fsattr_data *data)
{
  /* undefined block numbers */
  data->blocks = data->bfree = data->bavail = 0;
  /* undefined node numbers */
  data->files = data->ffree = 0;

  return ORFS_SUCCESS;
}


static INLINE
orfs_result_t process_fs_request(struct server_request *req)
{
  orfs_result_t res;
  orfs_req_subtype_t subtype = req->req.subtype;

  switch ( subtype ) {
  case ORFS_INIT_FS_REQ:
    LOG(3, "Request subtype is INIT\n");
    res = init_req(&req->ans.target.fs.data.init.data);
    break;
  case ORFS_FINI_FS_REQ:
    LOG(3, "Request subtype is FINI\n");
    res = ORFS_SUCCESS;
    break;
  case ORFS_PING_FS_REQ:
    LOG(3, "Request subtype is PING\n");
    res = ORFS_SUCCESS;
    break;
  case ORFS_STATFS_FS_REQ:
    LOG(3, "Request subtype is STATFS\n");
    res = statfs_req(&req->ans.target.fs.data.statfs.fsattr);
    break;
  default:
    ERROR("FS request subtype %d is UNKNOWN\n", (int)subtype);
    res = ORFS_BAD_REQUEST;
  }

  return res;
}


/******************************
 * General Request Processing *
 ******************************/

void ramfs_process_queue()
{
  struct server_request *req;
  orfs_req_type_t type;
  orfs_result_t res;
    
  while ( (req=DEQUEUE_REQUEST(&ready_request_queue)) != NULL ) {
    type = req->req.type;
    switch ( type ) {
    case ORFS_FILE_REQ:
      LOG(3, "Request type is FILE\n");
      res = process_file_request(req);
      break;
    case ORFS_PAGE_REQ:
      LOG(3, "Request type is PAGE\n");
      res = process_page_request(req);
      break;
    case ORFS_INODE_REQ:
      LOG(3, "Request type is INODE\n");
      res = process_inode_request(req);
      break;
    case ORFS_DIR_REQ:
      LOG(3, "Request type is DIR\n");
      res = process_dir_request(req);
      break;
    case ORFS_2DIR_REQ:
      LOG(3, "Request type is 2DIR\n");
      res = process_2dir_request(req);
      break;
    case ORFS_FS_REQ:
      LOG(3, "Request type is FS\n");
      res = process_fs_request(req);
      break;
    default:
      ERROR("Request type %d is UNKNOWN\n", (int)type);
      res = ORFS_BAD_REQUEST;
    }
    
    if ( res == RAMFS_REQUEST_BLOCKED )
      continue;

    req->ans.result = res;
    req->ans.type = type;
    req->ans.subtype = req->req.subtype;
    req->ans.tag = req->req.tag;
    
    put_io_vector(req->in_vect);

    QUEUE_REQUEST(&done_request_queue, req);
    counters.processed++;
  }
}


/******************
 * Initialization *
 ******************/

int ramfs_init()
{
  struct ramfs_node *node;
  struct ramfs_dentry *dot_dentry, *dotdot_dentry;
  struct ramfs_dir_data *data;
  ramfs_hash_t i;
  ramfs_fd_t j;

  /* init ino and dentry hash tables */
  for(i=0; i<HASH_TABLE_SIZE; i++) {
    LIST_INIT(&fs.ino_table[i]);
    LIST_INIT(&fs.name_table[i]);
  }

  /* alloc root node */
  node = get_new_node(ORFS_IFDIR | ORFS_IRWXUGO, 0, 0);
  if ( !node )
    goto error;

  /* alloc root dir data */
  data = &node->data.dir;
  data->entry = calloc(DIRENTS_DEFAULT_MAX, sizeof(*data->entry));
  if ( !data->entry ) {
    ERROR("Cannot allocate RAMFS dir entry list\n");
    goto error_with_empty_node;
  }
  data->entry_max = DIRENTS_DEFAULT_MAX;
  data->entry_nr = 0;

  /* fill trivial dirents */
  dot_dentry = get_new_dentry(node, node, ".", 1);
  if ( !dot_dentry )
    goto error_with_node;
  dotdot_dentry = get_new_dentry(node, node, "..", 2);
  if ( !dotdot_dentry )
    goto error_with_dot;
  /* these link cannot fail */
  link_dentry(dot_dentry);
  link_dentry(dotdot_dentry);

  /* set root */
  fs.root = node;

  /* init file descriptors */
  for(j=0; j<DESCRIPTOR_MAX; j++) {
    fd_table[j].node = NULL;
    fd_table[j].next_free = j+1;
  }
  first_free_fd = 0;

  /* alloc fake buffer */
  fake_buffer = network_malloc(MEMORY_BLOCK_SIZE);
  if ( !fake_buffer )
    goto error_with_fs;

  return 0;

 error_with_fs:
  put_old_dentry(dotdot_dentry);
 error_with_dot:
  put_old_dentry(dot_dentry);
 error_with_node:
  free(data->entry);
 error_with_empty_node:
  free(node);
 error:
  return -1;
}


void ramfs_exit()
{
  ramfs_hash_t i;

  network_free(fake_buffer);

  for(i=0; i<HASH_TABLE_SIZE; i++) {
    struct list_head *head = &fs.ino_table[i];
    struct list_head *elt, *next;

    for(elt=head; elt!=head; elt=next) {
      struct ramfs_node *node = LIST_HEAD_NODE(elt);
      next = elt->next;

      if ( ORFS_ISDIR(node->mode) ) {
	struct ramfs_dir_data *data = &node->data.dir;
	int i=0;
	orfs_doff_t off = 0;
	while ( i<data->entry_nr ) {
	  off = data->entry[off]->off;
	  put_old_dentry(data->entry[off]);
	  i++;
	}
      }
      free_unlinked_node(node);

    }

  }
}
