/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2004 Brice Goglin <Brice.Goglin@ens-lyon.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <getopt.h>

#include "orfs_common_network.h"
#include "ramfs_network.h"
#include "ramfs_fs.h"
#include "ramfs_queue.h"
#include "ramfs_messages.h"


/* Initialize messaging */
INIT_LOGS();

/* Initialize request counters */
struct ramfs_counters counters = { 0 };


/*********************
 * Network functions *
 *********************/

int (*network_init)() = NULL;
void (*network_exit)() = NULL;
void (*network_process)(int *) = NULL;
void* (*network_malloc)(size_t) = NULL;
void (*network_free)(void*) = NULL;


/************************
 * Command Line Parsing *
 ************************/

static char *callname;
static int verbose = 0;


static
void exit_with_usage()
{
  fprintf(stderr,
	  "Usage: %s [options]\n"
#if CONFIG_TCP
	  "  --tcp-port|-T N    set TCP server port\n"
#endif
#if CONFIG_GM
	  "  --gm-port|-G N     set GM server port\n"
#endif
#if CONFIG_MX
	  "  --mx-port|-M N     set MX server endpoint\n"
#endif
#if CONFIG_GM || CONFIG_MX
	  "  --myri-board|-B N  set server's Myrinet board (for GM/MX)\n"
#endif
	  , callname);
  exit(-1);
}


static inline
void parse_cmdline(int argc, char *const argv[])
{
  char c;
  static struct option long_opts[] = { 
#if CONFIG_TCP
    { "tcp", 0, NULL, 't' },
    { "tcp-port", 1, NULL, 'T' },
#endif
#if CONFIG_GM
    { "gm", 0, NULL, 'g' },
    { "gm-port", 1, NULL, 'G' },
#endif
#if CONFIG_MX
    { "mx", 0, NULL, 'm' },
    { "mx-port", 1, NULL, 'M' },
#endif
#if CONFIG_GM || CONFIG_MX
    { "myri-board", 1, NULL, 'B' },
#endif
    { NULL, 0, NULL, 0 },
  };

  callname = strrchr(argv[0], '/')+1;

  /* Default network setup */
#if CONFIG_TCP
  TCP_SETUP();
#elif CONFIG_GM
  GM_SETUP();
#elif CONFIG_MX
  MX_SETUP();
#endif

  /* option parsing loop */
  while ( (c=getopt_long(argc, argv,
#if CONFIG_TCP
			 "tT:"
#endif
#if CONFIG_GM
			 "gG:"
#endif
#if CONFIG_MX
			 "mM:"
#endif
#if CONFIG_GM || CONFIG_MX
			 "B:"
#endif
			 "v", long_opts, NULL))
	  != EOF ) {
    switch ( c ) {

#if CONFIG_TCP
    case 't':
      TCP_SETUP();
      if ( verbose )
	fprintf(stderr, "Using TCP network layer\n");
      break;

    case 'T': {
      long port;
      char *strtolres;

      port = strtol(optarg, &strtolres, 10);

      if ( !*optarg || *strtolres ) {
	fprintf(stderr, "Unrecognized TCP port %s\n", optarg);
	exit_with_usage();
      }	
      tcp_server_port = port;
      if ( verbose )
	fprintf(stderr,
		"Using TCP port %ld\n",
		tcp_server_port);
      break;
    }
#endif /* CONFIG_TCP */

#if CONFIG_GM
    case 'g':
      GM_SETUP();
      if ( verbose )
	fprintf(stderr, "Using GM network layer\n");
      break;

    case 'G': {
      long port;
      char *strtolres;

      port = strtol(optarg, &strtolres, 10);

      if ( !*optarg || *strtolres ) {
	fprintf(stderr, "Unrecognized GM port %s\n", optarg);
	exit_with_usage();
      }	
      gm_server_port = port;
      if ( verbose )
	fprintf(stderr,
		"Using GM port %ld\n",
		gm_server_port);
      break;
    }
#endif /* CONFIG_GM */

#if CONFIG_MX
    case 'm':
      MX_SETUP();
      if ( verbose )
	fprintf(stderr, "Using MX network layer\n");
      break;

    case 'M': {
      long ep;
      char *strtolres;

      ep = strtol(optarg, &strtolres, 10);

      if ( !*optarg || *strtolres ) {
	fprintf(stderr, "Unrecognized MX endpoint %s\n", optarg);
	exit_with_usage();
      }	
      mx_server_ep = ep;
      if ( verbose )
	fprintf(stderr,
		"Using MX endpoint %ld\n",
		mx_server_ep);
      break;
    }
#endif /* CONFIG_MX */

#if CONFIG_GM || CONFIG_MX
    case 'B': {
      long board;
      char *strtolres;

      board = strtol(optarg, &strtolres, 10);

      if ( !*optarg || *strtolres ) {
	fprintf(stderr, "Unrecognized Myrinet board %s\n", optarg);
	exit_with_usage();
      }	
#if CONFIG_GM
      gm_server_board = board;
#endif
#if CONFIG_MX
      mx_server_board = board;
#endif
      if ( verbose )
	fprintf(stderr,
		"Using Myrinet board %ld\n",
		board);
      break;
    }
#endif /* CONFIG_GM || CONFIG_MX */

    case 'v':
      verbose = 1;
      break;

    case '?':
    default:
      exit_with_usage();
    }
  }

  /* Check network */
  if ( !network_init ) {
    fprintf(stderr, "No network layer available\n");
    exit(-1);
  }
}


/******************
 * Manage Signals *
 ******************/

static
int exiting = 0;

static
void sigint_handler()
{
  exiting = 1;
}

static
void sigterm_handler()
{
  printf("Received %ld Processed %ld Sent %ld Pending %ld Sending %ld\n",
	 counters.received, counters.processed, counters.sent,
	 counters.received-counters.processed, counters.processed-counters.sent);
}

static
void fix_sigint_handler()
{
  struct sigaction sa;
  sa.sa_flags = 0;
  sigemptyset(&sa.sa_mask);
  sa.sa_handler = sigint_handler;
  sigaction(SIGINT, &sa, NULL);
  sa.sa_handler = sigterm_handler;
  sigaction(SIGTERM, &sa, NULL);
}


/****************
 * Main Program *
 ****************/

int main(int argc, char *const argv[])
{
  int res = -1;

  parse_cmdline(argc, argv);
  
  fix_sigint_handler();

  queue_init();

  res = network_init();
  if ( res < 0 )
    goto out;

  res = ramfs_init();
  if ( res < 0 )
    goto out_with_net;
  
  network_process(&exiting);

  ramfs_exit();
 out_with_net:
  network_exit();
 out:
  return res;
}
